/*
 * nasd_cheops_aio.c
 */
/*
 * io queue which manages multiple parallel ios for striped objects
 *
 *
 *
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */


#include <nasd/nasd_options.h>
#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <stdlib.h>

#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_cheops_common.h>
#include <nasd/nasd_cheops_types.h>
#include <nasd/nasd_cheops_raidmap.h> 
#include <nasd/nasd_cheops_ios.h>

#define _NASD_CHEOPS_AIO_DEBUG 0

_nasd_cheops_ioq_t *gl_io_queue;

extern void 
_nasd_cheops_print_asm (nasd_asm_t *asm);

/* local functions */
int _nasd_cheops_ioq_enq(_nasd_cheops_ioq_t *, _nasd_cheops_io_t *);
int _nasd_cheops_ioq_deq(_nasd_cheops_ioq_t *, _nasd_cheops_io_t **);
nasd_boolean_t _nasd_cheops_ioq_empty(_nasd_cheops_ioq_t *);
void _nasd_cheops_asm_freeall(nasd_asm_t *);
void _nasd_cheops_io_freeall(_nasd_cheops_aggr_io_t *);
void _nasd_cheops_init_io(NASD_pda_t *, _nasd_cheops_aggr_io_t *, 
			  _nasd_cheops_io_t *);
int _nasd_cheops_enq_io(_nasd_cheops_io_t *);

/* Queue operations */
int
_nasd_cheops_ioq_init()
{
	int rc;
	_nasd_cheops_ioq_t *ioq;

	NASD_Malloc(ioq, sizeof(_nasd_cheops_ioq_t), (_nasd_cheops_ioq_t *));
	if (ioq==NULL) {
	  (cheops_Msg "malloc failed..\n");
	  return ENOMEM;
	} else
	   gl_io_queue = ioq;

#if MJU_DEBUG
	fprintf(stderr, "_nasd_cheops_ioq_init(): gl_io_queue=0x%x\n",
		(unsigned int) gl_io_queue);
#endif /* MJU_DEBUG */

	ioq->q_head = NULL;
	ioq->q_tail = NULL;

	rc = nasd_mutex_init( &(ioq->q_mutex) );
	if (rc) {
	  (cheops_Msg "Error: can not initialize mutex\n");
	  return rc;
	}

	rc = nasd_cond_init( &(ioq->q_not_empty) );
	if (rc) {
	  (cheops_Msg "Error: can not initialize condition variable\n");
	  return rc;
	}

  	return 0;
}

int
_nasd_cheops_ioq_enq(_nasd_cheops_ioq_t *ioq,
		    _nasd_cheops_io_t       *new_item)
{

	NASD_LOCK_MUTEX(ioq->q_mutex);

	if ( ioq->q_head == NULL ){
	  new_item->qnext = NULL;
	  new_item->qprev= NULL;
	  ioq->q_head = new_item;
	  ioq->q_tail = new_item;
	  NASD_SIGNAL_COND(ioq->q_not_empty);
	} else {
	  new_item->qnext = ioq->q_head;
	  new_item->qprev = NULL;
 	  ioq->q_head->qprev = new_item;
	  ioq->q_head = new_item;
	}
	NASD_UNLOCK_MUTEX(ioq->q_mutex);
  return 0;
}

int
_nasd_cheops_ioq_deq(_nasd_cheops_ioq_t *ioq,
	  	    _nasd_cheops_io_t       **ret_item)
{
	_nasd_cheops_io_t *item_bef_last;

	NASD_LOCK_MUTEX(ioq->q_mutex);  

	if (ioq->q_head == NULL) {
	  *ret_item = NULL;
	  NASD_UNLOCK_MUTEX(ioq->q_mutex);
	  return _NASD_CHEOPS_ERR_ISEMPTY;
	}

	if (ioq->q_head == ioq->q_tail) {
	  *ret_item = ioq->q_head;
	  ioq->q_head = ioq->q_tail = NULL;
	} else {
	  *ret_item = ioq->q_tail;
	  item_bef_last = ioq->q_tail->qprev;
 	  item_bef_last->qnext = NULL;
	  ioq->q_tail = item_bef_last;
	  NASD_SIGNAL_COND(ioq->q_not_empty);
	}

	NASD_UNLOCK_MUTEX(ioq->q_mutex);
	return 0;
}
  
nasd_boolean_t 
_nasd_cheops_ioq_empty(_nasd_cheops_ioq_t *ioq)
{
	if (ioq->q_head==NULL) 
	   return NASD_TRUE;
	else
	   return NASD_FALSE;
}

void 
_nasd_cheops_asm_freeall(nasd_asm_t *asm)
{
         NASD_stripe_access_t *sac, *psac;
	 int i;

	 if (asm ==NULL)
	   return;

         sac = asm->first_stripe_access;	 

	 while (sac != NULL) {
	   psac = sac;
	   sac = sac->next;
	   NASD_Free(psac, sizeof(NASD_stripe_access_t));
	 }

	 NASD_Free(asm, sizeof(nasd_asm_t));

}

void 
_nasd_cheops_io_freeall(_nasd_cheops_aggr_io_t *aggr_io)
{
        int i=0;
	int in_size = aggr_io->in_datalen;
	int out_size = aggr_io->in_datalen;
        _nasd_cheops_io_t *cio, *pcio;


        cio = aggr_io->first_io;
        while (cio) {
	  pcio = cio;
	  cio= cio->cnext;
	  NASD_Free(pcio, sizeof(_nasd_cheops_io_t));
	  i++;
	}

	nasd_mutex_destroy(&(aggr_io->flag_mutex));	
	nasd_cond_destroy(&(aggr_io->ios_all_done));

        NASD_Free(aggr_io, sizeof(_nasd_cheops_aggr_io_t));

#if _NASD_CHEOPS_AIO_DEBUG > 0
        (cheops_Msg "freed io (size=%d/%d), num_ios=%d\n", in_size, out_size, i);
#endif

}

void 
_nasd_cheops_init_io(NASD_pda_t *pda, _nasd_cheops_aggr_io_t *pio, 
	            _nasd_cheops_io_t *io)
{
        io->ni = pda->data_obj_id.nasd_identifier;
        io->di = pda->data_obj_id.disk_identifier;
	io->pni = pda->parity_obj_id.nasd_identifier;
	io->pdi = pda->parity_obj_id.disk_identifier;
	io->in_offset = pda->start_address;
	io->in_datalen = pda->len;
	io->out_datalen = 0;
	io->bms_targ = 0;
	io->databuf = pda->buf+(pda->databuf_offset);
	io->parent_io = pio;	
	io->lblock = pda->lblock;
	io->qnext = io->qprev = NULL;
}

int 
_nasd_cheops_enq_io(_nasd_cheops_io_t *io)
{
       return ( _nasd_cheops_ioq_enq(gl_io_queue, io) );
}
