/*
 * nasd_cheops_mgr_main.c
 */
/*
 * Main program for Cheops storage manager
 *
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Khalil Amiri, CMU SCS/ECE, July 18 1997
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_cheops_mgr.h>
#include <dce/dce_error.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#ifdef LINUX
#include <getopt.h>
#endif /* LINUX */

extern int   nasd_cheops_mgr_init();
extern int nasd_cheops_mgr_shutdown();

/* Possible debugging */
#define NASD_TEST_DRIVE

/* Nasty global */
handle_t nasd_drive_handle;
extern int Partition_id;


#define REPORT_DCE_ERROR(status) \
{ \
    dce_error_inq_text (status, error_text, &error_status); \
\
    if (error_status != -1) \
    { \
        printf (">>> rpc runtime error: %s\n", error_text); \
    } \
    else \
    { \
        printf (">>> rpc runtime error: no error text translation\n"); \
    } \
}

char        error_text[dce_c_error_string_len];
int         error_status;


void signal_interrupt()
{
  fprintf(stderr,"Received SIGINT, cleaning up ...\n");
  nasd_cheops_mgr_shutdown();
  fprintf(stderr,"Received SIGINT, exiting\n");
  exit(0);
}

void usage(char *prog) {
  printf ("usage: %s [-u <stripe unit size>] [-r <raid level>] [-w <stripe width>]", prog);
  printf(" partition_number\n");
  exit (EXIT_FAILURE);
}

/* options */
int Default_Stripe_Unit_Size;
_NASD_CHEOPS_raid_level_t Default_RAID_Level;
int Default_Stripe_Width;

void cheops_mgr_get_args(int argc, char **argv) {

  int option;
  extern char *optarg;
  extern int optind;
  int raid_level;
  int rflag = 0;
  int test;

  Default_Stripe_Width = 4;
  Default_RAID_Level = RAID0;
  Default_Stripe_Unit_Size = 64*1024;

  while ((option = getopt(argc, argv, "hu:r:w:")) != EOF)
    switch (option) {
    case 'h': /* usage */
      usage(argv[0]);
      break;
    case 'u': /* stripe unit size */
      Default_Stripe_Unit_Size = atoi(optarg)*1024;
      break;
    case 'r': /* default raid level */
      raid_level = atoi(optarg);
      switch (raid_level) {
      case 0: 
	Default_RAID_Level = RAID0;
	fprintf(stderr, "--Cheops Manager Options--\n");
	fprintf(stderr, "Running with RAID0.\n");
	break;
      case 1:
	Default_RAID_Level = NORAID;
	fprintf(stderr, "--Cheops Manager Options--\n");
	fprintf(stderr, "Running with NORAID.\n");
	break;
      case 2:
	Default_RAID_Level = RAID1;
	fprintf(stderr, "--Cheops Manager Options--\n");
	fprintf(stderr, "Running with RAID1.\n");
	break;
      case 3:
	Default_RAID_Level = RAID5;
	fprintf(stderr, "--Cheops Manager Options--\n");
	fprintf(stderr, "Running with RAID5.\n");
	break;
      case 4:
	Default_RAID_Level = LAST_DEFINED_LAYOUT;
	break;
      }
      rflag++;
      break;
    case 'w': /* file size */
      test = atoi(optarg);
      Default_Stripe_Width = test;
      break;
    }

  if ((argc - optind)<1) {
    usage(argv[0]);
  }
  
  Partition_id = atoi(argv[optind]);
  optind++;
 
  if (!rflag) {
    fprintf(stderr, "--Cheops Manager Options--\n");
    fprintf(stderr, "Running with RAID0.\n");
  }
  fprintf(stderr, "Stripe size = %dKB.\n", Default_Stripe_Unit_Size/1024);
  fprintf(stderr, "Stripe width = %d.\n", Default_Stripe_Width);
  if (Default_Stripe_Width==1) {
    fprintf(stderr, "WARNING: Not striping over drives\n");
  }
  fprintf(stderr, "Partition = %d.\n", Partition_id);

  /* Rest of arguments */
  for ( ; optind < argc; optind++) {
    fprintf(stderr,"WARNING: Extra argument ignored %s\n", argv[optind]);
  }
}

int
main (argc, argv)
int     argc;
char    *argv[];

{
  boolean32           valid_protseq;
  unsigned_char_t     *string_binding;
  rpc_binding_vector_p_t binding_vector;
  int                 i;
  unsigned32          status;
  nasd_status_t nasd_status;	
  int rc;
  unsigned_char_t *obj_uuid,*protseq,*network_addr,*endpoint,*network_options;

  signal(SIGINT, (void (*)())signal_interrupt); /* register handler so we call exit() */
	
  cheops_mgr_get_args(argc, argv);
	
  /* initialize */
  rc = nasd_cheops_mgr_init();
  if(rc) {
    fprintf(stderr, "initialization failed\n");
    exit (EXIT_FAILURE);
  }

  /*
   * try to register with the handle in the idl file
   */
  rpc_server_use_protseq_if(
			    (unsigned_char_t *) "ncadg_ip_udp",
			    rpc_c_protseq_max_reqs_default,
			    nasd_cheops_mgr_v0_1_s_ifspec,
			    &status);
	
  if (status != rpc_s_ok) 
    {
      printf ("*** Can't register protocol sequence*\n"); 
      REPORT_DCE_ERROR (status);
      exit (EXIT_FAILURE);
    } else {
      printf ("=== Registered myself successfully\n"); 
    }
	
  /*
   * register the server interface
   */
  rpc_server_register_if (nasd_cheops_mgr_v0_1_s_ifspec, NULL, NULL, &status);
	
  if (status != rpc_s_ok) 
    {
      printf ("*** Can't register interface\n");
      REPORT_DCE_ERROR (status);
      exit (EXIT_FAILURE);
    }

  /*
   * find out what binding information is actually available
   */
  rpc_server_inq_bindings (&binding_vector, &status);
	
  if (status != rpc_s_ok) 
    {
      printf ("*** Can't inquire binding information\n");
      REPORT_DCE_ERROR (status);
      exit (EXIT_FAILURE);
    }
	
  /*
   * walk the binding vector and put the binding information in a file
   */
  for (i = 0; i < binding_vector->count; i++)
    {
      unsigned_char_t *uuid, *protseq, *netaddr, *endpoint, *netopt;
	    
	    
      rpc_binding_to_string_binding (
				     binding_vector->binding_h[i],
				     &string_binding,
				     &status);
	    
      if (status != rpc_s_ok) 
	{
	  printf ("*** Can't get string binding\n");
	  REPORT_DCE_ERROR (status);
	  exit (EXIT_FAILURE);
	}
	    
      rpc_string_binding_parse (string_binding,
				&uuid, &protseq, &netaddr, &endpoint, &netopt, &status);

      if (status != rpc_s_ok) 
	{
	  printf ("*** Can't parse string binding\n");
	  REPORT_DCE_ERROR (status);
	  exit (EXIT_FAILURE);
        }
	    
      printf ("%s %s %s\n",
	      (char *) protseq, (char *) netaddr, (char *) endpoint);
	    
    }
	
  /*
   * wait for incoming client requests
   */
  rpc_server_listen (5, &status);

  if (status != rpc_s_ok)
    {
      printf ("*** Abnormal termination of server listener\n");
      REPORT_DCE_ERROR (status);
      exit (EXIT_FAILURE);
    }

  exit(0);
}




