/*
 * nasd_pdrive_client.c
 *
 * Wrappers for new drive RPCs
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/*
 * if the status field is set in a call to these routines (ie:
 * an RPC failure occurs) then the nasd_status is considered meaningless.
 */

#include <nasd/nasd_options.h>
#include <dce/dce_error.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_nonce_mgmt.h>
#include <nasd/nasd_mem.h>
#include <sys/errno.h>

#define HBG_DEBUG 0
#define HBG_DEBUG_READ 0
#define NASD_MAX_FREE_FO_STATE 16
#define NASD_FO_STATE_INC      2
#define NASD_FREE_FO_STATE_INITIAL 8

#define NASD_MAX_FREE_FO_LIST_STATE 16
#define NASD_FO_LIST_STATE_INC      2
#define NASD_FREE_FO_LIST_STATE_INITIAL 8

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif



#ifdef _KERNEL

#include <dce/ker/pthread_exc.h>

/* this doesnt exist in the kernel DCE header files for whatever reason */
#define exc_get_status(e,s) ( \
   (e)->kind == _exc_kind_status ? \
        (*(s) = (e)->match.value, 0) : \
        -1)

#endif /* _KERNEL */


/*
 * This should really come from a well known object but we hardwire it for now,
 * this is the "default"
 */
#define NASD_OD_BASIC_BLOCKSIZE 8192

#define NASD_CL_P_BUCKET_LEN 8192
static char nasd_cl_p_zerobits[NASD_CL_P_BUCKET_LEN];
static char nasd_cl_p_bitbucket[NASD_CL_P_BUCKET_LEN];
nasd_freelist_t *nasd_cl_fo_state_freelist=NULL;
nasd_freelist_t *nasd_cl_fo_list_state_freelist=NULL;

void
nasd_cl_p_dce_shutdown_fo_state_fl(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_cl_fo_state_freelist,buf,(nasd_cl_fo_state_t *));
}

void
nasd_cl_p_dce_shutdown_fo_state_list_fl(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_cl_fo_list_state_freelist,push_cur,
    (nasd_cl_fo_list_state_t *));
}

nasd_status_t
nasd_cl_p_dce_init(
  nasd_shutdown_list_t  *sl)
{
  nasd_status_t rc;
  int ret;

  /* Initialize the nonce subsystem */
  rc = nasd_init_nonce_mgmt(sl);
  if (rc != NASD_SUCCESS) {
    return (rc);
  }


  NASD_FREELIST_CREATE(nasd_cl_fo_state_freelist, NASD_MAX_FREE_FO_STATE,
                     NASD_FO_STATE_INC, sizeof(nasd_cl_fo_state_t));
  if (nasd_cl_fo_state_freelist == NULL)
    return(NASD_NO_MEM);
  rc = nasd_shutdown_proc(sl, nasd_cl_p_dce_shutdown_fo_state_fl, NULL);
  if (rc) {
    nasd_cl_p_dce_shutdown_fo_state_fl(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_cl_fo_state_freelist, NASD_FREE_FO_STATE_INITIAL,
                          buf, (nasd_cl_fo_state_t *));

  NASD_FREELIST_CREATE(nasd_cl_fo_list_state_freelist, 
                       NASD_MAX_FREE_FO_LIST_STATE,
                       NASD_FO_LIST_STATE_INC,
                       sizeof(nasd_cl_fo_list_state_t));
  if (nasd_cl_fo_list_state_freelist == NULL)
    return(NASD_NO_MEM);
  rc = nasd_shutdown_proc(sl, nasd_cl_p_dce_shutdown_fo_state_list_fl, NULL);
  if (rc) {
    nasd_cl_p_dce_shutdown_fo_state_list_fl(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_cl_fo_list_state_freelist,
                      NASD_FREE_FO_LIST_STATE_INITIAL,
                      push_cur,
                      (nasd_cl_fo_list_state_t *));

  return(NASD_SUCCESS);
}




/* Cases:
   1: All the data should be sent before next digest
   2: Some of the data be sent and then we do the digest
   3: We need to do a digest
   4: Some of the digest has been sent but not all of it
   5:\
   * 
   * NOTE: having data in a digest to send takes precedence over
   *  being "s->done"
   */

#define HAS_INTEGRITY_DATA(_s_) (((_s_)->protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA)
#define HAS_PRIVACY_DATA(_s_) (((_s_)->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA)

void
nasd_cl_p_fr_alloc(state, bsize, buf, bcount)
  rpc_ss_pipe_state_t    state;
  unsigned int           bsize;
  nasd_byte_t          **buf;
  unsigned int          *bcount;
{
  nasd_cl_fo_state_t *s;
  int abytes;

  s = (nasd_cl_fo_state_t *)state;
  *buf = (nasd_byte_t *)(s->buf+s->alloc_off);
  abytes = NASD_MIN(bsize,s->len-s->alloc_off);
  if (abytes <= 0) {
    /*
     * A DCE pipe bug prevents us from being allowed to
     * indicate EOT here by having abytes go to zero.
     */
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
  }
  else {
    *bcount = abytes;
    s->alloc_off += abytes;
  }
}
 
void
nasd_cl_p_fw_alloc(state, bsize, buf, bcount)
  rpc_ss_pipe_state_t    state;
  unsigned int           bsize;
  nasd_byte_t          **buf;
  unsigned int          *bcount;
{
  nasd_cl_fo_state_t *s;
  int abytes;
  int i; 
#if NASD_SECURE_RPCS_ENABLE > 0
  HMAC_SHA1_CTX tmp_context;

  nasd_digest_t tdigest;

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  
  s = (nasd_cl_fo_state_t *)state;
  *buf = (nasd_byte_t *)(s->buf+s->alloc_off);
#if 1
  abytes = NASD_MIN(bsize,s->len-s->alloc_off);
#else
  /* You'll take the bytes whether or not you want them, dammit */
  abytes = s->len-s->alloc_off;
#endif

#if NASD_SECURE_RPCS_ENABLE > 0
  /* First we check if we have any bytes from a digest hanging around
     to be sent. */
  if( (HAS_INTEGRITY_DATA(s)) && ( s->digest_bytes_alloced > -1 )) {
  
    abytes = NASD_MIN(bsize,sizeof(nasd_digest_t)- s->digest_bytes_alloced);
    *bcount=abytes;
    *buf=((nasd_byte_t
*)s->digest[s->curr_alloc_digest])+s->digest_bytes_alloced;
    s->digest_bytes_alloced +=abytes;
    /* Check if this is the last bit of the digest */
    if (s->digest_bytes_alloced == sizeof (nasd_digest_t) ) {
      s->digest_bytes_alloced = -1;
      s->curr_alloc_digest = (s->curr_alloc_digest + 1 )
        % NASD_MAX_OUTSTANDING_DIGESTS;
    }
    
      return;
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  if (abytes <= 0) {
#if NASD_SECURE_RPCS_ENABLE > 0
    if (HAS_INTEGRITY_DATA(s) && !s->done) {
    
      /* We have no data to send BUT we have not finished so that means
         we need to send the final digest */
      
      if (s->digest_state[ s->curr_alloc_digest] != NASD_DIGEST_SLOT_FREE) {
        /* we ran out of slots -- What now? */
        NASD_PANIC();
      }
      s->digest_state[s->curr_alloc_digest] = NASD_DIGEST_SLOT_BUSY;    
      HMAC_SHA1_Final(s->digest[s->curr_alloc_digest],&s->context.HMAC_context);
      abytes=NASD_MIN(bsize,sizeof(nasd_digest_t));
      *bcount=abytes;
      *buf=(nasd_byte_t*)s->digest[s->curr_alloc_digest];
      s->done=1; /* after this digest, we will be done */
      s->digest_bytes_alloced=abytes;
      if (abytes == sizeof(nasd_digest_t)) {
        s->curr_alloc_digest++;
        s->curr_alloc_digest%=NASD_MAX_OUTSTANDING_DIGESTS;    
        s->digest_bytes_alloced=-1; /* all of it was sent */
      }

      return;
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */ 
    /*
     * A DCE pipe bug prevents us from being allowed to
     * indicate EOT here by having abytes go to zero.
     */
#if HBG_DEBUG > 2
      printf("Sending bitbucket\n");
#endif /* HBG_DEBUG > 1 */

    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
  }
  else {
#if HBG_DEBUG > 2
    printf("We actually alloced %d bytes[0x%p]- First 20
bytes\n[",abytes,*buf);
    for(i=0;i<20;i++)
      printf("%x",(*buf)[i]);
    printf("]\n");
#endif /* HBG_DEBUG > 1 */

#if NASD_SECURE_RPCS_ENABLE > 0
    if (HAS_INTEGRITY_DATA(s) && ((s->alloc_off+abytes+ s->start_off) >= s->alloc_next_digest_off)) {
      /* We are sending data past the next digest cutoff
         so we cut the alloc short and set up everything to
         send the digest next time we are called */
      abytes=s->alloc_next_digest_off - (s->alloc_off+s->start_off);
    }

    if (HAS_INTEGRITY_DATA(s) && (abytes > 0 )){
#if (NASD_USE_HMAC_ONLY == 0)
      SHA1_Update(FALSE, &s->context.SHA1_ctx, (unsigned char *)*buf, abytes);
      
      if ((s->alloc_off + abytes) % NASD_OD_BASIC_BLOCKSIZE  == 0) {
        
        SHA1_Final(FALSE, tdigest, &s->context.SHA1_ctx);
        SHA1_Init(FALSE, &s->context.SHA1_ctx);
        HMAC_SHA1_Update( &s->context.HMAC_context, tdigest, sizeof(nasd_digest_t));
       
      }
#else
      HMAC_SHA1_Update( &s->context.HMAC_context, *buf, abytes);

   
#endif /* (NASD_USE_HMAC_ONLY == 0) */
    }

    if (HAS_INTEGRITY_DATA(s) && ((s->alloc_off+abytes+ s->start_off) >= s->alloc_next_digest_off)) {
      /* Next thing we are going to do is send a digest */

      if (s->digest_state[ s->curr_alloc_digest] != NASD_DIGEST_SLOT_FREE) {
        /* we ran out of slots -- What now? */
        printf("PANIC! s->digest_state[%d]=%d\n",
               s->curr_alloc_digest, s->digest_state[s->curr_alloc_digest]);
        NASD_PANIC();
      }
      s->digest_state[s->curr_alloc_digest] = NASD_DIGEST_SLOT_BUSY;
      if (s->alloc_next_digest_off == (s->start_off + s->len ))
        s->done=1;
      s->alloc_next_digest_off+=s->digest_block_size;
      
      bcopy((char *)&(s->context.HMAC_context), (char *)&tmp_context, sizeof(HMAC_SHA1_CTX));
      HMAC_SHA1_Final(s->digest[s->curr_alloc_digest],&tmp_context);
      s->digest_bytes_alloced=0; /* causes the digest to be alloced next */
    }
    


#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  
    *bcount = abytes;
    s->alloc_off += abytes;

  }

}

void
nasd_cl_p_fr_push(state, buf, count)
  rpc_ss_pipe_state_t  *state;
  idl_char             *buf;
  unsigned32            count;
{
  nasd_cl_fo_state_t *s;
  char *b, *t;
  int a;
  
  if (count == 0)
    return;

  t = (char *)buf;

  s = (nasd_cl_fo_state_t *)state;
  b = s->buf + s->push_off;
  a = NASD_MIN(count,s->len-s->push_off);
  s->push_off += a;


  if (t != b) {
    bcopy(t, b, a);
  }
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_context_update_receive(&s->context, (nasd_byte_t*) b , a, s->curr_file_off);
  s->curr_file_off += a;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */  

}

    /* Cases:
       1. We pull some data
       2. We pull some digest bits
       
       
       NOTE: We are using the fact that the digests and the
       data are in seperate buffer so either the pull will
       be strictly for data or strictly for digest bits but
       never shall the two be in one pull call.
       */
void
nasd_cl_p_fw_pull(
  rpc_ss_pipe_state_t  *state,
  idl_char             *buf,
  unsigned32            in_count,
  unsigned32           *out_countp)
{
  nasd_cl_fo_state_t *s;
  char *b, *t; 
  int a;

#if HBG_DEBUG > 0
  printf("pull %d bytes  from  %p \n",in_count,buf);
#endif /* HBG_DEBUG > 0 */



  if (in_count == 0) {
    *out_countp = 0;
    return;
  }

  t = (char *)buf;
  s = (nasd_cl_fo_state_t *)state;
#if NASD_SECURE_RPCS_ENABLE > 0
  
    if (HAS_INTEGRITY_DATA(s) && (s->digest_bytes_pulled > 0)) {

      /* Continuing to pull a digest */
      a=NASD_MIN(in_count,sizeof(nasd_digest_t)-s->digest_bytes_pulled);
      b=(char*)s->digest[s->curr_pull_digest] + s->digest_bytes_pulled;
      s->digest_bytes_pulled+=a;
      s->last_pulled_data=0;
      if (s->digest_bytes_pulled == sizeof(nasd_digest_t)) {
        s->digest_state[s->curr_pull_digest]=NASD_DIGEST_SLOT_FREE;
        s->curr_pull_digest+=1;
        s->curr_pull_digest%=NASD_MAX_OUTSTANDING_DIGESTS;
        s->pull_next_digest_off+=s->digest_block_size;
      }
    } else if (HAS_INTEGRITY_DATA(s) &&
               (((((s->push_off+s->start_off) == s->pull_next_digest_off)
                && (s->last_pulled_data))
               || (s->push_off == s->len)))) {

      if ( (s->push_off == s->len) && (!s->last_pulled_data)) {
        *out_countp=0;


        return;
      }
      /* Starting to pull a digest */
      s->last_pulled_data=0;
      a=NASD_MIN(in_count,sizeof(nasd_digest_t));
      b=(char*)s->digest[s->curr_pull_digest];
      if (a != sizeof(nasd_digest_t)){ 
        s->digest_bytes_pulled=a;
      } else {
        s->digest_state[s->curr_pull_digest]=NASD_DIGEST_SLOT_FREE;
        s->curr_pull_digest+=1;
        s->curr_pull_digest%=NASD_MAX_OUTSTANDING_DIGESTS;
        s->pull_next_digest_off+=s->digest_block_size;
      } 
    } else {
    /* Normal Case */
      s->last_pulled_data=1;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
   
    b = s->buf + s->push_off;
    a = NASD_MIN(in_count,s->len-s->push_off);
    s->push_off += a;
#if NASD_SECURE_RPCS_ENABLE > 0
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  if (t != b) {
    bcopy(t, b, a);
  }

  *out_countp = a;
}

void
nasd_cl_p_free_mem_list(
  nasd_mem_list_t  *src)
{
  int maxoff;

  if (src == NULL) 
    return;

  nasd_cl_p_free_mem_list(src->next);
  maxoff= (src->nelem * src->stride) +  src->len;
  NASD_Free(src->addr, sizeof(nasd_byte_t) * maxoff);
  NASD_Free(src, sizeof(nasd_mem_list_t));
}

nasd_status_t
nasd_cl_p_clone_mem_list(
  nasd_mem_list_t   *src,
  nasd_mem_list_t  **clone)
{
  nasd_status_t rc;
  int maxoff;

  if (src == NULL){
    *clone=NULL;
    return (NASD_SUCCESS);
  }
  NASD_Malloc(*clone, sizeof(nasd_mem_list_t), (nasd_mem_list_t *));
  if (*clone == NULL) return (NASD_NO_MEM);
  (*clone)->len=src->len;
  (*clone)->stride=src->stride;
  (*clone)->nelem=src->nelem;
  NASD_ASSERT(src->addr != NULL);
  
  /* For now, assume stride > 0 */
  NASD_ASSERT(src->stride >= 0 );

  maxoff= (src->nelem * src->stride) +  src->len;
  NASD_Malloc((*clone)->addr, sizeof(nasd_byte_t) * maxoff, (void *));
  if ((*clone)->addr == NULL) {
    NASD_Free(*clone, sizeof(nasd_mem_list_t));
    return (NASD_NO_MEM);
  } else 
    bcopy(src->addr, (*clone)->addr, sizeof(nasd_byte_t) * maxoff);

  rc=nasd_cl_p_clone_mem_list(src->next, &((*clone)->next));
  if (rc != NASD_SUCCESS) {
    NASD_Free((*clone)->addr, sizeof(nasd_byte_t)*maxoff);
    NASD_Free(*clone, sizeof(nasd_mem_list_t));
  } 
#if HBG_DEBUG > 0
  printf("Cloned 0x%p into 0x%p and copied data from 0x%p to 0x%p\n",
         src, *clone, src->addr, (*clone)->addr);
#endif /* HBG_DEBUG > 0 */ 
 return rc;
}

void
nasd_cl_p_check_mem_list(
  nasd_mem_list_t  *memlist,
  nasd_status_t    *status)
{
  if (memlist == NULL)
    goto bad;
  if (memlist->len <= 0)
    goto bad;
  if (memlist->nelem <= 0)
    goto bad;
  if (memlist->addr == NULL)
    goto bad;

  *status = NASD_SUCCESS;
  return;
bad:
  *status = NASD_BAD_MEM_LIST;
}

void
nasd_cl_p_fr_list_alloc(
  rpc_ss_pipe_state_t    state,
  unsigned int           bsize,
  nasd_byte_t          **buf,
  unsigned int          *bcount)
{
  nasd_cl_fo_list_state_t *s;
  nasd_len_t out_len;
  int abytes;


  s = (nasd_cl_fo_list_state_t *)state;
  out_len = *s->out_lenp;
  if (s->badstate) {
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    return;
  }
  if (out_len > s->targetlen) {
    s->badstate = 1;
    s->status = NASD_REMOTE_OVERRUN;
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    return;
  }
  *buf = (nasd_byte_t *)(s->alloc_buf + s->alloc_off);
  abytes = NASD_MIN(bsize,s->alloc_len-s->alloc_off);
  if (abytes <= 0) {
    /*
     * A DCE pipe bug prevents us from being allowed to
     * indicate EOT here by having abytes go to zero.
     */
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
    return;
  }
  *bcount = abytes;
  s->alloc_off += abytes;
  if (s->alloc_off >= s->alloc_len) {

    /* next chunk */
    NASD_ASSERT(s->alloc_off == s->alloc_len);
    s->alloc_elem++;
    if (s->alloc_elem >= s->alloc_cur->nelem) {
      NASD_ASSERT(s->alloc_elem == s->alloc_cur->nelem);
      if ((out_len + abytes) >= s->targetlen) {
        if ((out_len + abytes) > s->targetlen) {
          s->badstate = 1;
          *bcount = 0;
          *buf = NULL;
          return;
        }
        s->alloc_cur = NULL;
        s->alloc_buf = NULL;
        s->alloc_off = 0;
        s->alloc_len = 0;
      }
      else {
        s->alloc_cur = s->push_cur->next;
        s->alloc_elem = 0;
        nasd_cl_p_check_mem_list(s->alloc_cur, &s->status);
        if (s->status) {
          s->badstate = 1;
          *bcount = 0;
          *buf = NULL;
          return;
        }
        s->alloc_buf = s->alloc_cur->addr;
        s->alloc_off = 0;
        s->alloc_len = s->alloc_cur->len;
      }
    }
    else {
      s->alloc_buf = s->alloc_buf + s->alloc_cur->stride;
      s->alloc_off = 0;
      s->alloc_len = s->alloc_cur->len;
    }
  }
}


void
nasd_cl_p_fw_list_alloc(
  rpc_ss_pipe_state_t    state,
  unsigned int           bsize,
  nasd_byte_t          **buf,
  unsigned int          *bcount)
{
  nasd_cl_fo_list_state_t *s;
  nasd_len_t out_len;
  int abytes;
#if NASD_SECURE_RPCS_ENABLE > 0

  HMAC_SHA1_CTX tmp_context;
  nasd_digest_t tdigest;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */


#if HBG_DEBUG > 0
  printf("write list alloc  %d bytes \n",bsize);
#endif /* HBG_DEBUG > 0 */
  s = (nasd_cl_fo_list_state_t *)state;
  out_len = *s->out_lenp;

  if (s->badstate) {
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    return;
  }

  if (out_len > s->targetlen) {
    s->badstate = 1;
    s->status = NASD_REMOTE_OVERRUN;
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    return;
  }
  *buf = (nasd_byte_t *)(s->alloc_buf + s->alloc_off);
  abytes = NASD_MIN(bsize,s->alloc_len-s->alloc_off);

#if NASD_SECURE_RPCS_ENABLE > 0
  /* First we check if we have any bytes from a digest hanging around
     to be sent. */
  if( (HAS_INTEGRITY_DATA(s)) && ( s->digest_bytes_alloced > -1 )) {
    abytes = NASD_MIN(bsize,sizeof(nasd_digest_t)- s->digest_bytes_alloced);
    *bcount=abytes;
    *buf=((nasd_byte_t
*)s->digest[s->curr_alloc_digest])+s->digest_bytes_alloced;
    s->digest_bytes_alloced +=abytes;
    /* Check if this is the last bit of the digest */
    if (s->digest_bytes_alloced == sizeof (nasd_digest_t) ) {
      s->digest_bytes_alloced = -1;
      s->curr_alloc_digest = (s->curr_alloc_digest + 1 ) % NASD_MAX_OUTSTANDING_DIGESTS;
    }
#if HBG_DEBUG > 2
      printf("end write list alloc %d\n",*bcount);
#endif /* HBG_DEBUG > 2 */
      return;
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  if (abytes <= 0) {
#if NASD_SECURE_RPCS_ENABLE > 0
      /* We have no data to send BUT we have not finished so that means
         we need to send the final digest */
    if (HAS_INTEGRITY_DATA(s) && !s->done) {
      if (s->digest_state[ s->curr_alloc_digest] != NASD_DIGEST_SLOT_FREE) {
        /* we ran out of slots -- What now? */
        NASD_PANIC();
      }
      s->digest_state[s->curr_alloc_digest] = NASD_DIGEST_SLOT_BUSY;
      HMAC_SHA1_Final(s->digest[s->curr_alloc_digest], &s->context.HMAC_context);
      abytes=NASD_MIN(bsize,sizeof(nasd_digest_t));
      *bcount=abytes;
    
      *buf=(nasd_byte_t*)s->digest[s->curr_alloc_digest];
      s->done=1; /* after this digest, we will be done */
      s->digest_bytes_alloced=abytes;
      
      if (abytes == sizeof(nasd_digest_t)) {
        s->curr_alloc_digest = (s->curr_alloc_digest + 1 ) % NASD_MAX_OUTSTANDING_DIGESTS;
        s->digest_bytes_alloced=-1; /* all of it was sent */
      }
#if HBG_DEBUG > 2
      
      printf("end write list alloc %d\n",*bcount);
      
#endif /* HBG_DEBUG > 1 */
      return;
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */ 



    /*
     * A DCE pipe bug prevents us from being allowed to
     * indicate EOT here by having abytes go to zero.
     */
    *bcount = NASD_MIN(bsize,NASD_CL_P_BUCKET_LEN);
    *buf = (nasd_byte_t *)nasd_cl_p_bitbucket;
  } else {
#if NASD_SECURE_RPCS_ENABLE > 0
    if (HAS_INTEGRITY_DATA(s) && ((s->total_alloc_off+abytes+ s->start_off) >= s->alloc_next_digest_off)) {
      /* We are sending data past the next digest cutoff
         so we cut the alloc short and set up everything to
         send the digest next time we are called */
      abytes=s->alloc_next_digest_off - (s->total_alloc_off + s->start_off);

    }
   
    if (HAS_INTEGRITY_DATA(s)  && (abytes > 0 )){
#if (NASD_USE_HMAC_ONLY == 0)
      SHA1_Update(FALSE, &s->context.SHA1_ctx, (unsigned char *)*buf, abytes);
      
      if ((s->total_alloc_off + abytes) % NASD_OD_BASIC_BLOCKSIZE == 0  ) {
        SHA1_Final(FALSE, tdigest, &s->context.SHA1_ctx);
        SHA1_Init(FALSE, &s->context.SHA1_ctx);
        HMAC_SHA1_Update(&s->context.HMAC_context, tdigest, sizeof(nasd_digest_t));
        
      }   
#else
      HMAC_SHA1_Update(&s->context.HMAC_context, *buf, abytes);

      
#endif /* (NASD_USE_HMAC_ONLY == 0) */
    }


    if (HAS_INTEGRITY_DATA(s) && ((s->total_alloc_off+abytes+ s->start_off) >= s->alloc_next_digest_off)) {

      if (s->digest_state[ s->curr_alloc_digest] != NASD_DIGEST_SLOT_FREE) {
        /* we ran out of slots -- What now? */
        printf("PANIC! s->digest_state[%d]=%d\n",
               s->curr_alloc_digest, s->digest_state[s->curr_alloc_digest]);
        NASD_PANIC();
      }

      s->digest_state[s->curr_alloc_digest] = NASD_DIGEST_SLOT_BUSY;
      if (s->alloc_next_digest_off == (s->start_off + s->targetlen ))
        s->done=1;
      s->alloc_next_digest_off+=s->digest_block_size;
      
      bcopy((char *)&s->context.HMAC_context, (char *)&tmp_context,
sizeof(HMAC_SHA1_CTX));
      HMAC_SHA1_Final(s->digest[s->curr_alloc_digest],&tmp_context);
      
      s->digest_bytes_alloced=0; /* causes the digest to be alloced next */
    }
    
    s->total_alloc_off += abytes;
 

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */  


  *bcount = abytes;
  s->alloc_off += abytes;
  if (s->alloc_off >= s->alloc_len) {

    /* next chunk */
    NASD_ASSERT(s->alloc_off == s->alloc_len);
    s->alloc_elem++;
    if (s->alloc_elem >= s->alloc_cur->nelem) {
      NASD_ASSERT(s->alloc_elem == s->alloc_cur->nelem);
      if ((out_len + abytes) >= s->targetlen) {
        if ((out_len + abytes) > s->targetlen) {
          s->badstate = 1;
          *bcount = 0;
          *buf = NULL;
          return;
        }
        s->alloc_cur = NULL;
        s->alloc_buf = NULL;
        s->alloc_off = 0;
        s->alloc_len = 0;
      }
      else {
        s->alloc_cur = s->push_cur->next;
        s->alloc_elem = 0;
        nasd_cl_p_check_mem_list(s->alloc_cur, &s->status);
        if (s->status) {
          s->badstate = 1;
          *bcount = 0;
          *buf = NULL;
          return;
        }
        s->alloc_buf = s->alloc_cur->addr;
        s->alloc_off = 0;
        s->alloc_len = s->alloc_cur->len;
      }
    }
    else {
      s->alloc_buf = s->alloc_buf + s->alloc_cur->stride;
      s->alloc_off = 0;
      s->alloc_len = s->alloc_cur->len;
    }
  }
  }

#if HBG_DEBUG > 2
      printf("end write list *buf=%p *bcount=%d bytes\n",*buf,*bcount);
#endif /* HBG_DEBUG > 2 */
}

void
nasd_cl_p_fr_list_push(
  rpc_ss_pipe_state_t  *state,
  idl_char             *buf,
  unsigned32            count)
{
  nasd_cl_fo_list_state_t *s;
  nasd_len_t out_len;
  char *b, *t;
  int a;

  if (count == 0)
    return;
  s = (nasd_cl_fo_list_state_t *)state;
  out_len = *s->out_lenp;
  if (s->badstate)
    return;
  if (out_len >= s->targetlen) {
    s->badstate = 1;
    s->status = NASD_REMOTE_OVERRUN;
    return;
  }
  t = (char *)buf;
  b = (char *)(s->push_buf + s->push_off);
  a = NASD_MIN(count,s->push_len-s->push_off);
  if (t != b) {
    bcopy(t, b, a);
  }
 
  s->push_off += a;
  out_len += a;
  *s->out_lenp = out_len;

#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_security_context_update_receive(&s->context, (nasd_byte_t*) b, a, s->curr_file_off);
  s->curr_file_off += a;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */  
  if (s->push_off >= s->push_len) {
    /* next chunk */
    NASD_ASSERT(s->push_off == s->push_len);
    s->push_elem++;
    if (s->push_elem >= s->push_cur->nelem) {
      NASD_ASSERT(s->push_elem == s->push_cur->nelem);
      if ((out_len) >= s->targetlen) {    
        if ((out_len) > s->targetlen) {   
          s->badstate = 1;
          return;
        }
        s->push_cur = NULL;
        s->push_buf = NULL;
        s->push_off = 0;
        s->push_len = 0;
      }
      else {
        s->push_cur = s->push_cur->next;
        s->push_elem = 0;
        nasd_cl_p_check_mem_list(s->push_cur, &s->status);
        if (s->status) {
          s->badstate = 1;
          return;
        }
        s->push_buf = s->push_cur->addr;
        s->push_off = 0;
        s->push_len = s->push_cur->len;
      }
    }
    else {
      s->push_buf = s->push_buf + s->push_cur->stride;
      s->push_off = 0;
      s->push_len = s->push_cur->len;
    }
    
  }
}

void
nasd_cl_p_fw_list_pull(
  rpc_ss_pipe_state_t  *state,
  idl_char             *buf,
  unsigned32            in_count,
  unsigned32           *out_countp)
{
  nasd_cl_fo_list_state_t *s;
  nasd_len_t out_len;
  char *b, *t;
  int a;
#if HBG_DEBUG > 0
  printf("list pull  %d bytes from %p  \n",in_count, buf );
#endif /* HBG_DEBUG > 0 */

  if (in_count == 0) {
    *out_countp = 0;
    return;
  }

  s = (nasd_cl_fo_list_state_t *)state;
  out_len = *s->out_lenp;

  if (s->badstate) {
    return;
  }


  t = (char *)buf;
  b = (char *)(s->push_buf + s->push_off);
#if NASD_SECURE_RPCS_ENABLE > 0
    if (HAS_INTEGRITY_DATA(s) && (s->digest_bytes_pulled > 0)) {

      /* Continuing to pull a digest */
      a=NASD_MIN(in_count,sizeof(nasd_digest_t)-s->digest_bytes_pulled);
      b=(char*)s->digest[s->curr_pull_digest] + s->digest_bytes_pulled;
      s->digest_bytes_pulled += a;
      s->last_pulled_data=0;
      if (s->digest_bytes_pulled == sizeof(nasd_digest_t)) {
        s->digest_state[s->curr_pull_digest]=NASD_DIGEST_SLOT_FREE;
        s->curr_pull_digest= (s->curr_pull_digest + 1 ) %
NASD_MAX_OUTSTANDING_DIGESTS;
        s->pull_next_digest_off+=s->digest_block_size;
      }
    } else if (HAS_INTEGRITY_DATA(s) &&
               (((((s->total_push_off + s->start_off) == s->pull_next_digest_off)
                && (s->last_pulled_data))
               || (s->total_push_off == s->targetlen)))) {

      if ( (s->total_push_off == s->targetlen) && (!s->last_pulled_data)) {
        *out_countp=0;
#if HBG_DEBUG > 1
        printf("end write list pull \n");
#endif /* HBG_DEBUG > 0 */
        return;
      }
      /* Starting to pull a digest */
      s->last_pulled_data=0;
      a=NASD_MIN(in_count,sizeof(nasd_digest_t));
      b=(char*)s->digest[s->curr_pull_digest];
      if (a != sizeof(nasd_digest_t)){ 
        s->digest_bytes_pulled=a;
      } else {
        s->digest_state[s->curr_pull_digest]=NASD_DIGEST_SLOT_FREE;
        s->curr_pull_digest= (s->curr_pull_digest + 1 ) % NASD_MAX_OUTSTANDING_DIGESTS;
        s->pull_next_digest_off+=s->digest_block_size;
      } 
    } else {
    /* Normal Case */
      s->last_pulled_data=1;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */


      if (out_len >= s->targetlen) {
        /* Is this legal? */
        *out_countp = 0;
        return;
      }

      
  a = NASD_MIN(in_count,s->push_len-s->push_off);
  s->push_off += a;
#if NASD_SECURE_RPCS_ENABLE > 0
  s->total_push_off += a;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */  
  out_len += a;
  *s->out_lenp = out_len;

  if (s->push_off >= s->push_len) {

    /* next chunk */
    NASD_ASSERT(s->push_off == s->push_len);
    s->push_elem++;
    if (s->push_elem >= s->push_cur->nelem) {
      NASD_ASSERT(s->push_elem == s->push_cur->nelem);
      if ((out_len) >= s->targetlen) {    
        if ((out_len) > s->targetlen) {   
          s->badstate = 1;
          *out_countp = 0;
          return;
        }
        s->push_cur = NULL;
        s->push_buf = NULL;
        s->push_off = 0;
        s->push_len = 0;
      }
      else {
        s->push_cur = s->push_cur->next;
        s->push_elem = 0;
        nasd_cl_p_check_mem_list(s->push_cur, &s->status);
        if (s->status) {
          s->badstate = 1;
          *out_countp = 0;
          return;
        }
        s->push_buf = s->push_cur->addr;
        s->push_off = 0;
        s->push_len = s->push_cur->len;
      }
    }
    else {
      s->push_buf = s->push_buf + s->push_cur->stride;
      s->push_off = 0;
      s->push_len = s->push_cur->len;
    }
  }
 
#if NASD_SECURE_RPCS_ENABLE > 0
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */


  if (t != b) {
    bcopy(t, b, a);
  }
  *out_countp = a;

  
#if HBG_DEBUG > 1
  printf("end write list pull \n");
#endif /* HBG_DEBUG > 0 */


}

/*
 * BEGIN RPC WRAPPERS
 *
 * These functions depend more than we would like on the sizes of fields
 * and the layout. When you add an API, pay attention to how your args
 * and res structure are layed out. -hgobioff
 */

void
nasd_cl_prpc_null_dr(
  rpc_binding_handle_t   handle,
  nasd_status_t         *nasd_status,
  error_status_t        *status)
{
  TRY {
    nasd_p_null_dr(handle, nasd_status, status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)status);
  } ENDTRY
}

void
nasd_cl_prpc_sync_dr(
  rpc_binding_handle_t   handle,
  nasd_status_t         *nasd_status,
  error_status_t        *status)
{
  TRY {
    nasd_p_sync_dr(handle, nasd_status, status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)status);
  } ENDTRY
}

void
nasd_cl_prpc_part_creat_dr(
  rpc_binding_handle_t          handle,
  nasd_key_t                    in_key,
  nasd_p_part_creat_dr_args_t  *args,
  nasd_p_part_creat_dr_res_t   *res,
  error_status_t               *status)
{
  nasd_p_part_creat_dr_args_otw_t args_otw;
  nasd_p_part_creat_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_part_creat_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_part_creat_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;

  nasd_p_part_creat_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t
*)&(((nasd_p_part_creat_dr_args_t*)args_otw)->in_digest), NULL);

  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_part_creat_otw_dr(handle, args_otw, res_otw, status);



      nasd_p_part_creat_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res,
      (nasd_byte_t*)&(((nasd_p_part_creat_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_create_dr(
  rpc_binding_handle_t      handle,
  nasd_key_t                in_key,
  nasd_p_create_dr_args_t  *args,
  nasd_p_create_dr_res_t   *res,
  error_status_t           *status)
{
  nasd_p_create_dr_args_otw_t args_otw;
  nasd_p_create_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_create_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_create_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;


  nasd_p_create_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_create_dr_args_t*)args_otw)->in_digest),NULL);

  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_create_otw_dr(handle, args_otw, res_otw, status);

      nasd_p_create_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key, (nasd_byte_t*)res_otw,
      res_len, (nasd_status_t*)res, 
      (nasd_byte_t*)&(((nasd_p_create_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_create_dr__otw_provided(
  rpc_binding_handle_t      handle,
  nasd_key_t                in_key,
  nasd_p_create_dr_args_t  *args,
  nasd_p_create_dr_res_t   *res,
  nasd_byte_t              *args_otw,
  nasd_byte_t              *res_otw,
  error_status_t           *status)
{
  nasd_args_len_t args_len = sizeof(nasd_p_create_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_create_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;

  nasd_p_create_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_create_dr_args_t*)args_otw)->in_digest),NULL);

  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_create_otw_dr(handle, args_otw, res_otw, status);

      nasd_p_create_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key, (nasd_byte_t*)res_otw,
      res_len, (nasd_status_t*)res, 
      (nasd_byte_t*)&(((nasd_p_create_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_getattr_dr(
  rpc_binding_handle_t       handle,
  nasd_key_t                 in_key,
  nasd_p_getattr_dr_args_t  *args,
  nasd_p_getattr_dr_res_t   *res,
  error_status_t            *status)
{
  nasd_p_getattr_dr_args_otw_t args_otw;
  nasd_p_getattr_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_getattr_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_getattr_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;

  nasd_p_getattr_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0
  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_getattr_dr_args_t*)args_otw)->in_digest), NULL);

  SETUP_IV(nasd_p_getattr_dr_args_t, args_otw, icv);


  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_getattr_otw_dr(handle, args_otw, res_otw, status);


      nasd_p_getattr_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res,
      (nasd_byte_t*)&(((nasd_p_getattr_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_write_simple_dr(
  rpc_binding_handle_t        handle,
  nasd_key_t                  in_key,
  nasd_p_smpl_op_dr_args_t   *args,
  void                       *buf,
  nasd_p_fastwrite_dr_res_t  *res,
  error_status_t             *status)
{
  nasd_p_smpl_op_dr_args_otw_t args_otw;
  nasd_p_fastwrite_dr_res_otw_t res_otw;
  nasd_cl_fo_state_t *statep; 
  nasd_byte_pipe_t pipe;
  nasd_len_t out_len;
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_status_t digested;
  int cnt;
  nasd_args_len_t args_len = sizeof(nasd_p_smpl_op_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_fastwrite_dr_res_otw_t);
  void *mybuff;
  nasd_iv_t icv;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

 
  NASD_FREELIST_GET(nasd_cl_fo_state_freelist, statep, buf,
                    (nasd_cl_fo_state_t *));


  statep->buf = buf;
  statep->len = args->in_len;
  statep->alloc_off = statep->push_off = 0;

  /*
    state.buf = buf;
    state.len = args->in_len;
    state.alloc_off = state.push_off = 0;
    */
  
  nasd_p_smpl_op_dr_marshall_args(args, args_otw);


#if NASD_SECURE_RPCS_ENABLE > 0
  statep->protection=args->in_security_param.actual_protection;
  
  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
                                       (nasd_byte_t*)args_otw, args_len, 
                                       (nasd_byte_t
*)&(((nasd_p_smpl_op_dr_args_t*)args_otw)->in_digest), 
                                       &statep->context);
  


  if (digested != NASD_SUCCESS) {
    res->nasd_status = digested;
  } else {

    if ((statep->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      /* XXX Quick fix now -- better solution later */

      NASD_Malloc(mybuff, args->in_len, (void *));
      bcopy(buf, mybuff, args->in_len);
      statep->buf=mybuff;
    }
    
    if ((statep->protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) {

      statep->count=0;
      statep->start_off=args->in_offset;
      statep->digest_block_size = NASD_DIGEST_BLOCK_SIZE;
      statep->last_pulled_data=0;
      statep->alloc_next_digest_off= args->in_offset / statep->digest_block_size;
      statep->alloc_next_digest_off=(1 +statep->alloc_next_digest_off)*statep->digest_block_size;
      statep->pull_next_digest_off =statep->alloc_next_digest_off;
      statep->curr_alloc_digest=0;
      statep->curr_pull_digest=0;
      statep->done=0;
      for(cnt=0;cnt<NASD_MAX_OUTSTANDING_DIGESTS; cnt++ ) 
        statep->digest_state[cnt]=NASD_DIGEST_SLOT_FREE;
      statep->digest_bytes_alloced=-1;
      statep->digest_bytes_pulled=-1;
      SHA1_Init(FALSE, &statep->context.SHA1_ctx);
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    
  pipe.alloc = (void (*)())nasd_cl_p_fw_alloc;
  pipe.pull = (void (*)())nasd_cl_p_fw_pull;
  pipe.state = (rpc_ss_pipe_state_t)statep;
  

  TRY {

    nasd_p_write_simple_otw_dr(handle, args_otw, pipe, res_otw, status);

#if NASD_SECURE_RPCS_ENABLE > 0
    if ((statep->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {

      NASD_Free(mybuff, args->in_len);

    }
    

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    nasd_p_fastwrite_dr_unmarshall_res(res_otw, res);
    
  }
  CATCH_ALL {
    res->nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)status);
  } ENDTRY


#if NASD_SECURE_RPCS_ENABLE > 0
      }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res, 
      (nasd_byte_t*)&(((nasd_p_fastwrite_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
 
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
 NASD_FREELIST_FREE(nasd_cl_fo_state_freelist, statep, buf);


}

void
nasd_cl_prpc_read_simple_dr(
  rpc_binding_handle_t       handle,
  nasd_key_t                 in_key,
  int                        is_read2,
  nasd_p_smpl_op_dr_args_t  *args,
  void                      *buf,
  nasd_p_fastread_dr_res_t  *res,
  error_status_t            *status)
{
  nasd_p_smpl_op_dr_args_otw_t args_otw;
  nasd_p_fastread_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_smpl_op_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_fastread_dr_res_otw_t);
  nasd_status_t digested;
  nasd_cl_fo_state_t *statep;
  nasd_byte_pipe_t pipe;
  nasd_len_t out_len;
  nasd_iv_t icv;
 
  NASD_FREELIST_GET(nasd_cl_fo_state_freelist, statep, buf,
                    (nasd_cl_fo_state_t *)); 
  statep->buf = buf;
  statep->len = args->in_len;
  statep->alloc_off = statep->push_off = 0;
#if NASD_SECURE_RPCS_ENABLE > 0
  statep->curr_file_off = args->in_offset;
#endif 
 

  pipe.alloc = (void (*)())nasd_cl_p_fr_alloc;
  pipe.push = (void (*)())nasd_cl_p_fr_push;
  pipe.state = (rpc_ss_pipe_state_t)statep;

  nasd_p_smpl_op_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0
  bzero(&statep->context, sizeof(nasd_security_context_t));
  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_smpl_op_dr_args_t*)args_otw)->in_digest), NULL);
 

  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }  else {
    SETUP_REPLY_IV(icv);

    nasd_security_reply_init((nasd_args_t *)args, in_key, icv,
                             &statep->context);
    
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  TRY {
    if (is_read2) {
      nasd_p_read2_simple_otw_dr(handle, args_otw, pipe, res_otw, status);
    }
    else {
      nasd_p_read_simple_otw_dr(handle, args_otw, pipe, res_otw, status);
    }

      nasd_p_fastread_dr_unmarshall_res(res_otw, res);
 
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res,    
      (nasd_byte_t*)&(((nasd_p_fastread_dr_res_t*)res_otw)->out_digest),
      &statep->context);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
   NASD_FREELIST_FREE(nasd_cl_fo_state_freelist, statep, buf);
}

void
nasd_cl_prpc_tread_simple_dr(
  rpc_binding_handle_t        handle,
  nasd_key_t                  in_key,
  nasd_p_thrtl_op_dr_args_t  *args,
  void                       *buf,
  nasd_p_fastread_dr_res_t   *res,
  error_status_t             *status)
{
  nasd_p_thrtl_op_dr_args_otw_t args_otw;
  nasd_p_fastread_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_thrtl_op_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_fastread_dr_res_otw_t);
  nasd_status_t digested;
  nasd_cl_fo_state_t *statep;
  nasd_byte_pipe_t pipe;
  nasd_len_t out_len;
  nasd_iv_t icv;

  NASD_FREELIST_GET(nasd_cl_fo_state_freelist, statep, buf,
                    (nasd_cl_fo_state_t *));
  statep->buf = buf;
  statep->len = args->in_len;
  statep->alloc_off = statep->push_off = 0;
#if NASD_SECURE_RPCS_ENABLE > 0
  statep->curr_file_off = args->in_offset;
#endif

  pipe.alloc = (void (*)())nasd_cl_p_fr_alloc;
  pipe.push = (void (*)())nasd_cl_p_fr_push;
  pipe.state = (rpc_ss_pipe_state_t)statep;

  nasd_p_thrtl_op_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_thrtl_op_dr_args_t*)args_otw)->in_digest), NULL);


  if (digested != NASD_SUCCESS) {
    res->nasd_status = digested;
  }
  else {
    SETUP_REPLY_IV(icv);
    bzero(&statep->context, sizeof(nasd_security_context_t));
    nasd_security_reply_init((nasd_args_t *)args, in_key, icv,
                             &statep->context);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_tread_simple_otw_dr(handle, args_otw, pipe, res_otw, status);


      nasd_p_fastread_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res, 
      (nasd_byte_t*)&(((nasd_p_fastread_dr_res_t*)res_otw)->out_digest),
      &statep->context);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  NASD_FREELIST_FREE(nasd_cl_fo_state_freelist, statep, buf);
}

void
nasd_cl_prpc_range_read(
  rpc_binding_handle_t     h,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_mem_list_t         *elems,
  nasd_offset_t            in_offset,
  int                      in_len,
  nasd_len_t              *out_lenp,
  nasd_status_t           *nasd_status,
  error_status_t          *status)
{
  nasd_p_smpl_op_dr_args_otw_t args_otw;
  nasd_p_fastread_dr_res_otw_t res_otw;
  nasd_p_smpl_op_dr_args_t args;
  nasd_cl_fo_list_state_t *statep;
  nasd_p_fastread_dr_res_t res;
  nasd_byte_pipe_t pipe;
  nasd_status_t rc;
  nasd_args_len_t args_len = sizeof(nasd_p_smpl_op_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_fastread_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;
  NASD_FREELIST_GET(nasd_cl_fo_list_state_freelist, statep, push_cur,
                    (nasd_cl_fo_list_state_t *));

  *status = 0;
  *out_lenp = 0;

  args.in_capability_args = cap_args;
  args.in_partnum = partnum;
  args.in_identifier = nasdid;
  args.in_offset = in_offset;
  args.in_len = in_len;
  args.in_security_param.type = cap_args.type;
  args.in_security_param.partnum = partnum;
  args.in_security_param.actual_protection = cap_args.min_protection;

  nasd_cl_p_check_mem_list(elems, &rc);
  if (rc) {
    *nasd_status = rc;
    return;
  }
  statep->out_lenp = out_lenp;
  statep->targetlen = in_len;
  statep->badstate = 0;
  statep->alloc_cur = statep->push_cur = elems;
  statep->alloc_elem = statep->push_elem = 0;
  statep->alloc_buf = statep->push_buf = elems->addr;
  statep->alloc_off = statep->push_off = 0;
  statep->alloc_len = statep->push_len = elems->len;
  statep->status = NASD_SUCCESS;
#if NASD_SECURE_RPCS_ENABLE > 0
  statep->curr_file_off= in_offset;
#endif
  pipe.alloc = (void (*)())nasd_cl_p_fr_list_alloc;
  pipe.push = (void (*)())nasd_cl_p_fr_list_push;
  pipe.state = (rpc_ss_pipe_state_t)statep;

  nasd_p_smpl_op_dr_marshall_args(&args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)&args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_smpl_op_dr_args_t*)args_otw)->in_digest), NULL);
  

  if (digested != NASD_SUCCESS) {
    res.nasd_status=digested;
  }
  else {
    bzero(&statep->context, sizeof(nasd_security_context_t));
    SETUP_REPLY_IV(icv);   
    nasd_security_reply_init((nasd_args_t *)&args, in_key, icv,
                             &statep->context);
    
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_read_simple_otw_dr(h, args_otw, pipe, res_otw, status);

      nasd_p_fastread_dr_unmarshall_res(res_otw, &res);
    }
    CATCH_ALL {
      res.nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)&args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)&res,
      (nasd_byte_t*)&(((nasd_p_fastread_dr_res_t*)res_otw)->out_digest),
      &statep->context);
  }
  /* Which nasd_status should take precendence - HG */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  if (statep->status || statep->badstate) {
    *nasd_status = NASD_MEM_LIST_ERR;
  }
  else {
    *nasd_status = res.nasd_status;
  }
  NASD_FREELIST_FREE(nasd_cl_fo_list_state_freelist, statep, push_cur);

}

void
nasd_cl_prpc_setattr_dr(
  rpc_binding_handle_t       handle,
  nasd_key_t                 in_key,
  nasd_p_setattr_dr_args_t  *args,
  nasd_p_setattr_dr_res_t   *res,
  error_status_t            *status)
{
  nasd_p_setattr_dr_args_otw_t args_otw;
  nasd_p_setattr_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_setattr_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_setattr_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;

  nasd_p_setattr_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_setattr_dr_args_t*)args_otw)->in_digest), NULL);



  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_setattr_otw_dr(handle, args_otw, res_otw, status);


      nasd_p_setattr_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key, (nasd_byte_t*)res_otw,
    res_len, (nasd_status_t*)res, 
    (nasd_byte_t*)&(((nasd_p_setattr_dr_res_t*)res_otw)->out_digest),
    NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_setattr_dr__otw_provided(
  rpc_binding_handle_t       handle,
  nasd_key_t                 in_key,
  nasd_p_setattr_dr_args_t  *args,
  nasd_p_setattr_dr_res_t   *res,
  nasd_byte_t               *args_otw,
  nasd_byte_t               *res_otw,
  error_status_t            *status)
{
  nasd_args_len_t args_len = sizeof(nasd_p_setattr_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_setattr_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;

  nasd_p_setattr_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_setattr_dr_args_t*)args_otw)->in_digest), NULL);



  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_setattr_otw_dr(handle, args_otw, res_otw, status);


      nasd_p_setattr_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key, (nasd_byte_t*)res_otw,
    res_len, (nasd_status_t*)res, 
    (nasd_byte_t*)&(((nasd_p_setattr_dr_res_t*)res_otw)->out_digest),
    NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_flush_obj_dr(
  rpc_binding_handle_t         handle,
  nasd_key_t                   in_key,
  nasd_p_flush_obj_dr_args_t  *args,
  nasd_p_flush_obj_dr_res_t   *res,
  error_status_t              *status)
{
  nasd_p_flush_obj_dr_args_otw_t args_otw;
  nasd_p_flush_obj_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_flush_obj_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_flush_obj_dr_res_otw_t);
  nasd_status_t digested;
  nasd_status_t rc;
 nasd_iv_t icv;
  nasd_p_flush_obj_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t*)&(((nasd_p_flush_obj_dr_args_t*)args_otw)->in_digest), NULL);


  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_flush_obj_otw_dr(handle, args_otw,res_otw, status);
      nasd_p_flush_obj_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res, 
      (nasd_byte_t*)&(((nasd_p_flush_obj_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_eject_obj_dr(
  rpc_binding_handle_t         handle,
  nasd_key_t                   in_key,
  nasd_p_eject_obj_dr_args_t  *args,
  nasd_p_eject_obj_dr_res_t   *res,
  error_status_t              *status)
{
  nasd_p_eject_obj_dr_args_otw_t args_otw;
  nasd_p_eject_obj_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_eject_obj_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_eject_obj_dr_res_otw_t);
  nasd_status_t digested;
  nasd_status_t rc;
 nasd_iv_t icv;
  nasd_p_eject_obj_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t*)&(((nasd_p_eject_obj_dr_args_t*)args_otw)->in_digest), NULL);

  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_eject_obj_otw_dr(handle, args_otw,res_otw, status);
      nasd_p_eject_obj_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res,
      (nasd_byte_t*)&(((nasd_p_eject_obj_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_remove_dr(
  rpc_binding_handle_t      handle,
  nasd_key_t                in_key,
  nasd_p_remove_dr_args_t  *args,
  nasd_p_remove_dr_res_t   *res,
  error_status_t           *status)
{
  nasd_p_remove_dr_args_otw_t args_otw;
  nasd_p_remove_dr_res_otw_t res_otw;
  nasd_args_len_t args_len = sizeof(nasd_p_remove_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_remove_dr_res_otw_t);
  nasd_status_t digested;
  nasd_status_t rc;
 nasd_iv_t icv;
  nasd_p_remove_dr_marshall_args(args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_remove_dr_args_t*)args_otw)->in_digest), NULL);


  if (digested != NASD_SUCCESS) {
    res->nasd_status=digested;
  }
  else {
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_remove_otw_dr(handle, args_otw, res_otw, status);

      nasd_p_remove_dr_unmarshall_res(res_otw, res);
    }
    CATCH_ALL {
      res->nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
  }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)res, 
      (nasd_byte_t*)&(((nasd_p_remove_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}

void
nasd_cl_prpc_initialize_dr(
  rpc_binding_handle_t          handle,
  nasd_p_initialize_dr_args_t  *args,
  nasd_p_initialize_dr_res_t   *res,
  error_status_t               *status)
{
  nasd_p_initialize_dr_args_otw_t args_otw;
  nasd_p_initialize_dr_res_otw_t res_otw;
  nasd_status_t rc;

  nasd_p_initialize_dr_marshall_args(args, args_otw);
  TRY {
    nasd_p_initialize_otw_dr(handle, args_otw, res_otw, status);
    nasd_p_initialize_dr_unmarshall_res(res_otw, res);
  }
  CATCH_ALL {
    res->nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)status);
  } ENDTRY
}

void
nasd_cl_prpc_range_write(
  rpc_binding_handle_t     h,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_mem_list_t         *memlist,
  nasd_offset_t            in_offset,
  int                      in_len,
  nasd_len_t              *out_lenp,
  nasd_status_t           *nasd_status,
  error_status_t          *status)
{
  nasd_p_smpl_op_dr_args_otw_t args_otw;
  nasd_p_fastwrite_dr_res_otw_t res_otw;
  nasd_p_smpl_op_dr_args_t args;
  nasd_cl_fo_list_state_t *statep;
  nasd_p_fastwrite_dr_res_t res;
  nasd_byte_pipe_t pipe;
  nasd_status_t rc;
#if NASD_SECURE_RPCS_ENABLE > 0
  nasd_mem_list_t         *tmp_memlist;
  nasd_status_t digested;
  int cnt;
  nasd_args_len_t args_len = sizeof(nasd_p_smpl_op_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_fastwrite_dr_res_otw_t);
  void *mybuff;
  nasd_iv_t icv;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  NASD_FREELIST_GET(nasd_cl_fo_list_state_freelist, statep, push_cur,
                    (nasd_cl_fo_list_state_t *));
  *status = 0;
  *out_lenp = 0;

  args.in_partnum = partnum;
  args.in_identifier = nasdid;
  args.in_offset = in_offset;
  args.in_len = in_len;
  args.in_security_param.type = cap_args.type;
  args.in_security_param.partnum = partnum;
  args.in_security_param.actual_protection = cap_args.min_protection;
  args.in_capability_args = cap_args;

  statep->out_lenp = out_lenp;
  statep->targetlen = in_len;
  statep->badstate = 0;
  statep->alloc_cur = statep->push_cur = memlist;
  statep->alloc_elem = statep->push_elem = 0;
  statep->alloc_buf = statep->push_buf = memlist->addr;
  statep->alloc_off = statep->push_off = 0;
#if NASD_SECURE_RPCS_ENABLE > 0
  statep->total_alloc_off = statep->total_push_off = 0;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  statep->alloc_len = statep->push_len = memlist->len;
  statep->status = NASD_SUCCESS;

  nasd_p_smpl_op_dr_marshall_args(&args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0
  digested = nasd_security_digest_args((nasd_args_t*)&args, in_key,
                                       (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_smpl_op_dr_args_t*)args_otw)->in_digest),
    &statep->context);


  if (digested != NASD_SUCCESS) {
    res.nasd_status = digested;
  } else {
    statep->protection=args.in_security_param.actual_protection;
    /* XXX Quick fix now -- better solution later */
    if ((statep->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      nasd_cl_p_clone_mem_list(memlist, &tmp_memlist);
      statep->alloc_cur = statep->push_cur  = tmp_memlist;
      statep->alloc_buf = statep->push_buf  = tmp_memlist->addr;
    }

    if ((statep->protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) {
      statep->count=0;
      statep->start_off=args.in_offset;
      statep->digest_block_size = NASD_DIGEST_BLOCK_SIZE;
      statep->last_pulled_data=0;
      statep->alloc_next_digest_off= args.in_offset / statep->digest_block_size;
      statep->alloc_next_digest_off=(1 +statep->alloc_next_digest_off)*statep->digest_block_size;
      statep->pull_next_digest_off =statep->alloc_next_digest_off;
      statep->curr_alloc_digest=0;
      statep->curr_pull_digest=0;
      statep->done=0;
      for(cnt=0;cnt<NASD_MAX_OUTSTANDING_DIGESTS; cnt++ ) 
        statep->digest_state[cnt]=NASD_DIGEST_SLOT_FREE;
      statep->digest_bytes_alloced=-1;
      statep->digest_bytes_pulled=-1;
    }
   SHA1_Init(FALSE, &statep->context.SHA1_ctx);
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  
  nasd_cl_p_check_mem_list(statep->alloc_cur, &rc);
  if (rc) {
    *nasd_status = rc;
    return;
  }
  
  pipe.alloc = (void (*)())nasd_cl_p_fw_list_alloc;
  pipe.pull = (void (*)())nasd_cl_p_fw_list_pull;
  pipe.state = (rpc_ss_pipe_state_t)statep;


  TRY {
    nasd_p_write_simple_otw_dr(h, args_otw, pipe, res_otw, status);
#if NASD_SECURE_RPCS_ENABLE > 0

    if ((statep->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
      nasd_cl_p_free_mem_list(tmp_memlist);
    }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    nasd_p_fastwrite_dr_unmarshall_res(res_otw, &res);
  }
  CATCH_ALL {
    res.nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)status);
  } ENDTRY

#if NASD_SECURE_RPCS_ENABLE > 0
      }


  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)&args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)&res,
      (nasd_byte_t*)&(((nasd_p_fastwrite_dr_res_t*)res_otw)->out_digest),
      NULL);
  }
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */


  if (statep->status || statep->badstate) {
    *nasd_status = NASD_MEM_LIST_ERR;
  }
  else {
    *nasd_status = res.nasd_status;
  }
  NASD_FREELIST_FREE(nasd_cl_fo_list_state_freelist, statep, push_cur);
}

void nasd_cl_prpc_range_tread(
  rpc_binding_handle_t     h,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_uint64              in_bms_targ,
  nasd_mem_list_t         *memlist,
  nasd_offset_t            in_offset,
  int                      in_len,
  nasd_len_t              *out_lenp,
  nasd_status_t           *nasd_status,
  error_status_t          *status)
{
  nasd_p_thrtl_op_dr_args_otw_t args_otw;
  nasd_p_fastread_dr_res_otw_t res_otw;
  nasd_p_thrtl_op_dr_args_t args;
  nasd_cl_fo_list_state_t *statep;
  nasd_p_fastread_dr_res_t res;
  nasd_byte_pipe_t pipe;
  nasd_status_t rc;
  nasd_args_len_t args_len = sizeof(nasd_p_thrtl_op_dr_args_otw_t);
  nasd_args_len_t res_len = sizeof(nasd_p_fastread_dr_res_otw_t);
  nasd_status_t digested;
  nasd_iv_t icv;
  *status = 0;
  *out_lenp = 0;

  NASD_FREELIST_GET(nasd_cl_fo_list_state_freelist, statep, push_cur,
                    (nasd_cl_fo_list_state_t *));
  args.in_partnum = partnum;
  args.in_identifier = nasdid;
  args.in_offset = in_offset;
  args.in_bms_targ = in_bms_targ;
  args.in_len = in_len;
  args.in_security_param.type = cap_args.type;
  args.in_security_param.partnum = partnum;
  args.in_security_param.actual_protection = cap_args.min_protection;

  nasd_cl_p_check_mem_list(memlist, &rc);
  if (rc) {
    *nasd_status = rc;
    return;
  }
  statep->out_lenp = out_lenp;
  statep->targetlen = in_len;
  statep->badstate = 0;
  statep->alloc_cur = statep->push_cur = memlist;
  statep->alloc_elem = statep->push_elem = 0;
  statep->alloc_buf = statep->push_buf = memlist->addr;
  statep->alloc_off = statep->push_off = 0;
  statep->alloc_len = statep->push_len = memlist->len;
  statep->status = NASD_SUCCESS;

  pipe.alloc = (void (*)())nasd_cl_p_fr_list_alloc;
  pipe.push = (void (*)())nasd_cl_p_fr_list_push;
  pipe.state = (rpc_ss_pipe_state_t)statep;

  nasd_p_thrtl_op_dr_marshall_args(&args, args_otw);

#if NASD_SECURE_RPCS_ENABLE > 0

  digested = nasd_security_digest_args((nasd_args_t*)&args, in_key,
    (nasd_byte_t*)args_otw, args_len, 
    (nasd_byte_t *)&(((nasd_p_thrtl_op_dr_args_t*)args_otw)->in_digest), NULL);

  if (digested != NASD_SUCCESS) {
    res.nasd_status=digested;
  }
  else {
    SETUP_REPLY_IV(icv);
    bzero(&statep->context, sizeof(nasd_security_context_t));
    nasd_security_reply_init((nasd_args_t *)&args, in_key, icv,
                             &statep->context);
    
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
    TRY {
      nasd_p_tread_simple_otw_dr(h, args_otw, pipe, res_otw, status);
      nasd_p_fastread_dr_unmarshall_res(res_otw, &res);

    }
    CATCH_ALL {
      res.nasd_status = NASD_RPC_TRAP;
      exc_get_status(THIS_CATCH,(int *)status);
    } ENDTRY
#if NASD_SECURE_RPCS_ENABLE > 0
   }
  /* If we don't have an RPC error */
  if (!(*status)) {
    nasd_security_verify_res((nasd_args_t*)&args, in_key,
      (nasd_byte_t*)res_otw, res_len, (nasd_status_t*)&res, 
      (nasd_byte_t*)&(((nasd_p_fastread_dr_res_t*)res_otw)->out_digest),
                             &statep->context);
  }
  /* Which nasd_status should take precendence - HG */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  if (statep->status || statep->badstate) {
    *nasd_status = NASD_MEM_LIST_ERR;
  }
  else {
    *nasd_status = res.nasd_status;
  }
  NASD_FREELIST_FREE(nasd_cl_fo_list_state_freelist, statep, push_cur);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */

