/*
 * nasd_pdrive_switch.c
 *
 * RPC wrappers, dispatch to correct binding type
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/*
 * if the status field is set in a call to these routines (ie:
 * an RPC failure occurs) then the nasd_status is considered meaningless.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_pdrive_client.h>

#define NASD_CL_P_SWITCH_NO_SAL 1
#define NASD_CL_P_MSGQ_DBG_ERRORS 1

#if NASD_CL_P_SWITCH_NO_SAL > 0
#include <dce/dce_error.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_mem.h>
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */

#include <nasd/nasd_pdev.h>
#include <sys/errno.h>
#include <sys/ioctl.h>

#ifndef KERNEL
#include <fcntl.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <nasd/nasd_mq.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>
#endif /* !KERNEL */

#define KEYCPY(_src_,_dest_) bcopy((char *)(_src_), (char *)(_dest_), NASD_KEY_SZ)

#if NASD_CL_P_SWITCH_NO_SAL > 0
#ifdef KERNEL
#include <nasdpd.h>
#define NASD_OD_RPC_NASD_RET res->nasd_status
#include <nasd/nasd_od_rpc.h>
#include <nasd/nasd_od.h>
#include <nasd/nasd_cache.h>

#if defined(DEC_OSF) && defined(KERNEL)
#include <mach/vm_param.h>
#include <nasd/dux/nasd_dux_opt.h>
#endif /* DEC_OSF && KERNEL */

typedef struct nasd_co_pipe_state_s {
  nasd_byte_t  *orig_buf;
  nasd_byte_t  *cur_buf;
  int           orig_buf_len;
  int           buf_len_remain;
} nasd_co_pipe_state_t;

nasd_status_t nasd_co_pull(void *state_arg, void *buf,
  nasd_len_t in_len, nasd_len_t *out_lenp);
nasd_status_t nasd_co_push(void *state_arg, void *buf,
  nasd_len_t in_len);

#endif /* KERNEL */

NASD_DECLARE_ONCE(nasd_cl_p_init_once)
NASD_DECLARE_MUTEX(nasd_cl_p_use_counter_lock)
int nasd_cl_p_use_counter = 0;
nasd_shutdown_list_t *nasd_cl_p_shutdown_list = NULL;

/* Don't use __FILE__ here to avoid getting kernel-mangled names */
#define DOBEGIN(_opname_) NASD_OD_RPC_DOBEGIN(_opname_,"nasd_pdrive_switch.c")
#define DORETURN(_opname_) NASD_OD_RPC_DORETURN(_opname_,"nasd_pdrive_switch.c")

#define LOCK_COUNTER_LOCK()   NASD_LOCK_MUTEX(nasd_cl_p_use_counter_lock)
#define UNLOCK_COUNTER_LOCK() NASD_UNLOCK_MUTEX(nasd_cl_p_use_counter_lock)

/*
 * nasd_cl_p_real_init
 *
 * Called when the system comes into use after not being
 * used (start-of-day call to nasd_cl_p_init(), or first
 * such call after the last call to nasd_cl_p_shutdown()
 * which actually deactivated the system).
 */
nasd_status_t
nasd_cl_p_real_init()
{
  nasd_status_t rc;

  /* already did threads_init() */


  rc = nasd_mem_init();
  if (rc) {
    nasd_threads_shutdown();
    return(rc);
  }


  rc = nasd_shutdown_sys_init();
  if (rc) {
    nasd_mem_shutdown();
    nasd_threads_shutdown();
    return(rc);
  }

  rc = nasd_shutdown_list_init(&nasd_cl_p_shutdown_list);
  if (rc) {
    nasd_shutdown_cleanup();
    nasd_mem_shutdown();
    nasd_threads_shutdown();
    return(rc);
  }

#if NASD_SECURITY_KEEP_STATS > 0 

  rc= nasd_3DES_init_stats();
  if (!rc) {
    rc= nasd_SHA1_init_stats();
  }
  if (!rc) {
    rc= nasd_HMAC_init_stats();
  }
  
  if (rc) {
    nasd_shutdown_cleanup();
    nasd_mem_shutdown();
    nasd_threads_shutdown();
    return(rc);
  }

#endif /* NASD_SECURITY_KEEP_STATS > 0 */


  rc = nasd_cl_p_dce_init(nasd_cl_p_shutdown_list);
  if (rc)
    goto bad;

#if defined(KERNEL) && NNASDPD > 0
  rc = nasdpd_init(nasd_cl_p_shutdown_list);
  if (rc)
    goto bad;
#endif /* KERNEL && NNASDPD > 0 */

  return(NASD_SUCCESS);

bad:
  rc = nasd_shutdown_list_shutdown(nasd_cl_p_shutdown_list);
  if (rc) {
    NASD_PANIC();
  }
  nasd_shutdown_cleanup();
  nasd_mem_shutdown();
  nasd_threads_shutdown();
  return(rc);
}

/*
 * nasd_cl_p_real_shutdown
 *
 * Called when last user of the cl_p system calls nasd_cl_p_shutdown().
 * Clean up and deallocate resources.
 */
void
nasd_cl_p_real_shutdown()
{
  nasd_status_t rc;

  rc = nasd_shutdown_list_shutdown(nasd_cl_p_shutdown_list);
  if (rc) {
    NASD_PANIC();
  }
  nasd_cl_p_shutdown_list = NULL;
  nasd_shutdown_cleanup();
  nasd_mem_shutdown();
  nasd_threads_shutdown();

}

/*
 * nasd_cl_p_sys_init
 *
 * Executed exactly once, the first time nasd_cl_p_init() is
 * called. Initialize counter tracking number of times system
 * is initted.
 */
void
nasd_cl_p_sys_init()
{
  nasd_status_t rc;

  rc = nasd_mutex_init(&nasd_cl_p_use_counter_lock);
  if (rc) {
    NASD_PANIC();
  }
  nasd_cl_p_use_counter = 0;
}

/*
 * nasd_cl_p_init
 *
 * Keep a counter of the number of times we're initted and
 * shutdown. When the last shutdown arrives, really deallocate.
 * This lets multiple subsystems use us without knowing about
 * one another.
 */
nasd_status_t
nasd_cl_p_init()
{
  nasd_status_t rc;

  rc = nasd_threads_init();
  if (rc)
    return(rc);
  nasd_once(&nasd_cl_p_init_once, nasd_cl_p_sys_init);
  LOCK_COUNTER_LOCK();
  nasd_cl_p_use_counter++;
  if (nasd_cl_p_use_counter == 1) {
    rc = nasd_cl_p_real_init();
    if (rc) {
      nasd_cl_p_use_counter = 0;
    }
  }
  else {
    /* get rid of extra threads init above */
    nasd_threads_shutdown();
    rc = NASD_SUCCESS;
  }
  UNLOCK_COUNTER_LOCK();

  return(rc);
}

/*
 * nasd_cl_p_shutdown
 *
 * Previous caller of nasd_cl_p_init() not using the cl_p
 * subsystem any more. Deallocate and cleanup iff necessary.
 */
void
nasd_cl_p_shutdown()
{
  NASD_ASSERT(nasd_cl_p_use_counter != 0);

  LOCK_COUNTER_LOCK();
  nasd_cl_p_use_counter--;
  if (nasd_cl_p_use_counter == 0) {
    nasd_cl_p_real_shutdown();
  }
  UNLOCK_COUNTER_LOCK();
}

#else /* NASD_CL_P_SWITCH_NO_SAL > 0 */

#define NASD_Malloc(_p_, _size_, _cast_) { \
  _p_ = _cast_ malloc(_size_); \
}

#define NASD_Free(_p_, _sz_) { \
  free(_p_); \
}

#define LOCK_COUNTER_LOCK()     { /* noop */ }
#define UNLOCK_COUNTER_LOCK()   { /* noop */ }

int nasd_cl_p_use_counter = 0;

nasd_status_t
nasd_cl_p_init()
{
  nasd_cl_p_use_counter++;
  return(NASD_SUCCESS);
}

void
nasd_cl_p_shutdown()
{
  nasd_cl_p_use_counter--;
}

#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */

nasd_status_t
nasd_bind_to_drive(
  char                 *drive_name,
  char                 *portnum,
  int                   binding_type,
  void                 *binding_param,
  int                   binding_param_len,
  nasd_drive_handle_t  *handlep)
{
  nasd_drive_handle_t handle;

  if (nasd_cl_p_use_counter == 0)
    return(NASD_SYS_NOT_INITIALIZED);
#ifndef KERNEL
{
  extern char *getenv();
  char *env;

  /* set this to something like "/dev/nasdkp0" */
  env = getenv("NASD_JIMZ_MONKEY_COLOCATE");
  /*
   * If we've already got a nasd_drive_param_dce_kpdev_t
   * argument, we use it anyway.
   */
  if (env) {
    if (binding_type != NASD_BIND_COLOCATE) {
      if (binding_type != NASD_BIND_DCE_KPDEV) {
        static nasd_drive_param_dce_kpdev_t kpdev_args;
        strcpy(kpdev_args.devname, env);
        binding_param = &kpdev_args;
        binding_param_len = sizeof(kpdev_args);
      }
      binding_type = NASD_BIND_COLOCATE;
    }
  }
}
#endif /* !KERNEL */

  *handlep = NULL;
#if NASD_CL_P_SWITCH_NO_SAL > 0
  if ((binding_type == NASD_BIND_DCE_DIRECT)
#ifndef KERNEL
    || (binding_type == NASD_BIND_DCE_DIRECT_TCP)
#endif /* !KERNEL */
    )
  {
    nasd_status_t rc;

    rc = NASD_FAIL;
    NASD_Malloc(handle, sizeof(struct nasd_drive_handle_s), (nasd_drive_handle_t));
    if (handle == NULL)
      return(NASD_NO_MEM);
    bzero((char *)handle, sizeof(struct nasd_drive_handle_s));
    handle->type = NASD_BIND_DCE_DIRECT;
    handle->real_type = binding_type;
    if (binding_type == NASD_BIND_DCE_DIRECT) {
      rc = nasd_dce_direct_bind_to_server((unsigned_char_t *)"ncadg_ip_udp",
        (unsigned_char_t *)drive_name, (unsigned_char_t *)portnum,
        &handle->h);
    }
#ifndef KERNEL
    else if (binding_type == NASD_BIND_DCE_DIRECT_TCP) {
      rc = nasd_dce_direct_bind_to_server((unsigned_char_t *)"ncacn_ip_tcp",
        (unsigned_char_t *)drive_name, (unsigned_char_t *)portnum,
        &handle->h);
    }
#endif /* !KERNEL */
    if (rc) {
      NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
      return(rc);
    }
    *handlep = handle;
    LOCK_COUNTER_LOCK();
    nasd_cl_p_use_counter++;
    UNLOCK_COUNTER_LOCK();
    return(NASD_SUCCESS);
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else /* my apologies for this construct */
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
  if ((binding_type == NASD_BIND_DCE_KPDEV)
    || (binding_type == NASD_BIND_DCE_KPDEV_FAKE_RW))
  {
    nasd_drive_param_dce_kpdev_t *param;
    struct hostent he, *hep;
    NASDPD_bind_t kpbind;
    struct in_addr addr;
    nasd_status_t rc;
    int ret;

    if (binding_param_len != sizeof(nasd_drive_param_dce_kpdev_t))
      return(NASD_BAD_BINDING_PARAM_LEN);
    param = (nasd_drive_param_dce_kpdev_t *)binding_param;

    if (isdigit(drive_name[0])) {
      strcpy(kpbind.drive_name, drive_name);
    }
    else {
      rc = nasd_gethostbyname_r(drive_name, &he);
      if (rc)
        return(rc);
      hep = &he;
      bcopy((char *)hep->h_addr_list[0], (char *)&addr.s_addr, 4);
      strcpy(kpbind.drive_name, inet_ntoa(addr));
    }

    NASD_Malloc(handle, sizeof(struct nasd_drive_handle_s), (nasd_drive_handle_t));
    if (handle == NULL)
      return(NASD_NO_MEM);
    bzero((char *)handle, sizeof(struct nasd_drive_handle_s));
    handle->type = NASD_BIND_DCE_KPDEV;
    handle->real_type = binding_type;
    handle->fd = open(param->devname, O_RDWR);
    if (handle->fd < 0) {
      NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
      return(NASD_FAIL);
    }
    strcpy(kpbind.drive_port, portnum);
    kpbind.binding_type = NASD_BIND_DCE_DIRECT;
    kpbind.binding_param = NULL;
    kpbind.binding_param_len = 0;
    ioctl(handle->fd, NASDPD_PUNBIND);
    ret = ioctl(handle->fd, NASDPD_PBIND, &kpbind);
    if (ret) {
      NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
      return(NASD_FAIL);
    }
    *handlep = handle;
    LOCK_COUNTER_LOCK();
    nasd_cl_p_use_counter++;
    UNLOCK_COUNTER_LOCK();
    return(NASD_SUCCESS);
  } else if (binding_type == NASD_BIND_MSGQ) {
    int mqid;

    NASD_Malloc(handle, sizeof(struct nasd_drive_handle_s),
                (nasd_drive_handle_t));
    if (handle == NULL)
      return(NASD_NO_MEM);
    memset(handle, 0, sizeof(struct nasd_drive_handle_s));
    handle->type = NASD_BIND_MSGQ;
    handle->real_type = binding_type;

    /* get id for drive's queue */
    mqid = msgget(NASD_MSGQ_KEY, 0);
    if(mqid < 0)
      return NASD_FAIL;
    handle->fd = mqid;

    /* create a queue for replies from drive */
    mqid = msgget(IPC_PRIVATE, IPC_CREAT | 0666);
    if(mqid < 0)
      return NASD_FAIL;
    handle->replyq = mqid;

#if NASD_CL_P_SWITCH_NO_SAL > 0
    /* XXX this seems like it will trigger a message queue leak,
       but it is unclear to me how to properly trigger this action
       on client shutdown with NO_SAL turned on. */
    nasd_shutdown_proc(nasd_cl_p_shutdown_list, nasd_cl_p_rm_msgq,
                       (void *) mqid);
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */

    /* initialize sequence numbers */
    nasd_mutex_init(&handle->seqlock);
    handle->seqnum = 1;

    *handlep = handle;
    LOCK_COUNTER_LOCK();
    nasd_cl_p_use_counter++;
    UNLOCK_COUNTER_LOCK();
    return(NASD_SUCCESS);
  }
#endif /* !KERNEL */
  else if (binding_type == NASD_BIND_COLOCATE) {
#ifndef KERNEL
    nasd_drive_param_dce_kpdev_t *param;
    NASDPD_bind_t kpbind;
    int ret, fd;

    if (binding_param_len != sizeof(nasd_drive_param_dce_kpdev_t))
      return(NASD_BAD_BINDING_PARAM_LEN);
    param = (nasd_drive_param_dce_kpdev_t *)binding_param;
    fd = open(param->devname, O_RDWR);
    if (fd < 0) {
      return(NASD_FAIL);
    }
#endif /* !KERNEL */

    NASD_Malloc(handle, sizeof(struct nasd_drive_handle_s), (nasd_drive_handle_t));
    if (handle == NULL)
      return(NASD_NO_MEM);
    bzero((char *)handle, sizeof(struct nasd_drive_handle_s));
#ifdef KERNEL
    handle->type = binding_type;
    handle->real_type = binding_type;
#else /* KERNEL */
    handle->type = NASD_BIND_DCE_KPDEV;
    handle->real_type = binding_type;
    handle->fd = fd;
    strcpy(kpbind.drive_name, "colocated drive");
    strcpy(kpbind.drive_port, portnum);
    kpbind.binding_type = binding_type;
    kpbind.binding_param = NULL;
    kpbind.binding_param_len = 0;
    ioctl(handle->fd, NASDPD_PUNBIND);
    ret = ioctl(handle->fd, NASDPD_PBIND, &kpbind);
    if (ret) {
      NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
      return(NASD_FAIL);
    }
#endif /* KERNEL */
    *handlep = handle;
    LOCK_COUNTER_LOCK();
    nasd_cl_p_use_counter++;
    UNLOCK_COUNTER_LOCK();
    return(NASD_SUCCESS);
  }
  return(NASD_BAD_HANDLE_TYPE);
}

nasd_status_t
nasd_unbind_drive(
  nasd_drive_handle_t  *handlep)
{
  nasd_drive_handle_t handle;

  handle = *handlep;
  if (handle == NULL) {
    return(NASD_BAD_HANDLE);
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    int rc;

    rc = nasd_dce_direct_unbind_server(&handle->h);
    if (rc)
      return(NASD_FAIL);
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_cl_p_shutdown();
    return(NASD_SUCCESS);
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if ((handle->type == NASD_BIND_DCE_KPDEV)
    || (handle->type == NASD_BIND_COLOCATE))
  {
    int rc;

    rc = ioctl(handle->fd, NASDPD_PUNBIND);
    rc = close(handle->fd);
    if (rc)
      return(NASD_FAIL);
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_cl_p_shutdown();
    return(NASD_SUCCESS);
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    int rc;

    rc = msgctl(handle->replyq, IPC_RMID, NULL);
    if(rc)
      return NASD_FAIL;
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_cl_p_shutdown();
    return NASD_SUCCESS;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_Free(handle, sizeof(struct nasd_drive_handle_s));
    *handlep = NULL;
    nasd_cl_p_shutdown();
    return(NASD_SUCCESS);
  }
#endif /* KERNEL */
  return(NASD_BAD_HANDLE_TYPE);
}

void
nasd_cl_p_rm_msgq(void *arg)
{
  int mqid = (int) arg;

  msgctl(IPC_RMID, mqid, (struct msqid_ds *) NULL);
}


void
nasd_cl_p_null_dr(
  nasd_drive_handle_t   handle,
  nasd_status_t        *nasd_status,
  nasd_rpc_status_t    *status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_null_dr(handle->h, nasd_status, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_null_t null;
    int rc;

    rc = ioctl(handle->fd, NASDPD_PNULL, &null);
    if (rc) {
      *nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = null.start_cc;
    handle->last_stop_cc = null.stop_cc;
    *nasd_status = null.nasd_status;
    *status = null.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_null_msg_t *msg = &(mb.un.null_msg);
    NASDMQ_null_rep_t *rep = &(mb.un.null_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_NULL;
    msg->replyq = handle->replyq;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      *nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      *nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *nasd_status = rep->nasd_status;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_od_rpc_basic_t out_res;
    nasd_od_rpc_basic_t *res = &out_res;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(null);
    *nasd_status = out_res.nasd_status = NASD_SUCCESS;
    DORETURN(null);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_sync_dr(
  nasd_drive_handle_t   handle,
  nasd_status_t        *nasd_status,
  nasd_rpc_status_t    *status)
{
  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_sync_dr(handle->h, nasd_status, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_sync_t sync;
    int rc;

    rc = ioctl(handle->fd, NASDPD_PSYNC, &sync);
    if (rc) {
      *nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = sync.start_cc;
    handle->last_stop_cc = sync.stop_cc;
    *nasd_status = sync.nasd_status;
    *status = sync.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_sync_msg_t *msg = &(mb.un.sync_msg);
    NASDMQ_sync_rep_t *rep = &(mb.un.sync_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_SYNC;
    msg->replyq = handle->replyq;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      *nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      *nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *nasd_status = rep->nasd_status;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_od_rpc_basic_t out_res;
    nasd_od_rpc_basic_t *res = &out_res;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(sync);
    *nasd_status = out_res.nasd_status = nasd_odc_flush_dirty(1);
    DORETURN(sync);
  }
#endif /* KERNEL */
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_part_creat_dr(
  nasd_drive_handle_t           handle,
  nasd_key_t                    in_key,
  nasd_p_part_creat_dr_args_t  *args,
  nasd_p_part_creat_dr_res_t   *res,
  nasd_rpc_status_t            *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_part_creat_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_crpart_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = *args;
    rc = ioctl(handle->fd, NASDPD_P_PCRPART, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_crpart_msg_t *msg = &(mb.un.p_crpart_msg);
    NASDMQ_p_crpart_rep_t *rep = &(mb.un.p_crpart_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PCRPART;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_identifier_t nid;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(part_creat);
    res->nasd_status = nasd_od_create_partition(args->in_partnum,
      args->in_blkcnt, args->in_min_protection, args->in_partition_key,
      args->in_red_key, args->in_black_key, &nid);
    DORETURN(part_creat);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_create_dr(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_p_create_dr_args_t  *args,
  nasd_p_create_dr_res_t   *res,
  nasd_rpc_status_t        *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_create_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_create_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = args;
    ioccmd.res = res;
    rc = ioctl(handle->fd, NASDPD_P_PCREATE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_create_msg_t *msg = &(mb.un.p_create_msg);
    NASDMQ_p_create_rep_t *rep = &(mb.un.p_create_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PCREATE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(create);
    res->nasd_status = nasd_obj_create(args->in_partnum,
      &args->in_attribute, args->in_fieldmask, &res->out_identifier,
      &res->out_attribute, 0);
    DORETURN(create);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_create_dr__otw_provided(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_p_create_dr_args_t  *args,
  nasd_p_create_dr_res_t   *res,
  nasd_byte_t              *args_otw,
  nasd_byte_t              *res_otw,
  nasd_rpc_status_t        *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_create_dr__otw_provided(handle->h, in_key,
      args, res, args_otw, res_otw, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_create_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = args;
    ioccmd.res = res;
    rc = ioctl(handle->fd, NASDPD_P_PCREATE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_create_msg_t *msg = &(mb.un.p_create_msg);
    NASDMQ_p_create_rep_t *rep = &(mb.un.p_create_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PCREATE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(create);
    res->nasd_status = nasd_obj_create(args->in_partnum,
      &args->in_attribute, args->in_fieldmask, &res->out_identifier,
      &res->out_attribute, 0);
    DORETURN(create);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_getattr_dr(
  nasd_drive_handle_t        handle,
  nasd_key_t                 in_key,
  nasd_p_getattr_dr_args_t  *args,
  nasd_p_getattr_dr_res_t   *res,
  nasd_rpc_status_t         *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_getattr_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_getattr_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = args;
    ioccmd.res = res;
    rc = ioctl(handle->fd, NASDPD_P_PGETATTR, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_getattr_msg_t *msg = &(mb.un.p_getattr_msg);
    NASDMQ_p_getattr_rep_t *rep = &(mb.un.p_getattr_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PGETATTR;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(getattr);
    res->nasd_status = nasd_obj_getattr(args->in_partnum,
      args->in_identifier, &res->out_attribute);
    DORETURN(getattr);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_write_simple_dr(
  nasd_drive_handle_t         handle,
  nasd_key_t                  in_key,
  nasd_p_smpl_op_dr_args_t   *args,
  void                       *buf,
  nasd_p_fastwrite_dr_res_t  *res,
  nasd_rpc_status_t          *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_write_simple_dr(handle->h, in_key, args, buf, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_write_simple_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = *args;
    ioccmd.buf = buf;
    if (handle->real_type == NASD_BIND_DCE_KPDEV_FAKE_RW)
      rc = ioctl(handle->fd, NASDPD_P_PWRITE_FAKE, &ioccmd);
    else
      rc = ioctl(handle->fd, NASDPD_P_PWRITE_SIMPLE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_write_simple_msg_t *msg = &(mb.un.p_write_simple_msg);
    NASDMQ_p_write_simple_rep_t *rep = &(mb.un.p_write_simple_rep);
    int seqnum;
    int rc;
    int shmid;
    nasd_byte_t *shmbuf;

    shmid = shmget(IPC_PRIVATE, args->in_len, IPC_CREAT | 0666);
    if(shmid < 0) {
#if NASD_CL_P_MSGQ_DBG_ERRORS > 0
      fprintf(stderr, "write_simple: shmget failed: rc=%d errno=%d (%s)\n",
              shmid, errno, strerror(errno));
#endif /* NASD_CL_P_MSGQ_DBG_ERRORS > 0 */
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    msg->shmid = shmid;
    shmbuf = shmat(shmid, NULL, 0);
    if(shmbuf == (nasd_byte_t *) -1) {
#if NASD_CL_P_MSGQ_DBG_ERRORS > 0
      fprintf(stderr, "write_simple: shmat failed: rc=0x%x errno=%d (%s)\n",
              shmbuf, errno, strerror(errno));
#endif /* NASD_CL_P_MSGQ_DBG_ERRORS > 0 */
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      shmctl(shmid, IPC_RMID, NULL);
      return;
    }
    memcpy(shmbuf, buf, args->in_len);

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);
    mb.mtype = NASD_MSGQ_PWRITE_SIMPLE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    shmdt(shmbuf);
    shmctl(shmid, IPC_RMID, NULL);
    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_co_pipe_state_t state;
    nasd_procpipe_t procpipe;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(write_simple);
    state.orig_buf = state.cur_buf = (nasd_byte_t *)buf;
    state.orig_buf_len = state.buf_len_remain = args->in_len;
    procpipe.state = &state;
    procpipe.push = NULL;
    procpipe.pull = nasd_co_pull;
    res->out_datalen = 0;

    res->nasd_status = nasd_obj_write_simple(args->in_partnum,
      args->in_identifier, args->in_offset, args->in_len, &procpipe,
      NULL, &res->out_datalen);

    DORETURN(write_simple);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_read_simple_dr(
  nasd_drive_handle_t         handle,
  nasd_key_t                  in_key,
  nasd_p_smpl_op_dr_args_t   *args,
  void                       *buf,
  nasd_p_fastread_dr_res_t   *res,
  nasd_rpc_status_t          *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_read_simple_dr(handle->h, in_key, 0, args, buf, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_read_simple_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.is_read2 = 0;
    ioccmd.args = *args;
    ioccmd.buf = buf;
    if (handle->real_type == NASD_BIND_DCE_KPDEV_FAKE_RW)
      rc = ioctl(handle->fd, NASDPD_P_PREAD_FAKE, &ioccmd);
    else
      rc = ioctl(handle->fd, NASDPD_P_PREAD_SIMPLE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_read_simple_msg_t *msg = &(mb.un.p_read_simple_msg);
    NASDMQ_p_read_simple_rep_t *rep = &(mb.un.p_read_simple_rep);
    int seqnum;
    int rc;
    int shmid;
    nasd_byte_t *shmbuf;

    shmid = shmget(IPC_PRIVATE, args->in_len, IPC_CREAT | 0666);
    if(shmid < 0) {
#if NASD_CL_P_MSGQ_DBG_ERRORS > 0
      fprintf(stderr, "read_simple: shmget failed: rc=%d errno=%d (%s)\n",
              shmid, errno, strerror(errno));
#endif /* NASD_CL_P_MSGQ_DBG_ERRORS > 0 */
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    msg->shmid = shmid;
    shmbuf = shmat(shmid, NULL, 0);
    if(shmbuf == (nasd_byte_t *) -1) {
#if NASD_CL_P_MSGQ_DBG_ERRORS > 0
      fprintf(stderr, "read_simple: shmat failed: rc=0x%x errno=%d (%s)\n",
              shmbuf, errno, strerror(errno));
#endif /* NASD_CL_P_MSGQ_DBG_ERRORS > 0 */
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      shmctl(shmid, IPC_RMID, NULL);
      return;
    }

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);
    mb.mtype = NASD_MSGQ_PREAD_SIMPLE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    msg->is_read2 = 0;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    memcpy(buf, shmbuf, args->in_len);

    shmdt(shmbuf);
    shmctl(shmid, IPC_RMID, NULL);
    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_co_pipe_state_t state;
    nasd_procpipe_t procpipe;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(read_simple);
    state.orig_buf = state.cur_buf = (nasd_byte_t *)buf;
    state.orig_buf_len = state.buf_len_remain = args->in_len;
    procpipe.state = &state;
    procpipe.push = nasd_co_push;
    procpipe.pull = NULL;
    res->out_datalen = 0;

    res->nasd_status = nasd_obj_read_simple(args->in_partnum,
      args->in_identifier, args->in_offset, args->in_len, NULL,
      0, &procpipe, &res->out_datalen, NULL);

    DORETURN(read_simple);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_read2_simple_dr(
  nasd_drive_handle_t         handle,
  nasd_key_t                  in_key,
  nasd_p_smpl_op_dr_args_t   *args,
  void                       *buf,
  nasd_p_fastread_dr_res_t   *res,
  nasd_rpc_status_t          *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_read_simple_dr(handle->h, in_key, 1, args, buf, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_read_simple_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.is_read2 = 1;
    ioccmd.args = *args;
    ioccmd.buf = buf;
    if (handle->real_type == NASD_BIND_DCE_KPDEV_FAKE_RW)
      rc = ioctl(handle->fd, NASDPD_P_PREAD_FAKE, &ioccmd);
    else
      rc = ioctl(handle->fd, NASDPD_P_PREAD_SIMPLE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_read_simple_msg_t *msg = &(mb.un.p_read_simple_msg);
    NASDMQ_p_read_simple_rep_t *rep = &(mb.un.p_read_simple_rep);
    int seqnum;
    int rc;
    int shmid;
    nasd_byte_t *shmbuf;

    shmid = shmget(IPC_PRIVATE, args->in_len, IPC_CREAT | 0666);
    if(shmid < 0) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    msg->shmid = shmid;
    shmbuf = shmat(shmid, NULL, 0);
    if(shmbuf == (nasd_byte_t *) -1) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      shmctl(shmid, IPC_RMID, NULL);
      return;
    }

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);
    mb.mtype = NASD_MSGQ_PREAD_SIMPLE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    msg->is_read2 = 1;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    memcpy(buf, shmbuf, args->in_len);

    shmdt(shmbuf);
    shmctl(shmid, IPC_RMID, NULL);
    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_co_pipe_state_t state;
    nasd_procpipe_t procpipe;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(read2_simple);
    state.orig_buf = state.cur_buf = (nasd_byte_t *)buf;
    state.orig_buf_len = state.buf_len_remain = args->in_len;
    procpipe.state = &state;
    procpipe.push = nasd_co_push;
    procpipe.pull = NULL;
    res->out_datalen = 0;

    res->nasd_status = nasd_obj_read_simple(args->in_partnum,
      args->in_identifier, args->in_offset, args->in_len, NULL,
      1, &procpipe, &res->out_datalen, NULL);

    DORETURN(read2_simple);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_tread_simple_dr(
  nasd_drive_handle_t         handle,
  nasd_key_t                  in_key,
  nasd_p_thrtl_op_dr_args_t  *args,
  void                       *buf,
  nasd_p_fastread_dr_res_t   *res,
  nasd_rpc_status_t          *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_tread_simple_dr(handle->h, in_key, args, buf, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_tread_simple_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = *args;
    ioccmd.buf = buf;
    rc = ioctl(handle->fd, NASDPD_P_PTREAD_SIMPLE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_tread_simple_msg_t *msg = &(mb.un.p_tread_simple_msg);
    NASDMQ_p_tread_simple_rep_t *rep = &(mb.un.p_tread_simple_rep);
    int seqnum;
    int rc;
    int shmid;
    nasd_byte_t *shmbuf;

    shmid = shmget(IPC_PRIVATE, args->in_len, IPC_CREAT | 0666);
    if(shmid < 0) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    msg->shmid = shmid;
    shmbuf = shmat(shmid, NULL, 0);
    if(shmbuf == (nasd_byte_t *) -1) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      shmctl(shmid, IPC_RMID, NULL);
      return;
    }

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);
    mb.mtype = NASD_MSGQ_PTREAD_SIMPLE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    memcpy(buf, shmbuf, args->in_len);

    shmdt(shmbuf);
    shmctl(shmid, IPC_RMID, NULL);
    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    nasd_co_pipe_state_t state;
    nasd_procpipe_t procpipe;
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(tread_simple);
    state.orig_buf = state.cur_buf = (nasd_byte_t *)buf;
    state.orig_buf_len = state.buf_len_remain = args->in_len;
    procpipe.state = &state;
    procpipe.push = nasd_co_push;
    procpipe.pull = NULL;
    res->out_datalen = 0;

    res->nasd_status = nasd_obj_read_simple(args->in_partnum,
      args->in_identifier, args->in_offset, args->in_len, &args->in_bms_targ,
      0, &procpipe, &res->out_datalen, NULL);

    DORETURN(tread_simple);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_range_read(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_mem_list_t         *memlist,
  nasd_offset_t            in_offset,
  nasd_len_t               in_len,
  nasd_len_t              *out_lenp,
  nasd_status_t           *nasd_status,
  nasd_rpc_status_t       *status)
{
  nasd_p_smpl_op_dr_args_t rd_args;
  nasd_p_fastread_dr_res_t rd_res;
  int emulate_range;

  emulate_range = 0;

  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_range_read(handle->h, in_key, cap_args, partnum, nasdid,
      memlist, in_offset, in_len, out_lenp, nasd_status, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV ||
           handle->type == NASD_BIND_MSGQ) {
    emulate_range = 1;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    emulate_range = 1;
  }
#endif /* KERNEL */
  if (emulate_range) {
    nasd_len_t len, total_len;
    nasd_offset_t offset;
    nasd_mem_list_t *m;
    nasd_byte_t *addr;
    nasd_status_t rc;
    int elem;

    offset = in_offset;
    total_len = 0;
    *out_lenp = 0;
    for(m=memlist;m;m=m->next) {
      addr = m->addr;
      for(elem=0;elem<m->nelem;elem++) {
        rd_args.in_partnum = partnum;
        rd_args.in_identifier = nasdid;
        rd_args.in_offset = offset;
        rd_args.in_len = m->len;
        rd_args.in_capability_args = cap_args;
        rd_args.in_security_param.type = cap_args.type;
        rd_args.in_security_param.partnum = partnum;
        rd_args.in_security_param.actual_protection = cap_args.min_protection;
        nasd_cl_p_read_simple_dr(handle, in_key, &rd_args, addr, &rd_res, status);
        len = rd_res.out_datalen;
        *nasd_status = rd_res.nasd_status;
        if (*nasd_status || *status)
          return;
        total_len += len;
        *out_lenp = total_len;
        if (len < m->len)
          return;
        addr += m->stride;
        offset += len;
      }
    }
    *nasd_status = NASD_SUCCESS;
    *status = 0;
    return;
  }
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_range_write(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_mem_list_t         *memlist,
  nasd_offset_t            in_offset,
  nasd_len_t               in_len,
  nasd_len_t              *out_lenp,
  nasd_status_t           *nasd_status,
  nasd_rpc_status_t       *status)
{
  nasd_p_smpl_op_dr_args_t wr_args;
  nasd_p_fastwrite_dr_res_t wr_res;
  int emulate_range;

  emulate_range = 0;

  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_range_write(handle->h, in_key, cap_args, partnum, nasdid,
      memlist, in_offset, in_len, out_lenp, nasd_status, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV ||
           handle->type == NASD_BIND_MSGQ) {
    emulate_range = 1;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    emulate_range = 1;
  }
#endif /* KERNEL */
  if (emulate_range) {
    nasd_len_t len, total_len;
    nasd_offset_t offset;
    nasd_mem_list_t *m;
    nasd_byte_t *addr;
    nasd_status_t rc;
    int elem;

    offset = in_offset;
    total_len = 0;
    *out_lenp = 0;
    for(m=memlist;m;m=m->next) {
      addr = m->addr;
      for(elem=0;elem<m->nelem;elem++) {
        wr_args.in_partnum = partnum;
        wr_args.in_identifier = nasdid;
        wr_args.in_offset = offset;
        wr_args.in_len = m->len;
        wr_args.in_capability_args = cap_args;
        wr_args.in_security_param.type = cap_args.type;
        wr_args.in_security_param.partnum = partnum;
        wr_args.in_security_param.actual_protection = cap_args.min_protection;
        nasd_cl_p_write_simple_dr(handle, in_key, &wr_args, addr, &wr_res, status);
        len = wr_res.out_datalen;
        *nasd_status = wr_res.nasd_status;
        if (*nasd_status || *status)
          return;
        total_len += len;
        *out_lenp = total_len;
        if (len < m->len)
          return;
        addr += m->stride;
        offset += len;
      }
    }
    *nasd_status = NASD_SUCCESS;
    *status = 0;
    return;
  }
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_range_tread(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_uint64              in_bms_targ,
  nasd_mem_list_t         *memlist,
  nasd_offset_t            in_offset,
  nasd_len_t               in_len,
  nasd_len_t              *out_lenp,
  nasd_status_t           *nasd_status,
  nasd_rpc_status_t       *status)
{
  nasd_p_thrtl_op_dr_args_t trd_args;
  nasd_p_fastread_dr_res_t rd_res;
  int emulate_range;

  emulate_range = 0;

  if (handle == NULL) {
    *nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_range_tread(handle->h, in_key, cap_args, partnum, nasdid,
      in_bms_targ, memlist, in_offset, in_len, out_lenp, nasd_status, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV ||
           handle->type == NASD_BIND_MSGQ) {
    emulate_range = 1;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    emulate_range = 1;
  }
#endif /* KERNEL */
  if (emulate_range) {
    nasd_len_t len, total_len;
    nasd_offset_t offset;
    nasd_mem_list_t *m;
    nasd_byte_t *addr;
    nasd_status_t rc;
    int elem;

    offset = in_offset;
    total_len = 0;
    *out_lenp = 0;
    for(m=memlist;m;m=m->next) {
      addr = m->addr;
      for(elem=0;elem<m->nelem;elem++) {
        trd_args.in_partnum = partnum;
        trd_args.in_identifier = nasdid;
        trd_args.in_offset = offset;
        trd_args.in_bms_targ = in_bms_targ;
        trd_args.in_len = m->len;
        trd_args.in_capability_args = cap_args;
        trd_args.in_security_param.type = cap_args.type;
        trd_args.in_security_param.partnum = partnum;
        trd_args.in_security_param.actual_protection = cap_args.min_protection;
        nasd_cl_p_tread_simple_dr(handle, in_key, &trd_args, addr, &rd_res, status);
        len = rd_res.out_datalen;
        *nasd_status = rd_res.nasd_status;
        if (*nasd_status || *status)
          return;
        total_len += len;
        *out_lenp = total_len;
        if (len < m->len)
          return;
        addr += m->stride;
        offset += len;
      }
    }
    *nasd_status = NASD_SUCCESS;
    *status = 0;
    return;
  }
  *nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_setattr_dr(
  nasd_drive_handle_t        handle,
  nasd_key_t                 in_key,
  nasd_p_setattr_dr_args_t  *args,
  nasd_p_setattr_dr_res_t   *res,
  nasd_rpc_status_t         *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_setattr_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_setattr_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = args;
    ioccmd.res = res;
    rc = ioctl(handle->fd, NASDPD_P_PSETATTR, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_setattr_msg_t *msg = &(mb.un.p_setattr_msg);
    NASDMQ_p_setattr_rep_t *rep = &(mb.un.p_setattr_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PSETATTR;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(setattr);
    res->nasd_status = nasd_obj_setattr(args->in_partnum,
      args->in_identifier, &args->in_attribute, args->in_fieldmask,
      &res->out_attribute);
    DORETURN(setattr);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_setattr_dr__otw_provided(
  nasd_drive_handle_t        handle,
  nasd_key_t                 in_key,
  nasd_p_setattr_dr_args_t  *args,
  nasd_p_setattr_dr_res_t   *res,
  nasd_byte_t               *args_otw,
  nasd_byte_t               *res_otw,
  nasd_rpc_status_t         *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_setattr_dr__otw_provided(handle->h, in_key,
      args, res, args_otw, res_otw, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_setattr_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = args;
    ioccmd.res = res;
    rc = ioctl(handle->fd, NASDPD_P_PSETATTR, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_setattr_msg_t *msg = &(mb.un.p_setattr_msg);
    NASDMQ_p_setattr_rep_t *rep = &(mb.un.p_setattr_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PSETATTR;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(setattr);
    res->nasd_status = nasd_obj_setattr(args->in_partnum,
      args->in_identifier, &args->in_attribute, args->in_fieldmask,
      &res->out_attribute);
    DORETURN(setattr);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_flush_obj_dr(
  nasd_drive_handle_t          handle,
  nasd_key_t                   in_key,
  nasd_p_flush_obj_dr_args_t  *args,
  nasd_p_flush_obj_dr_res_t   *res,
  nasd_rpc_status_t           *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_flush_obj_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_flush_obj_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = *args;
    rc = ioctl(handle->fd, NASDPD_P_PFLUSH_OBJ, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_flush_obj_msg_t *msg = &(mb.un.p_flush_obj_msg);
    NASDMQ_p_flush_obj_rep_t *rep = &(mb.un.p_flush_obj_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PFLUSH_OBJ;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(flush_obj);
    res->nasd_status = nasd_obj_flush(args->in_partnum,
      args->in_identifier);
    DORETURN(flush_obj);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_eject_obj_dr(
  nasd_drive_handle_t          handle,
  nasd_key_t                   in_key,
  nasd_p_eject_obj_dr_args_t  *args,
  nasd_p_eject_obj_dr_res_t   *res,
  nasd_rpc_status_t           *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_eject_obj_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_eject_obj_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = *args;
    rc = ioctl(handle->fd, NASDPD_P_PEJECT_OBJ, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_eject_obj_msg_t *msg = &(mb.un.p_eject_obj_msg);
    NASDMQ_p_eject_obj_rep_t *rep = &(mb.un.p_eject_obj_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PEJECT_OBJ;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(eject_obj);
    res->nasd_status = nasd_obj_eject(args->in_partnum,
      args->in_identifier);
    DORETURN(eject_obj);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_remove_dr(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_p_remove_dr_args_t  *args,
  nasd_p_remove_dr_res_t   *res,
  nasd_rpc_status_t        *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_remove_dr(handle->h, in_key, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_remove_t ioccmd;
    int rc;

    KEYCPY(in_key, ioccmd.in_key);
    ioccmd.args = *args;
    rc = ioctl(handle->fd, NASDPD_P_PREMOVE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_remove_msg_t *msg = &(mb.un.p_remove_msg);
    NASDMQ_p_remove_rep_t *rep = &(mb.un.p_remove_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PREMOVE;
    msg->replyq = handle->replyq;
    KEYCPY(in_key, msg->in_key);
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL

    DOBEGIN(remove);
    res->nasd_status = nasd_obj_remove(args->in_partnum,
      args->in_identifier);
    DORETURN(remove);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

void
nasd_cl_p_initialize_dr(
  nasd_drive_handle_t           handle,
  nasd_p_initialize_dr_args_t  *args,
  nasd_p_initialize_dr_res_t   *res,
  nasd_rpc_status_t            *status)
{
  if (handle == NULL) {
    res->nasd_status = NASD_BAD_HANDLE;
    *status = 0;
    return;
  }
#if NASD_CL_P_SWITCH_NO_SAL > 0
  else if (handle->type == NASD_BIND_DCE_DIRECT) {
    nasd_cl_prpc_initialize_dr(handle->h, args, res, status);
    return;
  }
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
#ifndef KERNEL
  else if (handle->type == NASD_BIND_DCE_KPDEV) {
    NASDPD_p_initialize_t ioccmd;
    int rc;

    ioccmd.args = *args;
    rc = ioctl(handle->fd, NASDPD_P_PINITIALIZE, &ioccmd);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    handle->last_start_cc = ioccmd.start_cc;
    handle->last_stop_cc = ioccmd.stop_cc;
    *res = ioccmd.res;
    *status = ioccmd.op_status;
    return;
  }
  else if (handle->type == NASD_BIND_MSGQ) {
    nasd_msgq_buf_t mb;
    NASDMQ_p_initialize_msg_t *msg = &(mb.un.p_initialize_msg);
    NASDMQ_p_initialize_rep_t *rep = &(mb.un.p_initialize_rep);
    int seqnum;
    int rc;

    NASD_LOCK_MUTEX(handle->seqlock);
    msg->seqnum = seqnum = handle->seqnum++;
    NASD_UNLOCK_MUTEX(handle->seqlock);

    mb.mtype = NASD_MSGQ_PINITIALIZE;
    msg->replyq = handle->replyq;
    msg->args = *args;
    rc = msgsnd(handle->fd, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN, 0);
    if (rc) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }
    rc = msgrcv(handle->replyq, (struct msgbuf *) &mb, NASD_MSGQ_MSGLEN,
                seqnum, 0);
    if (rc != NASD_MSGQ_MSGLEN) {
      res->nasd_status = NASD_RPC_LOCAL_FAIL;
      *status = errno;
      return;
    }

    *res = rep->res;
    *status = rep->op_status;
    return;
  }
#endif /* !KERNEL */
#ifdef KERNEL
  else if (handle->type == NASD_BIND_COLOCATE) {
    NASD_OD_RPC_OP_STATS_DECL
    
    DOBEGIN(initialize);
    res->nasd_status = nasd_security_initialize_drive(args->in_master_key,
      args->in_drive_key);
    DORETURN(initialize);
  }
#endif /* KERNEL */
  res->nasd_status = NASD_BAD_HANDLE_TYPE;
  *status = 0;
}

/*
 * END RPC WRAPPERS
 *
 * BEGIN CONTROL WRAPPERS
 */

/*
 * Read or write an info page
 */
nasd_status_t
nasd_cl_p_ctrl_rdwr_stats(
  nasd_drive_handle_t      handle,
  int                      partnum,
  nasd_identifier_t        nasdid,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  void                    *info,
  nasd_offset_t            offset,
  int                      read)
{
  nasd_p_smpl_op_dr_args_t op_args;
  nasd_p_fastwrite_dr_res_t wr_res;
  nasd_p_fastread_dr_res_t rd_res;
  nasd_len_t in_len, out_len;
  nasd_rpc_status_t status;
  nasd_status_t rc;

  if (read) {
    op_args.in_partnum = partnum;
    op_args.in_identifier = nasdid;
    op_args.in_offset = offset;
    op_args.in_len = NASD_INFO_PAGESIZE;
    op_args.in_capability_args = cap_args;
    op_args.in_security_param.type = cap_args.type;
    op_args.in_security_param.partnum = partnum;
    op_args.in_security_param.actual_protection = cap_args.min_protection;
    nasd_cl_p_read2_simple_dr(handle, in_key, &op_args, (void *)info,
      &rd_res, &status);
    out_len = rd_res.out_datalen;
    rc = rd_res.nasd_status;
  }
  else {
    op_args.in_partnum = partnum;
    op_args.in_identifier = nasdid;
    op_args.in_offset = offset;
    op_args.in_len = NASD_INFO_PAGESIZE;
    op_args.in_capability_args = cap_args;
    op_args.in_security_param.type = cap_args.type;
    op_args.in_security_param.partnum = partnum;
    op_args.in_security_param.actual_protection = cap_args.min_protection;
    nasd_cl_p_write_simple_dr(handle, in_key, &op_args, (void *)info,
      &wr_res, &status);
    out_len = wr_res.out_datalen;
    rc = wr_res.nasd_status;
  }
  if (status)
    rc = NASD_RPC_TRAP;
  else if (out_len != NASD_INFO_PAGESIZE)
    rc = NASD_FAIL;
  else if (read && *(nasd_uint64 *)info != nasdid) /* The ctrl_id field */
    rc = NASD_CTRL_ID_CHECK_FAILED;
  return(rc);
}

/*
 * Retrieve drive info struct
 */
nasd_status_t
nasd_cl_p_ctrl_get_drive_info(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  nasd_ctrl_drive_info_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 0, 1);
  return(rc);
}

/*
 * Retrieve partition info struct
 */
nasd_status_t
nasd_cl_p_ctrl_get_part_info(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  int                      partnum,
  nasd_ctrl_part_info_t   *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, partnum, NASD_CTRL_PART_INFO,
    in_key, cap_args, info, 0, 1);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_get_drive_op_stats(
  nasd_drive_handle_t         handle,
  nasd_key_t                  in_key,
  nasd_capability_args_t      cap_args,
  nasd_ctrl_drive_opstats_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 8192, 1);
  return(rc);
}


nasd_status_t
nasd_cl_p_ctrl_put_drive_op_stats(
  nasd_drive_handle_t         handle,
  nasd_key_t                  in_key,
  nasd_capability_args_t      cap_args,
  nasd_ctrl_drive_opstats_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 8192, 0);
  return(rc);
}


nasd_status_t
nasd_cl_p_ctrl_get_drive_cache_stats(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  nasd_ctrl_cache_stat_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 16384, 1);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_put_drive_cache_stats(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  nasd_ctrl_cache_stat_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 16384, 0);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_get_drive_io_stats(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  nasd_ctrl_io_stat_t     *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 24576, 1);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_put_drive_io_stats(
  nasd_drive_handle_t      handle,
  nasd_key_t               in_key,
  nasd_capability_args_t   cap_args,
  nasd_ctrl_io_stat_t     *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 24576, 0);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_get_drive_io_read_size_stats(
  nasd_drive_handle_t           handle,
  nasd_key_t                    in_key,
  nasd_capability_args_t        cap_args,
  nasd_ctrl_io_size_stat_t     *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 32768, 1);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_put_drive_io_read_size_stats(
  nasd_drive_handle_t           handle,
  nasd_key_t                    in_key,
  nasd_capability_args_t        cap_args,
  nasd_ctrl_io_size_stat_t     *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 32768, 0);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_get_drive_io_write_size_stats(
  nasd_drive_handle_t           handle,
  nasd_key_t                    in_key,
  nasd_capability_args_t        cap_args,
  nasd_ctrl_io_size_stat_t     *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 40960, 1);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_put_drive_io_write_size_stats(
  nasd_drive_handle_t           handle,
  nasd_key_t                    in_key,
  nasd_capability_args_t        cap_args,
  nasd_ctrl_io_size_stat_t     *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 40960, 0);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_get_drive_layout_stats(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_capability_args_t    cap_args,
  nasd_ctrl_layout_stat_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 49152, 1);
  return(rc);
}


nasd_status_t
nasd_cl_p_ctrl_put_drive_layout_stats(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_capability_args_t    cap_args,
  nasd_ctrl_layout_stat_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 49152, 0);
  return(rc);
}


nasd_status_t
nasd_cl_p_ctrl_get_drive_ioqueue_stats(
  nasd_drive_handle_t        handle,
  nasd_key_t                 in_key,
  nasd_capability_args_t     cap_args,
  nasd_ctrl_ioqueue_stat_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 57344, 1);
  return(rc);
}


nasd_status_t
nasd_cl_p_ctrl_put_drive_ioqueue_stats(
  nasd_drive_handle_t        handle,
  nasd_key_t                 in_key,
  nasd_capability_args_t     cap_args,
  nasd_ctrl_ioqueue_stat_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_DRIVE_INFO,
    in_key, cap_args, info, 57344, 0);
  return(rc);
}


nasd_status_t
nasd_cl_p_ctrl_get_drive_trace_basic(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_capability_args_t    cap_args,
  nasd_ctrl_trace_basic_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_TRACE_INFO,
    in_key, cap_args, info, 0, 1);
  return(rc);
}

nasd_status_t
nasd_cl_p_ctrl_put_drive_trace_basic(
  nasd_drive_handle_t       handle,
  nasd_key_t                in_key,
  nasd_capability_args_t    cap_args,
  nasd_ctrl_trace_basic_t  *info)
{
  nasd_status_t rc;

  rc = nasd_cl_p_ctrl_rdwr_stats(handle, 0, NASD_CTRL_TRACE_INFO,
    in_key, cap_args, info, 0, 0);
  return(rc);
}

/*
 * END CONTROL WRAPPERS
 *
 * BEGIN ERROR WRAPPERS
 */

char *
_nasd_cl_error_string(
  nasd_drive_handle_t   handle,
  nasd_rpc_status_t     status,
  nasd_error_string_t   str,
  char                 *in_file,
  int                   line)
{
  int dce_ps;
  char *file;
  int len;

  /*
   * Be paranoid about fitting filename in string. Truncate to 80
   * characters. Use the last 80 characters instead of the first
   * to try to be maximally informitive.
   */
  len = strlen(in_file);
  if (len > 80) {
    file = &in_file[len-80];
  }
  else {
    file = in_file;
  }

  if (handle == NULL) {
    sprintf(str,
      "NULL handle passed to nasd_cl_error_string file [%s] line %d",
      file, line);
  }
  else if (status == 999) {
    /*
     * !!! Nasty hack. See nasd_od_rpc.h for this
     * in action.
     */
    strcpy(str, "Drive offline");
  }
  else {
    switch(handle->real_type) {
      case NASD_BIND_DCE_DIRECT:
      case NASD_BIND_DCE_DIRECT_TCP:
      case NASD_BIND_DCE_KPDEV:
      case NASD_BIND_DCE_KPDEV_FAKE_RW:
#if NASD_CL_P_SWITCH_NO_SAL > 0
        dce_error_inq_text(status, (unsigned char *)str, &dce_ps);
        if (dce_ps) {
          sprintf(str, "Unknown DCE status 0x%x", status);
        }
#else /* NASD_CL_P_SWITCH_NO_SAL > 0 */
        sprintf(str,
          "DCE status 0x%x not translatable with standalone NASD library",
          status);
#endif /* NASD_CL_P_SWITCH_NO_SAL > 0 */
        break;
      case NASD_BIND_COLOCATE:
        switch(status) {
          case 0:
            strcpy(str, "Success");
            break;
          case 999:
            strcpy(str, "Drive offline");
            break;
          default:
            sprintf(str, "Unknown error %u", status);
            break;
        }
        break;
#ifndef KERNEL
      case NASD_BIND_MSGQ:
        switch(status) {
        case 0:
          strcpy(str, "Success");
          break;
        default:
          strcpy(str, strerror(status));
          break;
        }
        break;
#endif /* !KERNEL */
      default:
        sprintf(str,
          "Unknown handle type %d passed to "
          "nasd_cl_error_string file %s line %d",
          handle->real_type, file, line);
        break;
    }
  }

  return(str);
}

/*
 * Support routines
 */

#ifdef KERNEL

/*
 * Pipe routines for colocated drive
 */

nasd_status_t
nasd_co_pull(
  void        *state_arg,
  void        *buf,
  nasd_len_t   in_len,
  nasd_len_t  *out_lenp)
{
  nasd_co_pipe_state_t *state;
  nasd_len_t len;

  state = (nasd_co_pipe_state_t *)state_arg;
  len = NASD_MIN(in_len, state->buf_len_remain);
  if (len) {
#if defined(DEC_OSF) && defined(KERNEL) && 1
    register vm_offset_t s0, s1, partial, d0, d1;
    /*
     * Maintain this ordering of checks- client buffer more
     * likely to be unaligned, and AXP predicts branch-taken
     */
    if ((!page_aligned(state->cur_buf)) || (!page_aligned(buf))) {
      bcopy((char *)state->cur_buf, (char *)buf, len);
    }
    else {
      s0 = ((vm_offset_t)state->cur_buf) + (vm_offset_t)len;
      s1 = trunc_page(s0);
      partial = s0 - s1;
      if (partial) {
        d0 = ((vm_offset_t)buf) + (vm_offset_t)len;
        d1 = trunc_page(d0);
        NASD_ASSERT(((int)partial) >= 0);
        bcopy((char *)s1, (char *)d1, (int)partial);
      }
      nasd_dux_pgs_cpy((vm_offset_t)state->cur_buf, (vm_offset_t)buf, s1);
    }
#else /* DEC_OSF && KERNEL */
    bcopy((char *)state->cur_buf, (char *)buf, len);
#endif /* DEC_OSF && KERNEL */
    state->buf_len_remain -= len;
    state->cur_buf += len;
  }
  *out_lenp = len;
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_co_push(
  void        *state_arg,
  void        *buf,
  nasd_len_t   in_len)
{
  nasd_co_pipe_state_t *state;
  int len;

  state = (nasd_co_pipe_state_t *)state_arg;
  len = NASD_MIN(in_len, state->buf_len_remain);
  if (len) {
#if defined(DEC_OSF) && defined(KERNEL)
    register vm_offset_t s0, s1, partial, d0, d1;
    /*
     * Maintain this ordering of checks- client buffer more
     * likely to be unaligned, and AXP predicts branch-taken
     */
    if ((!page_aligned(state->cur_buf)) || (!page_aligned(buf))) {
      bcopy((char *)buf, (char *)state->cur_buf, len);
    }
    else {
      s0 = ((vm_offset_t)buf) + (vm_offset_t)len;
      s1 = trunc_page(s0);
      partial = s0 - s1;
      if (partial) {
        d0 = ((vm_offset_t)state->cur_buf) + (vm_offset_t)len;
        d1 = trunc_page(d0);
        NASD_ASSERT(((int)partial) >= 0);
        bcopy((char *)s1, (char *)d1, (int)partial);
      }
      nasd_dux_pgs_cpy((vm_offset_t)buf, (vm_offset_t)state->cur_buf, s1);
    }
#else /* DEC_OSF && KERNEL */
    bcopy((char *)buf, (char *)state->cur_buf, len);
#endif /* DEC_OSF && KERNEL */
    state->buf_len_remain -= len;
    state->cur_buf += len;
  }
  return(NASD_SUCCESS);
}

#endif /* KERNEL */

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
