/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Howard Gobioff
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/***********************************************************************
 *
 * nasd_security_capability.c
 *
 * This module contains functions to help file managers in generating
 * capabilities.
 *
 ***********************************************************************/

#include <nasd/nasd_options.h>
#if defined(_KERNEL) && defined(DEC_OSF)
#include <sys/kernel.h>
#endif /* _KERNEL && DEC_OSF */
#include <nasd/nasd_general.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_net_convert.h>
#include <nasd/nasd_convert_types.h>
#include <nasd/nasd_security_capability.h>
#include <nasd/nasd_keymgmt.h>
#include <cryptolib/libcrypt.h>

#include <sys/time.h>
#include <sys/resource.h>
#ifndef _KERNEL
#include <string.h>
#endif /* _KERNEL */
#define PRINT_ERROR

#ifdef DEBUG_ALL
#define PRINT_ERROR
#define DEBUG_BUILD_CAPABILITY
#endif

void
nasd_security_print_capability(
  nasd_cookie_t  *cookie)
{

printf("Capability {[%08" NASD_64x_FMT "],partnum=%d,ni=%ld,expire=%d,rights=%x,min_prot=%x,drive_name=%d,AuditID=%d,type=%d,region_start=%lu,region_end=%lu}\n ",((nasd_uint64 *)cookie->key)[0],
       cookie->capability_args.partnum,
       cookie->capability_args.ni,
       cookie->capability_args.expiration_seconds,
       cookie->capability_args.rights,

       cookie->capability_args.min_protection,
       cookie->capability_args.drive_name,
       cookie->capability_args.AuditID,
       cookie->capability_args.type,

       cookie->capability_args.region_start,
       cookie->capability_args.region_end      
);

}
void nasd_security_build_capability(nasd_drive_name_t  drive_name,
                                    nasd_partnum_t partnum,
				    nasd_identifier_t ni,
				    nasd_access_rights_t rights,
				    nasd_audit_t AuditID,
				    nasd_int32 expiration,
				    nasd_uint16 min_protection,
                                    nasd_uint16 type,
                                    nasd_offset_t region_start,
                                    nasd_offset_t region_end,
				    nasd_av_t av,
				    nasd_cookie_t* out_cookie)
{

  nasd_byte argsbuf[sizeof(nasd_capability_args_t)];
  nasd_byte avbuf[sizeof(nasd_av_t)];
  nasd_byte key_buff[SHA_DIGEST_LENGTH];
  nasd_key_t  basis_key;
  HMAC_SHA1_CTX HMAC_context;

  bzero((char *)out_cookie,sizeof(nasd_cookie_t));
  out_cookie->capability_args.ni=ni;
  out_cookie->capability_args.expiration_seconds=expiration;
  out_cookie->capability_args.rights=rights;
  out_cookie->capability_args.min_protection=min_protection;
  out_cookie->capability_args.drive_name=drive_name;
  out_cookie->capability_args.AuditID=AuditID;
  out_cookie->capability_args.type=type;
  out_cookie->capability_args.region_start=region_start;
  out_cookie->capability_args.region_end=region_end;
  out_cookie->capability_args.partnum=partnum;
#ifdef DEBUG_BUILD_CAPABILITY
  nasd_security_print_capability(out_cookie);
#endif
#if NASD_SECURE_RPCS_ENABLE > 0
  if (min_protection != NASD_NO_PROTECTION) {
 
    nasd_capability_args_t_to_net(&out_cookie->capability_args,argsbuf);
    nasd_av_t_to_net(&av,avbuf);
    
    if (type == NASD_RED_CAPABILITY)
      nasd_security_get_red_key(out_cookie->capability_args.partnum,
                       basis_key);
    else
      nasd_security_get_black_key(out_cookie->capability_args.partnum,
                       basis_key);
  
    
    HMAC_SHA1_Init(&HMAC_context,(char *)basis_key,sizeof(nasd_key_t));
    HMAC_SHA1_Update(&HMAC_context,(unsigned char *)argsbuf,sizeof(nasd_capability_args_t));
    HMAC_SHA1_Update(&HMAC_context,(unsigned char *)avbuf,sizeof(nasd_av_t));
    HMAC_SHA1_Final((unsigned char *)key_buff,&HMAC_context);
    
  

    bcopy(key_buff,out_cookie->key,NASD_MIN(SHA_DIGEST_LENGTH,sizeof(nasd_key_t))); 
    
#ifdef DEBUG_BUILD_CAPABILITY

    printf("DEBUG(nasd_security_build_capability): digest=[%08" NASD_64x_FMT "]\n",((nasd_uint64 *)out_cookie->key)[0]   );

    printf("DEBUG(nasd_security_build_capability): %d byte keys\n",sizeof(nasd_key_t));
#endif
  }
#else /* NASD_SECURE_RPCS_ENABLE > 0 */
  bzero(out_cookie->key,sizeof(nasd_key_t));
  
#endif /*NASD_SECURE_RPCS_ENABLE > 0*/

}


