/*
 * nasd_dux_basic.c
 *
 * Basic system support stuff
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_timer.h>
#include <sys/time.h>
#include <sys/sysinfo.h>
#include <machine/rpb.h>
#include <machine/hal/hal_sysinfo.h>

#ifdef _KERNEL
#include <sys/kernel.h>
#endif /* _KERNEL */

#ifndef KERNEL
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <nlist.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#endif /* !KERNEL */

#if NASD_CMU_PDL > 0
#include <pdllib.h>
#else /* NASD_CMU_PDL > 0 */
#include <sys/param.h>
#endif /* NASD_CMU_PDL > 0 */

nasd_uint64 nasd_ctimer_ticks;

extern int syscall(int number, int arg, ...);

static nasd_timespec_t nasd_dux_last_time = {0,0};
static nasd_timespec_t nasd_dux_uniq_time = {0,0};

void
nasd_gettime(
  nasd_timespec_t  *tm)
{
  struct timeval tv;

#ifdef KERNEL
  TIME_READ(tv);
#else /* KERNEL */
#if NASD_CMU_PDL > 0
  PDL_mmclock_gettime(&tv);
#else /* NASD_CMU_PDL > 0 */
  gettimeofday(&tv, NULL);
#endif /* NASD_CMU_PDL > 0 */
#endif /* KERNEL */
  tm->ts_sec = tv.tv_sec;
  tm->ts_nsec = tv.tv_usec*1000;
  if ((tm->ts_sec == nasd_dux_last_time.ts_sec)
    && (tm->ts_nsec == nasd_dux_last_time.ts_nsec))
  {
    nasd_dux_uniq_time.ts_nsec++;
    if (nasd_dux_uniq_time.ts_nsec >= NASD_NSEC_PER_SEC) {
      nasd_dux_uniq_time.ts_nsec -= NASD_NSEC_PER_SEC;
      nasd_dux_uniq_time.ts_sec++;
    }
    *tm = nasd_dux_uniq_time;
  }
  else {
    nasd_dux_last_time = *tm;
    nasd_dux_uniq_time = nasd_dux_last_time;
  }
}

#ifndef KERNEL

nasd_status_t
nasd_sys_init()
{
#if NASD_CMU_PDL > 0
  int ret;

  ret = PDL_get_clock_speed(&nasd_ctimer_ticks);
  if (ret) {
    perror("PDL_get_clock_speed");
    return(NASD_FAIL);
  }
#else /* NASD_CMU_PDL > 0 */
  char *kmemdevname, buf[sizeof(struct rpb)+8];
  char *memdevname, kernel_name[MAXPATHLEN+1];
  struct nlist nl[2], *np;
  unsigned long rpb_addr;
  nasd_status_t rc;
  int kfd, fd, ret;
  struct rpb rpb;
  off_t off;

  kmemdevname = "/dev/kmem";
  memdevname = "/dev/mem";

  np = &nl[0];
  bzero((char *)np, sizeof(nl));
  nl[0].n_name = "pmap_physhwrpb";
  nl[1].n_name = NULL;

  rc = NASD_SUCCESS;

  /* get running kernel name */
  bzero(kernel_name, MAXPATHLEN+1);
  kernel_name[0] = '/';
  ret = getsysinfo(GSI_BOOTEDFILE, &kernel_name[1], MAXPATHLEN, 0, 0);
  if (ret != 1) {
    /* cannot get booted kernel name */
    return(NASD_FAIL);
  }

  ret = nlist(kernel_name, np);
  if (ret) {
    /* cannot nlist kernel */
    return(NASD_FAIL);
  }

  if (np->n_type == 0) {
    /* didn't find the symbols we want */
    return(NASD_FAIL);
  }

  kfd = open(kmemdevname, O_RDONLY);
  if (kfd < 0) {
    perror(kmemdevname);
    return(NASD_FAIL);
  }
  fd = open(memdevname, O_RDONLY);
  if (fd < 0) {
    perror(memdevname);
    return(NASD_FAIL);
  }

  /*
   * pmap_physhwrpb is a variable in the kernel containing the physical
   * address of the hardware RPB. We'll just find that variable and
   * read it, then use that as a physical memory address to read the
   * rpb itself.
   */

  off = lseek(kfd, np->n_value, SEEK_SET);
  if (off != np->n_value) {
    /* cannot seek to address of hwrpb addr */
    return(NASD_FAIL);
  }

  ret = read(kfd, &rpb_addr, sizeof(rpb_addr));
  if (ret != sizeof(rpb_addr)) {
    /* cannot read address of hwrpb addr */
    rc = NASD_FAIL;
    goto isbad;
  }

  off = lseek(fd, rpb_addr, SEEK_SET);
  if (off != rpb_addr) {
    /* cannot seek to rpb addr */
    rc = NASD_FAIL;
    goto isbad;
  }

  ret = read(fd, &rpb, sizeof(rpb));
  if (ret != sizeof(rpb)) {
    /* cannot read rpb */
    rc = NASD_FAIL;
    goto isbad;
  }

  /*
   * One extra sanity check: the RPB is self-identifying.
   * This field is guaranteed to have the value
   * 0x0000004250525748, always.
   */
  if (rpb.rpb_string != 0x0000004250525748) {
    rc = NASD_FAIL;
    goto isbad;
  }

isbad:
  if (rc) {
    /*
     * rpb failed validation
     * most likely, the kernel binary has been modified since
     * it was booted
     */
    close(kfd);
    close(fd);
    return(rc);
  }

  nasd_ctimer_ticks = rpb.rpb_counter;

  close(kfd);
  close(fd);
#endif /* NASD_CMU_PDL > 0 */
  return(NASD_SUCCESS);
}

/*
 * nasd_srv
 * (user-level only)
 *
 * nasd_srv() system call: takes two arguments
 * int cmd, and void *arg
 */
int
nasd_srv(cmd, arg)
  int    cmd;
  void  *arg;
{
  int rc;

  rc = syscall(247, cmd, arg);
  return(rc);
}

/*
 * (user-level only)
 */
nasd_status_t
nasd_gethostbyname_r(
  char            *hostname,
  struct hostent  *hep)
{
  struct hostent_data hd;
  int rc;

  bzero((char *)&hd, sizeof(hd));
  rc = gethostbyname_r(hostname, hep, &hd);
  if (rc)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

#if NASD_CMU_PDL > 0
nasd_status_t
nasd_get_total_idle_time(
  nasd_timespec_t  *ts)
{
  struct timespec tt;
  int ret;

  ret = PDL_get_total_idle_time(&tt);
  if (ret)
    return(NASD_FAIL);

  ts->ts_sec = tt.tv_sec;
  ts->ts_nsec = tt.tv_nsec;

  return(NASD_SUCCESS);
}
#endif /* NASD_CMU_PDL > 0 */

#endif /* !KERNEL */

#ifdef KERNEL

nasd_status_t
nasd_sys_init()
{
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threads_init()
{
  /* nothing to do here */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threads_init_once()
{
  /* nothing to do here */
  return(NASD_SUCCESS);
}

void
nasd_sys_threads_shutdown()
{
  /* nothing to do here */
}

decl_simple_lock_data(,nasd_once_lock)

nasd_status_t
nasd_sys_mutex_init(m)
  decl_simple_lock_data(,*m)
{
  simple_lock_init(m);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_mutex_destroy(m)
  decl_simple_lock_data(,*m)
{
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_cond_init(c)
  NASD_DECLARE_COND(*c)
{
  *c = 0; /* no reason */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_cond_destroy(c)
  NASD_DECLARE_COND(*c)
{
  return(NASD_SUCCESS);
}

int
_nasd_once(
  nasd_once_t   *once_block,
  void         (*init_routine)(void),
  char          *file,
  int            line)
{
  int s;

  s = splextreme();
  simple_lock(&nasd_once_lock);
  if (once_block->done == 0) {
    once_block->done = 1;
    (*init_routine)();
  }
  simple_unlock(&nasd_once_lock);
  splx(s);
}

nasd_status_t
nasd_sys_threadattr_create(
  nasd_sys_threadattr_t  *attrp)
{
  attrp->dummy = 0;
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threadattr_destroy(
  nasd_sys_threadattr_t  *attrp)
{
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_threadattr_setstacksize(
  nasd_sys_threadattr_t  *attrp,
  int                     size)
{
  return(NASD_OP_NOT_SUPPORTED);
}

nasd_status_t
nasd_sys_threadattr_getstacksize(
  nasd_sys_threadattr_t  *attrp,
  int                    *sizep)
{
  *sizep = KERNEL_STACK_SIZE;
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_thread_create_attr(
  nasd_sys_thread_t      *threadp,
  nasd_sys_threadattr_t  *attrp,
  nasd_threadfunc_t       func,
  nasd_threadarg_t        arg)
{
  *threadp = kernel_thread_w_arg(first_task, (void (*)())func, arg);
  if (*threadp == NULL)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_sys_thread_create(
  nasd_sys_thread_t  *threadp,
  nasd_threadfunc_t   func,
  nasd_threadarg_t    arg)
{
  *threadp = kernel_thread_w_arg(first_task, (void (*)())func, arg);
  if (*threadp == NULL)
    return(NASD_FAIL);
  return(NASD_SUCCESS);
}

#if NASD_USE_DUX_RW_LOCKS > 0

nasd_status_t
nasd_rwlock_init(l)
  NASD_DECLARE_RWLOCK(*l)
{
  lock_setup(l, NULL, TRUE);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_rwlock_destroy(l)
  NASD_DECLARE_RWLOCK(*l)
{
  lock_terminate(l);
  return(NASD_SUCCESS);
}

#endif /* NASD_USE_DUX_RW_LOCKS > 0 */

#if TRACK_IDLE > 0
nasd_status_t
nasd_get_total_idle_time(
  nasd_timespec_t  *ts)
{
  extern timespec_t idle_time;

  ts->ts_sec = idle_time.tv_sec;
  ts->ts_nsec = idle_time.tv_nsec;
  return(NASD_SUCCESS);
}
#endif /* TRACK_IDLE > 0 */

#endif /* KERNEL */

nasd_status_t
nasd_get_clock_speed(
  nasd_uint64  *ticksp)
{
  *ticksp = nasd_ctimer_ticks;
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
