/*
 * nasd_marshall.c
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/***********************************************************************
 *
 * nasd_marshall.c
 *
 * To do our own marshalling and unmarshalling (for security and to be fast)
 *
 ***********************************************************************/

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>

#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_net_convert.h>
#ifdef _KERNEL
/* XXX - including this file causes a kernel compile to blow up */
#else /* _KERNEL */
#include <nasd/nasd_convert_types.h>
#endif /* _KERNEL */

#ifdef _KERNEL
/* XXX - what to do about assert() in the kernel? */
#define assert(ignore) ((void)0)
#else /* _KERNEL */
#include <assert.h>
#endif /* _KERNEL */

/*
 * There are three types of RPCs and two types of marshalling to support them
 *  (1) RPCs with fixed-size argument structures and fixed-size result 
 *      structures (e.g. GETATTR)
 *  (2) RPCs with variable-sizes argument structures and fixed-size result
 *      structures (e.g. GRANT).  NOTE: the argument list contains an element
 *      that is variable-sized, but has a defined maximum size that is
 *      defined at compile-time and enforced
 *  (3) RPCs of type (1) or (2) with an additional byte_pipe structure for
 *      a variable-sized data "buffer" used for large data chunks 
 *      (e.g. FASTREAD is of type (1) with a byte_pipe) the pipe is not
 *      marshalled and is a separate parameter to the RPC
 *
 * NOTE: given the decision to go with fixed-maximum-size for variable-sized
 * argument lists, it is also possible to support variable-sized result
 * structures.  There simply are no RPCs that currently do this
 *
 */

/* GETATTR */
void nasd_p_getattr_dr_marshall_args(nasd_p_getattr_dr_args_t *in,
				   nasd_p_getattr_dr_args_otw_t out)
{
  nasd_p_getattr_dr_args_t_to_net(in,out);
}

void nasd_p_getattr_dr_unmarshall_args(nasd_p_getattr_dr_args_otw_t in,
				     nasd_p_getattr_dr_args_t *out)
{
  nasd_p_getattr_dr_args_t_from_net(in,out);
}

void nasd_p_getattr_dr_marshall_res(nasd_p_getattr_dr_res_t *in,
				  nasd_p_getattr_dr_res_otw_t out)
{
  nasd_p_getattr_dr_res_t_to_net(in,out);
}

void nasd_p_getattr_dr_unmarshall_res(nasd_p_getattr_dr_res_otw_t in,
				    nasd_p_getattr_dr_res_t *out)
{
  nasd_p_getattr_dr_res_t_from_net(in,out);
}



/* SETATTR */
void nasd_p_setattr_dr_marshall_args(nasd_p_setattr_dr_args_t *in,
				   nasd_p_setattr_dr_args_otw_t out)
{
  nasd_p_setattr_dr_args_t_to_net(in,out);
}

void nasd_p_setattr_dr_unmarshall_args(nasd_p_setattr_dr_args_otw_t in,
				     nasd_p_setattr_dr_args_t *out)
{
  nasd_p_setattr_dr_args_t_from_net(in,out);
}

void nasd_p_setattr_dr_marshall_res(nasd_p_setattr_dr_res_t *in,
				  nasd_p_setattr_dr_res_otw_t out)
{
  nasd_p_setattr_dr_res_t_to_net(in,out);
}

void nasd_p_setattr_dr_unmarshall_res(nasd_p_setattr_dr_res_otw_t in,
				    nasd_p_setattr_dr_res_t *out)
{
  nasd_p_setattr_dr_res_t_from_net(in,out);
}

/* CREATE */

void nasd_p_create_dr_marshall_args(nasd_p_create_dr_args_t *in,
				   nasd_p_create_dr_args_otw_t out)
{
  nasd_p_create_dr_args_t_to_net(in,out);
}

void nasd_p_create_dr_unmarshall_args(nasd_p_create_dr_args_otw_t in,
				     nasd_p_create_dr_args_t *out)
{
  nasd_p_create_dr_args_t_from_net(in,out);
}

void nasd_p_create_dr_marshall_res(nasd_p_create_dr_res_t *in,
				  nasd_p_create_dr_res_otw_t out)
{
  nasd_p_create_dr_res_t_to_net(in,out);
}

void nasd_p_create_dr_unmarshall_res(nasd_p_create_dr_res_otw_t in,
				    nasd_p_create_dr_res_t *out)
{
  nasd_p_create_dr_res_t_from_net(in,out);
}



/* FASTREAD */
void nasd_p_fastread_dr_marshall_res(nasd_p_fastread_dr_res_t *in,
				   nasd_p_fastread_dr_res_otw_t out)
{
  nasd_p_fastread_dr_res_t_to_net(in,out);
}

void nasd_p_fastread_dr_unmarshall_res(nasd_p_fastread_dr_res_otw_t in,
				     nasd_p_fastread_dr_res_t *out)
{
  nasd_p_fastread_dr_res_t_from_net(in,out);
}



/* FASTWRITE */
void nasd_p_fastwrite_dr_marshall_res(nasd_p_fastwrite_dr_res_t *in,
				   nasd_p_fastwrite_dr_res_otw_t out)
{
  nasd_p_fastwrite_dr_res_t_to_net(in,out);
}

void nasd_p_fastwrite_dr_unmarshall_res(nasd_p_fastwrite_dr_res_otw_t in,
				     nasd_p_fastwrite_dr_res_t *out)
{
  nasd_p_fastwrite_dr_res_t_from_net(in,out);
}

/* PART_CREAT */

void
nasd_p_part_creat_dr_marshall_args(
  nasd_p_part_creat_dr_args_t      *in,
  nasd_p_part_creat_dr_args_otw_t   out)
{
  nasd_p_part_creat_dr_args_t_to_net(in, out);
}

void
nasd_p_part_creat_dr_unmarshall_args(
  nasd_p_part_creat_dr_args_otw_t   in,
  nasd_p_part_creat_dr_args_t      *out)
{
  nasd_p_part_creat_dr_args_t_from_net(in, out);
}

void
nasd_p_part_creat_dr_marshall_res(
  nasd_p_part_creat_dr_res_t      *in,
  nasd_p_part_creat_dr_res_otw_t   out)
{
  nasd_p_part_creat_dr_res_t_to_net(in, out);
}

void
nasd_p_part_creat_dr_unmarshall_res(
  nasd_p_part_creat_dr_res_otw_t   in,
  nasd_p_part_creat_dr_res_t      *out)
{
  nasd_p_part_creat_dr_res_t_from_net(in, out);
}

/* SMPL_OP */

void
nasd_p_smpl_op_dr_marshall_args(
  nasd_p_smpl_op_dr_args_t      *in,
  nasd_p_smpl_op_dr_args_otw_t   out)
{
  nasd_p_smpl_op_dr_args_t_to_net(in, out);
}

void
nasd_p_smpl_op_dr_unmarshall_args(
  nasd_p_smpl_op_dr_args_otw_t   in,
  nasd_p_smpl_op_dr_args_t      *out)
{
  nasd_p_smpl_op_dr_args_t_from_net(in, out);
}


/* THRTL_OP */

void
nasd_p_thrtl_op_dr_marshall_args(
  nasd_p_thrtl_op_dr_args_t      *in,
  nasd_p_thrtl_op_dr_args_otw_t   out)
{
  nasd_p_thrtl_op_dr_args_t_to_net(in, out);
}

void
nasd_p_thrtl_op_dr_unmarshall_args(
  nasd_p_thrtl_op_dr_args_otw_t   in,
  nasd_p_thrtl_op_dr_args_t      *out)
{
  nasd_p_thrtl_op_dr_args_t_from_net(in, out);
}


/* FLUSH_OBJ */

void
nasd_p_flush_obj_dr_marshall_args(
  nasd_p_flush_obj_dr_args_t      *in,
  nasd_p_flush_obj_dr_args_otw_t   out)
{
  nasd_p_flush_obj_dr_args_t_to_net(in, out);
}

void
nasd_p_flush_obj_dr_unmarshall_args(
  nasd_p_flush_obj_dr_args_otw_t   in,
  nasd_p_flush_obj_dr_args_t      *out)
{
  nasd_p_flush_obj_dr_args_t_from_net(in, out);
}

void
nasd_p_flush_obj_dr_marshall_res(
  nasd_p_flush_obj_dr_res_t      *in,
  nasd_p_flush_obj_dr_res_otw_t   out)
{
  nasd_p_flush_obj_dr_res_t_to_net(in, out);
}

void
nasd_p_flush_obj_dr_unmarshall_res(
  nasd_p_flush_obj_dr_res_otw_t   in,
  nasd_p_flush_obj_dr_res_t      *out)
{
  nasd_p_flush_obj_dr_res_t_from_net(in,out);
}

/* EJECT_OBJ */

void
nasd_p_eject_obj_dr_marshall_args(
  nasd_p_eject_obj_dr_args_t      *in,
  nasd_p_eject_obj_dr_args_otw_t   out)
{
  nasd_p_eject_obj_dr_args_t_to_net(in, out);
}

void
nasd_p_eject_obj_dr_unmarshall_args(
  nasd_p_eject_obj_dr_args_otw_t   in,
  nasd_p_eject_obj_dr_args_t      *out)
{
  nasd_p_eject_obj_dr_args_t_from_net(in, out);
}

void
nasd_p_eject_obj_dr_marshall_res(
  nasd_p_eject_obj_dr_res_t      *in,
  nasd_p_eject_obj_dr_res_otw_t   out)
{
  nasd_p_eject_obj_dr_res_t_to_net(in, out);
}

void
nasd_p_eject_obj_dr_unmarshall_res(
  nasd_p_eject_obj_dr_res_otw_t   in,
  nasd_p_eject_obj_dr_res_t      *out)
{
  nasd_p_eject_obj_dr_res_t_from_net(in,out);
}

/* REMOVE */

void
nasd_p_remove_dr_marshall_args(
  nasd_p_remove_dr_args_t      *in,
  nasd_p_remove_dr_args_otw_t   out)
{
  nasd_p_remove_dr_args_t_to_net(in, out);
}

void
nasd_p_remove_dr_unmarshall_args(
  nasd_p_remove_dr_args_otw_t   in,
  nasd_p_remove_dr_args_t      *out)
{
  nasd_p_remove_dr_args_t_from_net(in, out);
}

void
nasd_p_remove_dr_marshall_res(
  nasd_p_remove_dr_res_t      *in,
  nasd_p_remove_dr_res_otw_t   out)
{
  nasd_p_remove_dr_res_t_to_net(in, out);
}

void
nasd_p_remove_dr_unmarshall_res(
  nasd_p_remove_dr_res_otw_t   in,
  nasd_p_remove_dr_res_t      *out)
{
  nasd_p_remove_dr_res_t_from_net(in, out);
}







/* INITIALIZE */

void
nasd_p_initialize_dr_marshall_args(
  nasd_p_initialize_dr_args_t      *in,
  nasd_p_initialize_dr_args_otw_t   out)
{
  nasd_p_initialize_dr_args_t_to_net(in, out);
}

void
nasd_p_initialize_dr_unmarshall_args(
  nasd_p_initialize_dr_args_otw_t   in,
  nasd_p_initialize_dr_args_t      *out)
{
  nasd_p_initialize_dr_args_t_from_net(in, out);
}

void
nasd_p_initialize_dr_marshall_res(
  nasd_p_initialize_dr_res_t      *in,
  nasd_p_initialize_dr_res_otw_t   out)
{
  nasd_p_initialize_dr_res_t_to_net(in, out);
}

void
nasd_p_initialize_dr_unmarshall_res(
  nasd_p_initialize_dr_res_otw_t   in,
  nasd_p_initialize_dr_res_t      *out)
{
  nasd_p_initialize_dr_res_t_from_net(in, out);
}





/* CHANGE_KEY */

void
nasd_p_change_key_dr_marshall_args(
  nasd_p_change_key_dr_args_t      *in,
  nasd_p_change_key_dr_args_otw_t   out)
{
  nasd_p_change_key_dr_args_t_to_net(in, out);
}

void
nasd_p_change_key_dr_unmarshall_args(
  nasd_p_change_key_dr_args_otw_t   in,
  nasd_p_change_key_dr_args_t      *out)
{
  nasd_p_change_key_dr_args_t_from_net(in, out);
}

void
nasd_p_change_key_dr_marshall_res(
  nasd_p_change_key_dr_res_t      *in,
  nasd_p_change_key_dr_res_otw_t   out)
{
  nasd_p_change_key_dr_res_t_to_net(in, out);
}

void
nasd_p_change_key_dr_unmarshall_res(
  nasd_p_change_key_dr_res_otw_t   in,
  nasd_p_change_key_dr_res_t      *out)
{
  nasd_p_change_key_dr_res_t_from_net(in, out);
}

/* NFS -- LOOKUP */
void nasd_nfs_lookup_marshall_args(nasd_nfs_lookup_args_t *in,
				   nasd_nfs_lookup_args_otw_t out)
{
  nasd_nfs_lookup_args_t_to_net(in,out);
}

void nasd_nfs_lookup_unmarshall_args(nasd_nfs_lookup_args_otw_t in,
				     nasd_nfs_lookup_args_t *out)
{
  nasd_nfs_lookup_args_t_from_net(in,out);
}

void nasd_nfs_lookup_marshall_res(nasd_nfs_lookup_res_t *in,
				  nasd_nfs_lookup_res_otw_t out)
{
  nasd_nfs_lookup_res_t_to_net(in,out);
}

void nasd_nfs_lookup_unmarshall_res(nasd_nfs_lookup_res_otw_t in,
				    nasd_nfs_lookup_res_t *out)
{
  nasd_nfs_lookup_res_t_from_net(in,out);
}










/*
 * Control structures
 */

/* DRIVE_INFO */

void
nasd_drive_info_marshall_ctrl(
  nasd_ctrl_drive_info_t      *in,
  nasd_ctrl_drive_info_otw_t   out)
{
  nasd_ctrl_drive_info_to_net(in, out);
}

void
nasd_drive_info_unmarshall_ctrl(
  nasd_ctrl_drive_info_otw_t   in,
  nasd_ctrl_drive_info_t      *out)
{
  nasd_ctrl_drive_info_from_net(in, out);
}

/* PART_INFO */

void
nasd_part_info_marshall_ctrl(
  nasd_ctrl_part_info_t      *in,
  nasd_ctrl_part_info_otw_t   out)
{
  nasd_ctrl_part_info_to_net(in, out);
}

void
nasd_part_info_unmarshall_ctrl(
  nasd_ctrl_part_info_otw_t   in,
  nasd_ctrl_part_info_t      *out)
{
  nasd_ctrl_part_info_from_net(in, out);
}

/* DRIVE_OPSTATS */

void
nasd_drive_opstats_marshall_ctrl(
  nasd_ctrl_drive_opstats_t      *in,
  nasd_ctrl_drive_opstats_otw_t   out)
{
  nasd_ctrl_drive_opstats_to_net(in, out);
}

void
nasd_drive_opstats_unmarshall_ctrl(
  nasd_ctrl_drive_opstats_otw_t   in,
  nasd_ctrl_drive_opstats_t      *out)
{
  nasd_ctrl_drive_opstats_from_net(in, out);
}

/* CACHE_STAT */

void
nasd_cache_stat_marshall_ctrl(
  nasd_ctrl_cache_stat_t      *in,
  nasd_ctrl_cache_stat_otw_t   out)
{
  nasd_ctrl_cache_stat_to_net(in, out);
}

void
nasd_cache_stat_unmarshall_ctrl(
  nasd_ctrl_cache_stat_otw_t   in,
  nasd_ctrl_cache_stat_t      *out)
{
  nasd_ctrl_cache_stat_from_net(in, out);
}

/* LAYOUT_STAT */

void
nasd_layout_stat_marshall_ctrl(
  nasd_ctrl_layout_stat_t      *in,
  nasd_ctrl_layout_stat_otw_t   out)
{
  nasd_ctrl_layout_stat_to_net(in, out);
}

void
nasd_layout_stat_unmarshall_ctrl(
  nasd_ctrl_layout_stat_otw_t   in,
  nasd_ctrl_layout_stat_t      *out)
{
  nasd_ctrl_layout_stat_from_net(in, out);
}

/* IOQUEUE_STAT */

void
nasd_ioqueue_stat_marshall_ctrl(
  nasd_ctrl_ioqueue_stat_t      *in,
  nasd_ctrl_ioqueue_stat_otw_t   out)
{
  nasd_ctrl_ioqueue_stat_to_net(in, out);
}

void
nasd_ioqueue_stat_unmarshall_ctrl(
  nasd_ctrl_ioqueue_stat_otw_t   in,
  nasd_ctrl_ioqueue_stat_t      *out)
{
  nasd_ctrl_ioqueue_stat_from_net(in, out);
}

/* IO_STAT */

void
nasd_io_stat_marshall_ctrl(
  nasd_ctrl_io_stat_t      *in,
  nasd_ctrl_io_stat_otw_t   out)
{
  nasd_ctrl_io_stat_to_net(in, out);
}

void
nasd_io_stat_unmarshall_ctrl(
  nasd_ctrl_io_stat_otw_t   in,
  nasd_ctrl_io_stat_t      *out)
{
  nasd_ctrl_io_stat_from_net(in, out);
}

/* IO_SIZE_STAT */

void
nasd_io_size_stat_marshall_ctrl(
  nasd_ctrl_io_size_stat_t      *in,
  nasd_ctrl_io_size_stat_otw_t   out)
{
  nasd_ctrl_io_size_stat_to_net(in, out);
}

void
nasd_io_size_stat_unmarshall_ctrl(
  nasd_ctrl_io_size_stat_otw_t   in,
  nasd_ctrl_io_size_stat_t      *out)
{
  nasd_ctrl_io_size_stat_from_net(in, out);
}

/* TRACE_BASIC */

void
nasd_trace_basic_marshall_ctrl(
  nasd_ctrl_trace_basic_t      *in,
  nasd_ctrl_trace_basic_otw_t   out)
{
  nasd_ctrl_trace_basic_to_net(in, out);
}

void
nasd_trace_basic_unmarshall_ctrl(
  nasd_ctrl_trace_basic_otw_t   in,
  nasd_ctrl_trace_basic_t      *out)
{
  nasd_ctrl_trace_basic_from_net(in, out);
}

/* TRACE_HEADER */
void
nasd_trace_header_marshall(
  nasd_trace_header_t      *in,
  nasd_trace_header_otw_t   out)
{
  nasd_trace_header_to_net(in, out);
}

void
nasd_trace_header_unmarshall(
  nasd_trace_header_otw_t   in,
  nasd_trace_header_t      *out)
{
  nasd_trace_header_from_net(in, out);
}


/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
