/*
 * nasd_shutdown.c
 *
 * Shutdown management for NASD.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_shutdown.h>

NASD_DECLARE_ONCE(nasd_shutdown_init_once)
NASD_DECLARE_MUTEX(nasd_shutdown_use_counter_lock)
int nasd_shutdown_use_counter;

nasd_freelist_t *nasd_free_shutdown = NULL;
#define NASD_MAX_FREE_SHUTDOWN  64
#define NASD_SHUTDOWN_INC        8
#define NASD_SHUTDOWN_INITIAL   16

nasd_freelist_t *nasd_free_shutdown_list;
#define NASD_MAX_FREE_SHUTDOWN_LIST 256
#define NASD_SHUTDOWN_LIST_INC       32
#define NASD_SHUTDOWN_LIST_INITIAL  128

static nasd_shutdown_t *
nasd_shutdown_get()
{
  nasd_shutdown_t *s;

  NASD_FREELIST_GET(nasd_free_shutdown,s,next,(nasd_shutdown_t *));
  return(s);
}

static void
nasd_shutdown_give(
  nasd_shutdown_t *s)
{
  NASD_FREELIST_FREE(nasd_free_shutdown,s,next);
}

static nasd_status_t
init_shutdown_list(
  nasd_shutdown_list_t  *sl)
{
  nasd_status_t rc;

  rc = nasd_mutex_init(&sl->lock);
  return(rc);
}

static void
clean_shutdown_list(
  nasd_shutdown_list_t  *sl)
{
  nasd_mutex_destroy(&sl->lock);
}

/*
 * nasd_shutdown_sys_real_init
 *
 * Called when the system comes into use after not being
 * used (start-of-day call to nasd_shutdown_sys_init(), or first
 * such call after the last call to nasd_shutdown_cleanup()
 * which actually deactivated the system).
 */
nasd_status_t
nasd_shutdown_sys_real_init()
{
  if (nasd_free_shutdown)
    return(NASD_SUCCESS);

  NASD_FREELIST_CREATE(nasd_free_shutdown, NASD_MAX_FREE_SHUTDOWN,
    NASD_SHUTDOWN_INC, sizeof(nasd_shutdown_t));
  if (nasd_free_shutdown == NULL)
    return(NASD_NO_MEM);
  NASD_FREELIST_PRIME(nasd_free_shutdown, NASD_SHUTDOWN_INITIAL,next,
    (nasd_shutdown_t *));

  NASD_FREELIST_CREATE(nasd_free_shutdown_list, NASD_MAX_FREE_SHUTDOWN_LIST,
    NASD_SHUTDOWN_LIST_INC, sizeof(nasd_shutdown_list_t));
  if (nasd_free_shutdown_list == NULL) {
    NASD_FREELIST_DESTROY(nasd_free_shutdown,next,(nasd_shutdown_list_t *));
    return(NASD_NO_MEM);
  }
  NASD_FREELIST_PRIME_INIT(nasd_free_shutdown_list,
    NASD_SHUTDOWN_LIST_INITIAL,next,(nasd_shutdown_list_t *),
    init_shutdown_list);

  return(NASD_SUCCESS);
}

nasd_status_t
nasd_shutdown_list_init(
  nasd_shutdown_list_t  **slp)
{
  nasd_shutdown_list_t *sl;

  NASD_FREELIST_GET_INIT(nasd_free_shutdown_list,sl,next,
    (nasd_shutdown_list_t *),init_shutdown_list);
  if (sl == NULL) {
    return(NASD_NO_MEM);
  }
  sl->list = NULL;

  *slp = sl;

  return(NASD_SUCCESS);
}

nasd_status_t
nasd_shutdown_list_shutdown_internal(
  nasd_shutdown_list_t  *sl,
  int                    verbose)
{
  nasd_shutdown_t *s, *n;

  NASD_LOCK_MUTEX(sl->lock);
  for(s=sl->list;s;s=n) {
    n = s->next;
    if (s->proc) {
      if (verbose) {
        printf("Shutdown proc %lx(%lx) %s:%d\n", (unsigned long)s->proc,
          (unsigned long)s->arg, s->file, s->line);
      }
      s->proc(s->arg);
    }
    else {
      if (verbose) {
        printf("Shutdown mem %lx (%ld) %s:%d\n", (unsigned long)s->arg,
          s->size, s->file, s->line);
      }
      NASD_Free(s->arg, s->size);
    }
    nasd_shutdown_give(s);
  }
  NASD_UNLOCK_MUTEX(sl->lock);
  NASD_FREELIST_FREE_CLEAN(nasd_free_shutdown_list,sl,next,
    clean_shutdown_list);
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_shutdown_list_shutdown(
  nasd_shutdown_list_t  *sl)
{
  nasd_status_t rc;

  rc = nasd_shutdown_list_shutdown_internal(sl, NASD_DEBUG_SHUTDOWN);
  return(rc);
}

nasd_status_t
nasd_shutdown_list_shutdown_verbose(
  nasd_shutdown_list_t  *sl)
{
  nasd_status_t rc;

  rc = nasd_shutdown_list_shutdown_internal(sl, 1);
  return(rc);
}

nasd_status_t
_nasd_shutdown_proc(sl, proc, arg, file, line)
  nasd_shutdown_list_t   *sl;
  void                  (*proc)(void *);
  void                   *arg;
  char                   *file;
  int                     line;
{
  nasd_shutdown_t *s;

  s = nasd_shutdown_get();
  if (s == NULL)
    return(NASD_NO_MEM);

  s->proc = proc;
  s->arg = arg;
  s->size = 0;
  s->file = file;
  s->line = line;

  NASD_LOCK_MUTEX(sl->lock);
  s->next = sl->list;
  sl->list = s;
  NASD_UNLOCK_MUTEX(sl->lock);

  return(NASD_SUCCESS);
}

nasd_status_t
_nasd_shutdown_mem(
  nasd_shutdown_list_t   *sl,
  void                   *addr,
  int                     size,
  char                   *file,
  int                     line)
{
  nasd_shutdown_t *s;

  s = nasd_shutdown_get();
  if (s == NULL)
    return(NASD_NO_MEM);

  s->proc = NULL;
  s->arg = addr;
  s->size = size;
  s->file = file;
  s->line = line;

  NASD_LOCK_MUTEX(sl->lock);
  s->next = sl->list;
  sl->list = s;
  NASD_UNLOCK_MUTEX(sl->lock);

  return(NASD_SUCCESS);
}

/*
 * nasd_shutdown_real_cleanup
 *
 * Called when last user of the shutdown system calls
 * nasd_shutdown_real_cleanup(). Clean up and deallocate resources.
 */
void
nasd_shutdown_real_cleanup()
{
  NASD_FREELIST_DESTROY(nasd_free_shutdown,next,(nasd_shutdown_t *));
  NASD_FREELIST_DESTROY_CLEAN(nasd_free_shutdown_list,next,
    (nasd_shutdown_list_t *),clean_shutdown_list);
  nasd_free_shutdown = NULL;
  nasd_free_shutdown_list = NULL;
}

/*
 * nasd_shutdownsys_init
 *
 * Executed exactly once, the first time nasd_shutdown_sys_init() is
 * called. Initialize counter tracking number of times system
 * is initted.
 */
void
nasd_shutdownsys_init()
{
  nasd_status_t rc;

  rc = nasd_mutex_init(&nasd_shutdown_use_counter_lock);
  if (rc) {
    NASD_PANIC();
  }
  nasd_shutdown_use_counter = 0;
}

/*
 * nasd_shutdown_sys_init
 *
 * Keep a counter of the number of times we're initted and
 * shutdown. When the last shutdown arrives, really deallocate.
 * This lets multiple subsystems use us without knowing about
 * one another.
 */
nasd_status_t
nasd_shutdown_sys_init()
{
  nasd_status_t rc;

  nasd_once(&nasd_shutdown_init_once, nasd_shutdownsys_init);
  NASD_LOCK_MUTEX(nasd_shutdown_use_counter_lock);
  nasd_shutdown_use_counter++;
  if (nasd_shutdown_use_counter == 1) {
    rc = nasd_shutdown_sys_real_init();
    if (rc) {
      nasd_shutdown_use_counter = 0;
      nasd_free_shutdown = NULL;
      nasd_free_shutdown_list = NULL;
    }
  }
  else {
    rc = NASD_SUCCESS;
  }
  NASD_UNLOCK_MUTEX(nasd_shutdown_use_counter_lock);

  return(rc);
}

/*
 * nasd_shutdown_cleanup
 *
 * Previous caller of nasd_shutdown_sys_init() not using the shutdown
 * subsystem any more. Deallocate and cleanup iff necessary.
 */
void
nasd_shutdown_cleanup()
{
  NASD_ASSERT(nasd_shutdown_use_counter != 0);

  NASD_LOCK_MUTEX(nasd_shutdown_use_counter_lock);
  nasd_shutdown_use_counter--;
  if (nasd_shutdown_use_counter == 0) {
    nasd_shutdown_real_cleanup();
  }
  NASD_UNLOCK_MUTEX(nasd_shutdown_use_counter_lock);
}

void
nasd_shutdown_mutex(arg)
  void *arg;
{
  nasd_status_t rc;

  rc = nasd_mutex_destroy(arg);
  if (rc) {
    printf("WARNING: got 0x%x (%s) destroying NASD mutex 0x%lx\n",
      rc, nasd_error_string(rc), (unsigned long)arg);
  }
}

void
nasd_shutdown_cond(arg)
  void *arg;
{
  nasd_status_t rc;

  rc = nasd_cond_destroy(arg);
  if (rc) {
    printf("WARNING: got 0x%x (%s) destroying NASD cond 0x%lx\n",
      rc, nasd_error_string(rc), rc, (unsigned long)arg);
  }
}

void
nasd_shutdown_rwlock(arg)
  void *arg;
{
  nasd_status_t rc;

  rc = nasd_rwlock_destroy(arg);
  if (rc) {
    printf("WARNING: got 0x%x (%s) destroying NASD rwlock 0x%lx\n",
      rc, nasd_error_string(rc), rc, (unsigned long)arg);
  }
}

#if defined(DEC_OSF) && defined(KERNEL)
void
nasd_shutdown_page(arg)
  void  *arg;
{
  vm_page_t pp;

  pp = (vm_page_t)arg;
  vm_pg_free(pp);
}
#endif /* DEC_OSF && KERNEL */

void
nasd_shutdown_shutdown(
  void  *arg)
{
  nasd_shutdown_list_t *sl;
  nasd_status_t rc;

  sl = (nasd_shutdown_list_t *)arg;
  rc = nasd_shutdown_list_shutdown(sl);
  if (rc) {
    printf("WARNING: got 0x%x (%s) shutting down list 0x%lx\n",
      rc, nasd_error_string(rc), (unsigned long)sl);
  }
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
