/*
 * nasd_threadstuff.c
 *
 * Generic threading stuff
 */
/*
 * Copyright (c) 1995,1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Authors: Jim Zelenka, Marc Unangst
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_common.h>

/*
 * Shared stuff
 */

NASD_DECLARE_ONCE(nasd_threads_init_once)
NASD_SYS_DECLARE_MUTEX(nasd_threads_use_counter_lock)
int nasd_threads_use_counter;

#if LINUX
NASD_SYS_DECLARE_MUTEX(nasd_atomic_global_mutex)
#endif /* LINUX */

/*
 * nasd_threads_real_init
 *
 * Called when the system comes into use after not being
 * used (start-of-day call to nasd_threads_init(), or first
 * such call after the last call to nasd_threads_shutdown()
 * which actually deactivated the system).
 */
nasd_status_t
nasd_threads_real_init()
{
  nasd_status_t rc;

  NASD_ASSERT(nasd_threads_use_counter == 1);

  rc = nasd_sys_threads_init();
  if (rc)
    return(rc);

  return(NASD_SUCCESS);
}

/*
 * nasd_threads_real_shutdown
 *
 * Called when last user of the threads system calls nasd_threads_shutdown().
 * Clean up and deallocate resources.
 */
void
nasd_threads_real_shutdown()
{
  long total;
  int i;

  NASD_ASSERT(nasd_threads_use_counter == 0);

  nasd_sys_threads_shutdown();
}

/*
 * nasd_threadsys_init
 *
 * Executed exactly once, the first time nasd_threads_init() is
 * called. Initialize counter tracking number of times system
 * is initted.
 */
void
nasd_threadsys_init()
{
  nasd_status_t rc;

  rc = nasd_sys_init();
  if (rc) {
    NASD_PANIC();
  }

  rc = nasd_sys_threads_init_once();
  if (rc) {
    NASD_PANIC();
  }

  rc = nasd_sys_mutex_init(&nasd_threads_use_counter_lock);
  if (rc) {
    NASD_PANIC();
  }

#if LINUX
  rc = nasd_sys_mutex_init(&nasd_atomic_global_mutex);
  if (rc) {
    NASD_PANIC();
  }
#endif /* LINUX */

  nasd_threads_use_counter = 0;
}

/*
 * nasd_threads_init
 *
 * Keep a counter of the number of times we're initted and
 * shutdown. When the last shutdown arrives, really deallocate.
 * This lets multiple subsystems use us without knowing about
 * one another.
 */
nasd_status_t
nasd_threads_init()
{
  nasd_status_t rc;

  nasd_once(&nasd_threads_init_once, nasd_threadsys_init);
  NASD_SYS_LOCK_MUTEX(nasd_threads_use_counter_lock);
  nasd_threads_use_counter++;
  if (nasd_threads_use_counter == 1) {
    rc = nasd_threads_real_init();
    if (rc) {
      nasd_threads_use_counter = 0;
      nasd_threads_real_shutdown();
    }
  }
  else {
    rc = NASD_SUCCESS;
  }
  NASD_SYS_UNLOCK_MUTEX(nasd_threads_use_counter_lock);

  return(rc);
}

/*
 * nasd_threads_shutdown
 *
 * Previous caller of nasd_threads_init() not using the threads
 * subsystem any more. Deallocate and cleanup iff necessary.
 */
void
nasd_threads_shutdown()
{
  NASD_ASSERT(nasd_threads_use_counter != 0);

  NASD_SYS_LOCK_MUTEX(nasd_threads_use_counter_lock);
  nasd_threads_use_counter--;
  if (nasd_threads_use_counter == 0) {
    nasd_threads_real_shutdown();
  }
  NASD_SYS_UNLOCK_MUTEX(nasd_threads_use_counter_lock);
}



nasd_status_t
_nasd_destroy_threadgroup(
  nasd_threadgroup_t  *g,
  char                *file,
  int                  line)
{
  nasd_status_t rc1, rc2;

#if NASD_DEBUG_ATOMIC > 0
  rc1 = _nasd_mutex_destroy(&g->mutex, file, line);
  rc2 = _nasd_cond_destroy(&g->cond, file, line);
#else /* NASD_DEBUG_ATOMIC > 0 */
  rc1 = nasd_mutex_destroy(&g->mutex);
  rc2 = nasd_cond_destroy(&g->cond);
#endif /* NASD_DEBUG_ATOMIC > 0 */
  if (rc1)
    return(rc1);
  return(rc2);
}

nasd_status_t
_nasd_init_threadgroup(
  nasd_threadgroup_t  *g,
  char                *file,
  int                  line)
{
  nasd_status_t rc;

#if NASD_DEBUG_ATOMIC > 0
  rc = _nasd_mutex_init(&g->mutex, file, line);
  if (rc)
    return(rc);
  rc = _nasd_cond_init(&g->cond, file, line);
  if (rc) {
    _nasd_mutex_destroy(&g->mutex, file, line);
    return(rc);
  }
#else /* NASD_DEBUG_ATOMIC > 0 */
  rc = nasd_mutex_init(&g->mutex);
  if (rc)
    return(rc);
  rc = nasd_cond_init(&g->cond);
  if (rc) {
    nasd_mutex_destroy(&g->mutex);
    return(rc);
  }
#endif /* NASD_DEBUG_ATOMIC > 0 */
  g->created = g->running = g->shutdown = g->should_shutdown = 0;
  return(NASD_SUCCESS);
}

#if NASD_DEBUG_ATOMIC > 0

static NASD_ATEnt_t nasd_atent_list;
static NASD_ATEnt_t *nasd_atent_done_list = NULL;

NASD_SYS_DECLARE_MUTEX(nasd_atent_mutex)
NASD_DECLARE_ONCE(nasd_atent_once)

void 
nasd_atent_init()
{
  nasd_status_t rc;

  rc = nasd_sys_mutex_init(&nasd_atent_mutex);
  if (rc) {
    /*
     * Don't translate error code- translation system may
     * want mutexes!
     */
    printf("NASD ERROR: got 0x%x creating nasd_atent_mutex\n", rc);
    NASD_PANIC();
  }
	nasd_atent_list.next = nasd_atent_list.prev = &nasd_atent_list;
}

#define ATENT_TYPE(_e_) ((((_e_)->type == 0)||((_e_)->type > 2)) ? 0 : (_e_)->type)
#define ATENT_OTYPE(_e_) ((((_e_)->otype == 0)||((_e_)->otype > 2)) ? 0 : (_e_)->otype)

void
nasd_atent_shutdown()
{
	int rc, num_freed[3], num_not_freed[3];
	NASD_ATEnt_t *r, *n;

  nasd_once(&nasd_atent_once, nasd_atent_init);

	num_freed[0] = num_freed[1] = num_freed[2] = 0;
	num_not_freed[0] = num_not_freed[1] = num_not_freed[2] = 0;
	printf("nasd_atent_shutdown:\n");
	for(r=nasd_atent_list.next;r!=&nasd_atent_list;r=r->next) {
		printf("r=%lx type=%d file=%s line=%d\n", r, r->type, r->file, r->line);
		num_not_freed[ATENT_TYPE(r)]++;
	}
  rc = nasd_sys_mutex_destroy(&nasd_atent_mutex);
	if (rc) {
		printf("ERROR: rc=0x%x destroying nasd_atent_mutex\n", rc);
		NASD_PANIC();
	}
	for(r=nasd_atent_done_list;r;r=n) {
		n = r->next;
		num_freed[ATENT_OTYPE(r)]++;
		free(r);
	}
	printf("%d mutexes not freed %d conditions not freed %d bogus not freed\n",
		num_not_freed[1], num_not_freed[2], num_not_freed[0]);
	printf("%d mutexes freed %d conditions freed %d bogus freed\n",
		num_freed[1], num_freed[2], num_freed[0]);
}

static NASD_ATEnt_t *
AllocATEnt(
  char  *file,
  int    line)
{
  NASD_ATEnt_t *t;

  t = (NASD_ATEnt_t *)malloc(sizeof(NASD_ATEnt_t));
  if (t == NULL) {
    NASD_PANIC();
  }
  t->file = file;
  t->line = line;
  t->type = 0;
  t->lk_file = NULL;
  t->lk_line = 0;
  return(t);
}

static void FreeATEnt(
  NASD_ATEnt_t  *t)
{
  t->otype = t->type;
  t->type = 0;
  t->next = nasd_atent_done_list;
  nasd_atent_done_list = t;
}

nasd_status_t
_nasd_mutex_init(
  NASD_ATEnt_t  **m,
  char           *file,
  int             line)
{
	NASD_ATEnt_t *a;
	int rc;

  nasd_once(&nasd_atent_once, nasd_atent_init);

	a = AllocATEnt(file,line);
  rc = nasd_sys_mutex_init(&a->m);
  if (rc) {
    printf("NASD ERROR: got 0x%x allocating mutex %s:%d\n",
      rc, file, line);
    NASD_PANIC();
  }
  NASD_SYS_LOCK_MUTEX(nasd_atent_mutex);
  a->next = nasd_atent_list.next;
  a->prev = &nasd_atent_list;
  a->type = NASD_ATENT_M;
  a->next->prev = a;
  a->prev->next = a;
  NASD_SYS_UNLOCK_MUTEX(nasd_atent_mutex);

	*m = a;
	return(NASD_SUCCESS);
}

nasd_status_t
_nasd_mutex_destroy(
  NASD_ATEnt_t  **m,
  char           *file,
  int             line)
{
	NASD_ATEnt_t *r;
	int rc;

  nasd_once(&nasd_atent_once, nasd_atent_init);

	r = *m;
  rc = nasd_sys_mutex_destroy(&r->m);
	if (rc) {
		printf("ERROR: rc=0x%x destroying mutex %s:%d\n",
			rc, file, line);
		NASD_PANIC();
	}
  NASD_SYS_LOCK_MUTEX(nasd_atent_mutex);
	r->next->prev = r->prev;
	r->prev->next = r->next;
	FreeATEnt(r);
  NASD_SYS_UNLOCK_MUTEX(nasd_atent_mutex);

	*m = NULL;
	return(NASD_SUCCESS);
}

nasd_status_t
_nasd_cond_init(
  NASD_ATEnt_t  **c,
  char           *file,
  int             line)
{
	NASD_ATEnt_t *a;
	int rc;

  nasd_once(&nasd_atent_once, nasd_atent_init);

  a = AllocATEnt(file,line);
  rc = nasd_sys_cond_init(&a->c);
  if (rc) {
    printf("NASD ERROR: got 0x%x allocating cond %s:%d\n",
      rc, file, line);
    NASD_PANIC();
  }
  NASD_SYS_LOCK_MUTEX(nasd_atent_mutex);
  a->next = nasd_atent_list.next;
  a->prev = &nasd_atent_list;
  a->type = NASD_ATENT_C;
  a->next->prev = a;
  a->prev->next = a;
  NASD_SYS_UNLOCK_MUTEX(nasd_atent_mutex);

  *c = a;
  return(NASD_SUCCESS);
}

nasd_status_t
_nasd_cond_destroy(
  NASD_ATEnt_t  **c,
  char           *file,
  int             line)
{
  NASD_ATEnt_t *r;
  int rc;

  nasd_once(&nasd_atent_once, nasd_atent_init);

  r = *c;
  rc = nasd_sys_cond_destroy(&r->c);
	if (rc) {
		printf("ERROR: rc=0x%x destroying cond %s:%d\n",
			rc, file, line);
		NASD_PANIC();
	}
  NASD_SYS_LOCK_MUTEX(nasd_atent_mutex);
  r->next->prev = r->prev;
  r->prev->next = r->next;
  FreeATEnt(r);
  NASD_SYS_UNLOCK_MUTEX(nasd_atent_mutex);

  *c = NULL;
  return(NASD_SUCCESS);
}

int
_nasd_atent_try_lock_mutex(
  NASD_ATEnt_t  *c,
  char          *file,
  int            line)
{
  int rc;

  NASD_ASSERT(c->type == NASD_ATENT_M);
  rc = NASD_SYS_TRY_LOCK_MUTEX(c->m);
  if (rc) {
    c->lk_file = file;
    c->lk_line = line;
  }
  return(rc);
}

#endif /* NASD_DEBUG_ATOMIC > 0 */

/*
 * And now... cheezy readers-writers locks
 *
 * Where applicable, use system-provided, otherwise, use our
 * own cheezy ones
 */

#if NASD_TS_CHEESY_RWLOCKS > 0
/*
 * These are some really simple rw locks. Writers
 * always get priority (potentially starving readers).
 * These should only be used if there's nothing better
 * provided by the local platform. --jimz
 */
nasd_status_t
_nasd_rwlock_init(l, file, line)
  NASD_DECLARE_RWLOCK(*l)
  char  *file;
  int    line;
{
  nasd_status_t rc;

  rc = nasd_mutex_init(&l->lock);
  if (rc)
    return(rc);
  rc = nasd_cond_init(&l->cond);
  if (rc) {
    nasd_mutex_destroy(&l->lock);
    return(rc);
  }
  rc = nasd_cond_init(&l->wcond);
  if (rc) {
    nasd_mutex_destroy(&l->lock);
    nasd_cond_destroy(&l->cond);
    return(rc);
  }
  l->nreaders = 0;
  l->nwriters_waiting = 0;
  l->nreaders_waiting = 0;
  l->writing = 0;
  l->rd_file = NULL;
  l->rd_line = 0;
  l->wr_file = NULL;
  l->wr_line = 0;
  return(NASD_SUCCESS);
}

nasd_status_t
_nasd_rwlock_destroy(l, file, line)
  NASD_DECLARE_RWLOCK(*l)
  char  *file;
  int    line;
{
  nasd_status_t rc1, rc2, rc3;

  rc1 = nasd_mutex_destroy(&l->lock);
  rc2 = nasd_cond_destroy(&l->cond);
  rc3 = nasd_cond_destroy(&l->wcond);
  if (rc1)
    return(rc1);
  if (rc2)
    return(rc2);
  return(rc3);
}

void
nasd_rwlock_read(l, file, line)
  NASD_DECLARE_RWLOCK(*l)
  char  *file;
  int    line;
{
  NASD_LOCK_MUTEX(l->lock);
  if (l->writing || l->nwriters_waiting) {
    l->nreaders_waiting++;
    do {
      NASD_WAIT_COND(l->cond,l->lock);
    } while(l->writing || l->nwriters_waiting);
    l->nreaders_waiting--;
    NASD_ASSERT(l->nreaders_waiting >= 0);
  }
  l->nreaders++;
  l->rd_file = file;
  l->rd_line = line;
  NASD_UNLOCK_MUTEX(l->lock);
}

void
nasd_rwlock_read_done(l, file, line)
  NASD_DECLARE_RWLOCK(*l)
  char  *file;
  int    line;
{
  int dobc;

  dobc = 0;
  NASD_LOCK_MUTEX(l->lock);
  l->nreaders--;
  NASD_ASSERT(l->nreaders >= 0);
  if ((l->nreaders == 0) && (l->nwriters_waiting)) {
    dobc = 1;
  }
  l->rd_file = NULL;
  l->rd_line = 0;
  NASD_UNLOCK_MUTEX(l->lock);
  if (dobc) {
    NASD_BROADCAST_COND(l->wcond);
  }
}

void
nasd_rwlock_write(l, file, line)
  NASD_DECLARE_RWLOCK(*l)
  char  *file;
  int    line;
{
  NASD_LOCK_MUTEX(l->lock);
  if (l->nreaders || l->writing) {
    l->nwriters_waiting++;
    do {
      NASD_WAIT_COND(l->wcond,l->lock);
    } while (l->nreaders || l->writing);
    l->nwriters_waiting--;
    NASD_ASSERT(l->nwriters_waiting >= 0);
  }
  l->writing = 1;
  l->wr_file = file;
  l->wr_line = line;
  NASD_UNLOCK_MUTEX(l->lock);
}

void
nasd_rwlock_write_done(l, file, line)
  NASD_DECLARE_RWLOCK(*l)
  char  *file;
  int    line;
{
  int dobc;

  dobc = 0;
  NASD_LOCK_MUTEX(l->lock);
  l->writing = 0;
  if (l->nwriters_waiting) {
    NASD_SIGNAL_COND(l->wcond);
  }
  else if (l->nreaders_waiting) {
    dobc = 1;
  }
  l->wr_file = NULL;
  l->wr_line = 0;
  NASD_UNLOCK_MUTEX(l->lock);
  if (dobc) {
    NASD_BROADCAST_COND(l->cond);
  }
}

#endif /* NASD_TS_CHEESY_RWLOCKS > 0 */

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
