/*
 * nasd_cbasic.c
 *
 * Basics of on-disk cache.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_cache.h>
#include <nasd/nasd_common.h>

int nasd_odc_force_format = 0; /* force formatting */
int nasd_odc_need_format = 0;  /* need to format disk */
int nasd_odc_need_recover = 0; /* need to recover cache state */

nasd_odc_state_t *nasd_odc_state;

int nasd_odc_dirty_autokick; /* how many dirty blocks auto-starts flush */
int nasd_odc_refblocks;      /* number of refcnt blocks */

extern int nasd_odc_buckets; /* number of buckets */

nasd_blkcnt_t nasd_od_blocks; /* number of data blocks on disk */

nasd_odc_oq_t nasd_odc_unusedq;
nasd_odc_oq_t nasd_odc_lru[NASD_ODC_NLRUS];
nasd_odc_oq_t nasd_odc_wireq;

#define NASD_MAX_FREE_ODC_ENT  (nasd_odc_size*2)
#define NASD_ODC_ENT_INC        1
#define NASD_ODC_ENT_INITIAL   (nasd_odc_size+(2*NASD_ODC_ENT_INC))

static nasd_freelist_t *nasd_odc_ent_freelist;

static void clean_ent(void *arg);
static nasd_status_t init_ent(void *arg);

void
nasd_odc_uninit_ent(
  nasd_odc_ent_t *ent)
{
  nasd_mutex_destroy(&ent->lock);
  nasd_cond_destroy(&ent->cond);
  nasd_cond_destroy(&ent->acond);
  nasd_rwlock_destroy(&ent->rwlock);
  nasd_odc_io_release_page(ent);
}

nasd_status_t
nasd_odc_init_ent(
  nasd_odc_ent_t  *ent,
  int              type,
  int              alloc_page)
{
  nasd_status_t rc;

  bzero((char *)ent, sizeof(nasd_odc_ent_t));

  rc = nasd_mutex_init(&ent->lock);
  if (rc)
    return(rc);

  rc = nasd_cond_init(&ent->cond);
  if (rc) {
    nasd_shutdown_mutex(&ent->lock);
    return(rc);
  }

  rc = nasd_cond_init(&ent->acond);
  if (rc) {
    nasd_shutdown_mutex(&ent->lock);
    nasd_shutdown_cond(&ent->cond);
    return(rc);
  }

  rc = nasd_rwlock_init(&ent->rwlock);
  if (rc) {
    nasd_shutdown_mutex(&ent->lock);
    nasd_shutdown_cond(&ent->cond);
    nasd_shutdown_cond(&ent->acond);
    return(rc);
  }

  ent->data_flags = 0;
  ent->lru_flags = 0;
  ent->io_flags = 0;
  ent->dirty_flags = 0;
  ent->type = type;
  ent->lru_num = NASD_ODC_LRU_NONE;

  if (alloc_page) {
    rc = nasd_odc_io_alloc_page(ent);
    if (rc) {
      nasd_shutdown_mutex(&ent->lock);
      nasd_shutdown_cond(&ent->cond);
      nasd_shutdown_cond(&ent->acond);
      nasd_shutdown_rwlock(&ent->rwlock);
      return(rc);
    }
  }
  else {
    NASD_ODC_IO_ENT_CLEAR_PAGE(ent);
  }

  return(NASD_SUCCESS);
}

/*
 * Initialize a queue
 */
nasd_status_t
nasd_queue_init(
  nasd_odc_oq_t  *q)
{
  int rc;

  rc = nasd_mutex_init(&q->lock);
  if (rc)
    return(NASD_FAIL);
  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_mutex,
    &q->lock);
  if (rc) {
    nasd_mutex_destroy(&q->lock);
    return(rc);
  }

  q->lock_file = NULL;
  q->lock_line = (-1);

  q->head.type = NASD_ODC_T_HOLDER;
  q->head.hnext = q->head.hprev = &q->head;
  q->head.dnext = q->head.dprev = &q->head;
  q->head.lnext = q->head.lprev = &q->head;
  q->head.wnext = q->head.wprev = &q->head;
  q->head.data.buf = NULL;

  q->size = 0;

  return(NASD_SUCCESS);
}

void
nasd_odc_preinsert(
  nasd_odc_ent_t  *ent)
{
  ent->refcnt = 0;
  ent->data_flags = NASD_CD_INVALID;
  ent->iocb = NULL;
  ent->iocb_arg = NULL;
  ent->real_sectno = nasd_odc_real_sectno(ent->blkno, ent->type);
  nasd_odc_block_hash_ins(ent);
}

/*
 * Initialize cache subsystem logic
 * (no I/Os performed yet, just set up
 * datastructures)
 */
nasd_status_t
nasd_cache_init()
{
  nasd_odc_ent_t *ent;
  nasd_status_t nrc;
  int i, rc;

  /* init free queue */
  nrc = nasd_queue_init(&nasd_odc_unusedq);
  if (nrc)
    return(nrc);

  /* init wired queue */
  nrc = nasd_queue_init(&nasd_odc_wireq);
  if (nrc)
    return(nrc);

  /* init LRUs */
  for(i=0;i<NASD_ODC_NLRUS;i++) {
    nrc = nasd_queue_init(&nasd_odc_lru[i]);
    if (nrc)
      return(nrc);
  }

  NASD_FREELIST_CREATE(nasd_odc_ent_freelist, NASD_MAX_FREE_ODC_ENT,
    NASD_ODC_ENT_INC, sizeof(nasd_odc_ent_t));
  if (nasd_odc_ent_freelist == NULL)
    return(NASD_NO_MEM);

  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_odc_ent_destroy, NULL);
  if (rc) {
    nasd_odc_ent_destroy(NULL);
    return(rc);
  }

  NASD_FREELIST_PRIME_INIT(nasd_odc_ent_freelist, NASD_ODC_ENT_INITIAL,lnext,
    (nasd_odc_ent_t *),init_ent);

  /*
   * From here on out, any failures result in a call to
   * nasd_odc_shutdown_cache(). This can be used to cleanup
   * partially-initialized state, provided that any early
   * initialization is done above here so that it knows
   * what it needs to clean up.
   */

  /* init block lookup subsys */
  nrc = nasd_odc_blocksys_init();
  if (nrc) {
    printf("ERROR (%s:%d): cannot init block sys, rc=%lu\n",
      __FILE__, __LINE__, (long)nrc);
    nasd_odc_shutdown_cache(NULL);

    return(nrc);
  }

  nasd_odc_dirty_autokick = (2*nasd_odc_size)/5;

  printf("DRIVE: %d GP cache blocks\n", nasd_odc_size);
  printf("DRIVE: dirty autokick at %d blocks\n",
    nasd_odc_dirty_autokick);

  /*
   * See nasd_drive_options.h for an explanation of the settings
   */
#if NASD_DRIVE_WIRE_REFCNT > 0
  printf("DRIVE: refcnt blocks will be wired\n");
#endif /* NASD_DRIVE_WIRE_REFCNT > 0 */
#if NASD_DRIVE_WIRE_NPT == 1
  printf("DRIVE: NPT-2 blocks will be wired\n");
#endif /* NASD_DRIVE_WIRE_NPT == 1 */
#if NASD_DRIVE_WIRE_NPT == 2
  printf("DRIVE: all NPT blocks will be wired\n");
#endif /* NASD_DRIVE_WIRE_NPT == 2 */
#if NASD_DRIVE_WIRE_NPT > 2
  printf("DRIVE: unknown NPT wire mode!\n");
#endif /* NASD_DRIVE_WIRE_NPT > 2 */

  /* init data handles */
  for(i=0;i<nasd_odc_size;i++) {
    rc = nasd_odc_alloc_ent(&ent);
    if (rc) {
      nasd_odc_shutdown_cache(NULL);
      return(NASD_NO_MEM);
    }

    rc = nasd_odc_init_ent(ent, NASD_ODC_T_FREE, 1);
    if (rc) {
      nasd_odc_shutdown_cache(NULL);
      return(rc);
    }

    NASD_ODC_Q_INS_NOLOCK(&nasd_odc_unusedq,ent,l);
  }

  nrc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_odc_shutdown_cache, NULL);
  if (nrc) {
    nasd_odc_shutdown_cache(NULL);
    return(nrc);
  }

  rc = nasd_odc_dirtysys_init();
  if (rc) {
    printf("ERROR (%s:%d): cannot init dirty sys, rc=0x%x (%s)\n",
      __FILE__, __LINE__, rc, nasd_error_string(rc));
    return(rc);
  }

#if MJU_DEBUG
  printf("nasd_cache_init() done\n");
#endif /* MJU_DEBUG */

  return(NASD_SUCCESS);
}

/*
 * Call with ent lock held.
 *
 * "real" version is entirely macroized in nasd_cache.h
 */
void
_nasd_odc_wait_not_busy(
  nasd_odc_ent_t  *ent)
{
  nasd_status_t raise_rc;

  raise_rc = NASD_FAIL;
  while(ent->data_flags&NASD_CD_BUSY) {
    if (raise_rc != NASD_SUCCESS)
      raise_rc = nasd_od_io_try_raise_pri(ent, NASD_IO_PRI_MED);
    NASD_WAIT_COND(ent->cond,ent->lock);
  }
}

/*
 * Call with ent lock held.
 *
 * "real" version is entirely macroized in nasd_cache.h
 */
void
_nasd_odc_wait_not_busy_invalid(
  nasd_odc_ent_t  *ent)
{
  nasd_status_t raise_rc;

  raise_rc = NASD_FAIL;
  while(ent->data_flags&(NASD_CD_BUSY|NASD_CD_INVALID)) {
    if (raise_rc != NASD_SUCCESS)
      raise_rc = nasd_od_io_try_raise_pri(ent, NASD_IO_PRI_MED);
    NASD_WAIT_COND(ent->cond,ent->lock);
  }
}

void
nasd_odc_shutdown_cache(
  void  *ignored_arg)
{
  nasd_odc_ent_t *ent;
  int i;

  /*
   * The dirty system has already flushed all the
   * dirty pages, and made them clean. Likewise,
   * we've already registered functions that will
   * deallocate all the page ents. Our job is to
   * hunt down all the outstanding page ent structures
   * and deallocate them.
   */

  /*
   * The dirty system has already flushed all the
   * dirty pages, and made them clean. Here, we hunt
   * down all outstanding cache blocks and deallocate
   * them appropriately.
   */

  for(i=0;i<NASD_ODC_NLRUS;i++) {
    while(NASD_ODC_Q_SIZE(&nasd_odc_lru[i])) {
      NASD_ODC_Q_DEQ_TAIL_NOLOCK(&nasd_odc_lru[i],ent,l);
      nasd_odc_free_ent(ent);
    }
    NASD_ASSERT(nasd_odc_lru[i].head.lprev == &nasd_odc_lru[i].head);
    NASD_ASSERT(nasd_odc_lru[i].head.lnext == &nasd_odc_lru[i].head);
  }

  while(NASD_ODC_Q_SIZE(&nasd_odc_unusedq)) {
    NASD_ODC_Q_DEQ_TAIL_NOLOCK(&nasd_odc_unusedq,ent,l);
    nasd_odc_free_ent(ent);
  }
  NASD_ASSERT(nasd_odc_unusedq.head.lprev == &nasd_odc_unusedq.head);
  NASD_ASSERT(nasd_odc_unusedq.head.lnext == &nasd_odc_unusedq.head);

  while(NASD_ODC_Q_SIZE(&nasd_odc_wireq)) {
    NASD_ODC_Q_DEQ_TAIL_NOLOCK(&nasd_odc_wireq,ent,w);
    nasd_odc_free_ent(ent);
  }
  NASD_ASSERT(nasd_odc_wireq.head.lprev == &nasd_odc_wireq.head);
  NASD_ASSERT(nasd_odc_wireq.head.lnext == &nasd_odc_wireq.head);
}

static void
clean_ent(
  void  *arg)
{
  nasd_odc_ent_t *ent;

  ent = (nasd_odc_ent_t *)arg;
  nasd_odc_uninit_ent(ent);
}

static nasd_status_t
init_ent(
  void  *arg)
{
  nasd_odc_ent_t *ent;
  nasd_status_t rc;

  ent = (nasd_odc_ent_t *)arg;
  rc = nasd_odc_init_ent(ent, NASD_ODC_T_FREE, 0);

  return(rc);
}

void
nasd_odc_ent_destroy(
  void  *arg)
{
  NASD_FREELIST_DESTROY_CLEAN(nasd_odc_ent_freelist,lnext,
    (nasd_odc_ent_t *), clean_ent);
}

nasd_status_t
nasd_odc_alloc_ent(
  nasd_odc_ent_t  **entp)
{
  nasd_odc_ent_t *ent;

  NASD_FREELIST_GET_INIT(nasd_odc_ent_freelist,ent,lnext,
    (nasd_odc_ent_t *),init_ent);
  if (ent == NULL)
    return(NASD_NO_MEM);
  *entp = ent;
  return(NASD_SUCCESS);
}

void
nasd_odc_free_ent(
  nasd_odc_ent_t  *ent)
{
  nasd_odc_uninit_ent(ent);
  NASD_FREELIST_FREE_CLEAN(nasd_odc_ent_freelist,ent,lnext,clean_ent);
}

/*
 * Temporarily release empty entry to general-use pool.
 * ("Empty" in that it has no associated page.)
 */
void
nasd_odc_put_ent(
  nasd_odc_ent_t  *ent)
{
  NASD_FREELIST_FREE_CLEAN(nasd_odc_ent_freelist,ent,lnext,clean_ent);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
