/*
 * nasd_od_mq.c
 *
 * Drive side of a cheesy shared-memory and message queue based
 * "colocation" interface.
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Marc Unangst
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#ifndef KERNEL
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_cache.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_mq.h>
#include <nasd/nasd_od_rpc.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_keymgmt.h>

#include <errno.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>

#define DOBEGIN(_opname_) NASD_OD_RPC_DOBEGIN(_opname_,"nasd_od_mq.c")
#define DORETURN(_opname_) NASD_OD_RPC_DORETURN(_opname_,"nasd_od_mq.c")

#define NASD_OD_MQ_DEBUG		0
#define NASD_OD_MQ_LISTENER_THREADS	5

typedef struct nasd_mq_pipe_state_s {
  int shmid;
  nasd_byte_t *orig_buf;
  nasd_byte_t *cur_buf;
  int orig_buf_len;
  int buf_len_remain;
} nasd_mq_pipe_state_t;

int nasd_od_mq_ignore_eexist = 0;

static int nasd_od_mq_mqid;
static int nasd_od_mq_count = 0;
static int nasd_od_mq_shutdown_req = 0;

static void nasd_od_mq_listen(nasd_threadarg_t);
static nasd_status_t nasd_od_mq_pull(void *state_arg, void *buf,
				     nasd_len_t in_len, nasd_len_t *out_lenp);
static nasd_status_t nasd_od_mq_push(void *state_arg, void *buf,
				     nasd_len_t in_len);
static void nasd_od_mq_op_null(nasd_status_t *, nasd_rpc_status_t *);
static void nasd_od_mq_op_sync(nasd_status_t *, nasd_rpc_status_t *);
static void nasd_od_mq_op_part_creat(nasd_key_t,
				  nasd_p_part_creat_dr_args_t *,
				  nasd_p_part_creat_dr_res_t *,
				  nasd_rpc_status_t *);
static void nasd_od_mq_op_create(nasd_key_t,
			      nasd_p_create_dr_args_t *,
			      nasd_p_create_dr_res_t *,
			      nasd_rpc_status_t *);
static void nasd_od_mq_op_getattr(nasd_key_t,
			       nasd_p_getattr_dr_args_t *,
			       nasd_p_getattr_dr_res_t *,
			       nasd_rpc_status_t *);
static void nasd_od_mq_op_write_simple(int, nasd_key_t,
				       nasd_p_smpl_op_dr_args_t *,
				       nasd_p_fastwrite_dr_res_t *,
				       nasd_rpc_status_t *);
static void nasd_od_mq_op_read_simple(int, int, nasd_key_t,
				      nasd_p_smpl_op_dr_args_t *,
				      nasd_p_fastread_dr_res_t *,
				      nasd_rpc_status_t *);
static void nasd_od_mq_op_tread_simple(int, nasd_key_t,
				       nasd_p_thrtl_op_dr_args_t *,
				       nasd_p_fastread_dr_res_t *,
				       nasd_rpc_status_t *);
static void nasd_od_mq_op_setattr(nasd_key_t,
			       nasd_p_setattr_dr_args_t *,
			       nasd_p_setattr_dr_res_t *,
			       nasd_rpc_status_t *);
static void nasd_od_mq_op_flush_obj(nasd_key_t,
				 nasd_p_flush_obj_dr_args_t *,
				 nasd_p_flush_obj_dr_res_t *,
				 nasd_rpc_status_t *);
static void nasd_od_mq_op_eject_obj(nasd_key_t,
				 nasd_p_eject_obj_dr_args_t *,
				 nasd_p_eject_obj_dr_res_t *,
				 nasd_rpc_status_t *);
static void nasd_od_mq_op_remove(nasd_key_t,
			      nasd_p_remove_dr_args_t *,
			      nasd_p_remove_dr_res_t *,
			      nasd_rpc_status_t *);
static void nasd_od_mq_op_initialize(nasd_p_initialize_dr_args_t *,
				     nasd_p_initialize_dr_res_t *,
				     nasd_rpc_status_t *);

nasd_status_t
nasd_startup_msgq(void)
{
  int rc;
  nasd_thread_t mq_listen[NASD_OD_MQ_LISTENER_THREADS];
  int i;

#if NASD_OD_MQ_DEBUG > 0
  fprintf(stderr, "mq: nasd_startup_msgq\n");
#endif /* NASD_OD_MQ_DEBUG > 0 */

  for(i = 0; i < NASD_OD_MQ_LISTENER_THREADS; i++) {
    rc = nasd_thread_create(&mq_listen[i], nasd_od_mq_listen, NULL);
    if(rc) {
      printf("DRIVE: error creating message queue listener %d: rc=0x%x (%s)\n",
	     i, rc, nasd_error_string(rc));
      return rc;
    }
  }

  return NASD_SUCCESS;
}

/* XXX should have a matching nasd_shutdown_msgq() routine to kill
   listener threads. */

nasd_status_t
nasd_od_mq_init(void)
{
  nasd_status_t rc;

#if NASD_OD_MQ_DEBUG > 0
  fprintf(stderr, "mq: nasd_od_mq_init\n");
#endif /* NASD_OD_MQ_DEBUG > 0 */

  if(nasd_od_mq_count == 0) {
    int flags;

    if(nasd_od_mq_ignore_eexist)
      flags = IPC_CREAT | 0666;
    else
      flags = IPC_CREAT | IPC_EXCL | 0666;

    nasd_od_mq_mqid = msgget(NASD_MSGQ_KEY, flags);
#if NASD_OD_MQ_DEBUG > 0
    fprintf(stderr, "mq: queue created, id %d\n", nasd_od_mq_mqid);
#endif /* NASD_OD_MQ_DEBUG > 0 */
    if(nasd_od_mq_mqid < 0) {
      printf("DRIVE ERROR: cannot create nasd_od_mq_mqid (got %d errno %d)\n",
        nasd_od_mq_mqid, errno);
      return(NASD_FAIL);
    }

    rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_od_mq_rm_msgq,
		       (void *)((unsigned long)nasd_od_mq_mqid));
    if (rc) {
      nasd_od_mq_rm_msgq(NULL);
      return(rc);
    }
  }

  nasd_od_mq_count++;

  return NASD_SUCCESS;
}

void
nasd_od_mq_shutdown(void)
{
  nasd_od_mq_count--;
}

/* master listener/dispatcher thread */
static void
nasd_od_mq_listen(nasd_threadarg_t ignored)
{
  int ret;
  int replyq;
  nasd_msgq_buf_t mb;
  nasd_msgq_buf_t mr;

#if NASD_OD_MQ_DEBUG > 0
  fprintf(stderr, "mq: mq_listen started\n");
#endif /* NASD_OD_MQ_DEBUG > 0 */

  while(1) {
    ret = msgrcv(nasd_od_mq_mqid, &mb, NASD_MSGQ_MSGLEN, 0, 0);
    if(ret < 0) {
      if(errno == EINTR)
	continue;
      else if(errno == EIDRM) {
#if NASD_OD_MQ_DEBUG > 0
	printf("nasd_od_mq: message queue removed; returning\n");
#endif /* NASD_OD_MQ_DEBUG > 0 */
	return;
      } else if(errno == E2BIG) {
	printf("nasd_od_mq: WARNING: received oversize message\n");
	continue;
      } else {
	printf("nasd_od_mq: msgrcv returned %d; errno=%d (%s)\n",
	       ret, errno, strerror(errno));
	NASD_PANIC();
      }
    }

#if NASD_OD_MQ_DEBUG > 0
    printf("nasd_od_mq: received message type %ld\n", mb.mtype);
#endif /* NASD_OD_MQ_DEBUG > 0 */

    /* dispatch to appropriate handler function */
    switch(mb.mtype) {
    case NASD_MSGQ_NULL: {
      NASDMQ_null_msg_t *msg = &(mb.un.null_msg);
      NASDMQ_null_rep_t *rep = &(mr.un.null_rep);
      nasd_od_mq_op_null(&rep->nasd_status, &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_SYNC: {
      NASDMQ_sync_msg_t *msg = &(mb.un.sync_msg);
      NASDMQ_sync_rep_t *rep = &(mr.un.sync_rep);
      nasd_od_mq_op_sync(&rep->nasd_status, &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PCRPART: {
      NASDMQ_p_crpart_msg_t *msg = &(mb.un.p_crpart_msg);
      NASDMQ_p_crpart_rep_t *rep = &(mr.un.p_crpart_rep);
      nasd_od_mq_op_part_creat(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PCREATE: {
      NASDMQ_p_create_msg_t *msg = &(mb.un.p_create_msg);
      NASDMQ_p_create_rep_t *rep = &(mr.un.p_create_rep);
      nasd_od_mq_op_create(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PGETATTR: {
      NASDMQ_p_getattr_msg_t *msg = &(mb.un.p_getattr_msg);
      NASDMQ_p_getattr_rep_t *rep = &(mr.un.p_getattr_rep);
      nasd_od_mq_op_getattr(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PWRITE_SIMPLE: {
      NASDMQ_p_write_simple_msg_t *msg = &(mb.un.p_write_simple_msg);
      NASDMQ_p_write_simple_rep_t *rep = &(mr.un.p_write_simple_rep);
      nasd_od_mq_op_write_simple(msg->shmid, msg->in_key, &msg->args,
				 &rep->res, &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PREAD_SIMPLE: {
      NASDMQ_p_read_simple_msg_t *msg = &(mb.un.p_read_simple_msg);
      NASDMQ_p_read_simple_rep_t *rep = &(mr.un.p_read_simple_rep);
      nasd_od_mq_op_read_simple(msg->shmid, msg->is_read2, msg->in_key,
				&msg->args, &rep->res, &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PTREAD_SIMPLE: {
      NASDMQ_p_tread_simple_msg_t *msg = &(mb.un.p_tread_simple_msg);
      NASDMQ_p_tread_simple_rep_t *rep = &(mr.un.p_tread_simple_rep);
      nasd_od_mq_op_tread_simple(msg->shmid, msg->in_key, &msg->args,
				 &rep->res, &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PSETATTR: {
      NASDMQ_p_setattr_msg_t *msg = &(mb.un.p_setattr_msg);
      NASDMQ_p_setattr_rep_t *rep = &(mr.un.p_setattr_rep);
      nasd_od_mq_op_setattr(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PFLUSH_OBJ: {
      NASDMQ_p_flush_obj_msg_t *msg = &(mb.un.p_flush_obj_msg);
      NASDMQ_p_flush_obj_rep_t *rep = &(mr.un.p_flush_obj_rep);
      nasd_od_mq_op_flush_obj(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PEJECT_OBJ: {
      NASDMQ_p_eject_obj_msg_t *msg = &(mb.un.p_eject_obj_msg);
      NASDMQ_p_eject_obj_rep_t *rep = &(mr.un.p_eject_obj_rep);
      nasd_od_mq_op_eject_obj(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PREMOVE: {
      NASDMQ_p_remove_msg_t *msg = &(mb.un.p_remove_msg);
      NASDMQ_p_remove_rep_t *rep = &(mr.un.p_remove_rep);
      nasd_od_mq_op_remove(msg->in_key, &msg->args, &rep->res,
			    &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    case NASD_MSGQ_PINITIALIZE: {
      NASDMQ_p_initialize_msg_t *msg = &(mb.un.p_initialize_msg);
      NASDMQ_p_initialize_rep_t *rep = &(mr.un.p_initialize_rep);
      nasd_od_mq_op_initialize(&msg->args, &rep->res, &rep->op_status);
      mr.mtype = msg->seqnum;
      replyq = msg->replyq;
      break;
    }
    default:
      NASD_PANIC();
      /* unimplemented */
    }

    /* send reply */
    msgsnd(replyq, (struct msgbuf *) &mr, NASD_MSGQ_MSGLEN, 0);
  }
}

static void
nasd_od_mq_op_null(nasd_status_t *nasd_status,
		nasd_rpc_status_t *status)
{
  nasd_od_rpc_basic_t out_res;
  nasd_od_rpc_basic_t *res = &out_res;
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(null);
  *nasd_status = out_res.nasd_status = NASD_SUCCESS;
  DORETURN(null);
}

static void
nasd_od_mq_op_sync(nasd_status_t *nasd_status,
		nasd_rpc_status_t *status)
{
  nasd_od_rpc_basic_t out_res;
  nasd_od_rpc_basic_t *res = &out_res;
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(sync);
  *nasd_status = out_res.nasd_status = nasd_odc_flush_dirty(1);
  DORETURN(sync);
}

static void
nasd_od_mq_op_part_creat(nasd_key_t in_key,
		      nasd_p_part_creat_dr_args_t *args,
		      nasd_p_part_creat_dr_res_t *res,
		      nasd_rpc_status_t *status)
{
  nasd_identifier_t nid;
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(part_creat);
  res->nasd_status = nasd_od_create_partition(args->in_partnum,
					      args->in_blkcnt,
					      args->in_min_protection,
					      args->in_partition_key,
					      args->in_red_key,
					      args->in_black_key, &nid);
  DORETURN(part_creat);
}

static void
nasd_od_mq_op_create(nasd_key_t in_key,
		  nasd_p_create_dr_args_t *args,
		  nasd_p_create_dr_res_t *res,
		  nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(create);
  res->nasd_status = nasd_obj_create(args->in_partnum, &args->in_attribute,
				     args->in_fieldmask, &res->out_identifier,
				     &res->out_attribute, 0);
  DORETURN(create);
}

static void
nasd_od_mq_op_getattr(nasd_key_t in_key,
		   nasd_p_getattr_dr_args_t *args,
		   nasd_p_getattr_dr_res_t *res,
		   nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(getattr);
  res->nasd_status = nasd_obj_getattr(args->in_partnum, args->in_identifier,
				      &res->out_attribute);
  DORETURN(getattr);
}

static void
nasd_od_mq_op_write_simple(int shmid,
			   nasd_key_t in_key,
			   nasd_p_smpl_op_dr_args_t *args,
			   nasd_p_fastwrite_dr_res_t *res,
			   nasd_rpc_status_t *status)
{
  nasd_mq_pipe_state_t state;
  nasd_procpipe_t procpipe;
  nasd_byte_t *buf;
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(write_simple);
  buf = shmat(shmid, NULL, 0);
  if(buf == (nasd_byte_t *) -1) {
    res->nasd_status = NASD_FAIL;
    DORETURN(write_simple);
  }
  state.orig_buf = state.cur_buf = buf;
  state.orig_buf_len = state.buf_len_remain = args->in_len;
  procpipe.state = &state;
  procpipe.push = NULL;
  procpipe.pull = nasd_od_mq_pull;
  res->out_datalen = 0;
  res->nasd_status = nasd_obj_write_simple(args->in_partnum,
					   args->in_identifier,
					   args->in_offset, args->in_len,
					   &procpipe, NULL, &res->out_datalen);
  shmdt(buf);
  DORETURN(write_simple);
}

static void
nasd_od_mq_op_read_simple(int shmid,
			  int is_read2,
			  nasd_key_t in_key,
			  nasd_p_smpl_op_dr_args_t *args,
			  nasd_p_fastread_dr_res_t *res,
			  nasd_rpc_status_t *status)
{
  nasd_mq_pipe_state_t state;
  nasd_procpipe_t procpipe;
  nasd_byte_t *buf;
  NASD_OD_RPC_OP_STATS_DECL ;

  if(is_read2) {
    DOBEGIN(read2_simple);
  } else {
    DOBEGIN(read_simple);
  }

  buf = shmat(shmid, NULL, 0);
  if(buf == (nasd_byte_t *) -1) {
    res->nasd_status = NASD_FAIL;
    if(is_read2) {
      DORETURN(read2_simple);
    } else {
      DORETURN(read_simple);
    }
  }
  state.orig_buf = state.cur_buf = buf;
  state.orig_buf_len = state.buf_len_remain = args->in_len;
  procpipe.state = &state;
  procpipe.push = nasd_od_mq_push;
  procpipe.pull = NULL;
  res->out_datalen = 0;
  res->nasd_status = nasd_obj_read_simple(args->in_partnum,
					  args->in_identifier,
					  args->in_offset, args->in_len, NULL,
					  0, &procpipe, &res->out_datalen,
					  NULL);
  shmdt(buf);
  if(is_read2) {
    DORETURN(read2_simple);
  } else {
    DORETURN(read_simple);
  }
}

static void
nasd_od_mq_op_tread_simple(int shmid,
			   nasd_key_t in_key,
			   nasd_p_thrtl_op_dr_args_t *args,
			   nasd_p_fastread_dr_res_t *res,
			   nasd_rpc_status_t *status)
{
  nasd_mq_pipe_state_t state;
  nasd_procpipe_t procpipe;
  nasd_byte_t *buf;
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(tread_simple);

  buf = shmat(shmid, NULL, 0);
  if(buf == (nasd_byte_t *) -1) {
    res->nasd_status = NASD_FAIL;
    DORETURN(tread_simple);
  }
  state.orig_buf = state.cur_buf = buf;
  state.orig_buf_len = state.buf_len_remain = args->in_len;
  procpipe.state = &state;
  procpipe.push = nasd_od_mq_push;
  procpipe.pull = NULL;
  res->out_datalen = 0;
  res->nasd_status = nasd_obj_read_simple(args->in_partnum,
					  args->in_identifier,
					  args->in_offset, args->in_len,
					  &args->in_bms_targ, 0, &procpipe,
					  &res->out_datalen, NULL);
  shmdt(buf);
  DORETURN(tread_simple);
}

static void
nasd_od_mq_op_setattr(nasd_key_t in_key,
		   nasd_p_setattr_dr_args_t *args,
		   nasd_p_setattr_dr_res_t *res,
		   nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(setattr);
  res->nasd_status = nasd_obj_setattr(args->in_partnum, args->in_identifier,
				      &args->in_attribute, args->in_fieldmask,
				      &res->out_attribute);
  DORETURN(setattr);
}

static void
nasd_od_mq_op_flush_obj(nasd_key_t in_key,
		     nasd_p_flush_obj_dr_args_t *args,
		     nasd_p_flush_obj_dr_res_t *res,
		     nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(flush_obj);
  res->nasd_status = nasd_obj_flush(args->in_partnum, args->in_identifier);
  DORETURN(flush_obj);
}

static void
nasd_od_mq_op_eject_obj(nasd_key_t in_key,
		     nasd_p_eject_obj_dr_args_t *args,
		     nasd_p_eject_obj_dr_res_t *res,
		     nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(eject_obj);
  res->nasd_status = nasd_obj_eject(args->in_partnum, args->in_identifier);
  DORETURN(eject_obj);
}

static void
nasd_od_mq_op_remove(nasd_key_t in_key,
		  nasd_p_remove_dr_args_t *args,
		  nasd_p_remove_dr_res_t *res,
		  nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(remove);
  res->nasd_status = nasd_obj_remove(args->in_partnum, args->in_identifier);
  DORETURN(remove);
}

static void
nasd_od_mq_op_initialize(nasd_p_initialize_dr_args_t *args,
			 nasd_p_initialize_dr_res_t *res,
			 nasd_rpc_status_t *status)
{
  NASD_OD_RPC_OP_STATS_DECL ;

  DOBEGIN(initialize);
  res->nasd_status = nasd_security_initialize_drive(args->in_master_key,
						    args->in_drive_key);
  DORETURN(initialize);
}

static nasd_status_t
nasd_od_mq_pull(void *state_arg, void *buf, nasd_len_t in_len,
		nasd_len_t *out_lenp)
{
  nasd_mq_pipe_state_t *state = (nasd_mq_pipe_state_t *) state_arg;
  nasd_len_t len;

  len = NASD_MIN(in_len, state->buf_len_remain);
  if(len) {
    memcpy(buf, state->cur_buf, len);
    state->buf_len_remain -= len;
    state->cur_buf += len;
  }
  *out_lenp = len;
  return NASD_SUCCESS;
}

static nasd_status_t
nasd_od_mq_push(void *state_arg, void *buf, nasd_len_t in_len)
{
  nasd_mq_pipe_state_t *state = (nasd_mq_pipe_state_t *) state_arg;
  nasd_len_t len;

  len = NASD_MIN(in_len, state->buf_len_remain);
  if(len) {
    memcpy(state->cur_buf, buf, len);
    state->buf_len_remain -= len;
    state->cur_buf += len;
  }
  return NASD_SUCCESS;
}

void
nasd_od_mq_rm_msgq(void *arg)
{
  int m = (int)((unsigned long)arg);

  msgctl(m, IPC_RMID, NULL);
}
#endif /* !KERNEL */
