/*
 * nasd_security_dr.c
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Howard Gobioff
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/***********************************************************************
 *
 * nasd_security_dr.c
 *
 * Put the Security in NASD! drive component
 * 
 * These are calls only the drive should make.
 ***********************************************************************/

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#include <sys/errno.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_cache.h>
#if NASD_SECURE_RPCS_ENABLE > 0
#include <nasd/nasd_net_convert.h>
#include <nasd/nasd_convert_types.h>
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
#include <nasd/nasd_od.h>

#include <nasd/nasd_security_dr.h>

#if NASD_SECURE_RPCS_ENABLE > 0
#include <nasd/nasd_keymgmt.h>
#include <nasd/nasd_nonce_mgmt_dr.h>
#include <cryptolib/libcrypt.h>
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
#include <nasd/nasd_security_time.h>
#ifndef NULL
#define NULL 0
#endif

/* For the capability cache) */
nasd_security_cache_el_t *cache[NASD_SECURITY_CACHE_BUCKETS];
NASD_DECLARE_MUTEX(cache_lock[NASD_SECURITY_CACHE_BUCKETS]);

nasd_security_cache_el_t *cache_freelist;
NASD_DECLARE_MUTEX(cache_freelist_lock);

int cache_entry_count=0;
int cache_freelist_size=0;

extern char nasd_odc_zeroblk[NASD_OD_BASIC_BLOCKSIZE];
nasd_digest_t zero_digest; /*precomputed digest of a zero block */
#define PRINT_ERROR


nasd_status_t nasd_security_verify_args(nasd_args_t *in_args,
					nasd_byte_t *in_args_otw,
					nasd_args_len_t in_args_len,
					nasd_byte_t *in_data,
					nasd_int32 in_data_len,
					nasd_digest_nonce_otw_t in_digest,
                                        nasd_key_t in_key,
                                        nasd_security_context_t *contextp)
{
#if NASD_SECURE_RPCS_ENABLE > 0
  HMAC_SHA1_CTX HMAC_context;
  nasd_args_t *otw_args = (nasd_args_t *)in_args_otw;
  nasd_status_t rc;
  nasd_byte md_out[SHA_DIGEST_LENGTH]; 
  nasd_timespec_t now,then;
  nasd_uint64 now_l,then_l,diff;
  nasd_digest_nonce_t* otw_digest=(nasd_digest_nonce_t*)in_digest;
  nasd_key_t integrity_key;
#ifdef DEBUG_VERIFY_ARGS
  int i,cnt;
#endif

#ifdef DEBUG_VERIFY_ARGS
  printf("DEBUG(nasd_security_verify_args) entered in_args = %p\n",in_args);
#endif
  if (contextp!=NULL) {
    contextp->protection=in_args->in_security_param.actual_protection;
  } 
  
  if (in_args->in_security_param.actual_protection != NASD_NO_PROTECTION ) {

    if (in_args->in_security_param.actual_protection & NASD_PRIVACY_CAPABILITY) {
      /* At this point, we would need to decrypt that capability
         using the key indicated by security_param.type and 
         security_param.partnum */
      printf("This code should not be executed\n");
    }
    /* Now we can assume that  in_args.capability_args is readable */

    if (in_args->in_security_param.actual_protection &
        in_args->in_capability_args.min_protection !=
        in_args->in_capability_args.min_protection) {
      return (NASD_CAP_PROTECTION);
    }

#ifdef DEBUG_VERIFY_ARGS
  printf("DEBUG(nasd_security_verify_args): "
         "key=[%08" NASD_64x_FMT " %08" NASD_64x_FMT "]\n",
         ((nasd_uint64*)in_key)[0],((nasd_uint64*)in_key)[1]);
#endif /* DEBUG_VERIFY_ARGS */

  /* Check the nonce */ 
  
  nasd_timespec_t_from_net((nasd_byte *)&otw_digest->nonce,&then);
  rc=nasd_check_nonce_dr(&(in_args->in_capability_args),&then);
  if (rc)
    return(rc);

#ifdef DEBUG_VERIFY_ARGS
  printf("DEBUG(nasd_security_verify_args) Nonce is ok, move on to the digest\n");
#endif


  if (((in_args->in_security_param.actual_protection & NASD_INTEGRITY_ARGS) == 
      NASD_INTEGRITY_ARGS) ||((in_args->in_security_param.actual_protection &
			       NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA))
 {
   /* Captain, we've gotta go looking for a digest */

#ifdef DEBUG_VERIFY_ARGS
  printf("DEBUG(nasd_security_verify_args) start up the digest \n");
#endif
  nasd_security_generate_integrity_key(in_key, integrity_key);
   HMAC_SHA1_Init(&HMAC_context, (char *)integrity_key, sizeof(nasd_key_t));
   if ((in_args->in_security_param.actual_protection & NASD_INTEGRITY_DATA)
       == NASD_INTEGRITY_DATA) {
     /* This should never happen anymore -- get rid on next cleanup */
     HMAC_SHA1_Update(&HMAC_context, in_data, in_data_len);
    }
   if ((in_args->in_security_param.actual_protection & NASD_INTEGRITY_ARGS)
       == NASD_INTEGRITY_ARGS) {
 
     HMAC_SHA1_Update(&HMAC_context, (unsigned char *) in_args_otw+sizeof(nasd_args_t), in_args_len-(sizeof(nasd_digest_nonce_t)+sizeof(nasd_args_t)));
     HMAC_SHA1_Update(&HMAC_context, (unsigned char *) &otw_digest->nonce, sizeof(nasd_timespec_t));
   }
   if (contextp!=NULL) {
     bcopy( (char *)&HMAC_context, (char *) &contextp->HMAC_context, 
            sizeof(HMAC_SHA1_CTX));
     contextp->started=NASD_TRUE;
   }
   HMAC_SHA1_Final((unsigned char *)md_out, &HMAC_context);
   

#ifdef DEBUG_VERIFY_ARGS
  printf("DEBUG(nasd_security_verify_args)  digest done\n");
     
#ifdef DEBUG_VERIFY_ARGS_VERBOSE
	  printf("DEBUG(nasd_security_verify_args): first digest [");
	  for(cnt=0;cnt<NASD_DIGEST_SIZE;cnt++)
	    printf("%x",md_out[cnt]);
	  printf("]\n");
#endif
#endif
   if (bcmp( (char *) md_out,  (char *)otw_digest->digest, 
             NASD_MIN(SHA_DIGEST_LENGTH,NASD_DIGEST_SIZE))) {
#ifdef PRINT_ERROR
	  printf("DEBUG(nasd_security_verify_args) FAILED NASD_BAD_DIGEST_REQ\n");
#endif
	  return(NASD_BAD_DIGEST_REQ);
   }
 }

 
  }
 
#ifdef DEBUG_VERIFY_ARGS
  printf("DEBUG(nasd_security_verify_args) exit\n");
#endif
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  return(NASD_SUCCESS); /* see, nice and fast */
}
 
nasd_status_t nasd_security_digest_res(nasd_args_t *in_args,
                                       nasd_key_t  in_key,
				       nasd_byte_t *out_res_otw,
				       nasd_args_len_t out_res_len,
				       nasd_byte_t *out_data,
				       nasd_int32 out_data_len,
				       nasd_security_context_t *partial,
				       nasd_digest_nonce_otw_t out_digest)
{
#if NASD_SECURE_RPCS_ENABLE > 0
  HMAC_SHA1_CTX *HMAC_context_p;
  HMAC_SHA1_CTX HMAC_context;
  nasd_byte md_out[SHA_DIGEST_LENGTH];
  nasd_digest_nonce_t *otw_digest=(nasd_digest_nonce_t*)out_digest;
  nasd_timespec_t now;  
  nasd_int32 started=NASD_FALSE;
  nasd_digest_t tdigest;
  nasd_key_t integrity_key;
  
#ifdef DEBUG_DIGEST_RES
  printf("DEBUG(nasd_security_digest_res): entering\n");
  printf("partial = %p, out_data = %p\n",partial,out_data);
#endif

  NASD_ASSERT(! ((partial!=NULL) && ( out_data !=NULL)) );

  if (in_args->in_security_param.actual_protection != NASD_NO_PROTECTION) {
    nasd_security_get_drive_time(in_args->in_capability_args.drive_name,&now);
    nasd_timespec_t_to_net(&now,(nasd_byte *)&otw_digest->nonce);    


    /* First, we decide if we have a security context being handed to us
       or if we need to generate a new one. If we have to generate one, 
       our performance is going to be a lot slower since we are doing all
       the crypto at the last minute */
  
#ifdef DEBUG_DIGEST_RES
    printf("DEBUG(nasd_security_digest_res): "
           "key=[%08" NASD_64x_FMT " %08" NASD_64x_FMT "]\n",
           ((nasd_uint64*)in_key)[0],((nasd_uint64*)in_key)[1]);
#endif /* DEBUG_DIGEST_RES */
    /* If partial is non null, we have a partial digest to start us off.
       This also implies that the "data" has been digested allready 
       otherwise we start off the digest if one is needed*/

    if ( partial == NULL) {
#ifdef DEBUG_DIGEST_RES
      printf("DEBUG(nasd_security_digest_res): no partial digest\n");
#endif /* DEBUG_DIGEST_RES */
  

      if ((in_args->in_security_param.actual_protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) {
        /* Calculate the digest */
        /* This should never get excecuted -- data is now handled in 
           the pipes */
        nasd_security_generate_integrity_key(in_key, integrity_key);
	HMAC_SHA1_Init(&HMAC_context, (char *)integrity_key,
                       sizeof(nasd_key_t));
        started=NASD_TRUE;
	if ( (out_data != NULL) && (out_data_len>0) ) 
	  HMAC_SHA1_Update(&HMAC_context, out_data, out_data_len);
      }
      HMAC_context_p=&HMAC_context;


    } else {
#ifdef DEBUG_DIGEST_RES
      printf("DEBUG(nasd_security_digest_res):  partial digest to start with\n");
#endif
      HMAC_context_p=&partial->HMAC_context;
      if (partial->started)
	started=NASD_TRUE;
      
      NASD_ASSERT(in_args->in_security_param.actual_protection == partial->protection);
    }
  


   
    
  if ((in_args->in_security_param.actual_protection & NASD_INTEGRITY_ARGS) == NASD_INTEGRITY_ARGS) {
    /* Calculate the digest */
    if (!started) {
      /* this should never happen */
      started=NASD_TRUE;
      nasd_security_generate_integrity_key(in_key, integrity_key);
      HMAC_SHA1_Init(&HMAC_context, (char *) integrity_key,
                     sizeof(nasd_key_t));
      HMAC_context_p=&HMAC_context;
    }
  }



    /* At this point, we have an HMAC context established and initialized */
    
  
  /* Now we handle the finishing up the data */
#if (NASD_USE_HMAC_ONLY == 0)
    if (((in_args->in_security_param.actual_protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) && (partial != NULL) && (started== NASD_TRUE)){
      
      if (partial->curr_off % NASD_OD_BLOCK_SIZE) {
        /* Leftover data in an unkeyed digest (a trailing partial datablock) */
        SHA1_Final(NASD_FALSE, tdigest, &partial->SHA1_ctx);
#if DEBUG_DIGEST_RES > 0
        printf("digest_res (COMPUTED from running digest): "
               "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
               ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]   );
#endif /* DEBUB_DIGEST_RES */
        HMAC_SHA1_Update( &partial->HMAC_context, tdigest, NASD_DIGEST_SIZE);    
      }
    }
#endif /* (NASD_USE_HMAC_ONLY == 0) */
    /* Now we handle the result and the nonce */
    if ((in_args->in_security_param.actual_protection & NASD_INTEGRITY_ARGS) == NASD_INTEGRITY_ARGS) {
      HMAC_SHA1_Update(HMAC_context_p, (unsigned char *) out_res_otw,
                       out_res_len-sizeof(nasd_digest_nonce_t));
      HMAC_SHA1_Update(HMAC_context_p, (unsigned char *) &otw_digest->nonce,
                       sizeof(nasd_timespec_t));
    }
    
  }
    /* Generate a final digest */
  if (started) {  
    HMAC_SHA1_Final((unsigned char *)otw_digest->digest, HMAC_context_p);
#ifdef DEBUG_DIGEST_RES
    printf("DEBUG(nasd_security_digest_res): "
           "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n",
           ((nasd_uint64*)otw_digest->digest)[0],
           ((nasd_uint64*)otw_digest->digest)[1]);
#endif /* DEBUG_DIGEST_RES */
  } else {
    /* XXX - I zero fill when we don't use digests. This probably
       isn't necessary */
    bzero(( char *)otw_digest->digest, NASD_DIGEST_SIZE);
  }
  
#ifdef DEBUG_DIGEST_RES
  printf("DEBUG(nasd_security_digest_res): exiting\n");
#endif
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  return(NASD_SUCCESS); /* see, nice and fast */

}


 

nasd_status_t nasd_security_init_dr() {
#if NASD_SECURE_RPCS_ENABLE > 0

  nasd_status_t rc;
  SHA1_CTX ctx;
  char dummy[NASD_OD_BLOCK_SIZE];

  nasd_security_cache_init();
  bzero(dummy, NASD_OD_BLOCK_SIZE);
 
#if NASD_SECURITY_KEEP_STATS > 0

  rc= nasd_3DES_init_stats();
  if (rc) return rc;

  rc= nasd_SHA1_init_stats();
  if (rc) return rc;

  rc= nasd_HMAC_init_stats();
  if (rc) return rc;

#endif /* NASD_SECURITY_KEEP_STATS > 0 */

#if (NASD_USE_HMAC_ONLY == 0)
  /* Generate a precomputed digest on a zero-filled block for
     when we send clients a zero block */

  SHA1_Init(NASD_FALSE, &ctx);
  SHA1_Update(NASD_FALSE, &ctx, (unsigned char*) dummy, NASD_OD_BLOCK_SIZE);
  SHA1_Final(NASD_FALSE, zero_digest, &ctx);
#endif /*  (NASD_USE_HMAC_ONLY == 0) */

#if NASD_SECURITY_KEEP_STATS > 0

  /* Zero out the stats after we are done setting up
     because setup may have used the crypto functions */
  rc= nasd_3DES_zero_stats();
  if (rc) return rc;

  rc= nasd_SHA1_zero_stats();
  if (rc) return rc;

  rc= nasd_HMAC_zero_stats();
  if (rc) return rc;

#endif /* NASD_SECURITY_KEEP_STATS > 0 */

#if NASD_SECURITY_PRINT_STATS > 0
   rc = nasd_shutdown_proc(nasd_odc_shutdown,
                           nasd_HMAC_print_stats,NULL);
   if (rc) {
     return(rc);
   }
 
   rc = nasd_shutdown_proc(nasd_odc_shutdown,
                           nasd_SHA1_print_stats,NULL);
   if (rc) {
     return(rc);
   }
 
   rc = nasd_shutdown_proc(nasd_odc_shutdown,
                           nasd_3DES_print_stats,NULL);
   if (rc) {
     return(rc);
   }
#endif /* NASD_SECURITY_PRINT_STATS > 0 */

#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

  return(NASD_SUCCESS);
}




/* Capability Caching Code */

/* This is more or less a quick hack to to something reasonable.
   It doesn't expire entries when their expiration is passed nor
   does it boot entries when keys are changed (which is wrong but fine
   for our experiments which never last that long) --  
   it only flushes them out when the space is needed.
   */

nasd_status_t nasd_security_cache_init() {
#if NASD_SECURE_RPCS_ENABLE > 0
  int i;
  nasd_status_t rc=NASD_SUCCESS;
  nasd_security_cache_el_t * curr;
#ifdef DEBUG_CACHE_INIT
  printf("DEBUG(nasd_security_cache_init): entered\n");
#endif


  /* Initialize locks on each cache bucket and setup cleanup code*/

  for(i=0;(i<NASD_SECURITY_CACHE_BUCKETS) && (!rc);i++) {
    cache[i]=NULL;
    rc=nasd_mutex_init(&(cache_lock[i]));
    if (rc) {
      printf("ERROR: %s:%d nasd_mutex_init() failed, rc=%d\n", __FILE__, __LINE__, rc);
      return(rc);
    }
    rc=nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_mutex,
                           &(cache_lock[i]));
    if (rc) {
      nasd_mutex_destroy(&(cache_lock[i]));
      printf("ERROR: %s:%d cannot establish mutex autodestroy, rc=%d\n", __FILE__, __LINE__, rc);
      return(rc);
    }
  }

  /* Initialize freelist -- after the drive runs a while it will empty out */
  rc=nasd_mutex_init(&(cache_freelist_lock));
  if (rc) {
      printf("ERROR: %s:%d nasd_mutex_init() failed, rc=%d\n", __FILE__, __LINE__, rc);
    return(rc);
  }
  rc = nasd_shutdown_proc(nasd_odc_shutdown, nasd_shutdown_mutex,
                          &cache_freelist_lock);
  if (rc) {
      printf("ERROR: %s:%d cannot establish mutex autodestroy, rc=%d\n", __FILE__, __LINE__, rc);
    nasd_mutex_destroy(&cache_freelist_lock);
    return(rc);
  }

  NASD_Malloc(cache_freelist, sizeof(nasd_security_cache_el_t), 
              (nasd_security_cache_el_t *));
  


  if (cache_freelist == NULL) {
    printf("ERROR: %s:%d out of memory\n", __FILE__, __LINE__);
    return(NASD_NO_MEM);
  }
  
  rc = nasd_shutdown_mem(nasd_odc_shutdown, cache_freelist, 
                         sizeof(nasd_security_cache_el_t));
  if (rc) {
    NASD_Free(cache_freelist, sizeof(nasd_security_cache_el_t));
    cache_freelist=NULL;
    printf("ERROR (%s:%d): cannot establish autofree, rc=%lu\n",
           __FILE__, __LINE__, (long)rc);
    return(rc);
  }
    

  /* When I have time, I need to make this cleaner and more efficient 
     -hgobioff */
  cache_freelist->id=1;
  curr=cache_freelist;
  cache_entry_count=1;
  for(i=1;i<NASD_SECURITY_CACHE_ELEMENTS_DEFAULT;(i++)&(curr!=NULL)) {
    NASD_Malloc(curr->next, sizeof(nasd_security_cache_el_t),
                (nasd_security_cache_el_t *));
   
    if (curr->next == NULL) {
      printf("ERROR: %s:%d out of memory\n", __FILE__, __LINE__);
      return(NASD_NO_MEM);
    }
    curr->next->id=i+1;
    rc = nasd_shutdown_mem(nasd_odc_shutdown, curr->next, sizeof(nasd_security_cache_el_t));
    if (rc) {
      NASD_Free(curr->next, sizeof(nasd_security_cache_el_t));
      curr->next=NULL;
      printf("ERROR (%s:%d): cannot establish autofree, rc=%lu\n",
             __FILE__, __LINE__, (long)rc);
      return(rc);
      
    }
    cache_entry_count++;
    cache_freelist_size++;
    curr=curr->next;
  }
 
  if (curr!=NULL)
    curr->next=NULL;
  
  if (cache_entry_count != NASD_SECURITY_CACHE_ELEMENTS_DEFAULT) 
    {
#ifdef PRINT_ERROR
    printf("DEBUG(nasd_security_cache_init) NASD_NO_MEM allocated %d entries\n",cache_entry_count);
#endif
      return(NASD_NO_MEM);
    }

#ifdef DEBUG_CACHE_INIT
  printf("DEBUG(nasd_security_cache_init): exited with %d entries allocated\n",cache_entry_count);
#endif
#endif /* NASD_SECURE_RPCS_ENABLE >0*/
  return(NASD_SUCCESS);


}

nasd_status_t nasd_security_cache_invalidate(nasd_partnum_t partnum,
                                             nasd_identifier_t ni){

#if NASD_SECURE_RPCS_ENABLE > 0
int bucket;
nasd_security_cache_el_t *tmp;
int do_head=NASD_TRUE;
nasd_security_cache_el_t *curr;

#ifdef DEBUG_CACHE_INVALIDATE
 printf("DEBUG(nasd_security_cache_invalidate): entered\n");
#endif

bucket=NASD_SECURITY_NI_HASH(ni);

NASD_LOCK_MUTEX(cache_lock[bucket]);
/* Check the first element */
while (do_head==NASD_TRUE) {
  if (cache[bucket] != NULL ) {
    if ((cache[bucket]->cap_args.ni == ni ) &&
       (cache[bucket]->cap_args.partnum == partnum ))
 { 
#ifdef DEBUG_CACHE_INVALIDATE
 printf("DEBUG(nasd_security_cache_invalidate): freeing something at the start of the chain\n");
#endif
      /* Remove this element */
      NASD_LOCK_MUTEX(cache_freelist_lock);
      tmp=cache[bucket];
      cache[bucket]=cache[bucket]->next;
      tmp->next=cache_freelist;
      cache_freelist=tmp;
      cache_freelist_size++;
      NASD_UNLOCK_MUTEX(cache_freelist_lock);
      
    } else
      do_head=NASD_FALSE;
  } else
    do_head=NASD_FALSE;
}
/* Check the rest */

#ifdef DEBUG_CACHE_INVALIDATE
 printf("DEBUG(nasd_security_cache_invalidate): checking the rest of the chain\n");
#endif

if (cache[bucket] != NULL ) {
  curr=cache[bucket];

  while(curr->next!=NULL) {
    if ((curr->next->cap_args.ni == ni ) &&
        (curr->next->cap_args.partnum == partnum )) {
#ifdef DEBUG_CACHE_INVALIDATE
      printf("DEBUG(nasd_security_cache_invalidate): removing something along the chain\n");
#endif
      
      /* Remove this element */
      NASD_LOCK_MUTEX(cache_freelist_lock);
      tmp=curr->next;
      curr->next=curr->next->next;
      tmp->next=cache_freelist;
      cache_freelist=tmp;
      cache_freelist_size++;
      NASD_UNLOCK_MUTEX(cache_freelist_lock);
    }
    else
      curr=curr->next;
  }
}

NASD_UNLOCK_MUTEX(cache_lock[bucket]);
#ifdef DEBUG_CACHE_INVALIDATE
 printf("DEBUG(nasd_security_cache_invalidate): exited\n");
#endif
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
return (NASD_SUCCESS);
}


/* This should cleanup expired stuff but it doesn't yet - xxx */
nasd_status_t nasd_security_cache_lookup(nasd_capability_args_t *cap_args,
                                         nasd_key_t out_key) {
#if NASD_SECURE_RPCS_ENABLE > 0  
  int bucket;
  nasd_security_cache_el_t *candidate;
  int done=NASD_FALSE;
  nasd_security_cache_el_t *curr;
  int i;
#ifdef DEBUG_CACHE_LOOKUP
 printf("DEBUG(nasd_security_cache_lookup): entered\n");
#endif
   
  bucket=NASD_SECURITY_NI_HASH(cap_args->ni);
#ifdef DEBUG_CACHE_LOOKUP
 printf("DEBUG(nasd_security_cache_lookup): bucket %d\n", bucket);
#endif  
  NASD_LOCK_MUTEX(cache_lock[bucket]);
#ifdef DEBUG_CACHE_LOOKUP
 printf("DEBUG(nasd_security_cache_lookup): locked %d\n", bucket);
#endif  

  curr=cache[bucket];
  candidate=NULL;

  while ((curr!=NULL)&(!done)) {
#ifdef DEBUG_CACHE_LOOKUP
 printf("DEBUG(nasd_security_cache_lookup): checking %d\n", curr->id);
#endif  
    if (memcmp(&(curr->cap_args) , cap_args, sizeof(nasd_capability_args_t))== 0) {
      candidate=curr;
      done=NASD_TRUE;
#ifdef DEBUG_CACHE_LOOKUP
      printf("DEBUG(nasd_security_cache_lookup): candidate found\n");

#endif
     } 
    
    
    
#ifdef DEBUG_CACHE_LOOKUP
 printf("DEBUG(nasd_security_cache_lookup): step\n", bucket);
#endif
    curr=curr->next;
  }
     
  if (candidate!=NULL) {
    bcopy( (char *) candidate->the_key,  (char *)out_key,
           sizeof(nasd_key_t));
    NASD_UNLOCK_MUTEX(cache_lock[bucket]);
   
#ifdef DEBUG_CACHE_LOOKUP
    printf("DEBUG(nasd_security_cache_lookup): exited with a match on %d\n",
           candidate->id);
#endif
    return(NASD_SUCCESS);
  }
  NASD_UNLOCK_MUTEX(cache_lock[bucket]);

#ifdef DEBUG_CACHE_LOOKUP
 printf("DEBUG(nasd_security_cache_lookup): exited without a match\n");
#endif  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
 return(NASD_FAIL);
}

nasd_status_t nasd_security_cache_new_daily_key() {

#if NASD_SECURE_RPCS_ENABLE > 0  
  /* this should change all the which_key fields to the cached
     keys and kick out all the old ones. Should be done atomically
     so we want to lock the entire cache */

  /* XXX TODO */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  return(NASD_FAIL);
}


nasd_status_t nasd_security_cache_insert(nasd_capability_args_t *cap_args,nasd_key_t in_key){

#if NASD_SECURE_RPCS_ENABLE > 0  
  nasd_status_t rc;
  int bucket,src_bucket,i;
  nasd_security_cache_el_t *tmp, *prev;
#ifdef DEBUG_CACHE_INSERT
 printf("*DEBUG(nasd_security_cache_insert): entered\n");
#endif   

  bucket=NASD_SECURITY_NI_HASH(cap_args->ni);

  NASD_LOCK_MUTEX(cache_freelist_lock);
  if (cache_freelist!=NULL) {
    tmp=cache_freelist;
    cache_freelist=cache_freelist->next;
    NASD_UNLOCK_MUTEX(cache_freelist_lock);
    cache_freelist_size--;
  } else {

#ifdef DEBUG_CACHE_INSERT
    printf("**DEBUG(nasd_security_cache_insert): stealing entry\n");
#endif   
    NASD_UNLOCK_MUTEX(cache_freelist_lock);  
  
    src_bucket=bucket;  

    /* Find a bucket to steal from */
    if ( cache[src_bucket] == NULL) {
   
      for(src_bucket=(src_bucket+1) % NASD_SECURITY_CACHE_BUCKETS;  src_bucket != bucket; src_bucket = (src_bucket+1) % NASD_SECURITY_CACHE_BUCKETS) {
   
       
        NASD_LOCK_MUTEX(cache_lock[src_bucket]);    
        if (cache[src_bucket] != NULL)  break;
        /* Release the lock since the bucket is empty */
        NASD_UNLOCK_MUTEX(cache_lock[src_bucket]);       
      }
    } 
    
#ifdef DEBUG_CACHE_INSERT
 printf("*DEBUG(nasd_security_cache_insert): from bucket %d\n", src_bucket);
#endif   
    /* We exit holding a lock on bucket[src_bucket] */

    /* Find the last element in the bucket */
    NASD_ASSERT(cache[src_bucket] != NULL);
    
    tmp=cache[src_bucket];
    prev=NULL;
    while(tmp->next != NULL) {
      prev=tmp;
      tmp=tmp->next;
    }
    
    /* We know we are removing tmp from bucket[src_bucket]. 
       If tmp is the head of the bucket, prev == NULL 
    */
    
    /* remove tmp from the bucket */
    if (prev == NULL) {
      cache[src_bucket]=NULL;
    } else {
      prev->next=NULL;
    }
    tmp->next=NULL;
#ifdef DEBUG_CACHE_INSERT
  printf("DEBUG(nasd_security_cache_insert): stole %d\n", tmp->id);
#endif 
    NASD_UNLOCK_MUTEX(cache_lock[src_bucket]);    
  }

  bcopy(cap_args, &tmp->cap_args, sizeof(nasd_capability_args_t));
  bcopy(in_key, tmp->the_key, sizeof(nasd_key_t));

  NASD_LOCK_MUTEX(cache_lock[bucket]);
  tmp->next=cache[bucket];
  cache[bucket]=tmp;
  NASD_UNLOCK_MUTEX(cache_lock[bucket]);
#ifdef DEBUG_CACHE_INSERT
  printf("DEBUG(nasd_security_cache_insert): exited w/%d in bucket %d\n", tmp->id, bucket);
#endif 
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  
  return(NASD_SUCCESS);
}



void nasd_security_dr_print_capability(nasd_key_t in_key,
                                       nasd_capability_args_t *cap_args)
{
  int i;
       printf("Capability {[%08" NASD_64x_FMT "],ni=%ld,expire=%d,"
              "rights=%x,min_prot=%d,drive_name=%d,AuditID=%d,type=%d,"
              "region_start=%lu,region_end=%lu,partnum=%u, pad =[",
       ((nasd_uint64 *)in_key)[0],
       cap_args->ni,
       cap_args->expiration_seconds,
       cap_args->rights,

       cap_args->min_protection,
       cap_args->drive_name,
       cap_args->AuditID,
       cap_args->type,

       cap_args->region_start,
       cap_args->region_end,
       cap_args->partnum);
       for(i=0;i<6;i++) 
         printf("%x",cap_args->pad[i]);
       printf("]\n");
}



nasd_status_t nasd_security_get_request_key(nasd_args_t *in_args,
                                            nasd_byte_t *in_args_otw,
                                            nasd_args_len_t in_args_len,
                                            nasd_key_t out_key){
#if NASD_SECURE_RPCS_ENABLE > 0
  HMAC_SHA1_CTX HMAC_context;
  nasd_args_t *otw_args = (nasd_args_t *)in_args_otw;
  nasd_attribute_t tmp_attr;
  nasd_status_t rc;
  nasd_byte avbuf[sizeof(nasd_av_t)];
  nasd_key_t tmp_key;
  nasd_byte key_buff[SHA_DIGEST_LENGTH];
  
  if (in_args->in_security_param.actual_protection != NASD_NO_PROTECTION ) {
    
    if (in_args->in_security_param.actual_protection & NASD_PRIVACY_CAPABILITY) {
      /* At this point, we would need to decrypt that capability
         using the key indicated by security_param.type and 
         security_param.partnum */
      printf("This code should not be executed\n");
    }
  
    if ((in_args->in_security_param.type == NASD_BLACK_CAPABILITY )||
       (in_args->in_security_param.type == NASD_RED_CAPABILITY ))
      {
       
#if NASD_USE_KEY_CACHE > 0
      
        rc= nasd_security_cache_lookup(&in_args->in_capability_args, out_key);
       if (rc == NASD_SUCCESS) 
         {
          
         return (NASD_SUCCESS);
         }
#endif     
  
       /* We hope this will hit in the cache */
      rc = nasd_obj_getattr(in_args->in_capability_args.partnum,in_args->in_capability_args.ni,&tmp_attr);

      if (rc) {
#ifdef PRINT_ERROR_
	printf("DEBUG(nasd_security_get_request_key) received error %d when getting av for object %lx on partition %d\n",rc,in_args->in_capability_args.ni,in_args->in_capability_args.partnum);
#endif      
	return rc;
      }

      nasd_av_t_to_net(&tmp_attr.av,avbuf);

      if (in_args->in_security_param.type == NASD_BLACK_CAPABILITY)
        nasd_security_get_black_key(in_args->in_capability_args.partnum,tmp_key);
      else
        nasd_security_get_red_key(in_args->in_capability_args.partnum,tmp_key);
        
      /* Build the capability key */
      HMAC_SHA1_Init(&HMAC_context,(char *)tmp_key,sizeof(nasd_key_t));
      HMAC_SHA1_Update(&HMAC_context,(unsigned char *)&otw_args->in_capability_args,sizeof(nasd_capability_args_t));
      HMAC_SHA1_Update(&HMAC_context,(unsigned char *)avbuf,sizeof(nasd_av_t));
      HMAC_SHA1_Final((unsigned char *)key_buff,&HMAC_context);
      /* Stash the key so we have it again for the reply */
      bcopy( (char *)key_buff,  (char *)out_key, MIN(SHA_DIGEST_LENGTH,sizeof(nasd_key_t)));
 
#if NASD_USE_KEY_CACHE > 0
      /* I should cache the integrity and privacy keys rather than
         the credential key but this gets the right idea across for now
         - hgobioff */
      nasd_security_cache_insert(&in_args->in_capability_args, out_key);
#endif
#ifdef DEBUG_GET_KEY
      nasd_security_dr_print_capability(out_key,&in_args->in_capability_args);
      printf("w/AV = %d\n",tmp_attr.av);
#endif      

    } else if (in_args->in_security_param.type==NASD_RED_KEY) {
#ifdef DEBUG_GET_KEY
      printf("DEBUG(nasd_security_get_request_key) red key \n");
#endif
      nasd_security_get_red_key(in_args->in_security_param.partnum,out_key);
    } else if (in_args->in_security_param.type==NASD_BLACK_KEY) {
#ifdef DEBUG_GET_KEY
      printf("DEBUG(nasd_security_get_request_key) black key \n");
#endif
      nasd_security_get_black_key(in_args->in_security_param.partnum,out_key);
    } else if (in_args->in_security_param.type==NASD_PARTITION_KEY) {
#ifdef DEBUG_GET_KEY
  printf("DEBUG(nasd_security_get_request_key) partition key \n");
#endif
      nasd_security_get_partition_key(in_args->in_security_param.partnum,out_key);
    } else if (in_args->in_security_param.type==NASD_DRIVE_KEY) {
#ifdef DEBUG_GET_KEY
      printf("DEBUG(nasd_security_get_request_key) drive key \n");
#endif
      nasd_security_get_drive_key(out_key);
    } else if (in_args->in_security_param.type==NASD_MASTER_KEY) {
#ifdef DEBUG_GET_KEY
      printf("DEBUG(nasd_security_get_request_key) master key \n");
#endif
      nasd_security_get_drive_key(out_key);
    } else {
#ifdef  PRINT_ERROR
      printf("DEBUG(nasd_security_get_request_key) FAILED NASD_BAD_KEYTYPE \n");
#endif
      return(NASD_BAD_KEYTYPE);
    }
  /* Stash the key for use on the reply */

#ifdef DEBUG_GET_KEY
  printf("DEBUG(nasd_security_get_request_key): "
         "key=[%08" NASD_64x_FMT " %08" NASD_64x_FMT "]\n",
         ((nasd_uint64*)out_key)[0],((nasd_uint64*)out_key)[1]);
#endif /* DEBUG_GET_KEY */

}
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
  
 return(NASD_SUCCESS); 
}






/* Updates the context on a read call -
   performs the digest and then encryption. The
   caller will then ship off the data to a client */

void nasd_security_context_update_send_dr(nasd_security_context_t *sec_ctx,
                                       nasd_byte_t *data, nasd_len_t len, 
                                       nasd_blkrec_t* blkp,
                                       nasd_int32 use_zero_digest)
{
#if NASD_SECURE_RPCS_ENABLE > 0
 nasd_digest_t tdigest;
 SHA1_CTX tctx;

#ifdef DEBUG_CONTEXT_UPDATE
  printf("DEBUG(nasd_security_context_update_send): entered \n");
#endif

 NASD_ASSERT(sec_ctx->started==NASD_TRUE);
 if ((data!=NULL) && (len>0) ) {
     if ((sec_ctx->protection & NASD_INTEGRITY_DATA) == NASD_INTEGRITY_DATA) {
#if ((NASD_OD_EXT_PTR > 0) && (NASD_USE_HMAC_ONLY == 0)) 
       if (blkp != NULL) {
         /* Someone handed us a pointer to a digest cache location */

         /* We assert that this should ONLY happened for aligned blocks */

         NASD_ASSERT(len  == NASD_OD_BLOCK_SIZE  );
         NASD_ASSERT( (sec_ctx->curr_off % NASD_OD_BLOCK_SIZE) == 0 );

         NASD_ODC_LOCK_BLOCK(blkp->odc_entp);
         bcopy( blkp->digest, tdigest, NASD_DIGEST_SIZE);
         NASD_ODC_UNLOCK_BLOCK(blkp->odc_entp);

#if  DEBUG_UPDATE_SEND > 0
  printf("update_send (precomputed): "
         "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
         ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]   );
#endif /* DEBUG_UPDATE_SEND */
         /* XXX Make this check better -- ie: add a flag */
         if (bcmp((char*)tdigest, nasd_odc_zeroblk , NASD_DIGEST_SIZE) == 0) {
           SHA1_Init(NASD_FALSE, &tctx);
           SHA1_Update(NASD_FALSE, &tctx, data, len);
           SHA1_Final(NASD_FALSE, tdigest, &tctx);
#if DEBUG_UPDATE_SEND > 0
  printf("update_send (COMPUTED for datablock): "
         "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
         ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]   );
#endif /* DEBUB_UPDATE_SEND */
           NASD_ODC_LOCK_BLOCK(blkp->odc_entp);
           bcopy( tdigest, blkp->digest, NASD_DIGEST_SIZE);
           NASD_ODC_UNLOCK_BLOCK(blkp->odc_entp);
         }
         HMAC_SHA1_Update(&sec_ctx->HMAC_context,tdigest, NASD_DIGEST_SIZE);         
       } else {
#endif /* ((NASD_OD_EXT_PTR > 0) && (NASD_USE_HMAC_ONLY == 0))  */

#if (NASD_USE_HMAC_ONLY == 0)
         /* We were not given a cache location -- we need to discover what to do */
         if ( use_zero_digest &&  (len  ==  NASD_OD_BLOCK_SIZE ) &&
              ( (sec_ctx->curr_off  %  NASD_OD_BLOCK_SIZE) == 0 )) {
#if DEBUG_UPDATE_SEND > 0
  printf("update_send (using zerodigest): "
         "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
         ((nasd_uint64*)zero_digest)[0], ((nasd_uint64*)zero_digest)[1]   );
#endif /* DEBUB_UPDATE_SEND */


           HMAC_SHA1_Update(&sec_ctx->HMAC_context, zero_digest, NASD_DIGEST_SIZE);         
        } else {
           /* Not a full zero block -- proceed normally */
           
           SHA1_Update(NASD_FALSE, &sec_ctx->SHA1_ctx, data, len);
           if ((sec_ctx->curr_off+len) % NASD_OD_BLOCK_SIZE == 0) {
             /* We are ending on a block boundary so we add this digest to the keyed digest */
             SHA1_Final(NASD_FALSE, tdigest, &sec_ctx->SHA1_ctx);
             SHA1_Init(NASD_FALSE, &sec_ctx->SHA1_ctx);
#if DEBUG_UPDATE_SEND > 0
             printf("update_send (COMPUTED from running unkeyed): "
                    "digest=[%08" NASD_64x_FMT "%08" NASD_64x_FMT "]\n\n",
                    ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]   );
#endif /* DEBUB_UPDATE_SEND */
             HMAC_SHA1_Update(&sec_ctx->HMAC_context,tdigest, NASD_DIGEST_SIZE);         
           }
         }
#else
         HMAC_SHA1_Update(&sec_ctx->HMAC_context, data, len);         
#endif  /*(NASD_USE_HMAC_ONLY == 0) */

#if ((NASD_OD_EXT_PTR > 0) && (NASD_USE_HMAC_ONLY == 0)) 
       }
#endif /* ((NASD_OD_EXT_PTR > 0) && (NASD_USE_HMAC_ONLY == 0))  */     
     }
     if ((sec_ctx->protection & NASD_PRIVACY_DATA) == NASD_PRIVACY_DATA) {
     }
  

     sec_ctx->curr_off+= len;     
 }
#ifdef DEBUG_CONTEXT_UPDATE
  printf("DEBUG(nasd_security_context_update_send): exited \n");
#endif
  
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
} 
     




/* We set up the digest to partial digest to hold the 
   message digest context and the key for our running
   digest and encryption */
void nasd_security_reply_init_dr(nasd_args_t *in_args, 
                                 nasd_key_t in_key, 
                                 nasd_byte_t *icv,
                                 nasd_security_context_t *sec_ctx,
                                 nasd_offset_t start_off)
{
#if NASD_SECURE_RPCS_ENABLE > 0
  unsigned char enc_key[3][8];
  unsigned char *keys[3];
  nasd_key_t integrity_key, priv_key;
#ifdef DEBUG_REPLY_INIT
  printf("DEBUG(nasd_security_reply_init): entered \n");
  printf("DEBUG(nasd_security_reply_init): "
         "in_key=%08" NASD_64x_FMT " %08" NASD_64x_FMT "\n",
         ((nasd_uint64*)in_key)[0], ((nasd_uint64*)in_key)[1]);
#endif /* DEBUG_REPLY_INIT */
  
  sec_ctx->curr_off=start_off;
  sec_ctx->protection=in_args->in_security_param.actual_protection;
  sec_ctx->started=NASD_TRUE;
  if ((sec_ctx->protection & (NASD_INTEGRITY_DATA|NASD_INTEGRITY_ARGS) ) !=0 ) {
    /* We are digesting something */
#if  (NASD_USE_HMAC_ONLY == 0) 
    SHA1_Init(NASD_FALSE, &sec_ctx->SHA1_ctx);
#endif /*  (NASD_USE_HMAC_ONLY == 0) */
    nasd_security_generate_integrity_key(in_key, integrity_key);
    HMAC_SHA1_Init(&sec_ctx->HMAC_context,(char *)integrity_key,sizeof(nasd_key_t));
  }

 if ((in_args->in_security_param.actual_protection & (NASD_PRIVACY_DATA | NASD_PRIVACY_ARGS)) > 0) {
    NASD_ASSERT( sizeof(nasd_key_t) > 8 );
 
    nasd_security_generate_privacy_key(in_key, priv_key); 
    bcopy(&priv_key[0], enc_key[0], 8);
    bcopy(&priv_key[sizeof(nasd_key_t)-8], enc_key[1], 8);
    keys[0]=enc_key[0];
    keys[1]=enc_key[1];
    keys[2]=enc_key[0];
 }
#ifdef DEBUG_REPLY_INIT
  printf("DEBUG(nasd_security_reply_init): exited \n");
#endif /* DEBUG_REPLY_INIT */
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
}


nasd_status_t nasd_security_context_update_receive_dr(nasd_security_context_t *sec_ctx,
                                                      nasd_byte_t *data,
                                                      nasd_len_t len, 
                                                      nasd_procpipe_t *byte_pipe,
                                                      nasd_len_t remain,
                                                      nasd_blkrec_t *blkrec,
                                                      nasd_timespec_t *pipe_stall) {

  nasd_digest_t tdigest, tdigest_SHA1;
  nasd_digest_t in_digest;
  int r;
  nasd_len_t count;
  HMAC_SHA1_CTX tcontext;
  nasd_status_t prc;
  nasd_timer_t tm;

  /* No context == No work */
  if (sec_ctx) {


    /* decrypt data iff encrypted */
    if (sec_ctx->protection & NASD_PRIVACY_DATA) {
    }
   
    if (sec_ctx->protection & NASD_INTEGRITY_DATA) {
#if  (NASD_USE_HMAC_ONLY == 0) 
      SHA1_Update(NASD_FALSE, &sec_ctx->SHA1_ctx, data, len);
#else        
      HMAC_SHA1_Update(&sec_ctx->HMAC_context, data, len);

#endif /*  (NASD_USE_HMAC_ONLY == 0)  */
            sec_ctx->curr_off+=len;
#if  (NASD_USE_HMAC_ONLY == 0) 
            if ((sec_ctx->curr_off % NASD_OD_BLOCK_SIZE) == 0) {
              SHA1_Final(NASD_FALSE, tdigest_SHA1, &sec_ctx->SHA1_ctx);
              SHA1_Init(NASD_FALSE, &sec_ctx->SHA1_ctx);
              HMAC_SHA1_Update(&sec_ctx->HMAC_context, tdigest_SHA1, sizeof(nasd_digest_t));
            }
#endif /*  (NASD_USE_HMAC_ONLY == 0) */            

#if  DEBUG_UPDATE_RECEIVE > 0
            printf("current offset in ctx =%08" NASD_64u_FMT "\n",sec_ctx->curr_off);
#endif 
            
            if ((sec_ctx->curr_off % NASD_DIGEST_BLOCK_SIZE == 0) || (remain == 0)) {
              
              /* Calculate and intermediate digest */
              bcopy((char *) &sec_ctx->HMAC_context, (char *)&tcontext, 
                    sizeof(HMAC_SHA1_CTX));
              
              HMAC_SHA1_Final((unsigned char*)tdigest, &tcontext);
              
#if  DEBUG_UPDATE_RECEIVE > 0
              printf(" Digest caclulated was:\t%08" NASD_64u_FMT "%08" NASD_64u_FMT " \n", 
                     ((nasd_uint64*)tdigest)[0], ((nasd_uint64*)tdigest)[1]);
#endif 
              /* Suck in a digest */
              r=sizeof(nasd_digest_t);
              if (pipe_stall != NULL)
                NASD_TM_START(&tm);
              while(r) {
                prc = byte_pipe->pull(byte_pipe->state,
                                      (nasd_byte_t *)in_digest, r, &count);
                if (prc)
                  break;
                r -= count;
              }
              if (pipe_stall != NULL)
                NASD_TM_STOP_ACCUM_TS(&tm,pipe_stall);
              
              if (prc)
                NASD_PANIC();  
#if  DEBUG_UPDATE_RECEIVE > 0
              printf(" Digest received   was:\t%08" NASD_64u_FMT "%08" NASD_64u_FMT " \n", 
                     ((nasd_uint64*)in_digest)[0], ((nasd_uint64*)in_digest)[1]);
#endif 
              
              if (bcmp( (char *)in_digest,  (char *)tdigest, 
                        sizeof(nasd_digest_t)))
                {
                  printf("Received BAD DIGEST \n");
                  return(NASD_BAD_DIGEST);
                }
              
              return (NASD_GOOD_DIGEST);
            }
    }

#if  ((NASD_OD_EXT_PTR > 0) && (NASD_USE_HMAC_ONLY == 0))        
    /* We won't have a blkrec if this is an atomic block */
    if (blkrec != NULL) {
      NASD_ODC_LOCK_BLOCK(blkrec->odc_entp);
      blkrec->odc_entp->data_flags |= (NASD_CD_BUSY | NASD_CD_SECURITY);
      /* If we just finished an unkeyed digest block AND we know that we
         have digested a  full blocks worth of data */
      if (((sec_ctx->curr_off % NASD_OD_BLOCK_SIZE) == 0)  && ( (sec_ctx->curr_off - sec_ctx->start_off) >= NASD_OD_BLOCK_SIZE)){
        bcopy(tdigest_SHA1, (char *)blkrec->digest, sizeof(nasd_digest_t));
      } else {
        /* Eventually this wil be something more intelligent than zeroing */
        bzero((char *)blkrec->digest, sizeof(nasd_digest_t));
      }
      NASD_ODC_UNLOCK_BLOCK(blkrec->odc_entp);
    }
#endif /*  ((NASD_OD_EXT_PTR > 0) && (NASD_USE_HMAC_ONLY == 0)) */


  }
  return(NASD_SUCCESS);
}


