/*
 * pikdrive.c
 *
 * in-kernel pdrive user binary
 * (makes server syscall)
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_drive_options.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_sys.h>
#include <nasd/nasd_timer.h>
#include <nasd/nasd_drive_defaults.h>
#include <sys/stat.h>
#if LINUX
#include <getopt.h>
#endif /* LINUX */
#include <stdio.h>
#include <string.h>

extern void sync(void);

char *progname;
int doformat = 0;

int is_secure = 0;

int svc_threads = NASD_DEF_SERVICE_THREADS;
int cache_blocks = 1024;
int stack_size = 0;
int use_tcp = 0;

nasd_layout_type_t layout = NASD_DEF_LAYOUT;

nasd_ioqueue_type_t ioqueue = NASD_DEF_IOQUEUE;

int io_outstanding = NASD_DEF_IO_OUTSTANDING;

unsigned long max_len = 0;

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] devicename\n", progname);
  fprintf(stderr, "  -c   cache blocks [default %d]\n", cache_blocks);
  fprintf(stderr, "  -f   format\n");
  fprintf(stderr, "  -i   I/Os outstanding [default %d]\n",
    NASD_DEF_IO_OUTSTANDING);
  fprintf(stderr, "  -l   layout (default '%c')\n", NASD_DEF_LAYOUT);
  fprintf(stderr, "  -m   maximum disk length to use\n");
  fprintf(stderr, "  -q   disk queue policy (default '%c')\n",
    NASD_DEF_IOQUEUE);
  fprintf(stderr, "  -s   stack size (0 = system default) [default %d]\n",
    stack_size);
  fprintf(stderr, "  -t   service threads [default %d]\n",
    NASD_DEF_SERVICE_THREADS);
  fprintf(stderr, "  -T   use TCP\n");
  fflush(stderr);
  exit(1);
}

int
main(
  int     argc,
  char  **argv)
{
  nasd_delaycounter_t delayer;
  int port, rc, md_level;
  struct nasd_serv srv;
  nasd_options_t opts;
  char c, *root_path;
  struct stat s;
  dev_t dev;

  progname = argv[0];
  md_level = 0;

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  if (sscanf(NASD_PDRIVE_PORT, "%d", &port) != 1) {
    fprintf(stderr, "ERROR: %s internal error, cannot read port\n", progname);
    fflush(stderr);
    exit(1);
  }

  rc = nasd_srv(NASD_GET_OPTIONS, &opts);
  if (rc) {
    fprintf(stderr, "ERROR: %s cannot get nasd options from kernel\n",
      progname);
    fflush(stderr);
    exit(1);
  }

  if (opts.opts1 & NASD_OPTS1_SECURITY)
    is_secure = 1;
  else
    is_secure = 0;

  while ((c = getopt(argc, argv, "c:fi:l:m:q:s:t:T")) != EOF) {
    switch(c) {
      case 'c':
        if (sscanf(optarg, "%d", &cache_blocks) != 1)
          usage();
        break;
      case 'f':
        doformat = 1;
        break;
      case 'i':
        if (sscanf(optarg, "%d", &io_outstanding) != 1)
          usage();
        break;
      case 'l':
        if (strlen(optarg) != 1) {
          usage();
        }
        layout = optarg[0];
        break;
      case 'm':
        if (sscanf(optarg, "%lu", &max_len) != 1)
          usage();
        if (max_len == 0)
          usage();
        break;
      case 'q':
        if (strlen(optarg) != 1) {
          usage();
        }
        ioqueue = optarg[0];
        break;
      case 's':
        if (sscanf(optarg, "%d", &stack_size) != 1)
          usage();
        break;
      case 't':
        if (sscanf(optarg, "%d", &svc_threads) != 1)
          usage();
        break;
      case 'T':
        use_tcp = 1;
        break;
      default:
        usage();
    }
  }
  if (optind >= argc)
    usage();
  root_path = argv[optind];

  optind++;
  if (optind < argc)
    usage();

  if (svc_threads < 1)
    usage();
  if (cache_blocks < 1)
    usage();
  if (stack_size && (stack_size < 8192))
    usage();
  if (io_outstanding < 1)
    usage();

  if (strlen((char *)root_path) >= MAXPATHLEN) {
    fprintf(stderr, "ERROR: device path name too long\n");
    usage();
  }

  rc = stat(root_path, &s);
  if (rc) {
    perror(root_path);
    exit(1);
  }

  if (!S_ISCHR(s.st_mode)) {
    fprintf(stderr, "%s: %s is not a character device\n",
      progname, root_path);
    exit(1);
  }

  dev = s.st_rdev;    
 
  if (is_secure)
    printf("\nDrive compiled WITH security\n\n");
  else
    printf("\nDrive compiled WITHOUT security\n\n");

#if NASD_CMU_PDL > 0
  printf("Clearing local cache\n");
  rc = PDL_cflush();
  if (rc) {
    perror("PDL_cflush.1");
    fflush(stdout);
    exit(1);
  }
  sync(); sync(); sync();

  NASD_BEGIN_DELAYCNT(&delayer);
  NASD_DELAY_FROM(&delayer, NASD_USEC_PER_SEC/2);

  rc = PDL_cflush();
  if (rc) {
    perror("PDL_cflush.1");
    fflush(stdout);
    exit(1);
  }
  sync(); sync(); sync();
#endif /* NASD_CMU_PDL > 0 */

  printf("Ready to launch drive\n");
  fflush(stdout);

  strcpy(srv.root_path, root_path);
  srv.dev = dev;
  srv.do_format = doformat;
  srv.cache_blocks = cache_blocks;
  srv.svc_threads = svc_threads;
  srv.stack_size = stack_size;
  srv.port = port;
  srv.verbose = 1;
  srv.use_tcp = use_tcp;
  srv.security_enabled = 0;
  srv.max_disk_len = max_len;
  srv.config.layout_type = layout;
  srv.config.ioqueue_type = ioqueue;
  srv.config.ios_outstanding = io_outstanding;

  fflush(stdout);
  fflush(stderr);
  rc = nasd_srv(NASD_PSRV_GO, &srv);
  perror("nasd_srv(NASD_PSRV_GO)");

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
