/*
 * nasd_mem.h
 *
 * Generic memory allocation/deallocation
 */
/*
 * Copyright (c) 1995,1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Authors: Daniel Stodolsky, Mark Holland, Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#ifndef _NASD_MEM_H_
#define _NASD_MEM_H_

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>

#if __alpha
#define NASD_MEM_FENCEPOST_PATTERN 0xDEADBEEFDEADBEEFUL
#else /* __alpha */
#define NASD_MEM_FENCEPOST_PATTERN 0xDEADBEEFUL
#endif /* __alpha */

#if NASD_TRACK_ALLOCATIONS > 0
#define NASD_TRACK_ALLOC(_addr_,_sz_)   nasd_mem_track_alloc(_addr_,_sz_,__FILE__,__LINE__)
#define NASD_TRACK_DEALLOC(_addr_,_sz_) nasd_mem_track_dealloc(_addr_,_sz_,__FILE__,__LINE__)
#else /* NASD_TRACK_ALLOCATIONS > 0 */
#define NASD_TRACK_ALLOC(_addr_,_sz_)
#define NASD_TRACK_DEALLOC(_addr_,_sz_)
#endif /* NASD_TRACK_ALLOCATIONS > 0 */

#if NASD_MEM_FILL == 0
#define NASD_MEM_DO_FILL(_addr_,_nbytes_) { /* noop */ }
#else /* NASD_MEM_FILL == 0 */
extern nasd_uint32 nasd_mem_filler;
extern void nasd_mem_do_fill(void *addr, int nbytes);
#define NASD_MEM_DO_FILL(_addr_,_nbytes_) nasd_mem_do_fill(_addr_,_nbytes_)
#endif /* NASD_MEM_FILL == 0 */

#if NASD_MEM_FENCEPOST > 0
#define NASD_Fencepost(_p_,_sz_,_n_) { \
  unsigned long *_ll_; \
  void *_pp_, *_lpp_; \
  int _i; \
  _pp_ = (void *)(((unsigned long)_p_) - (NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
  _lpp_ = (void *)(((unsigned long)_p_) + _sz_); \
  _ll_ = _pp_; \
  if (_ll_[0] != _sz_) { \
    printf("%s:%d _ll_=%lx _ll_[0]=%lu _sz_=%lu n=%d\n", \
      __FILE__, __LINE__, _ll_, _ll_[0], _sz_, _n_); \
    NASD_PANIC(); \
  } \
  for(_i=1;_i<NASD_MEM_FENCEPOST;_i++) { \
    if (_ll_[_i] != NASD_MEM_FENCEPOST_PATTERN) { \
      printf("1-%d _ll_=0x%lx _ll_[1] = 0x%lx %s:%d n=%d\n", _i, _ll_, _ll_[1], \
        __FILE__, __LINE__, _n_); \
      NASD_PANIC(); \
    } \
  } \
  _ll_ = _lpp_; \
  for(_i=0;_i<NASD_MEM_FENCEPOST;_i++) { \
    if (_ll_[_i] != NASD_MEM_FENCEPOST_PATTERN) { \
      printf("2-%d _ll_=0x%lx _ll_[1] = 0x%lx %s:%d n=%d\n", _i, _ll_, _ll_[1], \
        __FILE__, __LINE__, _n_); \
      NASD_PANIC(); \
    } \
  } \
}
#else /* NASD_MEM_FENCEPOST > 0 */
#define NASD_MEM_Fencepost(_p_, _sz_, _n_) { /* noop */ }
#endif /* NASD_MEM_FENCEPOST > 0 */

#ifndef KERNEL

#include <stdlib.h>
#include <string.h>

#if NASD_MEM_COUNT_ALLOC > 0

extern nasd_int64 nasd_mem_allocated;

#define NASD_MEM_ALLOCATED(_addr_,_nbytes_) { \
  nasd_atomic_add64(&nasd_mem_allocated, _nbytes_); \
  NASD_TRACK_ALLOC(_addr_,_nbytes_); \
  NASD_MEM_DO_FILL(_addr_,_nbytes_); \
}

#define NASD_MEM_DEALLOCATED(_addr_,_nbytes_) { \
  nasd_atomic_sub64(&nasd_mem_allocated, _nbytes_); \
  NASD_TRACK_DEALLOC(_addr_,_nbytes_); \
}

#else /* NASD_MEM_COUNT_ALLOC > 0 */

#define NASD_MEM_ALLOCATED(_addr_,_nbytes_) { \
  NASD_MEM_DO_FILL(_addr_,_nbytes_); \
}
#define NASD_MEM_DEALLOCATED(_addr_,_nbytes_) { /* noop */ }

#endif /* NASD_MEM_COUNT_ALLOC > 0 */

#if NASD_MEM_FENCEPOST > 0

#define NASD_Malloc(_p_, _size_, _cast_) { \
  void *_ppp_, *_pp_, *_lpp_; \
  unsigned long *_ll_; \
  int _i; \
  _pp_ = _cast_ malloc(_size_ + (2*NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
  if (_pp_) { \
    _ppp_ = (void *)(((unsigned long)_pp_) + (NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
    _p_ = _ppp_; \
    NASD_MEM_ALLOCATED(_p_,_size_); \
    _lpp_ = (void *)(((unsigned long)_p_) + (_size_)); \
    _ll_ = _pp_; \
    _ll_[0] = _size_; \
    for(_i=1;_i<NASD_MEM_FENCEPOST;_i++) \
      _ll_[_i] = NASD_MEM_FENCEPOST_PATTERN; \
    _ll_ = _lpp_; \
    for(_i=0;_i<NASD_MEM_FENCEPOST;_i++) \
      _ll_[_i] = NASD_MEM_FENCEPOST_PATTERN; \
    NASD_Fencepost(_p_, _size_, -1); \
  } \
}

#define NASD_Free(_p_, _sz_) { \
  void *_pp_; \
  _pp_ = (void *)(((unsigned long)_p_) - (NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
  NASD_Fencepost(_p_, _sz_, -3); \
  free(_pp_); \
  NASD_MEM_DEALLOCATED(_p_,_sz_); \
}

#else /* NASD_MEM_FENCEPOST > 0 */

#define NASD_Malloc(_p_, _size_, _cast_) { \
  _p_ = _cast_ malloc(_size_); \
  NASD_MEM_ALLOCATED(_p_,_size_); \
}

#define NASD_Free(_p_, _sz_) { \
  free(_p_); \
  NASD_MEM_DEALLOCATED(_p_,_sz_); \
}

#endif /* NASD_MEM_FENCEPOST > 0 */

#else /* !KERNEL */

#include <sys/types.h>
#include <io/common/iotypes.h>
#include <sys/malloc.h>

#if NASD_MEM_COUNT_ALLOC > 0

extern nasd_int64 nasd_mem_allocated;

#define NASD_MEM_ALLOCATED(_addr_,_nbytes_) { \
  atomic_addq(&nasd_mem_allocated, _nbytes_); \
  NASD_TRACK_ALLOC(_addr_,_nbytes_); \
  NASD_MEM_DO_FILL(_addr_,_nbytes_); \
}

#define NASD_MEM_DEALLOCATED(_addr_,_nbytes_) { \
  atomic_addq(&nasd_mem_allocated, -(_nbytes_)); \
  NASD_TRACK_DEALLOC(_addr_,_nbytes_); \
}

#else /* NASD_MEM_COUNT_ALLOC > 0 */

#define NASD_MEM_ALLOCATED(_addr_,_nbytes_) { \
  NASD_MEM_DO_FILL(_addr_,_nbytes_); \
}
#define NASD_MEM_DEALLOCATED(_addr_,_nbytes_) { /* noop */ }

#endif /* NASD_MEM_COUNT_ALLOC > 0 */

extern vm_map_t kernel_map;

#define NASD_DK_MALLOC(_p_,_cast_,_size_) { \
  kern_return_t ret; \
  vm_offset_t kaddr; \
  if ((_size_)%PAGE_SIZE == 0) { \
    ret = vm_allocate(kernel_map, &kaddr, _size_, TRUE); \
    if (ret != KERN_SUCCESS) { \
      _p_ = _cast_ 0; \
    } \
    else { \
      _p_ = _cast_ kaddr; \
    } \
  } \
  else { \
    _p_ = _cast_ malloc((u_long)_size_, BUCKETINDEX(_size_), M_DEVBUF, M_WAITOK); \
  } \
}

#define NASD_DK_FREE(_p_,_size_) { \
  if ((_size_)%PAGE_SIZE == 0) { \
    vm_deallocate(kernel_map, (vm_offset_t)(_p_), _size_); \
  } \
  else { \
    free((void *)(_p_), M_DEVBUF); \
  } \
}

#if NASD_MEM_FENCEPOST > 0

#define NASD_Malloc(_p_, _rsize_, _cast_) { \
  void *_ppp_, *_pp_, *_lpp_; \
  unsigned long *_ll_; \
  int _size_, _i; \
  _size_ = (_rsize_ + (2*NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
  NASD_DK_MALLOC(_pp_,_cast_,_size_); \
  if (_pp_) { \
    _ppp_ = (void *)(((unsigned long)_pp_) + (NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
    _p_ = _ppp_; \
    NASD_MEM_ALLOCATED(_p_,_rsize_); \
    _lpp_ = (void *)(((unsigned long)_p_) + (_rsize_)); \
    _ll_ = _pp_; \
    _ll_[0] = _rsize_; \
    for(_i=1;_i<NASD_MEM_FENCEPOST;_i++) \
      _ll_[_i] = NASD_MEM_FENCEPOST_PATTERN; \
    _ll_ = _lpp_; \
    for(_i=0;_i<NASD_MEM_FENCEPOST;_i++) \
      _ll_[_i] = NASD_MEM_FENCEPOST_PATTERN; \
    NASD_Fencepost(_p_, _rsize_, -1); \
  } \
}

#define NASD_Free(_p_, _sz_) { \
  void *_pp_; \
  int _size_; \
  _size_ = (_sz_ + (2*NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
  _pp_ = (void *)(((unsigned long)_p_) - (NASD_MEM_FENCEPOST * sizeof(unsigned long))); \
  NASD_Fencepost(_p_, _sz_, -3); \
  NASD_DK_FREE(_pp_,_size_); \
  NASD_MEM_DEALLOCATED(_p_,_sz_); \
}

#else /* NASD_MEM_FENCEPOST > 0 */

#define NASD_Malloc(_p_, _size_, _cast_) { \
  NASD_DK_MALLOC(_p_,_cast_,_size_); \
  NASD_MEM_ALLOCATED(_p_,_size_); \
  /* bzero((char *)_p_, _size_); */ \
}

#define NASD_Free(_p_, _sz_) { \
  NASD_DK_FREE(_p_,_sz_); \
  NASD_MEM_DEALLOCATED(_p_,_sz_); \
}

#endif /* NASD_MEM_FENCEPOST > 0 */

#endif /* !KERNEL */

/*
 * Must directly invoke alloc/dealloc operations
 * for this list to avoid recursion.
 */
typedef struct nasd_mem_track_s nasd_mem_track_t;

struct nasd_mem_track_s {
  void              *addr;
  int                size;
  char              *file;
  int                line;
  nasd_mem_track_t  *next;
};

/*
 * Memory subsystem funcs
 */
extern nasd_status_t nasd_mem_dump_unfreed(void);
extern nasd_status_t nasd_mem_real_init(void);
extern void nasd_mem_real_shutdown(void);
extern void nasd_memsys_init(void);
extern nasd_status_t nasd_mem_init(void);
extern void nasd_mem_shutdown(void);
extern void nasd_mem_track_alloc(void *addr, int size,
  char *file, int line);
extern void nasd_mem_track_dealloc(void *addr, int size,
  char *file, int line);
extern nasd_mem_track_t *nasd_mem_gettrack(void);
extern void nasd_mem_freetrack(nasd_mem_track_t *t);

#endif /* !_NASD_MEM_H_ */

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
