/*
 * nasd_nfs_internal.h
 *
 * Internal structures and definitions for NASD NFS server.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#ifndef _NASD__NASD_NFS_INTERNAL_H_
#define _NASD__NASD_NFS_INTERNAL_H_

#include <nasd/nasd_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_shutdown.h>
#include <nasd/nasd_pdrive_client.h>

#include <nfs/nfs.h>

#ifdef LINUX
#define NF3DIR NFDIR
#define NF3REG NFREG
#define NF3LNK NFLNK
#endif /* LINUX */

typedef struct nasd_nfs_attr_s          nasd_nfs_attr_t;
typedef struct nasd_nfs_core_dir_s      nasd_nfs_core_dir_t;
typedef struct nasd_nfs_core_dirpage_s  nasd_nfs_core_dirpage_t;
typedef struct nasd_nfs_createpair_s    nasd_nfs_createpair_t;
typedef struct nasd_nfs_dirdata_s       nasd_nfs_dirdata_t;
typedef struct nasd_nfs_dirent_s        nasd_nfs_dirent_t;
typedef struct nasd_nfs_dirheader_s     nasd_nfs_dirheader_t;
typedef struct nasd_nfs_dirpage_s       nasd_nfs_dirpage_t;
typedef struct nasd_nfs_drive_s         nasd_nfs_drive_t;
typedef struct nasd_nfs_infopage_s      nasd_nfs_infopage_t;
typedef struct nasd_nfs_lookupstuff_s   nasd_nfs_lookupstuff_t;
typedef struct nasd_nfs_objhash_s       nasd_nfs_objhash_t;
typedef struct nasd_nfs_setattrpair_s   nasd_nfs_setattrpair_t;

#define NASD_NFS_NAMEHASH_BUCKETS 127

#if 1
#define NASD_NFS_OBJHASH_BUCKETS  127
#define nasd_nfs_obj_hash(_nid_) NASD_ABS((int)((_nid_)%NASD_NFS_OBJHASH_BUCKETS))
#else
#define NASD_NFS_OBJHASH_BUCKETS  129
#define nasd_nfs_obj_hash(_nid_) \
  (NASD_ABS(((((unsigned long)(_nid_))&0xfe00000000000000UL)>>57)|((((unsigned long)(_nid_))&0xffffffffUL)<<7))%NASD_NFS_OBJHASH_BUCKETS)
#endif

#define NASD_NFS_DIHASH_BUCKETS    16
#define NASD_NFS_DIHASH(_di_) ((_di_)&0xf)

#define NASD_NFS_DRIVENAME_LEN 64
#define NASD_NFS_DRPATH_LEN    1024

#define NASD_NFS_ROOT_UID 0

#define NASD_NFS_DH_SLOTFULL(_dh_,_slot_) \
  (((_dh_)->usemap[(_slot_)/8])&(1<<((_slot_)%8)))

/*
 * nfs_dirheader: 64 bytes
 */
struct nasd_nfs_dirheader_s {
  nasd_uint8  usemap[16];
  nasd_uint8  freecnt;
  char        spare[47];
};

#define NASD_NFS_DIRD_EXLEN(_dd_)  ((int)(((_dd_)->flags)&0x1f))
#define NASD_NFS_DIRD_TYPEOF(_dd_) ((((_dd_)->flags)&0x60)>>NASD_NFS_DIRD_SHIFT)

#define NASD_NFS_DIRD_SHIFT      5

#define NASD_NFS_DIRD_TYPE_BOGUS 0x0
#define NASD_NFS_DIRD_TYPE_DIR   0x1
#define NASD_NFS_DIRD_TYPE_LINK  0x2
#define NASD_NFS_DIRD_TYPE_DATA  0x3

#define NASD_NFS_MEMLIST_LEN 8

#define NASD_NFS_DIRDATA_SIZE               64
#define NASD_NFS_DIRDATA_EMBEDDED_NAME_SIZE 44

#define NASD_NFS_DIRPAGE_SIZE 8192

/*
 * nfs_dirdata: 64 bytes
 */
struct nasd_nfs_dirdata_s {
  nasd_uint64        flags;
  nasd_identifier_t  nasdid;
  nasd_uint32        name_len;
  char               name[NASD_NFS_DIRDATA_EMBEDDED_NAME_SIZE];
};

/*
 * nfs_dirpage: 8192 bytes
 */
struct nasd_nfs_dirpage_s {
  nasd_nfs_dirheader_t  header;
  nasd_nfs_dirdata_t    data_slots[127];
};

struct nasd_nfs_attr_s {
  nasd_p_getattr_dr_res_t   attr_res;
  int                       refcnt;
  nasd_nfs_attr_t          *next;
};
#define NASD_NFS_ATTR_ATTR(_a_) ((_a_)->attr_res.out_attribute)

struct nasd_nfs_drive_s {
  nasd_drive_handle_t    handle;
  int                    partnum;
  char                   drname[NASD_NFS_DRIVENAME_LEN];
  char                   pname[NASD_NFS_DRPATH_LEN];
  nasd_disk_ident_t      di;
  nasd_int32             portnum;
  nasd_uint32            net_addr;
  nasd_uint64            blocksize;
  nasd_nfs_core_dir_t   *root;
  nasd_nfs_drive_t      *next;
  nasd_nfs_drive_t      *dinext; /* for dihash */
};

struct nasd_nfs_dirent_s {
  char                     *name;
  nasd_nfs_core_dirpage_t  *page;
  nasd_nfs_dirdata_t       *data;
  nasd_nfs_attr_t          *attr;
  nasd_nfs_core_dir_t      *dir;
  nasd_nfs_dirent_t        *hnext;
  nasd_nfs_dirent_t        *hprev;
  nasd_nfs_dirent_t        *ihnext;
  nasd_nfs_dirent_t        *ihprev;
};

struct nasd_nfs_core_dir_s {
  nasd_identifier_t         nid;
  nasd_disk_ident_t         di;
  nasd_cookie_t             cookie;
  int                       refcnt;
  int                       deleted;
  int                       nentries;
  nasd_nfs_drive_t         *drive;
  nasd_nfs_core_dir_t      *lnext;
  nasd_nfs_core_dir_t      *lprev;
  nasd_nfs_core_dir_t      *dnext;
  nasd_nfs_core_dir_t      *dprev;
  nasd_nfs_core_dir_t      *hnext;
  nasd_nfs_core_dir_t      *hprev;
  nasd_nfs_attr_t          *attr;
  int                       isroot;
  NASD_DECLARE_MUTEX(lock)
  NASD_DECLARE_COND(cond)
  int                       busy;
  int                       ndirty;
  int                       npages;
  nasd_timespec_t           dirty_time;
  nasd_nfs_dirent_t        *buckets[NASD_NFS_NAMEHASH_BUCKETS];
  nasd_nfs_dirent_t        *idbuckets[NASD_NFS_OBJHASH_BUCKETS];
  nasd_nfs_core_dirpage_t  *first_page;
  nasd_nfs_core_dirpage_t  *last_page;
  int                       visited; /* debugging */
};

struct nasd_nfs_core_dirpage_s {
  int                       dirty;
  nasd_offset_t             offset;
  nasd_nfs_dirpage_t       *page;
  nasd_nfs_core_dirpage_t  *next;
  nasd_nfs_core_dirpage_t  *prev;
};

#if NASD_NFS_KEEP_CACHE_STATS > 0
#define NASD_NFS_CSINC(_stat_) \
  NASD_ATOMIC_INC64(&nasd_nfs_cache_stats._stat_)
#endif /* NASD_NFS_KEEP_CACHE_STATS > 0 */

extern nasd_shutdown_list_t *nasd_nfs_shutdown;

struct nasd_nfs_createpair_s {
  nasd_p_create_dr_args_t       args;
  nasd_p_create_dr_res_t        res;
  nasd_p_create_dr_args_otw_t   args_otw;
  nasd_p_create_dr_res_otw_t    res_otw;
  nasd_nfs_createpair_t        *next;
  nasd_nfs3_attributes_t       *nfs3attr;
  nasd_nfs3_attributes_t       *new_nfs3attr;
  nasd_fieldmask_t              fieldmask;
  nasd_nfs_attr_t              *dh_attr;
  nasd_nfs_core_dir_t          *cd;
  nasd_attribute_t             *attr;
  nasd_uint64                   flags;
  int                           i;
};

struct nasd_nfs_setattrpair_s {
  nasd_p_setattr_dr_args_t       sa_args;
  nasd_p_setattr_dr_res_t        sa_res;
  nasd_p_getattr_dr_res_t        ga_res;
  nasd_p_setattr_dr_args_otw_t   sa_args_otw;
  nasd_p_setattr_dr_res_otw_t    sa_res_otw;
  nasd_nfs_setattrpair_t        *next;
};

struct nasd_nfs_infopage_s {
  nasd_info_page_t      info_page;
  nasd_nfs_infopage_t  *next;
};

struct nasd_nfs_lookupstuff_s {
  nasd_nfs_lookup_args_t   in_args;
  nasd_nfs_lookup_res_t    out_res;
  nasd_nfs_lookupstuff_t  *next;
};

NASD_DECLARE_EXTERN_MUTEX(nasd_nfs_dircache_lock)

#define NASD_NFS_LOCK_DIRCACHE()   NASD_LOCK_MUTEX(nasd_nfs_dircache_lock)
#define NASD_NFS_UNLOCK_DIRCACHE() NASD_UNLOCK_MUTEX(nasd_nfs_dircache_lock)

#define NASD_NFS_LOCK_DIR(_d_,_dircache_lock_held_) { \
  if (_dircache_lock_held_ == 0) { \
    NASD_NFS_LOCK_DIRCACHE(); \
  } \
  NASD_LOCK_MUTEX((_d_)->lock); \
  if (_dircache_lock_held_ == 0) { \
    NASD_NFS_UNLOCK_DIRCACHE(); \
  } \
}

#define NASD_NFS_UNLOCK_DIR(_d_) { \
  NASD_UNLOCK_MUTEX((_d_)->lock); \
}

#if 0
#define nasd_nfs_dir_hash_check(_verbose_) _nasd_nfs_dir_hash_check(_verbose_)
#else
#define nasd_nfs_dir_hash_check(_verbose_)
#endif

/*
 * from nfs_core.c
 */
nasd_status_t nasd_nfs_really_shutdown(void);
void nsad_nfs_shutdown_timeoutsys(void *ignored);
nasd_status_t nasd_nfs_init(void);
nasd_nfs_createpair_t *nasd_nfs_get_createpair(void);
void nasd_nfs_free_createpair(nasd_nfs_createpair_t *cp);
nasd_status_t nasd_nfs_createpair_init(void);
void nasd_nfs_createpair_shutdown(void *ignored);
nasd_nfs_setattrpair_t *nasd_nfs_get_setattrpair(void);
void nasd_nfs_free_setattrpair(nasd_nfs_setattrpair_t *cp);
nasd_status_t nasd_nfs_setattrpair_init(void);
void nasd_nfs_setattrpair_shutdown(void *ignored);
nasd_nfs_infopage_t *nasd_nfs_get_infopage(void);
void nasd_nfs_free_infopage(nasd_nfs_infopage_t *cp);
nasd_status_t nasd_nfs_infopage_init(void);
void nasd_nfs_infopage_shutdown(void *ignored);
nasd_nfs_lookupstuff_t *nasd_nfs_get_lookupstuff(void);
void nasd_nfs_free_lookupstuff(nasd_nfs_lookupstuff_t *cp);
nasd_status_t nasd_nfs_lookupstuff_init(void);
void nasd_nfs_lookupstuff_shutdown(void *ignored);

/*
 * from nfs_dir.c
 */
extern nasd_nfs_core_dirpage_t *nasd_nfs_free_dirpages;
extern int nasd_nfs_num_free_dirpages;
void nasd_nfs_dir_cleanup(void);
void nasd_nfs_dir_hash_ins(nasd_nfs_core_dir_t *cd);
void _nasd_nfs_dir_hash_check(int verbose);
nasd_status_t nasd_nfs_dir_hash_lookup(nasd_nfs_identifier_t nfs_id,
  int force_load, nasd_nfs_core_dir_t **out_dirp,
  nasd_nfs_attr_t **out_attrp);
void nasd_nfs_dir_dirty_page(nasd_nfs_core_dir_t *dir,
  nasd_nfs_core_dirpage_t *cdp);
nasd_status_t nasd_nfs_dir_add_entry(nasd_nfs_core_dir_t *dir,
  char *name, nasd_identifier_t nasdid, nasd_uint64 flags,
  nasd_nfs_dirent_t **dep);
void nasd_nfs_dir_release(nasd_nfs_core_dir_t *dir);
void nasd_nfs_dir_release_post_attribute(nasd_nfs_core_dir_t *cd,
  nasd_nfs_identifier_t in_identifier, nasd_post_attribute_t *post_attribute);
void nasd_nfs_dir_writer_loop(void *arg);
void nasd_nfs_dir_writer_shutdown(void *ignored);
nasd_status_t nasd_nfs_dir_init(void);
void nasd_nfs_dealloc_coredir(nasd_nfs_core_dir_t *cd, int dircache_lock_held);
void nasd_nfs_ditch_coredir(nasd_nfs_core_dir_t *cd, int dircache_lock_held);
nasd_nfs_core_dir_t *nasd_nfs_dir_get_coredir(int dircache_lock_held);
void nasd_nfs_dir_mk_freepages(int npages);
nasd_status_t nasd_nfs_load_dir(nasd_nfs_core_dir_t *cd);
void nasd_nfs_post_attribute(nasd_nfs_core_dir_t *cd,
  nasd_nfs_identifier_t in_identifier, nasd_post_attribute_t *post_attribute);
nasd_status_t nasd_nfs_lookup_post_attribute(
  nasd_nfs_identifier_t in_identifier, nasd_post_attribute_t *post_attribute);
nasd_status_t nasd_nfs_dir_markerverf(nasd_nfs_core_dir_t *dir,
  nasd_markerverf_t *mvp);
nasd_status_t nasd_nfs_dir_remove_entry(nasd_nfs_core_dir_t *dir,
  nasd_nfs_dirent_t *de);

/*
 * from nfs_drive.c
 */
nasd_status_t nasd_nfs_drsys_init(void);
nasd_status_t nasd_nfs_mark_dirs_clean(void);
void nasd_nfs_shutdown_drives(void);
nasd_status_t nasd_nfs_add_mount(char *path, char *drive,
  char *drive_bind_name, int portnum, int partnum, nasd_uint32 net_addr);
nasd_status_t nasd_nfs_mount_lookup(char *dirpath,
  nasd_list_len_t in_listlen, nasd_net_address_t out_drivelist[],
  nasd_list_len_t *out_listlen, nasd_nfs_identifier_t *out_identifier);
nasd_status_t nasd_nfs_dir_write_dirty(nasd_nfs_core_dir_t *dir);
nasd_status_t nasd_nfs_dir_load_attr(nasd_nfs_core_dir_t *dir);
nasd_status_t nasd_nfs_do_getattr(nasd_nfs_drive_t *dr,
  nasd_identifier_t nasdid, nasd_p_getattr_dr_res_t *ga_res);
nasd_status_t nasd_nfs_do_setattr(nasd_nfs_drive_t *dr,
  nasd_p_setattr_dr_args_t *args, nasd_p_setattr_dr_res_t *res,
  nasd_byte_t *args_otw, nasd_byte_t *res_otw);
nasd_status_t nasd_nfs_get_drive(nasd_nfs_identifier_t in_identifier,
  nasd_nfs_drive_t **drp);
nasd_status_t nasd_nfs_do_create(nasd_nfs_drive_t *dr,
  nasd_p_create_dr_args_t *cr_args, nasd_p_create_dr_res_t *cr_res,
  nasd_byte_t *cr_args_otw, nasd_byte_t *cr_res_otw);
nasd_status_t nasd_nfs_do_remove(nasd_nfs_drive_t *dr,
  nasd_identifier_t nasdid);

/*
 * from nfs_name.c
 */
int nasd_nfs_name_hash(char *str);
void nasd_nfs_name_shutdown_freelist(void *arg);
void nasd_nfs_attr_shutdown_freelist(void *arg);
nasd_status_t nasd_nfs_name_init(void);
void nasd_nfs_name_cleandir(nasd_nfs_core_dir_t *dir);
nasd_status_t nasd_nfs_name_hash_ins(nasd_nfs_core_dir_t *dir,
  nasd_nfs_dirent_t *de);
nasd_nfs_dirent_t *nasd_nfs_name_add(nasd_nfs_core_dir_t *dir,
  nasd_nfs_core_dirpage_t *page, nasd_nfs_dirdata_t *data);
nasd_nfs_dirent_t *nasd_nfs_name_lookup(nasd_nfs_core_dir_t *dir,
  char *name);
nasd_status_t nasd_nfs_name_remove(nasd_nfs_core_dir_t *dir,
  nasd_nfs_dirent_t *de, int unhash_only);
nasd_nfs_attr_t *nasd_nfs_get_attr(void);
void nasd_nfs_release_attr(nasd_nfs_attr_t *attr);
void nasd_nfs_debug_name_cache(nasd_nfs_core_dir_t *dir);
nasd_status_t nasd_nfs_valid_name(char *name);

/*
 * from nfs_ops.c
 */
extern nasd_nfs_cachestats_t nasd_nfs_cache_stats;
void nasd_nfs_begin_rpc(nasd_opstat_t *opstat);
void nasd_nfs_end_rpc(nasd_opstat_t *opstat);
void nasd_nfs_wait_shutdown_rpc(void);
void nasd_nfs_stop_rpc(void);
void nasd_nfs_shutdown_rpc(void *arg);
nasd_status_t nasd_nfs_rpc_init(void);
nasd_status_t nasd_nfs_startup_rpc(void);

/*
 * frame stubs
 */
nasd_status_t nasd_nfs_do_drive_bind(char *drive_name,
  nasd_drive_handle_t *dhp);
nasd_status_t nasd_nfs_frame_init(void);

#endif /* !_NASD__NASD_NFS_INTERNAL_H_ */

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
