/*
 * nasd_drive_client.h
 *
 * defs/structs used by clients of NASD drives
 */
/*
 * Copyright (c) 1996,1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#ifndef _NASD__NASD_PDRIVE_CLIENT_H_
#define _NASD__NASD_PDRIVE_CLIENT_H_

#include <sys/param.h>

#include <nasd/nasd_options.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_threadstuff.h>
#include <dce/dce_error.h>
#include <nasd/nasd_control.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_shutdown.h>

typedef struct nasd_cl_fo_state_s       nasd_cl_fo_state_t;
typedef struct nasd_cl_fo_list_state_s  nasd_cl_fo_list_state_t;

/*
 * must be >= dce_c_error_string_len (160)
 */
typedef char nasd_error_string_t[160];
typedef nasd_uint32 nasd_rpc_status_t;

#include <nasd/nasd_security.h>

#define NASD_MAX_OUTSTANDING_DIGESTS 8
#define NASD_DIGEST_LOCKED 1
#define NASD_DIGEST_FREE 0
#define NASD_DIGEST_SENDING 2
#define NASD_DATA_BUFFER 0
#define NASD_DIGEST_BUFFER 1

#define NASD_BIND_BOGUS             0
#define NASD_BIND_DCE_DIRECT        1
#define NASD_BIND_DCE_KPDEV         2
#define NASD_BIND_DCE_DIRECT_TCP    3
#define NASD_BIND_COLOCATE          4
#define NASD_BIND_DCE_KPDEV_FAKE_RW 5
#define NASD_BIND_MSGQ		    6


/*
 * !!! note that nasd_drive_handle_t is a pointer
 * to the structure, not the structure itself, which
 * is not the NASD standard. This is for dealing with
 * older client code, and code which isn't aware of the
 * structured nature of this type. This also allows
 * NULL/non-NULL checks, and lets us more easily pass
 * pointers to the structure on the stack, without
 * changing legacy code.
 */
typedef struct nasd_drive_handle_s {
  int                   type;      /* binding type */
  int                   real_type; /* "real" binding type */
  int                   fd;        /* fd for kpdev or server msgq id */
  int			replyq;	   /* client msgq id */
  int			seqnum;	   /* message sequence number */
  NASD_DECLARE_MUTEX(seqlock)	   /* mutex for sequence number */
  rpc_binding_handle_t  h;         /* DCE handle */

  /*
   * kpdev will track in-kernel optimes for us
   * this is the "last one"
   * no guarantees of synchronization
   */
  unsigned int          last_start_cc; 
  unsigned int          last_stop_cc;

} *nasd_drive_handle_t;

#define NASD_DRIVE_HANDLE_LAST_START_CC(_h_) ((_h_)->last_start_cc)
#define NASD_DRIVE_HANDLE_LAST_STOP_CC(_h_)  ((_h_)->last_stop_cc)

typedef struct nasd_drive_param_dce_kpdev_s {
  char  devname[MAXPATHLEN];
} nasd_drive_param_dce_kpdev_t;

#define NASD_DIGEST_SLOT_FREE 1
#define NASD_DIGEST_SLOT_BUSY 2

struct nasd_cl_fo_state_s {
  char          *buf;
  unsigned int   buffer_is;          /* is the current buffer a digest or data */
  unsigned int   len;                /* total amount of data being sent */
  unsigned int   alloc_off;          /* offset into the datastream that has been alloced*/
 
  unsigned int   push_off;           /* offset in the datastreeam  has been pushed */
                                   
#if NASD_SECURE_RPCS_ENABLE > 0     
  /* Security Related Pipe state - these variables maintain a record
   of the current security context, outstanding message digests, where
   we are in the data or digest for both pushing and pulling, did 
   we last pull data, etc.*/

  /* used for reading and writing */

  nasd_security_context_t  context;
  
  /* all of this uglines if for writing */

  unsigned int   digest_block_size;
   nasd_digest_t digest[NASD_MAX_OUTSTANDING_DIGESTS];
  int digest_state[NASD_MAX_OUTSTANDING_DIGESTS];
  int curr_alloc_digest; /* indexes into the digest/ status arrays */
  int curr_pull_digest;  /* indexes into the digest/ status arrays */
  int digest_bytes_pulled; /* offset into the current digest */
  int digest_bytes_alloced;/* offset into the current digest */
  int last_pulled_data; /* Did we last pull some data ? */
  int done; /* Did we compute our final digest */
  int count; /* How many digests did we do */
  nasd_int16 protection; /* What protection are we using on this pipe */
  /* These offsets are all in terms of byte offset within the OBJECT */
  int  alloc_next_digest_off, pull_next_digest_off, start_off;


  /* all of this ugliness is for reading */
  nasd_offset_t curr_file_off;
    
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */
};

struct NASD_drive_comm_s {
  nasd_drive_handle_t  h;
  nasd_cookie_t        cookie;
  nasd_disk_ident_t    di;
};

struct nasd_cl_fo_list_state_s {
  nasd_len_t       *out_lenp;
  nasd_len_t        targetlen;
  int               badstate;
  nasd_mem_list_t  *alloc_cur;
  int               alloc_elem;
  char             *alloc_buf;
  unsigned int      alloc_off;
  unsigned int      alloc_len;
  nasd_mem_list_t  *push_cur;
  int               push_elem;
  char             *push_buf;
  unsigned int      push_off;
  unsigned int      push_len;
  nasd_status_t     status;
#if NASD_SECURE_RPCS_ENABLE > 0     
/* Security Related Pipe state - these variables maintain a record
   of the current security context, outstanding message digests, where
   we are in the data or digest for both pushing and pulling, did 
   we last pull data, etc.*/

  unsigned int   total_push_off;
  unsigned int   total_alloc_off;
  unsigned int   digest_block_size;
  nasd_security_context_t  context;
  nasd_digest_t digest[NASD_MAX_OUTSTANDING_DIGESTS];
  int digest_state[NASD_MAX_OUTSTANDING_DIGESTS];
  int curr_alloc_digest; /* indexes into the digest/ status arrays */
  int curr_pull_digest;  /* indexes into the digest/ status arrays */
  int digest_bytes_pulled; /* offset into the current digest */
  int digest_bytes_alloced;/* offset into the current digest */
  int last_pulled_data; /* Did we last pull some data ? */
  int done; /* Did we compute our final digest */
  int count; /* How many digests did we do */
  nasd_int16 protection; /* What protection are we using on this pipe */
  /* These offsets are all in terms of byte offset within the OBJECT */
  int  alloc_next_digest_off, pull_next_digest_off, start_off; 
  
  nasd_offset_t curr_file_off;
#endif /* NASD_SECURE_RPCS_ENABLE > 0 */

};

#define nasd_cl_error_string(_handle_,_status_,_str_) \
  _nasd_cl_error_string(_handle_,_status_,_str_,__FILE__,__LINE__)

/*
 * init/shutdown
 */
nasd_status_t nasd_cl_p_real_init(void);
void nasd_cl_p_real_shutdown(void);
void nasd_cl_p_sys_init(void);
nasd_status_t nasd_cl_p_init(void);
void nasd_cl_p_shutdown(void);
void nasd_cl_p_rm_msgq(void *);

/*
 * handle management
 */
extern nasd_status_t nasd_bind_to_drive(char *servername, char *portnum,
  int binding_type, void *binding_param, int binding_param_len,
  nasd_drive_handle_t *handle);

extern nasd_status_t nasd_unbind_drive(nasd_drive_handle_t *handle);


/*
 * RPC error wrappers
 */
extern char *_nasd_cl_error_string(nasd_drive_handle_t handle,
  nasd_rpc_status_t status, nasd_error_string_t str,
  char *file, int line);

/*
 * RPC wrappers
 */

extern void nasd_cl_p_null_dr(nasd_drive_handle_t handle,
  nasd_status_t *nasd_status, nasd_rpc_status_t *status);

extern void nasd_cl_p_sync_dr(nasd_drive_handle_t handle,
  nasd_status_t *nasd_status, nasd_rpc_status_t *status);

extern void nasd_cl_p_part_creat_dr(nasd_drive_handle_t handle, 
  nasd_key_t req_key, nasd_p_part_creat_dr_args_t *args,
  nasd_p_part_creat_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_create_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_create_dr_args_t *args, 
  nasd_p_create_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_create_dr__otw_provided(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_p_create_dr_args_t *args, nasd_p_create_dr_res_t *res,
  nasd_byte_t *args_otw, nasd_byte_t *res_otw,
  nasd_rpc_status_t *status);

extern void nasd_cl_p_getattr_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_getattr_dr_args_t *args,
  nasd_p_getattr_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_write_simple_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_smpl_op_dr_args_t *args, void *buf, 
  nasd_p_fastwrite_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_read_simple_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_smpl_op_dr_args_t *args, void *buf,
  nasd_p_fastread_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_read2_simple_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_smpl_op_dr_args_t *args, void *buf,
  nasd_p_fastread_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_tread_simple_dr(nasd_drive_handle_t handle, 
  nasd_key_t req_key, nasd_p_thrtl_op_dr_args_t *args, void *buf,
  nasd_p_fastread_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_range_read(nasd_drive_handle_t h,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum,
  nasd_identifier_t nasdid, nasd_mem_list_t *memlist, nasd_offset_t in_offset,
  nasd_len_t in_len, nasd_len_t *out_lenp, nasd_status_t *nasd_status,
  nasd_rpc_status_t *status);

extern void nasd_cl_p_range_tread(nasd_drive_handle_t h,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum, 
  nasd_identifier_t nasdid, nasd_uint64 in_bms_targ, nasd_mem_list_t *memlist,
  nasd_offset_t in_offset, nasd_len_t in_len, nasd_len_t *out_lenp,
  nasd_status_t *nasd_status, nasd_rpc_status_t *status);

extern void nasd_cl_p_range_write(nasd_drive_handle_t h,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum, 
  nasd_identifier_t nasdid, nasd_mem_list_t *memlist, nasd_offset_t in_offset,
  nasd_len_t in_len, nasd_len_t *out_lenp, nasd_status_t *nasd_status,
  nasd_rpc_status_t *status);

extern void nasd_cl_p_setattr_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key,
  nasd_p_setattr_dr_args_t *args, nasd_p_setattr_dr_res_t *res,
  nasd_rpc_status_t *status);

extern void nasd_cl_p_setattr_dr__otw_provided(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_p_setattr_dr_args_t *args, nasd_p_setattr_dr_res_t *res,
  nasd_byte_t *args_otw, nasd_byte_t *res_otw,
  error_status_t *status);

extern void nasd_cl_p_flush_obj_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_flush_obj_dr_args_t *args, 
  nasd_p_flush_obj_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_eject_obj_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_eject_obj_dr_args_t *args, 
  nasd_p_eject_obj_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_remove_dr(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_p_remove_dr_args_t *args,
  nasd_p_remove_dr_res_t *res, nasd_rpc_status_t *status);

extern void nasd_cl_p_initialize_dr(nasd_drive_handle_t handle,
  nasd_p_initialize_dr_args_t *args, nasd_p_initialize_dr_res_t *res,
  nasd_rpc_status_t *status);

extern nasd_status_t nasd_cl_p_rdwr_stats(
  nasd_drive_handle_t handle, int partnum, nasd_identifier_t nasdid,
  nasd_key_t in_key, nasd_capability_args_t capargs,
  void *info, nasd_offset_t offset, int read);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_info(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_drive_info_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_info(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_drive_info_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_cache_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_cache_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_op_stats(
  nasd_drive_handle_t handle, nasd_key_t in_key,
  nasd_capability_args_t cap_args, nasd_ctrl_drive_opstats_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_op_stats(
  nasd_drive_handle_t handle, nasd_key_t in_key,
  nasd_capability_args_t cap_args, nasd_ctrl_drive_opstats_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_cache_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_cache_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_io_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_io_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_io_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_io_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_io_read_size_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_io_size_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_io_read_size_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_io_size_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_io_write_size_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_io_size_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_io_write_size_stats(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_io_size_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_layout_stats(
  nasd_drive_handle_t handle, nasd_key_t in_key,
  nasd_capability_args_t cap_args, nasd_ctrl_layout_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_layout_stats(
  nasd_drive_handle_t handle, nasd_key_t in_key,
  nasd_capability_args_t cap_args, nasd_ctrl_layout_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_ioqueue_stats(
  nasd_drive_handle_t handle, nasd_key_t in_key,
  nasd_capability_args_t cap_args, nasd_ctrl_ioqueue_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_ioqueue_stats(
  nasd_drive_handle_t handle, nasd_key_t in_key,
  nasd_capability_args_t cap_args, nasd_ctrl_ioqueue_stat_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_part_info(nasd_drive_handle_t handle,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum,
  nasd_ctrl_part_info_t *info);

extern nasd_status_t nasd_cl_p_ctrl_get_drive_trace_basic(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_trace_basic_t *info);

extern nasd_status_t nasd_cl_p_ctrl_put_drive_trace_basic(
  nasd_drive_handle_t handle, nasd_key_t req_key,
  nasd_capability_args_t capargs, nasd_ctrl_trace_basic_t *info);

/*
 * DCE pipe monkey magic
 */

extern nasd_status_t nasd_cl_p_dce_init(nasd_shutdown_list_t *sl);

extern void nasd_cl_fo_alloc(rpc_ss_pipe_state_t state,
  unsigned int bsize, nasd_byte_t **buf, unsigned int *bcount);

extern void nasd_cl_fr_push(rpc_ss_pipe_state_t *state,
  idl_char *buf, unsigned32 count);

extern void nasd_cl_fw_pull(rpc_ss_pipe_state_t *state,
  idl_char *buf, unsigned32 in_count, unsigned32 *out_countp);

extern void nasd_cl_p_check_mem_list(nasd_mem_list_t *memlist,
  nasd_status_t *status);

extern void nasd_cl_p_fo_list_alloc(rpc_ss_pipe_state_t state,
  unsigned int bsize, nasd_byte_t **buf, unsigned int *bcount);

extern void nasd_cl_p_fr_list_push(rpc_ss_pipe_state_t *state,
  idl_char *buf, unsigned32 count);

/*
 * DCE-direct wrappers
 */

extern void nasd_cl_prpc_null_dr(rpc_binding_handle_t handle,
  nasd_status_t *nasd_status, error_status_t *status);

extern void nasd_cl_prpc_sync_dr(rpc_binding_handle_t handle,
  nasd_status_t *nasd_status, error_status_t *status);

extern void nasd_cl_prpc_part_creat_dr(rpc_binding_handle_t handle, 
  nasd_key_t req_key, nasd_p_part_creat_dr_args_t *args,
  nasd_p_part_creat_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_create_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_create_dr_args_t *args, 
  nasd_p_create_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_create_dr__otw_provided(
  rpc_binding_handle_t handle, nasd_key_t req_key,
  nasd_p_create_dr_args_t *args, nasd_p_create_dr_res_t *res,
  nasd_byte_t *args_otw, nasd_byte_t *res_otw,
  error_status_t *status);

extern void nasd_cl_prpc_getattr_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_getattr_dr_args_t *args,
  nasd_p_getattr_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_write_simple_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_smpl_op_dr_args_t *args, void *buf, 
  nasd_p_fastwrite_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_read_simple_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, int is_read2, nasd_p_smpl_op_dr_args_t *args,
  void *buf, nasd_p_fastread_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_tread_simple_dr(rpc_binding_handle_t handle, 
  nasd_key_t req_key, nasd_p_thrtl_op_dr_args_t *args, void *buf,
  nasd_p_fastread_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_range_read(rpc_binding_handle_t h,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum,
  nasd_identifier_t nasdid, nasd_mem_list_t *memlist, nasd_offset_t in_offset,
  int in_len, nasd_len_t *out_lenp, nasd_status_t *nasd_status,
  error_status_t *status);

extern void nasd_cl_prpc_range_tread(rpc_binding_handle_t h,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum, 
  nasd_identifier_t nasdid, nasd_uint64 in_bms_targ, nasd_mem_list_t *memlist,
  nasd_offset_t in_offset, int in_len, nasd_len_t *out_lenp,
  nasd_status_t *nasd_status, error_status_t *status);

extern void nasd_cl_prpc_range_write(rpc_binding_handle_t h,
  nasd_key_t req_key, nasd_capability_args_t capargs, int partnum, 
  nasd_identifier_t nasdid, nasd_mem_list_t *memlist, nasd_offset_t in_offset,
  int in_len, nasd_len_t *out_lenp, nasd_status_t *nasd_status,
  error_status_t *status);

extern void nasd_cl_prpc_setattr_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_setattr_dr_args_t *args,
  nasd_p_setattr_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_setattr_dr__otw_provided(
  rpc_binding_handle_t handle, nasd_key_t req_key,
  nasd_p_setattr_dr_args_t *args, nasd_p_setattr_dr_res_t *res,
  nasd_byte_t *args_otw, nasd_byte_t *res_otw,
  error_status_t *status);

extern void nasd_cl_prpc_flush_obj_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_flush_obj_dr_args_t *args, 
  nasd_p_flush_obj_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_eject_obj_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_eject_obj_dr_args_t *args, 
  nasd_p_eject_obj_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_remove_dr(rpc_binding_handle_t handle,
  nasd_key_t req_key, nasd_p_remove_dr_args_t *args,
  nasd_p_remove_dr_res_t *res, error_status_t *status);

extern void nasd_cl_prpc_initialize_dr(rpc_binding_handle_t handle,
  nasd_p_initialize_dr_args_t *args, nasd_p_initialize_dr_res_t *res,
  error_status_t *status);

/*
 * DCE-direct binding stuff
 */

extern nasd_status_t nasd_dce_direct_bind_to_server(
  unsigned_char_t *protseq, unsigned_char_t *host,
  unsigned_char_t *endpoint, rpc_binding_handle_t *binding_handle);

extern nasd_status_t nasd_dce_direct_unbind_server(
  rpc_binding_handle_t *binding_handle);

#endif /* !_NASD__NASD_PDRIVE_CLIENT_H_ */
