/*
 * nasd_threads_pthread.h
 *
 * pthread variant of threads portability layer.
 */
/*
 * Copyright (c) 1995,1996,1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Authors: Mark Holland, Daniel Stodolsky, Jim Zelenka, Marc Unangst
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#ifndef _NASD__NASD_THREADS_PTHREAD_H_
#define _NASD__NASD_THREADS_PTHREAD_H_

#include <nasd/nasd_options.h>

#ifdef DEC_OSF
#if NASD_PTHREAD_EXC > 0
#include <pthread_exc.h>
#else /* NASD_PTHREAD_EXC > 0 */
#include <pthread.h>
#endif /* NASD_PTHREAD_EXC > 0 */
#endif /* DEC_OSF */

#if defined(AIX) || defined(LINUX)
#include <pthread.h>
#endif /* AIX || LINUX */

#ifdef LINUX
/* don't get exception-handling stuff by default on Linux */
#include <dce/exc_handling.h>
#endif /* LINUX */

/* Note that on Linux, we could be running with one of two different
   pthreads APIs.  The stock pthreads API that LinuxThreads and glibc2
   provide is compliant with the POSIX Pthreads Draft 7 spec.
   However, DCE is written to the (obsolete) Draft 4 spec, and the API
   changed considerably between Draft 4 and Draft 7.  If we are
   building with DCE, then we need to use the Draft 4 API as well, and
   the FreeDCE package for Linux includes a compatibility overlay that
   redefines the appropriate pthreads calls to look like Draft 4.

   This is *different* from the API changes introduced by the
   exception-handling pthreads interface that DCE uses.  On Linux, you
   can get that interface by including <dce/pthread_exc.h> instead of
   <pthread.h>.  We prefer to use the non-exception-handling
   interface. */

nasd_status_t nasd_sys_mutex_init(pthread_mutex_t *m);
nasd_status_t nasd_sys_mutex_destroy(pthread_mutex_t *m);
nasd_status_t nasd_sys_cond_init(pthread_cond_t *c);
nasd_status_t nasd_sys_cond_destroy(pthread_cond_t *c);

#define NASD_SYS_DECLARE_MUTEX(_m_)                pthread_mutex_t _m_;
#define NASD_SYS_DECLARE_STATIC_MUTEX(_m_)  static pthread_mutex_t _m_;
#define NASD_SYS_DECLARE_EXTERN_MUTEX(_m_)  extern pthread_mutex_t _m_;
#define NASD_SYS_DECLARE_COND(_c_)                 pthread_cond_t  _c_;
#define NASD_SYS_DECLARE_STATIC_COND(_c_)   static pthread_cond_t  _c_;
#define NASD_SYS_DECLARE_EXTERN_COND(_c_)   extern pthread_cond_t  _c_;

#define NASD_SYS_LOCK_MUTEX(_m_)     {pthread_mutex_lock(&(_m_));}
#define NASD_SYS_UNLOCK_MUTEX(_m_)   pthread_mutex_unlock(&(_m_))
#define NASD_SYS_TRY_LOCK_MUTEX(_m_) pthread_mutex_trylock(&(_m_))

#define NASD_SYS_WAIT_COND(_c_,_m_)  pthread_cond_wait( &(_c_), &(_m_) )
#define NASD_SYS_SIGNAL_COND(_c_)    pthread_cond_signal( &(_c_) )
#define NASD_SYS_BROADCAST_COND(_c_) pthread_cond_broadcast(&(_c_))

#if LINUX
#define NASD_DECLARE_ONCE(_o_) pthread_once_t _o_ = PTHREAD_ONCE_INIT;
#else /* LINUX */
#define NASD_DECLARE_ONCE(_o_) pthread_once_t _o_ = {0,0,0};
#endif /* LINUX */
extern int _nasd_once(pthread_once_t *once_block,
  pthread_initroutine_t init_routine, char *file, int line);
#define nasd_once(_o_,_i_) _nasd_once(_o_, (pthread_initroutine_t)_i_, __FILE__, __LINE__)

typedef pthread_t               nasd_sys_thread_t;
#ifdef DEC_OSF
typedef pthread_addr_t          nasd_threadarg_t;
#else /* DEC_OSF */
typedef void *                  nasd_threadarg_t;
#endif /* DEC_OSF */
typedef pthread_attr_t          nasd_sys_threadattr_t;

#ifdef DEC_OSF
#define NASD_EXIT_THREAD(_status_)                  pthread_exit( (pthread_addr_t) (_status_) )
#else /* DEC_OSF */
#define NASD_EXIT_THREAD(_status_)                  pthread_exit( (void *) (_status_) )
#endif /* DEC_OSF */

#define NASD_DELAY_THREAD(_secs_,_nsecs_) { \
  struct timespec _interval; \
  _interval.tv_sec = (_secs_); \
  _interval.tv_nsec = (_nsecs_); \
  pthread_delay_np(&_interval); \
}

typedef unsigned long       nasd_thread_id_t;
#define NASD_THREAD_ID_NULL 0

#if defined(__alpha)
#define NASD_THREAD_ID_FMT  NASD_64u_FMT
#else /* __alpha */
#define NASD_THREAD_ID_FMT  "u"
#endif /* __alpha */



#ifdef DEC_OSF
#define nasd_thread_self() ((nasd_thread_id_t)(pthread_self().field1))
#endif /* DEC_OSF */

#ifdef LINUX
#define nasd_thread_self() ((nasd_thread_id_t)(pthread_self()))
#endif /* LINUX */

#ifdef LINUX
/* DCE forgets to prototype these, so we do so here (to avoid warnings) */
extern int pthd4_setspecific(pthread_key_t key, pthread_addr_t value);
extern int pthd4_getspecific(pthread_key_t key, pthread_addr_t *value);
#endif /* LINUX */

/* XXX only works if already at toplevel */
#define NASD_THREAD_KILL_SELF() return

#endif /* !_NASD__NASD_THREADS_PTHREAD_H_ */

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
