/*
 * nasd_pdev.c
 *
 * pseudo-device for performing nasd operations in the kernel on
 * behalf of a user-process
 *
 * Intended for use by AFS server, which can't make DCE RPC calls
 * for itself.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasdpd.h>

#if NNASDPD > 0

#include <nasd/nasd_options.h>
#include <nasd/nasd_pdrive_client.h>
#include <kern/lock.h>
#include <vm/vm_map.h>
#include <vm/vm_kern.h>
#include <vm/vm_lock.h>
#include <dce/ker/pthread_exc.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_nfs_client.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_freelist.h>
#include <sys/time.h>
#include <rpc/rpc.h>
#include <nfs/nfs.h>
#include <nfs/nfs_clnt.h>
#include <sys/mode.h>
#include <nasd/nasd_pdev.h>
#include <nasd/nasd_sys.h>
#include <nasd/nasd_timer.h>
#include <io/common/devdriver.h>

#define nasdpd_unit(_x_) minor(_x_)

typedef struct NASDpd_s   NASDpd_t;

char nasdpd_bogusbuf[8192];

struct NASDpd_s {
  int                   state;
  nasd_drive_handle_t   h;
};

NASDpd_t NASDPDs[NNASDPD];

#define GETDEV(_npd_,_unit_) { \
  _unit_ = nasdpd_unit(dev); \
  if (((_unit_) < 0) || ((_unit_) >= NNASDPD)) \
    return(ENODEV); \
  _npd_ = &NASDPDs[_unit_]; \
}

#define NPD_STATE_BOGUS   0
#define NPD_STATE_BOUND   1
#define NPD_STATE_UNBOUND 2

#define NPD_MAXLOOPS 5

int nasdpd_do_fake_byte_ops = 1;

nasd_status_t
nasdpd_init(
  nasd_shutdown_list_t  *sl)
{
  nasd_status_t rc;
  NASDpd_t *npd;
  int i;

  for(i=0;i<NNASDPD;i++) {
    npd = &NASDPDs[i];
    NASD_ASSERT((npd->state == NPD_STATE_BOGUS) || (npd->state == NPD_STATE_UNBOUND));
    npd->state = NPD_STATE_UNBOUND;
  }

  return(NASD_SUCCESS);
}

int
nasdpd_unbind(
  NASDpd_t *npd)
{
  nasd_status_t rc;

  if (npd->state != NPD_STATE_BOUND) {
    /* already done */
    return(0);
  }
  rc = nasd_unbind_drive(&npd->h);
  npd->state = NPD_STATE_UNBOUND;
  nasd_cl_p_shutdown();
  if (rc)
    return(EIO);
  return(0);
}

int
nasdpd_open(
  dev_t dev,
  int   flags,
  int   format)
{
  NASDpd_t *npd;
  int unit;

  GETDEV(npd,unit);

  return(0);
}

int
nasdpd_close(
  dev_t dev,
  int   flags)
{
  NASDpd_t *npd;
  int unit;

  GETDEV(npd,unit);
  /*
   * GETDEV has already returned ENODEV if unit is bad
   */
  return(0);
}

/*
 * XXX
 *
 * Note lack of serialization!
 */

int
nasdpd_ioctl(
  dev_t        dev,
  int          cmd,
  caddr_t      data,
  int          flags,
  struct ucred *cred)
{
  nasd_status_t nrc;
  int unit, rc, dir;
  NASDpd_t *npd;
  void *iobuf;

  GETDEV(npd,unit);

  switch(cmd) {

    /*
     * New drive operations begin here.
     */

    case NASDPD_PBIND: {
      NASDPD_bind_t *bind;
      char *binding_param;

      bind = (NASDPD_bind_t *)data;
      if (npd->state == NPD_STATE_BOUND) {
        /* already bound */
        return(EBUSY);
      }
      nrc = nasd_cl_p_init();
      if (nrc != NASD_SUCCESS) {
        printf("nasd_init_pdrive_client failed: %s:%d rc=0x%x (%s)\n",
          __FILE__, __LINE__, rc, nasd_error_string(rc));
        return (rc);
      }
      if (bind->binding_param_len) {
        if (bind->binding_param_len < 0) {
          nasd_cl_p_shutdown();
          return(EINVAL);
        }
        NASD_Malloc(binding_param, bind->binding_param_len+1, (char *));
        rc = copyin(bind->binding_param, binding_param,
          bind->binding_param_len);
        binding_param[bind->binding_param_len] = '\0';
        nrc = nasd_bind_to_drive(bind->drive_name, bind->drive_port,
          bind->binding_type, binding_param, bind->binding_param_len,
          &npd->h);
        NASD_Free(binding_param, bind->binding_param_len+1);
      }
      else {
        nrc = nasd_bind_to_drive(bind->drive_name, bind->drive_port,
          bind->binding_type, NULL, 0, &npd->h);
      }
      if (nrc) {
        nasd_cl_p_shutdown();
        return(EIO);
      }

      npd->state = NPD_STATE_BOUND;
    }
    break;
    case NASDPD_PUNBIND: {
      return(nasdpd_unbind(npd));
    }
    break;
    case NASDPD_PNULL: {
      NASDPD_null_t *null;
      null = (NASDPD_null_t *)data;
      if (npd->state != NPD_STATE_BOUND) {
        /* not bound */
        return(EBUSY);
      }
      /* _cl_ functions do their own exception handling */
      null->start_cc = NASD_RPCC();
      nasd_cl_p_null_dr(npd->h, &null->nasd_status, &null->op_status);
      null->stop_cc = NASD_RPCC();
    }
    break;
    case NASDPD_PSYNC: {
      NASDPD_sync_t *sync;

      sync = (NASDPD_sync_t *)data;
      if (npd->state != NPD_STATE_BOUND) {
        /* not bound */
        return(EBUSY);
      }
      sync->start_cc = NASD_RPCC();
      nasd_cl_p_sync_dr(npd->h, &sync->nasd_status, &sync->op_status);
      sync->stop_cc = NASD_RPCC();
      return(0);
    }
    break;

    case NASDPD_P_PCRPART: {
      NASDPD_p_crpart_t *cmd;
      cmd = (NASDPD_p_crpart_t *)data;
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_part_creat_dr(npd->h, cmd->in_key, &cmd->args,
        &cmd->res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
    }
    break;
    case NASDPD_P_PCREATE: {
      nasd_p_create_dr_args_t args;
      nasd_p_create_dr_res_t res;
      NASDPD_p_create_t *cmd;
      cmd = (NASDPD_p_create_t *)data;
      rc = copyin(cmd->args, &args, sizeof(args));
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_create_dr(npd->h, cmd->in_key, &args,
        &res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
      rc = copyout(&res, cmd->res, sizeof(res));
      if (rc)
        return(rc);
    }
    break;
    case NASDPD_P_PGETATTR: {
      nasd_p_getattr_dr_args_t args;
      nasd_p_getattr_dr_res_t res;
      NASDPD_p_getattr_t *cmd;
      cmd = (NASDPD_p_getattr_t *)data;
      rc = copyin(cmd->args, &args, sizeof(args));
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_getattr_dr(npd->h, cmd->in_key, &args,
        &res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
      rc = copyout(&res, cmd->res, sizeof(res));
      if (rc)
        return(rc);
    }
    break;
    case NASDPD_P_PWRITE_SIMPLE: {
      NASDPD_p_write_simple_t *cmd;
      cmd = (NASDPD_p_write_simple_t *)data;
      rc = nasd_wire_buf(cmd->buf, cmd->args.in_len,
        &iobuf, NASD_WIRE_DIR_WRITE);
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_write_simple_dr(npd->h, cmd->in_key, &cmd->args,
        iobuf, &cmd->res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
      rc = nasd_unwire_buf(cmd->buf, cmd->args.in_len, iobuf,
        NASD_WIRE_DIR_WRITE);
    }
    break;
    case NASDPD_P_PREAD_SIMPLE: {
      NASDPD_p_read_simple_t *cmd;
      cmd = (NASDPD_p_read_simple_t *)data;
      rc = nasd_wire_buf(cmd->buf, cmd->args.in_len,
        &iobuf, NASD_WIRE_DIR_READ);
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      if (cmd->is_read2) {
        nasd_cl_p_read2_simple_dr(npd->h, cmd->in_key, &cmd->args,
          iobuf, &cmd->res, &cmd->op_status);
      }
      else {
        nasd_cl_p_read_simple_dr(npd->h, cmd->in_key, &cmd->args,
          iobuf, &cmd->res, &cmd->op_status);
      }
      cmd->stop_cc = NASD_RPCC();
      rc = nasd_unwire_buf(cmd->buf, cmd->args.in_len, iobuf,
        NASD_WIRE_DIR_READ);
    }
    break;
    case NASDPD_P_PTREAD_SIMPLE: {
      NASDPD_p_tread_simple_t *cmd;
      cmd = (NASDPD_p_tread_simple_t *)data;
      rc = nasd_wire_buf(cmd->buf, cmd->args.in_len,
        &iobuf, NASD_WIRE_DIR_TREAD);
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_tread_simple_dr(npd->h, cmd->in_key, &cmd->args,
        iobuf, &cmd->res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
      rc = nasd_unwire_buf(cmd->buf, cmd->args.in_len, iobuf,
        NASD_WIRE_DIR_TREAD);
    }
    break;
    case NASDPD_P_PWRITE_FAKE: {
      NASDPD_p_write_simple_t *cmd;
      int i, j, t;
      cmd = (NASDPD_p_write_simple_t *)data;
      rc = nasd_wire_buf(cmd->buf, cmd->args.in_len,
        &iobuf, NASD_WIRE_DIR_WRITE);
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      t = cmd->args.in_len;
      if (nasdpd_do_fake_byte_ops) {
        for(i=0;i<t;i+=j) {
          j = NASD_MIN(sizeof(nasdpd_bogusbuf),t-i);
          bcopy((char *)iobuf, (char *)nasdpd_bogusbuf, j);
        }
      }
      cmd->res.out_datalen = t;
      cmd->res.nasd_status = NASD_SUCCESS;
      cmd->stop_cc = NASD_RPCC();
      rc = nasd_unwire_buf(cmd->buf, cmd->args.in_len, iobuf,
        NASD_WIRE_DIR_WRITE);
    }
    break;
    case NASDPD_P_PREAD_FAKE: {
      NASDPD_p_read_simple_t *cmd;
      int i, j, t;
      cmd = (NASDPD_p_read_simple_t *)data;
      rc = nasd_wire_buf(cmd->buf, cmd->args.in_len,
        &iobuf, NASD_WIRE_DIR_READ);
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      t = cmd->args.in_len;
      if (nasdpd_do_fake_byte_ops) {
        for(i=0;i<t;i+=j) {
          j = NASD_MIN(sizeof(nasdpd_bogusbuf),t-i);
          bcopy((char *)nasdpd_bogusbuf, (char *)iobuf, j);
        }
      }
      cmd->res.out_datalen = t;
      cmd->res.nasd_status = NASD_SUCCESS;
      cmd->stop_cc = NASD_RPCC();
      rc = nasd_unwire_buf(cmd->buf, cmd->args.in_len, iobuf,
        NASD_WIRE_DIR_READ);
    }
    break;
    case NASDPD_P_PSETATTR: {
      nasd_p_setattr_dr_args_t args;
      nasd_p_setattr_dr_res_t res;
      NASDPD_p_setattr_t *cmd;
      cmd = (NASDPD_p_setattr_t *)data;
      rc = copyin(cmd->args, &args, sizeof(args));
      if (rc)
        return(rc);
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_setattr_dr(npd->h, cmd->in_key, &args,
        &res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
      rc = copyout(&res, cmd->res, sizeof(res));
      if (rc)
        return(rc);
    }
    break;
    case NASDPD_P_PFLUSH_OBJ: {
      NASDPD_p_flush_obj_t *cmd;
      cmd = (NASDPD_p_flush_obj_t *)data;
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_flush_obj_dr(npd->h, cmd->in_key, &cmd->args,
        &cmd->res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
    }
    break;
    case NASDPD_P_PEJECT_OBJ: {
      NASDPD_p_eject_obj_t *cmd;
      cmd = (NASDPD_p_eject_obj_t *)data;
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_eject_obj_dr(npd->h, cmd->in_key, &cmd->args,
        &cmd->res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
    }
    break;
    case NASDPD_P_PREMOVE: {
      NASDPD_p_remove_t *cmd;
      cmd = (NASDPD_p_remove_t *)data;
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_remove_dr(npd->h, cmd->in_key, &cmd->args,
        &cmd->res, &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
    }
    break;
    case NASDPD_P_PINITIALIZE: {
      NASDPD_p_initialize_t *cmd;
      cmd = (NASDPD_p_initialize_t *)data;
      cmd->start_cc = NASD_RPCC();
      nasd_cl_p_initialize_dr(npd->h, &cmd->args, &cmd->res,
        &cmd->op_status);
      cmd->stop_cc = NASD_RPCC();
    }
    break;
    case NASDPD_T_TIMEOUT_TEST: {
      NASDPD_t_timeout_test_t *test;
      test = (NASDPD_t_timeout_test_t *)data;
#if NASD_TIMEOUT_TEST > 0
      printf("NASD KPDEV: begin timeout test\n");
      test->nasd_status = nasd_timeout_test(test->delta);
      printf("NASD KPDEV: done timeout test, got 0x%x (%s)\n",
        test->nasd_status, nasd_error_string(test->nasd_status));
#else /* NASD_TIMEOUT_TEST > 0 */
      printf("NASD KPDEV: timeout test scaffold not compiled in\n");
      test->nasd_status = NASD_OP_NOT_SUPPORTED;
#endif /* NASD_TIMEOUT_TEST > 0 */
    }
    break;
    default:
      return(EINVAL);
      break;
  }

  return(0);
}

#endif /* NNASDPD > 0 */

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
