/*
 * nasd_nfs_client_dce.c
 *
 * DCE dispatch functions for NASD NFS client
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <dce/dce_error.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_nfs_client.h>
#include <nasd/nasd_nfs_server.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_nonce_mgmt.h>
#include <nasd/nasd_mem.h>
#include <sys/errno.h>

#ifdef _KERNEL

#include <dce/ker/pthread_exc.h>

/* this doesnt exist in the kernel DCE header files for whatever reason */
#define exc_get_status(e,s) ( \
   (e)->kind == _exc_kind_status ? \
        (*(s) = (e)->match.value, 0) : \
        -1)

#endif /* _KERNEL */


nasd_status_t
nasd_nfscli_dce_init(
  nasd_shutdown_list_t  *sl)
{
  return(NASD_SUCCESS);
}

void
nasd_nfsdce_null(
  rpc_binding_handle_t   handle,
  nasd_status_t         *nasd_status,
  error_status_t        *op_status)
{
  TRY {
    nasd_null_fm(handle, nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_mount(
  rpc_binding_handle_t    handle,
  nasd_credential_t       in_credential,
  nasd_string_t           dirpath[],
  nasd_list_len_t         in_listlen,
  nasd_net_address_t      out_drivelist[],
  nasd_list_len_t        *out_listlen,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_mount_fm(handle, in_credential, dirpath,
      in_listlen, out_drivelist, out_listlen,
      out_cookie, out_identifier,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_fsstat(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsstat_t          *out_stat,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_fsstat_fm(handle, in_cookie, in_identifier,
      in_credential, out_stat, post_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_fsinfo(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsinfo_t          *out_info,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_fsinfo_fm(handle, in_cookie, in_identifier,
      in_credential, out_info, post_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_lookup_otw_provided(
  rpc_binding_handle_t     handle,
  nasd_nfs_lookup_args_t  *args,
  nasd_nfs_lookup_res_t   *res,
  nasd_byte_t             *args_otw_p,
  nasd_byte_t             *res_otw_p,
  error_status_t          *op_status)
{
  nasd_nfs_lookup_marshall_args(args, args_otw_p);

  TRY {
    nasd_lookup_otw_fm(handle, args_otw_p, res_otw_p, op_status);
    nasd_nfs_lookup_unmarshall_res(res_otw_p, res);
  }
  CATCH_ALL {
    res->nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_lookup(
  rpc_binding_handle_t     handle,
  nasd_nfs_lookup_args_t  *args,
  nasd_nfs_lookup_res_t   *res,
  error_status_t          *op_status)
{
  nasd_nfs_lookup_args_otw_t args_otw;
  nasd_nfs_lookup_res_otw_t res_otw;

  nasd_nfs_lookup_marshall_args(args, args_otw);

  TRY {
    nasd_lookup_otw_fm(handle, args_otw, res_otw, op_status);
    nasd_nfs_lookup_unmarshall_res(res_otw, res);
  }
  CATCH_ALL {
    res->nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_readdir(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_marker_t           in_marker,    
  nasd_markerverf_t       in_markerverf,
  nasd_len_t              in_count,
  nasd_marker_t          *out_marker,
  nasd_markerverf_t      *out_markerverf,
  nasd_len_t             *out_count,
  nasd_boolean_t         *out_eof,
  nasd_dirent_t           out_entries[],
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_readdir_fm(handle, in_cookie, in_identifier,
      in_credential, in_marker, in_markerverf, in_count,
      out_marker, out_markerverf, out_count, out_eof,
      out_entries, post_attribute, nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_access(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_access_t           in_access,
  nasd_access_t          *out_access,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_access_fm(handle, in_cookie, in_identifier,
      in_credential, in_access, out_access, post_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_setattr(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_guard_t            in_guard,
  nasd_attribute_t       *out_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_setattr_fm(handle, in_cookie, in_identifier,
      in_credential, in_attribute, in_fieldmask,
      in_guard, out_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_create(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_createmask_t       in_createmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_create_fm(handle, in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, in_createmask,
      out_cookie, out_identifier, out_attribute, post_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_symlink(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_string_t           in_symlink[],
  nasd_attribute_t        in_attribute,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_symlink_fm(handle, in_cookie, in_directory, in_credential,
      in_dirpath, in_symlink, in_attribute, out_cookie, out_identifier,
      out_attribute, post_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_remove(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_remove_fm(handle, in_cookie, in_directory, in_credential,
      in_dirpath, nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_mkdir(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_mkdir_fm(handle, in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, out_cookie,
      out_identifier, out_attribute, post_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_rmdir(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_rmdir_fm(handle, in_cookie, in_directory,
      in_credential, in_dirpath,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_newcookie(
  rpc_binding_handle_t    handle,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_cookie_t          *out_cookie,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_newcookie_fm(handle, in_identifier, in_credential,
      out_cookie, nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_rename(
  rpc_binding_handle_t    handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_from_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_from_path[],
  nasd_nfs_identifier_t   in_to_directory,
  nasd_string_t           in_to_path[],
  nasd_post_attribute_t  *post_from_attribute,
  nasd_post_attribute_t  *post_to_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  TRY {
    nasd_rename_fm(handle, in_cookie, in_from_directory,
      in_credential, in_from_path, in_to_directory, in_to_path,
      post_from_attribute, post_to_attribute,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

void
nasd_nfsdce_getstats(
  rpc_binding_handle_t      handle,
  nasd_nfs_opstats_t       *opstats,
  nasd_nfs_cachestats_t    *cachestats,
  nasd_nfs_rpc_opdepths_t  *opdepths,
  nasd_status_t            *nasd_status,
  error_status_t           *op_status)
{
  TRY {
    nasd_getstats_fm(handle, opstats, cachestats, opdepths,
      nasd_status, op_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_RPC_TRAP;
    exc_get_status(THIS_CATCH,(int *)op_status);
  } ENDTRY
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
