/*
 * nasd_knfs_server.c
 *
 * in-kernel NASD NFS server user binary
 * (reads mounts file and makes syscall to start server)
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <getopt.h>
#include <nasd/nasd_options.h>
#include <nasd/nasd_general.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_sys.h>
#include <sys/stat.h>

char *progname;
char *mountfile = "./nasd_nfs_mounts";

int svc_threads = 15;
int stack_size = 0;
int colocate = 0;

void
usage()
{
  fprintf(stderr, "USAGE: %s [options]\n", progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -l colocate with drive\n");
  fprintf(stderr, "  -m mountfile [default %s]\n", mountfile);
  fprintf(stderr, "  -s stack size (0 = DCE default) [default %d]\n",
    stack_size);
  fprintf(stderr, "  -t service threads [default %d]\n", svc_threads);
  fflush(stderr);
  exit(1);
}

int
main(
  int     argc,
  char  **argv)
{
  char buf[1024], *path, *drive, *part;
  struct nasd_nfs_config nfscfg;
  int ret, line, p, portnum;
  struct nasd_nfs_mount *m;
  nasd_uint32 net_addr;
  struct hostent *he;
  nasd_status_t rc;
  char c, *env;
  FILE *f;

  progname = argv[0];

  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  env = getenv("NASD_NFS_MNTFILE");
  if (env) {
    mountfile = env;
  }

  while ((c = getopt(argc, argv, "lm:s:t:")) != EOF) {
    switch(c) {
      case 'l':
        colocate = 1;
        break;
      case 'm':
        mountfile = optarg;
        break;
      case 's':
        if (sscanf(optarg, "%d", &stack_size) != 1)
          usage();
        break;
      case 't':
        if (sscanf(optarg, "%d", &svc_threads) != 1)
          usage();
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }
  if (optind < argc)
    usage();

  if (svc_threads < 1)
    usage();

  nfscfg.mount_array_len = 0;
  nfscfg.mount_array = (struct nasd_nfs_mount *)
    malloc((nfscfg.mount_array_len+1)*sizeof(struct nasd_nfs_mount));
  nfscfg.verbose = 1;
  nfscfg.svc_threads = svc_threads;
  nfscfg.stack_size = stack_size;
  nfscfg.colocate = colocate;

  ret = sscanf(NASD_PDRIVE_PORT, "%d", &portnum);
  if (ret != 1) {
    NASD_PANIC();
  }
  f = fopen(mountfile, "r");
  if (f == NULL) {
    fprintf(stderr, "NFS: Could not open %s for reading\n", mountfile);
    fflush(stderr);
    return(NASD_FAIL);
  }

  line = 0;
  while(fgets(buf, 1024, f) != NULL) {
    line++;
    buf[1023] = '\0';
    if ((buf[0] == '\0') || (buf[0] == '#'))
      continue;
    path = strtok(buf, " \t\n");
    if (path == NULL)
      continue;
    drive = strtok(NULL, " \t\n");
    if (drive == NULL) {
      fprintf(stderr, "Error parsing %s at line %d: no drive name\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    part = strtok(NULL, " \t\n");
    if (part == NULL) {
      fprintf(stderr, "Error parsing %s at line %d: no partition number\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    if (sscanf(part, "%d", &p) != 1) {
      fprintf(stderr, "Error parsing %s at line %d: bad partition number\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    if ((p < 0) || (p > 64)) {
      fprintf(stderr, "Error parsing %s at line %d: bad partition number\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    he = gethostbyname(drive);
    if (he == NULL) {
      fprintf(stderr, "Error resolving hostname %s (file %s line %d)\n",
        drive, mountfile, line);
      fflush(stderr);
      exit(1);
    }
    bcopy(he->h_addr_list[0], (char *)&net_addr, 4);

    nfscfg.mount_array = (struct nasd_nfs_mount *)realloc(nfscfg.mount_array,
      (nfscfg.mount_array_len+1)*sizeof(struct nasd_nfs_mount));
    m = &nfscfg.mount_array[nfscfg.mount_array_len];
    nfscfg.mount_array_len++;

#ifdef __alpha
    sprintf(m->drive_ip_addr, "%d.%d.%d.%d", (net_addr&0x000000ff),
      (net_addr&0x0000ff00)>>8, (net_addr&0x00ff0000)>>16,
      (net_addr&0xff000000)>>24);
#else /* __alpha */
    NASD_PANIC();
#endif /* __alpha */
    bcopy(path, m->path, NASD_MIN(MAXPATHLEN,strlen(path)));
    m->path[NASD_MIN(MAXPATHLEN-1,strlen(path))] = '\0';
    bcopy(drive, m->name, NASD_MIN(MAXPATHLEN,strlen(drive)));
    m->name[NASD_MIN(MAXPATHLEN-1,strlen(drive))] = '\0';
    m->portnum = portnum;
    m->partnum = p;
    m->net_addr = net_addr;
  }
  fclose(f);
  if (nfscfg.mount_array_len == 0) {
    fprintf(stderr, "ERROR: no mounts found in %s\n", mountfile);
    fflush(stderr);
    exit(1);
  }

  printf("Ready to launch NFS server.\n");
  fflush(stdout);
  fflush(stderr);
  rc = nasd_srv(NASD_NFS_SRV_GO, &nfscfg);
  perror("nasd_srv(NASD_NFS_SRV_GO)");

  exit(0);
}
