/*
 * nfs_core.c
 *
 * Core functionality for NASD NFS.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_timeout.h>
#include <nasd/nasd_nfs_internal.h>

/*
 * Server-global shutdown.
 */
nasd_shutdown_list_t *nasd_nfs_shutdown = NULL;

nasd_freelist_t *nasd_nfs_createpair_freelist = NULL;
#define NASD_MAX_FREE_CREATEPAIR 64
#define NASD_CREATEPAIR_INC       2
#define NASD_CREATEPAIR_INITIAL  16

nasd_freelist_t *nasd_nfs_setattrpair_freelist = NULL;
#define NASD_MAX_FREE_SETATTRPAIR 64
#define NASD_SETATTRPAIR_INC       2
#define NASD_SETATTRPAIR_INITIAL  16

nasd_freelist_t *nasd_nfs_infopage_freelist = NULL;
#define NASD_MAX_FREE_INFOPAGE 32
#define NASD_INFOPAGE_INC       2
#define NASD_INFOPAGE_INITIAL  16

nasd_freelist_t *nasd_nfs_lookupstuff_freelist = NULL;
#define NASD_MAX_FREE_LOOKUPSTUFF 32
#define NASD_LOOKUPSTUFF_INC       2
#define NASD_LOOKUPSTUFF_INITIAL  16

nasd_status_t
nasd_nfs_really_shutdown()
{
  nasd_status_t rc;
  int e;

  printf("NFS: really shutting down\n");

  /* synchronize any dirty data here (currently, there isn't any) */

  /* now get rid of clean dirs on the LRU */
  nasd_nfs_dir_cleanup();

  printf("NFS: shutdown drives\n");

  nasd_nfs_shutdown_drives();

  printf("NFS: begin automatic shutdown\n");

  e = 0;
  rc = nasd_shutdown_list_shutdown(nasd_nfs_shutdown);
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_shutdown_list_shutdown()\n",
      rc, nasd_error_string(rc));
    e = 1;
  }

  printf("NFS: automatic shutdown complete\n");

  nasd_shutdown_cleanup();

  nasd_cl_p_shutdown();

  printf("NFS: shutdown complete\n");

  return(NASD_SUCCESS);
}

void
nasd_nfs_shutdown_timeoutsys(
  void  *ignored)
{
  nasd_timeout_shutdown();
}

nasd_status_t
nasd_nfs_init()
{
  nasd_status_t rc;

  NASD_ASSERT(sizeof(nasd_nfs_dirpage_t) == 8192);
  NASD_ASSERT(sizeof(nasd_nfs_dirdata_t) == 64);
  NASD_ASSERT(sizeof(nasd_nfs_dirheader_t) == 64);

  rc = nasd_shutdown_sys_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_shutdown_sys_init()\n",
      rc, nasd_error_string(rc));
    return(rc);
  }

  rc = nasd_shutdown_list_init(&nasd_nfs_shutdown);
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_shutdown_list_init()\n",
      rc, nasd_error_string(rc));
    return(rc);
  }

  rc = nasd_timeout_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_timeout_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_shutdown_timeoutsys,
    NULL);
  if (rc) {
    nasd_nfs_shutdown_timeoutsys(NULL);
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_createpair_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_createpair_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_setattrpair_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_setattrpair_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_infopage_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_infopage_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_lookupstuff_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_lookupstuff_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_frame_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_frame_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_name_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_name_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_dir_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_dir_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_drsys_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_drsys_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_nfs_rpc_init();
  if (rc) {
    printf("ERROR: got 0x%x (%s) from nasd_nfs_rpc_init()\n",
      rc, nasd_error_string(rc));
    nasd_nfs_really_shutdown();
    return(rc);
  }

  rc = nasd_cl_p_init();
  if (rc) {
    printf("NFS: got 0x%x (%s) when initializing client library",
      rc, nasd_error_string(rc));
    return(rc);
  }

  return(NASD_SUCCESS);
}

nasd_nfs_createpair_t *
nasd_nfs_get_createpair()
{
  nasd_nfs_createpair_t *cp;

  NASD_FREELIST_GET(nasd_nfs_createpair_freelist,cp,next,(nasd_nfs_createpair_t *));
  return(cp);
}

void
nasd_nfs_free_createpair(
  nasd_nfs_createpair_t  *cp)
{
  NASD_FREELIST_FREE(nasd_nfs_createpair_freelist,cp,next);
}

nasd_status_t
nasd_nfs_createpair_init()
{
  nasd_status_t rc;

  NASD_FREELIST_CREATE(nasd_nfs_createpair_freelist, NASD_MAX_FREE_CREATEPAIR,
    NASD_CREATEPAIR_INC, sizeof(nasd_nfs_createpair_t));
  if (nasd_nfs_createpair_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_createpair_shutdown,
    NULL);
  if (rc) {
    nasd_nfs_createpair_shutdown(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_nfs_createpair_freelist,
    NASD_CREATEPAIR_INITIAL,next, (nasd_nfs_createpair_t *));
  return(NASD_SUCCESS);
}

void
nasd_nfs_createpair_shutdown(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_nfs_createpair_freelist,next,(nasd_nfs_createpair_t *));
}

nasd_nfs_setattrpair_t *
nasd_nfs_get_setattrpair()
{
  nasd_nfs_setattrpair_t *cp;

  NASD_FREELIST_GET(nasd_nfs_setattrpair_freelist,cp,next,(nasd_nfs_setattrpair_t *));
  return(cp);
}

void
nasd_nfs_free_setattrpair(
  nasd_nfs_setattrpair_t  *cp)
{
  NASD_FREELIST_FREE(nasd_nfs_setattrpair_freelist,cp,next);
}

nasd_status_t
nasd_nfs_setattrpair_init()
{
  nasd_status_t rc;

  NASD_FREELIST_CREATE(nasd_nfs_setattrpair_freelist, NASD_MAX_FREE_SETATTRPAIR,
    NASD_SETATTRPAIR_INC, sizeof(nasd_nfs_setattrpair_t));
  if (nasd_nfs_setattrpair_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_setattrpair_shutdown,
    NULL);
  if (rc) {
    nasd_nfs_setattrpair_shutdown(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_nfs_setattrpair_freelist,
    NASD_SETATTRPAIR_INITIAL,next, (nasd_nfs_setattrpair_t *));
  return(NASD_SUCCESS);
}

void
nasd_nfs_setattrpair_shutdown(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_nfs_setattrpair_freelist,next,(nasd_nfs_setattrpair_t *));
}

nasd_nfs_infopage_t *
nasd_nfs_get_infopage()
{
  nasd_nfs_infopage_t *cp;

  NASD_FREELIST_GET(nasd_nfs_infopage_freelist,cp,next,(nasd_nfs_infopage_t *));
  return(cp);
}

void
nasd_nfs_free_infopage(
  nasd_nfs_infopage_t  *cp)
{
  NASD_FREELIST_FREE(nasd_nfs_infopage_freelist,cp,next);
}

nasd_status_t
nasd_nfs_infopage_init()
{
  nasd_status_t rc;

  NASD_FREELIST_CREATE(nasd_nfs_infopage_freelist, NASD_MAX_FREE_INFOPAGE,
    NASD_INFOPAGE_INC, sizeof(nasd_nfs_infopage_t));
  if (nasd_nfs_infopage_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_infopage_shutdown,
    NULL);
  if (rc) {
    nasd_nfs_infopage_shutdown(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_nfs_infopage_freelist,
    NASD_INFOPAGE_INITIAL,next, (nasd_nfs_infopage_t *));
  return(NASD_SUCCESS);
}

void
nasd_nfs_infopage_shutdown(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_nfs_infopage_freelist,next,(nasd_nfs_infopage_t *));
}

nasd_nfs_lookupstuff_t *
nasd_nfs_get_lookupstuff()
{
  nasd_nfs_lookupstuff_t *cp;

  NASD_FREELIST_GET(nasd_nfs_lookupstuff_freelist,cp,next,(nasd_nfs_lookupstuff_t *));
  return(cp);
}

void
nasd_nfs_free_lookupstuff(
  nasd_nfs_lookupstuff_t  *cp)
{
  NASD_FREELIST_FREE(nasd_nfs_lookupstuff_freelist,cp,next);
}

nasd_status_t
nasd_nfs_lookupstuff_init()
{
  nasd_status_t rc;

  NASD_FREELIST_CREATE(nasd_nfs_lookupstuff_freelist, NASD_MAX_FREE_LOOKUPSTUFF,
    NASD_LOOKUPSTUFF_INC, sizeof(nasd_nfs_lookupstuff_t));
  if (nasd_nfs_lookupstuff_freelist == NULL) {
    return(NASD_NO_MEM);
  }
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_lookupstuff_shutdown,
    NULL);
  if (rc) {
    nasd_nfs_lookupstuff_shutdown(NULL);
    return(rc);
  }
  NASD_FREELIST_PRIME(nasd_nfs_lookupstuff_freelist,
    NASD_LOOKUPSTUFF_INITIAL,next, (nasd_nfs_lookupstuff_t *));
  return(NASD_SUCCESS);
}

void
nasd_nfs_lookupstuff_shutdown(
  void  *ignored)
{
  NASD_FREELIST_DESTROY(nasd_nfs_lookupstuff_freelist,next,(nasd_nfs_lookupstuff_t *));
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
