/*
 * nfs_kmain.c
 *
 * in-dux-kernel statup and support for NASD NFS server
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd_server.h>

#include <sys/secdefines.h>
#if SEC_BASE
#include <sys/security.h>
#include <sys/audit.h>
#include <sys/secpolicy.h>
#endif /* SEC_BASE */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/user.h>
#include <sys/kernel.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/buf.h>
#include <sys/proc.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/conf.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/specdev.h>
#include <sys/flock.h>
#include <sys/syslimits.h>
#include <kern/assert.h>
#include <kern/parallel.h>
#include <mach/mach_types.h>
#include <vm/vm_page.h>
#include <vm/vm_vppage.h>
#include <sys/vfs_ubc.h>
#include <vm/vm_mmap.h>
#include <vm/vm_vp.h>
#include <vm/vm_debug.h>
#include <sys/malloc.h>
#include <machine/rpb.h>

#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_sys.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_nfs_internal.h>

#include <dce/dce_error.h>
#include <dce/ker/pthread_exc.h>
#include <rpc/rpc.h>
#include <dce/rpc.h>
#include <dce/stubbase.h>

#include <dce/dce_error.h>

extern struct nasd_nfs_svinfo *nasd_nfs_sva;

nasd_status_t
nasd_nfs_frame_init()
{
  /* no kernel-specific stuff to init, here */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_nfs_do_drive_bind(
  char                 *bind_name,
  nasd_drive_handle_t  *dhp)
{
  nasd_status_t rc;

  if (nasd_nfs_sva->config.colocate) {
    rc = nasd_bind_to_drive(bind_name, NASD_PDRIVE_PORT,
      NASD_BIND_COLOCATE, NULL, 0, dhp);
  }
  else {
    rc = nasd_bind_to_drive(bind_name, NASD_PDRIVE_PORT,
      NASD_BIND_DCE_DIRECT, NULL, 0, dhp);
  }
  return(rc);
}

#define STOP_PSRV_THREAD() { \
  thread_terminate(current_thread()); \
  thread_halt_self(); \
}

void
nasd_nfs_srv_thread(
  nasd_threadarg_t  arg)
{
  struct nasd_nfs_svinfo *svinfo;
  struct nasd_nfs_config *nfscfg;
  dce_error_string_t err_str;
  error_status_t ps, status;
  nasd_status_t rc;

  svinfo = (struct nasd_nfs_svinfo *)arg;
  nfscfg = &svinfo->config;

  if (nfscfg->verbose) {
    printf("NFS: setup RPC\n");
  }

  rc = nasd_nfs_startup_rpc();
  if (rc) {
    printf("NFS: got 0x%x (%s) from nasd_nfs_startup_rpc()\n",
      rc, nasd_error_string(rc));
    STOP_PSRV_THREAD();
  }

  if (nfscfg->verbose) {
    printf("NFS: begin listening\n");
  }

  if (nfscfg->stack_size) {
    rpc_mgmt_set_server_stack_size(nfscfg->stack_size, &status);
    if (status) {
      dce_error_inq_text(status, err_str, &ps);
      if (ps)
        strcpy(err_str, "unknown");
      printf("NFS: status 0x%x (%s) setting stack size to %d\n",
        status, err_str, nfscfg->stack_size);
    }
  }

  rpc_server_listen(nfscfg->svc_threads, &status);
  switch(status) {
    case rpc_s_already_listening:
      svinfo->became_listener = 0;
      break;
    default:
      dce_error_inq_text(status, err_str, &ps);
      if (ps)
        strcpy(err_str, "unknown");
      printf("NFS: status 0x%x (%s) from listen\n", status, err_str);
      svinfo->became_listener = 1;
  }

  STOP_PSRV_THREAD();
}

/*
 * Call with NASD_D_LOCK held, consumes the lock.
 */
int
nasd_nfs_srv_go(
  struct nasd_nfs_svinfo  *svinfo)
{
  nasd_thread_t nasd_k_handler_thread;
  int verbose, abort, error, ret, i;
  struct nasd_nfs_config *nfscfg;
  struct nasd_nfs_mount *nfsmnt;
  nasd_status_t rc;

  nfscfg = &svinfo->config;
  verbose = nfscfg->verbose;
  abort = 1;

  /* GC unused pages */
  pmap_collect(vm_map_pmap(current_thread()->task->map));

  rc = nasd_nfs_init();
  if (rc) {
    printf("NFS: got 0x%x (%s) from nasd_nfs_init()\n",
      rc, nasd_error_string(rc));
    error = EIO;
    goto done;
  }

  /* deal with all the mounts given us by the user */
  for(i=0;i<nfscfg->mount_array_len;i++) {
    nfsmnt = &nfscfg->mount_array[i];
    rc = nasd_nfs_add_mount(nfsmnt->path, nfsmnt->name,
      nfsmnt->drive_ip_addr, nfsmnt->portnum, nfsmnt->partnum,
      nfsmnt->net_addr);
    if (rc) {
      printf("NFS: got 0x%x (%s) adding mount for %s@%s\n",
        rc, nasd_error_string(rc),
        nfsmnt->path, nfsmnt->name);
      rc = nasd_nfs_really_shutdown();
      if (rc) {
        printf("NFS: nasd_nfs_really_shutdown() returned 0x%x (%s)!\n",
          rc, nasd_error_string(rc));
        NASD_PANIC();
      }
      error = EIO;
      goto done;
    }
  }

  rc = nasd_thread_create(&nasd_k_handler_thread,
    (void (*)())nasd_nfs_srv_thread, (void *)svinfo);
  if (rc) {
    printf("nasd_nfs_srv_go(): cannot create handler thread, rc=0x%x (%s)\n",
      rc, nasd_error_string(rc));
    error = EIO;
    goto done;
  }

  /* wait for kill */
  assert_wait(0, TRUE);
  NASD_D_UNLOCK();

  if (verbose) {
    printf("NFS: thread away\n");
  }

  thread_block();

  if (verbose) {
    printf("NFS: begin shutdown\n");
  }

  /* cleanup */
  NASD_D_LOCK();

  if (verbose) {
    printf("NFS: stop RPC subsystem\n");
  }

  nasd_nfs_stop_rpc();

  if (verbose) {
    printf("NFS: sequence shutdown\n");
  }

  rc = nasd_nfs_really_shutdown();
  if (rc) {
    printf("NFS: nasd_nfs_really_shutdown() returned 0x%x (%s)!\n",
      rc, nasd_error_string(rc));
    NASD_PANIC();
  }

  abort = 0;
  error = EINTR;

done:

#ifdef free
#define dce_sucks_goats free
#undef free
#endif /* free */
  NASD_Free(nfscfg->mount_array,
    sizeof(struct nasd_nfs_mount)*nfscfg->mount_array_len);
  NASD_Free(svinfo, sizeof(struct nasd_nfs_svinfo));
#ifdef dce_sucks_goats
#define free dce_sucks_goats
#undef dce_sucks_goats
#endif /* dce_sucks_goats */

  nasd_nfs_sva = NULL;

  NASD_D_UNLOCK();

  if (verbose && (abort == 0)) {
    printf("NFS: done shutdown\n");
  }

  return(error);
}
