/*
 * nfs_main.c
 *
 * main() and friends for NASD NFS server.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_nfs_internal.h>
#include <nasd/nasd_pdrive.h>
#include <netdb.h>

char *progname;
char *mountfile = "./nasd_nfs_mounts";

int svc_threads = 15;
int stack_size  = 32768;

static int stopped = 0;
#if LINUX
NASD_DECLARE_STATIC_MUTEX(nasd_nfs_shutdown_mutex);
#endif /* LINUX */

int                           binding_type;
int                           binding_args_len;
void                         *binding_args;
char                         *binding_port = NASD_PDRIVE_PORT;
nasd_drive_param_dce_kpdev_t  kpdev_args;
int                           nondefault_binding = 0;

nasd_status_t
nasd_nfs_do_drive_bind(
  char                 *bind_name,
  nasd_drive_handle_t  *dhp)
{
  nasd_status_t rc;

  rc = nasd_bind_to_drive(bind_name, binding_port,
    binding_type, binding_args, binding_args_len, dhp);
  return(rc);
}

nasd_status_t
nasd_nfs_frame_init()
{
  /* no longer any user-specific stuff to init here */
  return(NASD_SUCCESS);
}

nasd_status_t
nasd_nfs_load_mounts()
{
  char buf[1024], *path, *drive, *part;
  int p, line, nmounts, ret, portnum;
  nasd_uint32 net_addr;
  struct hostent *he;
  nasd_status_t rc;
  FILE *f;

  ret = sscanf(NASD_PDRIVE_PORT, "%d", &portnum);
  if (ret != 1) {
    NASD_PANIC();
  }
  f = fopen(mountfile, "r");
  if (f == NULL) {
    fprintf(stderr, "NFS: Could not open %s for reading\n", mountfile);
    fflush(stderr);
    return(NASD_FAIL);
  }

  nmounts = 0;
  line = 0;
  while(fgets(buf, 1024, f) != NULL) {
    line++;
    buf[1023] = '\0';
    if ((buf[0] == '\0') || (buf[0] == '#'))
      continue;
    path = strtok(buf, " \t\n");
    if (path == NULL)
      continue;
    drive = strtok(NULL, " \t\n");
    if (drive == NULL) {
      fprintf(stderr, "Error parsing %s at line %d: no drive name\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    part = strtok(NULL, " \t\n");
    if (part == NULL) {
      fprintf(stderr, "Error parsing %s at line %d: no partition number\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    if (sscanf(part, "%d", &p) != 1) {
      fprintf(stderr, "Error parsing %s at line %d: bad partition number\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    if ((p < 0) || (p > 64)) {
      fprintf(stderr, "Error parsing %s at line %d: bad partition number\n",
        mountfile, line);
      fflush(stderr);
      exit(1);
    }
    he = gethostbyname(drive);
    if (he == NULL) {
      fprintf(stderr, "Error resolving hostname %s (file %s line %d)\n",
        drive, mountfile, line);
      fflush(stderr);
      exit(1);
    }
    bcopy(he->h_addr_list[0], (char *)&net_addr, 4);
    rc = nasd_nfs_add_mount(path, drive, drive, portnum, p, net_addr);
    if (rc) {
      fclose(f);
      return(rc);
    }
    nmounts++;
  }
  fclose(f);
  if (nmounts == 0) {
    fprintf(stderr, "NFS WARNING: no mounts found in %s\n", mountfile);
    fflush(stderr);
  }
  return(NASD_SUCCESS);
}

void
handle_sigint()
{
  nasd_status_t rc;

#if LINUX
  NASD_LOCK_MUTEX(nasd_nfs_shutdown_mutex);
  if(stopped)
    return;
  stopped = 1;
  NASD_UNLOCK_MUTEX(nasd_nfs_shutdown_mutex);
#else /* LINUX */
  stopped = 1;
#endif /* LINUX */

  fprintf(stderr, "NFS INTERRUPTED\n");
  fflush(stderr);

  nasd_nfs_stop_rpc();

  rc = nasd_nfs_really_shutdown();
  if (rc) {
    fprintf(stderr,
      "NFS ERROR: got 0x%x (%s) from nasd_nfs_really_shutdown()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  nasd_mem_shutdown();

  printf("Completed shutdown.\n");
#if NASD_MEM_COUNT_ALLOC > 0
  if (nasd_mem_allocated) {
    extern int nasd_mem_use_counter;
    fprintf(stderr,
      "WARNING: %ld bytes of memory still outstanding, something leaked core.\n",
      nasd_mem_allocated);
    fprintf(stderr, "memory subsystem use counter at %d\n", nasd_mem_use_counter);
    fflush(stderr);
  }
  else {
    printf("%ld bytes of memory still outstanding.\n", nasd_mem_allocated);
  }
#endif /* NASD_MEM_COUNT_ALLOC > 0 */
  fflush(stdout);

  exit(0);
}

void
usage()
{
  fprintf(stderr, "USAGE: %s [options]\n", progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -k use kernel device\n");
  fprintf(stderr, "  -l use colocated drive\n");
  fprintf(stderr, "  -m mountfile [default %s]\n", mountfile);
  fprintf(stderr, "  -s stack size [default %d]\n", stack_size);
  fprintf(stderr, "  -t service threads [default %d]\n", svc_threads);
  fflush(stderr);
  exit(1);
}

int
main(
  int     argc,
  char  **argv)
{
  dce_error_string_t err_str;
  error_status_t status;
  int ps;
  nasd_status_t rc;
  char c, *env;

  progname = argv[0];

  binding_type = NASD_BIND_DCE_DIRECT;
  binding_args = NULL;
  binding_args_len = 0;

  env = getenv("NASD_NFS_MNTFILE");
  if (env) {
    mountfile = env;
  }

  while ((c = getopt(argc, argv, "klm:s:t:")) != EOF) {
    switch(c) {
      case 'k':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_KPDEV;
        binding_args = &kpdev_args;
        binding_args_len = sizeof(kpdev_args);
        strcpy(kpdev_args.devname, "/dev/nasdkp0");
        break;
      case 'l':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_COLOCATE;
        binding_args = &kpdev_args;
        binding_args_len = sizeof(kpdev_args);
        strcpy(kpdev_args.devname, "/dev/nasdkp0");
        break;
      case 'm':
        mountfile = optarg;
        break;
      case 's':
        if (sscanf(optarg, "%d", &stack_size) != 1)
          usage();
        break;
      case 't':
        if (sscanf(optarg, "%d", &svc_threads) != 1)
          usage();
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }
  if (optind < argc)
    usage();

  rc = nasd_threads_init();
  if (rc) {
    fprintf(stderr, "NFS ERROR: got 0x%x (%s) from nasd_threads_init()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_mem_init();
  if (rc) {
    fprintf(stderr, "NFS ERROR: got 0x%x (%s) from nasd_mem_init()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_nfs_init();
  if (rc) {
    fprintf(stderr, "NFS ERROR: got 0x%x (%s) from nasd_nfs_init()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  rc = nasd_nfs_load_mounts();
  if (rc) {
    fprintf(stderr, "NFS ERROR: got 0x%x (%s) from nasd_nfs_load_mounts()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

  printf("NFS: setup RPC subsystem\n");
 
  rc = nasd_nfs_startup_rpc();
  if (rc) {
    fprintf(stderr, "NFS ERROR: got 0x%x (%s) from nasd_nfs_startup_rpc()\n",
      rc, nasd_error_string(rc));
    fflush(stderr);
    exit(1);
  }

#if LINUX
  nasd_mutex_init(&nasd_nfs_shutdown_mutex);
#endif /* LINUX */

  signal(SIGINT, handle_sigint);
#ifndef LINUX
  signal(SIGUSR1, handle_sigint);
#endif /* LINUX */

  printf("Running with pid %d\n", getpid());
  fflush(stdout);

  rpc_mgmt_set_server_stack_size(stack_size, &status);
  if (status) {
    dce_error_inq_text(status, err_str, &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("NFS: status 0x%x (%s) setting stack size to %d\n",
      status, err_str, stack_size);
  }

  rpc_server_listen(svc_threads, &status);

  printf("Listener aborted! status=0x%x\n", status);
  fflush(stdout);

  if (!stopped) {
    nasd_nfs_stop_rpc();

    rc = nasd_nfs_really_shutdown();
    if (rc) {
      fprintf(stderr,
        "NFS ERROR: got 0x%x (%s) from nasd_nfs_really_shutdown()\n",
        rc, nasd_error_string(rc));
      fflush(stderr);
      exit(1);
    }
  }

  exit(2);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
