/*
 * nfs_ops.c
 *
 * RPC dispatch for NASD NFS server.
 */
/*
 * Copyright (c) 1997,1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#undef  NASD_PTHREAD_EXC
#define NASD_PTHREAD_EXC 1
#include <nasd/nasd_types.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_freelist.h>
#include <nasd/nasd_itypes.h>
#include <nasd/nasd_mem.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_nfs_internal.h>
#include <nasd/nasd_nfs_server.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_control.h>
#include <nasd/nasd_timer.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_nfs_rpc.h>
#include <nasd/nasd_timeout.h>

#include <sys/errno.h>

#ifdef KERNEL
#include <dce/ker/pthread_exc.h>
#endif /* KERNEL */
#include <rpc/rpc.h>
#include <dce/stubbase.h>

nasd_uint64 nasd_nfs_threads_in_rpc = 0;
int nasd_nfs_wait_shutdown = 0;
NASD_DECLARE_COND(nasd_nfs_shutdown_cond)
NASD_DECLARE_MUTEX(nasd_nfs_rpc_cnt_lock)
NASD_DECLARE_MUTEX(nasd_nfs_rpc_qdepth_lock)

int nasd_nfs_protseq_installed = 0;
int nasd_nfs_active = 0;

int nasd_nfs_rpc_qdepth = 0;
nasd_timer_t nasd_nfs_rpc_qdepth_timer;
nasd_nfs_rpc_opdepths_t nasd_nfs_rpc_opdepths;

nasd_uint64 nasd_nfs_shutdown_timeouts = 0;
nasd_timeout_handle_t nasd_nfs_shutdown_timeout_handle;

nasd_nfs_opstats_t nasd_nfs_opstats;
nasd_nfs_cachestats_t nasd_nfs_cache_stats;

nasd_timespec_t nasd_nfs_shutdown_timeout_interval = {1, 0};

/* Don't use __FILE__ here to avoid getting kernel-mangled names */
#define DOBEGIN(_opname_) NASD_NFS_RPC_DOBEGIN(_opname_,"nasd_nfs_ops.c")
#define DORETURN(_opname_) NASD_NFS_RPC_DORETURN(_opname_,"nasd_nfs_ops.c")

nasd_nfs_lookup_res_t nasd_nfs_lookup_failmem;

/*
 * RPC handlers begin here
 */

void
nasd_snfs_null_fm(
  handle_t         client_handle, 
  nasd_status_t   *nasd_status,
  error_status_t  *op_status)
{
  nasd_nfs_rpc_basic_t out_res;
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(null);
  *nasd_status = out_res.nasd_status = NASD_SUCCESS;
  DORETURN(null);
}

void
nasd_snfs_mount_fm(
  handle_t                client_handle,
  nasd_credential_t       in_credential,
  nasd_string_t           dirpath[],
  nasd_list_len_t         in_listlen,
  nasd_net_address_t      out_drivelist[],
  nasd_list_len_t        *out_listlen,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  nasd_status_t rc;
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(mount);

  TRY {
    nasd_nfs_real_mount(in_credential, dirpath,
      in_listlen, out_drivelist, out_listlen, out_cookie,
      out_identifier, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing mount\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(mount);
}

void
nasd_snfs_fsstat_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsstat_t          *out_stat,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(fsstat);

  TRY {
    nasd_nfs_real_fsstat(in_cookie, in_identifier, in_credential,
      out_stat, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing fsstat\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(fsstat);
}

void
nasd_snfs_fsinfo_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_fsinfo_t          *out_info,    
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(fsinfo);

  TRY {
    nasd_nfs_real_fsinfo(in_cookie, in_identifier, in_credential,
      out_info, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing fsinfo\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(fsinfo);
}

void
nasd_snfs_lookup_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_post_attribute_t  *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(lookup);

  TRY {
    nasd_nfs_real_lookup(in_cookie, in_identifier, in_credential,
      (char *)in_dirpath, out_cookie, out_identifier,
      out_attribute, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing lookup\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(lookup);
}

void
nasd_snfs_readdir_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_marker_t           in_marker,    
  nasd_markerverf_t       in_markerverf,
  nasd_len_t              in_count,
  nasd_marker_t          *out_marker,
  nasd_markerverf_t      *out_markerverf,
  nasd_len_t             *out_count,
  nasd_boolean_t         *out_eof,
  nasd_dirent_t           out_entries[],
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(readdir);

  TRY {
    nasd_nfs_real_readdir(in_cookie, in_identifier, in_credential,
      in_marker, in_markerverf, in_count, out_marker, out_markerverf,
      out_count, out_eof, out_entries, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing readdir\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(readdir);
}

void
nasd_snfs_readdirplus_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_marker_t           in_marker,    
  nasd_markerverf_t       in_markerverf,
  nasd_len_t              in_count,
  nasd_marker_t          *out_marker,
  nasd_markerverf_t      *out_markerverf,
  nasd_len_t             *out_count,
  nasd_boolean_t         *out_eof,
  nasd_dirent_t           out_entries[],
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(readdirplus);

  *nasd_status = NASD_OP_NOT_SUPPORTED;

  DORETURN(readdirplus);
}

void
nasd_snfs_access_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_access_t           in_access,
  nasd_access_t          *out_access,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(access);

  TRY {
    nasd_nfs_real_access(in_cookie, in_identifier, in_credential,
      in_access, out_access, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing access\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(access);
}

void
nasd_snfs_setattr_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_guard_t            in_guard,
  nasd_attribute_t       *out_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(setattr);

  TRY {
    nasd_nfs_real_setattr(in_cookie, in_identifier,
      in_credential, in_attribute, in_fieldmask, in_guard,
      out_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing setattr\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(setattr);
}

void
nasd_snfs_create_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_createmask_t       in_createmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(create);

  TRY {
    nasd_nfs_real_create(in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, in_createmask,
      out_cookie, out_identifier, out_attribute, post_attribute,
      nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing create\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(create);
}

void
nasd_snfs_symlink_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_string_t           in_symlink[],
  nasd_attribute_t        in_attribute,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(symlink);

  TRY {
    nasd_nfs_real_symlink(in_cookie, in_directory, in_credential,
      in_dirpath, in_symlink, in_attribute, out_cookie, out_identifier,
      out_attribute, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing symlink\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(symlink);
}

void
nasd_snfs_remove_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(remove);

  TRY { 
    nasd_nfs_real_remove(in_cookie, in_directory, in_credential,
      in_dirpath, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing remove\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(remove);
}

void
nasd_snfs_mkdir_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_attribute_t        in_attribute,
  nasd_fieldmask_t        in_fieldmask,
  nasd_cookie_t          *out_cookie,
  nasd_nfs_identifier_t  *out_identifier,
  nasd_attribute_t       *out_attribute,
  nasd_post_attribute_t  *post_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(mkdir);

  TRY {
    nasd_nfs_real_mkdir(in_cookie, in_directory, in_credential,
      in_dirpath, in_attribute, in_fieldmask, out_cookie, out_identifier,
      out_attribute, post_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing mkdir\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(mkdir);
}

void
nasd_snfs_rmdir_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_dirpath[],
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(rmdir);

  TRY {
    nasd_nfs_real_rmdir(in_cookie, in_directory, in_credential,
      in_dirpath, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing rmdir\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(rmdir);
}

void
nasd_snfs_newcookie_fm(
  handle_t                client_handle,
  nasd_nfs_identifier_t   in_identifier,
  nasd_credential_t       in_credential,
  nasd_cookie_t          *out_cookie,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(newcookie);
  
  TRY {
    nasd_nfs_real_newcookie(in_identifier, in_credential,
      out_cookie, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing newcookie\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(newcookie);
}

void
nasd_snfs_rename_fm(
  handle_t                client_handle,
  nasd_cookie_t           in_cookie,
  nasd_nfs_identifier_t   in_from_directory,
  nasd_credential_t       in_credential,
  nasd_string_t           in_from_path[],
  nasd_nfs_identifier_t   in_to_directory,
  nasd_string_t           in_to_path[],
  nasd_post_attribute_t  *post_from_attribute,
  nasd_post_attribute_t  *post_to_attribute,
  nasd_status_t          *nasd_status,
  error_status_t         *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(rename);

  TRY {
    nasd_nfs_real_rename(in_cookie, in_from_directory, in_credential,
      in_from_path, in_to_directory, in_to_path, post_from_attribute,
      post_to_attribute, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing rename\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(rename);
}

void
nasd_snfs_getstats_fm(
  handle_t                  client_handle,
  nasd_nfs_opstats_t       *opstats,
  nasd_nfs_cachestats_t    *cachestats,
  nasd_nfs_rpc_opdepths_t  *opdepths,
  nasd_status_t            *nasd_status,
  error_status_t           *op_status)
{
  NASD_NFS_RPC_OP_STATS_DECL

  DOBEGIN(getstats);

  TRY {
    nasd_nfs_real_getstats(opstats, cachestats, opdepths, nasd_status);
  }
  CATCH_ALL {
    *nasd_status = NASD_REMOTE_TRAP;
#if NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0
    printf("NFS: caught exception performing getstats\n");
#endif /* NASD_NFS_RPC_DEBUG_DCE_EXCEPTION > 0 */
  } ENDTRY

  DORETURN(getstats);
}

void
nasd_snfs_lookup_otw_fm(
  handle_t                     client_handle, 
  nasd_nfs_lookup_args_otw_t   in_args_otw,
  nasd_nfs_lookup_res_otw_t    out_res_otw,
  error_status_t              *op_status)
{
  nasd_status_t *nasd_status, rc;
  nasd_nfs_lookupstuff_t *lus;
  NASD_NFS_RPC_OP_STATS_DECL

  nasd_status = &rc;
  DOBEGIN(lookup_otw);

  lus = nasd_nfs_get_lookupstuff();
  if (lus == NULL) {
    /* marshall a failure */
    nasd_nfs_lookup_marshall_res(&nasd_nfs_lookup_failmem, out_res_otw);
    goto really_done_lookup_otw;
  }

  nasd_nfs_lookup_unmarshall_args(in_args_otw, &lus->in_args);

  nasd_nfs_real_lookup(lus->in_args.in_cookie, lus->in_args.in_identifier,
    lus->in_args.in_credential, (char *)lus->in_args.in_dirpath,
    &lus->out_res.out_cookie, &lus->out_res.out_identifier,
    &lus->out_res.out_attribute, &lus->out_res.post_attribute,
    &rc);
  lus->out_res.nasd_status = rc;

  nasd_nfs_lookup_marshall_res(&lus->out_res, out_res_otw);

really_done_lookup_otw:
#if (DEBUG_LOOKUP_DETAIL > 0) || (DEBUG_RPC_ERR_DETAIL > 0)
  if (lus == NULL) {
    printf("lookup_otw complete lus NULL\n");
  }
  else if (rc) {
    printf("lookup_otw returning nasd status 0x%x (%s) for in_dirpath=\"%s\"\n", rc, nasd_error_string(rc), (char *)lus->in_args.in_dirpath);
  }
#if DEBUG_LOOKUP_DETAIL > 0
  else if (lus->out_res.out_attribute.valid != NASD_ATTRIBUTE_VALID) {
    printf("lookup_otw Successful with invalid attribute for in_dirpath=\"%s\" nasdid 0x%"
           NASD_ID_FMT "\n",
      (char *)lus->in_args.in_dirpath,
      lus->out_res.out_identifier.nasd_identifier);
    if (lus->out_res.out_attribute.valid == NASD_ATTRIBUTE_INVALID) {
      printf("lus->out_res.out_attribute.valid = INVALID\n");
    }
    else {
      printf("lus->out_res.out_attribute.valid = 0x%x\n", ((int)lus->out_res.out_attribute.valid)&0xff);
    }
  }
  else {
    printf("lookup_otw Successful for in_dirpath=\"%s\" nasdid 0x%"
           NASD_ID_FMT "\n%s attributes: ",
           (char *)lus->in_args.in_dirpath,
           lus->out_res.out_identifier.nasd_identifier,
           (char *)lus->in_args.in_dirpath);
    nasd_dump_nfs3_attributes((nasd_nfs3_attributes_t *)lus->out_res.out_attribute.attribute.fs_specific);
    printf("\n");
#ifndef KERNEL
    fflush(stdout);
#endif /* !KERNEL */
  }
#endif /* DEBUG_LOOKUP_DETAIL > 0 */
#endif /* (DEBUG_LOOKUP_DETAIL > 0) || (DEBUG_RPC_ERR_DETAIL > 0) */

  if (lus) {
    nasd_nfs_free_lookupstuff(lus);
  }

  DORETURN(lookup_otw);
}

/*
 * RPC mechanism begins here
 */

/*
 * !!! Do not hand-modify! Generate this stanza with "make mepv"
 */

static nasd_nfs_server_v0_1_epv_t nasd_nfs_server_epv = {
  nasd_snfs_null_fm
  ,nasd_snfs_mount_fm
  ,nasd_snfs_fsstat_fm
  ,nasd_snfs_fsinfo_fm
  ,nasd_snfs_lookup_fm
  ,nasd_snfs_readdir_fm
  ,nasd_snfs_readdirplus_fm
  ,nasd_snfs_access_fm
  ,nasd_snfs_setattr_fm
  ,nasd_snfs_create_fm
  ,nasd_snfs_symlink_fm
  ,nasd_snfs_remove_fm
  ,nasd_snfs_mkdir_fm
  ,nasd_snfs_rmdir_fm
  ,nasd_snfs_newcookie_fm
  ,nasd_snfs_rename_fm
  ,nasd_snfs_getstats_fm
  ,nasd_snfs_lookup_otw_fm
};

void
nasd_nfs_begin_rpc(
  nasd_opstat_t  *opstat)
{
  NASD_ATOMIC_INC64(&nasd_nfs_threads_in_rpc);
}

void
nasd_nfs_end_rpc(
  nasd_opstat_t  *opstat)
{
  NASD_ATOMIC_DEC64(&nasd_nfs_threads_in_rpc);
  if (nasd_nfs_wait_shutdown && (nasd_nfs_threads_in_rpc == 0)) {
    NASD_BROADCAST_COND(nasd_nfs_shutdown_cond);
  }
}

void
nasd_nfs_dump_opstat(
  nasd_opstat_t  *os,
  char           *name)
{
  unsigned long s, ns;

  s = os->op_nsecs / 1000000000UL;
  ns = os->op_nsecs % 1000000000UL;
  printf("Operation: %s\n", name);
  printf("  time       %lu:%09lu\n", s, ns);
  printf("  num_ops    %5lu\n", os->num_ops);
  printf("  in_flight  %5lu\n", os->in_flight);
  printf("  invalid    %5ld\n", os->invalid);
}

#define DUMP_OPSTAT(_opname_) \
  nasd_nfs_dump_opstat(&nasd_nfs_opstats._opname_,NASD_STRING(_opname_))

#define NASD_NFS_CSDUMP(_stat_) { \
  printf("  %24s %5lu\n", NASD_STRING(_stat_), nasd_nfs_cache_stats._stat_); \
}

void
nasd_nfs_timeout_shutdown_rpc(
  nasd_timeout_handle_t   tmh,
  void                   *arg1,
  void                   *arg2)
{
  nasd_timespec_t ts;
  char str[128];

#if NASD_NFS_RPC_DEBUG_RPC_SHUTDOWN > 0
  nasd_gettime(&ts);
  printf("NFS: shutdown timeout at %s, nasd_nfs_threads_in_rpc=%"
    NASD_64u_FMT "\n", nasd_timestr_r(ts, str), nasd_nfs_threads_in_rpc);
#endif /* NASD_NFS_RPC_DEBUG_RPC_SHUTDOWN > 0 */
  if (nasd_nfs_threads_in_rpc == 0) {
    NASD_BROADCAST_COND(nasd_nfs_shutdown_cond);
  }
  nasd_nfs_shutdown_timeouts++;
}

void
nasd_nfs_wait_shutdown_rpc()
{
  nasd_status_t rc;

  nasd_nfs_wait_shutdown = 1;

  rc = NASD_FAIL;

  NASD_LOCK_MUTEX(nasd_nfs_rpc_cnt_lock);
  while(nasd_nfs_threads_in_rpc) {
    if (rc) {
      rc = nasd_timeout_add(&nasd_nfs_shutdown_timeout_handle,
        nasd_nfs_timeout_shutdown_rpc, NULL, NULL,
        nasd_nfs_shutdown_timeout_interval,
        nasd_nfs_shutdown_timeout_interval,
        NASD_TIMEOUT_F_PERIODIC);
    }
    if (rc) {
      printf("NFS: got 0x%x (%s) adding timeout for shutdown-quiesce\n",
        rc, nasd_error_string(rc));
    }
    NASD_WAIT_COND(nasd_nfs_shutdown_cond,nasd_nfs_rpc_cnt_lock);
  }
  NASD_UNLOCK_MUTEX(nasd_nfs_rpc_cnt_lock);
}

void
nasd_nfs_stop_rpc()
{
  rpc_if_handle_t nasd_nfs_ifspec;
  dce_error_string_t err_str;
  error_status_t status;
  int ps;

  nasd_nfs_active = 0;

  nasd_nfs_ifspec = nasd_nfs_server_v0_1_s_ifspec;

#if 0
  /*
   * This just doesn't work. Hurray DCE.
   */
  rpc_mgmt_stop_server_listening(NULL, &status);
  if (status) {
    dce_error_inq_text(status, err_str, &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("NFS: cannot stop server listener status %d (%s)\n",
      status, err_str);
  }
#endif /* 0 */

  rpc_server_unregister_if(nasd_nfs_ifspec, NULL, &status);
  if (status) {
    dce_error_inq_text(status, err_str, &ps);
    printf("NFS: doubleplusungood, cannot unregister server_if status %d (%s)\n",
      status, err_str);
    NASD_PANIC();
  }
}

void
nasd_nfs_shutdown_rpc(arg)
  void  *arg;
{
#if DEBUG_RPC_SHUTDOWN > 0
  printf("nasd_nfs_shutdown_rpc(): nasd_nfs_threads_in_rpc=%ld\n",
    nasd_nfs_threads_in_rpc);
#endif /* DEBUG_RPC_SHUTDOWN > 0 */
  nasd_nfs_wait_shutdown_rpc();

#if DUMP_OP_STATS > 0
  DUMP_OPSTAT(null);
  DUMP_OPSTAT(mount);
  DUMP_OPSTAT(fsstat);
  DUMP_OPSTAT(fsinfo);
  DUMP_OPSTAT(lookup);
  DUMP_OPSTAT(readdir);
  DUMP_OPSTAT(readdirplus);
  DUMP_OPSTAT(access);
  DUMP_OPSTAT(setattr);
  DUMP_OPSTAT(create);
  DUMP_OPSTAT(remove);
  DUMP_OPSTAT(mkdir);
  DUMP_OPSTAT(rmdir);
  DUMP_OPSTAT(newcookie);
  DUMP_OPSTAT(rename);
  DUMP_OPSTAT(getstats);
#endif /* DUMP_OP_STATS > 0 */
#if NASD_NFS_DUMP_CACHE_STATS > 0
  NASD_NFS_CSDUMP(dir_lookup);
  NASD_NFS_CSDUMP(dir_force);
  NASD_NFS_CSDUMP(dirc_load_attr);
  NASD_NFS_CSDUMP(dirc_not_dir);
  NASD_NFS_CSDUMP(dirc_load_dir);
  NASD_NFS_CSDUMP(dirc_dirty);
  NASD_NFS_CSDUMP(dirc_new_dirty);
  NASD_NFS_CSDUMP(dirc_mkfree_loop);
  NASD_NFS_CSDUMP(dirc_mkfree_wait);
  NASD_NFS_CSDUMP(dir_load);
  NASD_NFS_CSDUMP(post_attribute);
  NASD_NFS_CSDUMP(post_attribute_valid);
  NASD_NFS_CSDUMP(post_attribute_invalid);
  NASD_NFS_CSDUMP(lookup_post_attribute);
  NASD_NFS_CSDUMP(dir_writer_fls);
  NASD_NFS_CSDUMP(dir_writer_pages);
  NASD_NFS_CSDUMP(dir_load_attr);
  NASD_NFS_CSDUMP(do_getattr);
  NASD_NFS_CSDUMP(do_setattr);
  NASD_NFS_CSDUMP(access_lookup_fail);
  NASD_NFS_CSDUMP(access_not_dir);
  NASD_NFS_CSDUMP(access_fault_attr);
  NASD_NFS_CSDUMP(access_get_attr);
  NASD_NFS_CSDUMP(lookup_get_attr);
  NASD_NFS_CSDUMP(setattr_get_attr);
#endif /* NASD_NFS_DUMP_CACHE_STATS > 0 */
}

nasd_status_t
nasd_nfs_rpc_init()
{
  nasd_status_t rc;

  nasd_nfs_rpc_qdepth = 0;

  bzero((char *)&nasd_nfs_opstats, sizeof(nasd_nfs_opstats));

  if (nasd_nfs_threads_in_rpc) {
    /* already/still running */
    return(NASD_FAIL);
  }

  nasd_nfs_wait_shutdown = 0;
  nasd_nfs_threads_in_rpc = 0;

  rc = nasd_cond_init(&nasd_nfs_shutdown_cond);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_cond,
    &nasd_nfs_shutdown_cond);
  if (rc) {
    nasd_cond_destroy(&nasd_nfs_shutdown_cond);
    return(rc);
  }

  rc = nasd_mutex_init(&nasd_nfs_rpc_cnt_lock);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_mutex,
    &nasd_nfs_rpc_cnt_lock);
  if (rc) {
    nasd_mutex_destroy(&nasd_nfs_rpc_cnt_lock);
    return(rc);
  }

  rc = nasd_mutex_init(&nasd_nfs_rpc_qdepth_lock);
  if (rc)
    return(rc);
  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_shutdown_mutex,
    &nasd_nfs_rpc_qdepth_lock);
  if (rc) {
    nasd_mutex_destroy(&nasd_nfs_rpc_qdepth_lock);
    return(rc);
  }

  bzero((char *)&nasd_nfs_lookup_failmem, sizeof(nasd_nfs_lookup_failmem));
  nasd_nfs_lookup_failmem.nasd_status = NASD_NO_MEM;

  return(NASD_SUCCESS);
}

nasd_status_t
nasd_nfs_startup_rpc()
{
  rpc_binding_vector_p_t binding_vector;
  rpc_if_handle_t nasd_nfs_ifspec;
  rpc_if_rep_p_t nasd_nfs_ifrep;
  dce_error_string_t err_str;
  error_status_t status;
  int ps;
  nasd_status_t rc;

  nasd_nfs_active = 1;

  nasd_nfs_ifrep = (rpc_if_rep_p_t)nasd_nfs_server_v0_1_s_ifspec;
  nasd_nfs_ifrep->manager_epv = (rpc_mgr_epv_t)&nasd_nfs_server_epv;
  nasd_nfs_ifspec = nasd_nfs_server_v0_1_s_ifspec;

  if (nasd_nfs_protseq_installed == 0) {
    rpc_server_use_protseq_if((unsigned_char_t *)"ncadg_ip_udp",
      rpc_c_protseq_max_reqs_default, nasd_nfs_ifspec, &status);
    if (status != rpc_s_ok) {
      dce_error_inq_text(status, err_str, &ps);
      if (ps)
        strcpy(err_str, "unknown");
      printf("NFS ERROR: cannot register protocol sequence 0x%x (%s)\n",
        (unsigned)status, err_str);
      return(NASD_FAIL);
    }
    nasd_nfs_protseq_installed = 1;
  }

  rpc_server_register_if(nasd_nfs_ifspec, NULL, NULL, &status);
  if (status != rpc_s_ok) {
    dce_error_inq_text(status, err_str, &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("NFS ERROR: cannot register interface 0x%x (%s)\n",
      (unsigned)status, err_str);
    return(NASD_FAIL);
  }

#if 0
  rpc_server_inq_bindings(&binding_vector, &status);
  if (status != rpc_s_ok) {
    dce_error_inq_text(status, err_str, &ps);
    if (ps)
      strcpy(err_str, "unknown");
    printf("NFS ERROR: cannot retrieve binding information 0x%x (%s)\n",
      (unsigned)status, err_str);
    return(NASD_FAIL);
  }
#endif

  rc = nasd_shutdown_proc(nasd_nfs_shutdown, nasd_nfs_shutdown_rpc, NULL);
  if (rc) {
    nasd_nfs_shutdown_rpc(NULL);
    return(NASD_FAIL);
  }

  return(NASD_SUCCESS);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
