/*
 * nasd_nfs_dumpdir.c
 *
 * Read a NASD object, parse it as a directory
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: Jim Zelenka
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_security.h>
#include <nasd/nasd_security_capability.h>
#include <nasd/nasd_security_time.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_nfs_internal.h>
#include <nasd/nasd_mem.h>

char *progname;

extern char *optarg;
extern int   optind;

int                           binding_type;
int                           binding_args_len;
void                         *binding_args;
char                         *binding_port = NASD_PDRIVE_PORT;
nasd_drive_param_dce_kpdev_t  kpdev_args;
int                           nondefault_binding = 0;

nasd_key_t part_key,master_key,drive_key,red_key,black_key;
nasd_error_string_t error_text;

nasd_status_t
nasd_security_get_red_key(
  nasd_partnum_t  partnum,
  nasd_key_t      key)
{
  bcopy(red_key,key,sizeof(nasd_key_t));
  return 0;
}

nasd_status_t
nasd_security_get_black_key(
  nasd_partnum_t  partnum,
  nasd_key_t      key)
{
  bcopy(black_key,key,sizeof(nasd_key_t));
  return 0;
}

char *
typestr(
  nasd_nfs_dirdata_t  *dd)
{
  static char str[128];

  switch(NASD_NFS_DIRD_TYPEOF(dd)) {
    case NASD_NFS_DIRD_TYPE_BOGUS:
      return("b");
    case NASD_NFS_DIRD_TYPE_DIR:
      return("d");
    case NASD_NFS_DIRD_TYPE_LINK:
      return("l");
    case NASD_NFS_DIRD_TYPE_DATA:
      return("n");
    default:
      sprintf(str, "0x%02x", NASD_NFS_DIRD_TYPEOF(dd)&0xff);
  }
  return(str);
}

void
usage()
{
  fprintf(stderr, "USAGE: %s [options] servername partnum nasdid master_password\n", progname);
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -k use kernel device\n");
  fprintf(stderr, "  -T use TCP\n");
  fflush(stderr);
  exit(1);
}

int
main(
  int     argc,
  char  **argv)
{
  int p, i, j, k, used, free, npages, mapbyte, mapent, l, slot;
  char *master_password, *server_name, name[1024], c;
  nasd_nfs3_attributes_t *nfs3attr;
  nasd_p_smpl_op_dr_args_t rd_args;
  nasd_p_fastread_dr_res_t rd_res;
  nasd_p_getattr_dr_args_t args;
  nasd_p_getattr_dr_res_t res;
  nasd_nfs_dirpage_t *pages;
  nasd_rpc_status_t status;
  nasd_attribute_t *attr;
  nasd_nfs_dirdata_t *dd;
  nasd_drive_handle_t h;
  nasd_cookie_t cookie;
  nasd_timespec_t tm;
  nasd_status_t rc;

  progname = argv[0];

  binding_type = NASD_BIND_DCE_DIRECT;
  binding_args = NULL;
  binding_args_len = 0;

  while ((c = getopt(argc, argv, "kT")) != EOF) {
    switch(c) {
      case 'k':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_KPDEV;
        binding_args = &kpdev_args;
        binding_args_len = sizeof(kpdev_args);
        strcpy(kpdev_args.devname, "/dev/nasdkp0");
        break;
      case 'T':
        if (nondefault_binding)
          usage();
        nondefault_binding = 1;
        binding_type = NASD_BIND_DCE_DIRECT_TCP;
        break;
      default:
        fprintf(stderr, "Unknown option '%c'\n", c);
        usage();
    }
  }

  if (optind >= argc)
    usage();
  server_name = argv[optind];
  optind++;

  bzero((char *)&args, sizeof(args));

  if (optind >= argc)
    usage();
  if (sscanf(argv[optind], "%d", &p) != 1)
    usage();
  optind++;
  args.in_partnum = p;

  if (optind >= argc)
    usage();
  rc = nasd_str_to_nasd_id(argv[optind], &args.in_identifier);
  optind++;
  if (rc)
    usage();

  if (optind >= argc)
    usage();
  master_password=argv[optind];
  optind++;

  if (optind < argc)
    usage();

  rc = nasd_cl_p_init();
  if (rc) {
    fprintf(stderr, "ERROR: cannot init client library, rc=0x%x (%s)\n",
      rc, nasd_error_string(rc));
    exit(1);
  }

  rc = nasd_bind_to_drive(server_name, binding_port,
    binding_type, binding_args, binding_args_len, &h);
  if (rc) {
    fprintf(stderr, "ERROR: cannot bind to server %s\n", server_name);
    fflush(stderr);
    exit(1);
  }

  /* Fill in capability and security param */
  nasd_security_password_to_keys(master_password,
    master_key, drive_key, args.in_partnum,
    part_key, red_key, black_key);
 
  args.in_security_param.type = NASD_BLACK_KEY;

  nasd_security_get_drive_time(0, &tm);
  tm.ts_sec += (60*60);
  nasd_security_build_capability(0, args.in_partnum,
    args.in_identifier,
    NASD_ACCESS_RIGHTS_GETATTR | NASD_ACCESS_RIGHTS_READ,
    0, tm.ts_sec, NASD_SECURE_OPS,NASD_BLACK_CAPABILITY,
    0, 0, 7, &cookie);
  bcopy((char *)&cookie.capability_args,
    (char *)&args.in_capability_args, sizeof(nasd_capability_args_t));
  args.in_security_param.type = args.in_capability_args.type; 

  args.in_security_param.partnum = args.in_partnum;
  args.in_security_param.actual_protection = NASD_SECURE_OPS;

  /*
   * Read NASD attributes
   */

  nasd_cl_p_getattr_dr(h, cookie.key, &args, &res, &status);

  if (res.nasd_status || status) {
    fprintf(stderr, "ERROR: nasd_status=0x%x (%s) status=0x%x (%s)\n",
      res.nasd_status, nasd_error_string(res.nasd_status),
      status, nasd_cl_error_string(h, status, error_text));
    exit(1);
  }

  attr = &res.out_attribute;
  nfs3attr = (nasd_nfs3_attributes_t *)attr->fs_specific;
  if (nfs3attr->type != 0x2) {
    fprintf(stderr, "Unexpected type 0x%02x\n", nfs3attr->type);
    exit(1);
  }

  printf("type 0x%02x mode 0x%x nlink %u uid %u gid %u\n",
    nfs3attr->type, nfs3attr->mode, nfs3attr->nlink,
    nfs3attr->uid, nfs3attr->gid);
  if (nfs3attr->name_debug[0]) {
    printf("debugging name: \"%s\"\n", nfs3attr->name_debug);
  }
  else {
    printf("no debugging name\n");
  }

  /*
   * Read in the contents
   */

  if (attr->object_len % NASD_NFS_DIRPAGE_SIZE) {
    fprintf(stderr, "WARNING: object len %" NASD_64u_FMT
      " not a multiple of %d\n", attr->object_len, NASD_NFS_DIRPAGE_SIZE);
  }
  npages = attr->object_len / NASD_NFS_DIRPAGE_SIZE;
  if (npages < 1) {
    fprintf(stderr, "ERROR: %d pages\n", npages);
    exit(1);
  }

  NASD_Malloc(pages, attr->object_len, (nasd_nfs_dirpage_t *));
  if (pages == NULL) {
    fprintf(stderr, "ERROR: could not allocate %" NASD_64u_FMT
      " bytes for directory\n", attr->object_len);
    exit(1);
  }

  rd_args.in_partnum = args.in_partnum;
  rd_args.in_identifier = args.in_identifier;
  rd_args.in_capability_args = cookie.capability_args;
  rd_args.in_security_param.type = cookie.capability_args.type;
  rd_args.in_security_param.partnum = args.in_partnum;
  rd_args.in_security_param.actual_protection
    = cookie.capability_args.min_protection;

  rd_args.in_offset = 0;
  while(rd_args.in_offset < attr->object_len) {
    rd_args.in_len = (nasd_len_t)attr->object_len
      - (nasd_len_t)rd_args.in_offset;

    nasd_cl_p_read_simple_dr(h, cookie.key, &rd_args, pages, &rd_res, &status);

    if (status || rd_res.nasd_status) {
      fprintf(stderr, "ERROR: nasd_status=0x%x (%s) status=0x%x (%s)\n",
        rd_res.nasd_status, nasd_error_string(rd_res.nasd_status),
        status, nasd_cl_error_string(h, status, error_text));
      exit(1);
    }

    rd_args.in_offset += (nasd_offset_t)rd_res.out_datalen;
  }

  /*
   * Dump contents
   */
  printf("Usage report:\n");
  used = 0;
  for(k=0;k<npages;k++) {
    for(i=0;i<sizeof(pages[k].header.usemap);i++) {
      for(j=0;j<8;j++) {
        if (pages[k].header.usemap[i]&(1<<j))
          used++;
      }
    }
    free = ((int)pages[k].header.freecnt)&0xff;
    if ((used + free) == 127) {
      printf("   page %2d  used %3d  free %3d\n", k, used,
        ((int)pages[k].header.freecnt)&0xff);
    }
    else {
      printf(" ! page %2d  used %3d  free %3d  (sum %d)\n", k, used,
        free, used+free);
    }
  }

  printf("Names:\n");
  mapbyte = 0;
  mapent = 0;
  for(k=0;k<npages;) {
    while((pages[k].header.usemap[mapbyte]&(1<<mapent)) == 0) {
      mapent++;
      if (mapent == 8) {
        mapent = 0;
        mapbyte++;
      }
      if (mapbyte == 16) {
        break;
      }
    }
    if (mapbyte == 16) {
      k++;
      mapbyte = 0;
      mapent = 0;
      continue;
    }
    slot = (mapbyte*8) + mapent;
    dd = &pages[k].data_slots[slot];
    l = 1 + NASD_NFS_DIRD_EXLEN(dd);
    if (dd->name_len >= 1024) {
      fprintf(stderr, "WARNING: slot %d name len %d\n",
        slot, dd->name_len);
    }
    bcopy(dd->name, name, dd->name_len);
    name[dd->name_len] = '\0';
    if (dd->name[dd->name_len] != '\0') {
      fprintf(stderr, "WARNING: slot %d name not null-terminated\n",
        slot);
    }

    printf(" page %2d  slot %3d  l %2d  type %s  %s\n", k, slot, l, typestr(dd), name);

    mapent += l;

    while (mapent >= 8) {
      mapent -= 8;
      mapbyte++;
    }
    if (mapbyte > 16) {
      fprintf(stderr, "WARNING: slot %d l=%d extends past end of page\n",
        slot, l);
      break;
    }
    if (mapbyte == 16) {
      if (mapent != 0) {
        fprintf(stderr, "WARNING: slot %d jumped to mapbyte %d mapent %d\n",
          slot, mapbyte, mapent);
        break;
      }
      k++;
      mapbyte = 0;
      mapent = 0;
    }
  }

  /*
   * Clean up and exit
   */

  rc = nasd_unbind_drive(&h);
  if (rc) {
    fprintf(stderr, "ERROR: got 0x%x (%s) unbinding drive\n",
      rc, nasd_error_string(rc));
    exit(1);
  }
  nasd_cl_p_shutdown();

  exit(0);
}

/* Local Variables:  */
/* indent-tabs-mode: nil */
/* tab-width: 2 */
/* End: */
