/*
 * marshall_tester.c
 */
/*
 * Copyright (c) 1997-1998,1999 Carnegie Mellon University.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon the
 * rights to redistribute these changes.
 */

/***********************************************************************
 *
 * marshall_tester.c
 *
 * Test our marshalling/unmarshalling code
 * Rewritten to check more carefully and return a meaningful value.
 *
 ***********************************************************************/

#include <nasd/nasd_options.h>
#include <nasd/nasd_types.h>
#include <nasd/nasd_error.h>
#include <nasd/nasd_nfs_types.h>
#include <nasd/nasd_drive_types.h>
#include <nasd/nasd_marshall.h>
#include <nasd/nasd_common.h>
#include <nasd/nasd_site.h>

#include <stdio.h>
#include <string.h>
#include <assert.h>

/*
 * A certain whiner compiler, which shall at the moment remain nameless,
 * generates warnings if the result of sizeof is not cast to int
 * when it is printed with %d. This compiler, commonly known as gcc,
 * shall remain nameless.
 */
#define SizeOf(_x_) ((int)sizeof(_x_))

int check_sizes(void);
int check_drive_info(void);
int check_part_info(void);
int check_trace_basic(void);
int check_part_creat(void);
int check_nasd_status(void);
int check_smpl_op(void);
int check_thrtl_op(void);
int check_getattr_args(void);
int check_getattr_res(void);
int check_setattr_args(void);
int check_setattr_res(void);
int check_create_args(void);
int check_create_res(void);
int check_flush_obj_args(void);
int check_flush_obj_res(void);
int check_eject_obj_args(void);
int check_eject_obj_res(void);
int check_initialize_args(void);
int check_initialize_res(void);
int check_change_key_args(void);
int check_change_key_res(void);
int check_trace_header(void);
int check_lookup_args(void);
int check_lookup_res(void);

int main(int argc,char* argv[]) {
  int fail = 0;
  
  if (check_sizes()) fail = 1;
  if (check_drive_info()) fail = 1;
  if (check_part_info()) fail = 1;
/*
 * XXX check:
 * drive_opstats
 * cache_stat
 * layout_stat
 * ioqueue_stat
 * io_stat
 */
  if (check_trace_basic()) fail = 1;
  if (check_part_creat()) fail = 1;
  if (check_nasd_status()) fail = 1;
  if (check_smpl_op()) fail = 1;
  if (check_thrtl_op()) fail = 1;
  if (check_getattr_args()) fail = 1;
  if (check_getattr_res()) fail = 1;
  if (check_setattr_args()) fail = 1;
  if (check_setattr_res()) fail = 1;
  if (check_create_args()) fail = 1;
  if (check_create_res()) fail = 1;
  if (check_flush_obj_args()) fail = 1;
  if (check_flush_obj_res()) fail = 1;
  if (check_eject_obj_args()) fail = 1;
  if (check_eject_obj_res()) fail = 1;
  if (check_initialize_args()) fail = 1;
  if (check_initialize_res()) fail = 1;
  if (check_change_key_args()) fail = 1;
  if (check_change_key_res()) fail = 1;
  if (check_trace_header()) fail = 1;
  if (check_lookup_args()) fail = 1;
  if (check_lookup_res()) fail = 1;

  if (fail) {
    printf("Marshalling tests failed.\n");
    return 1;
  } else {
    printf("Marshalling tests succeeded.\n");
    return 0;
  }
}

int check_sizes(void) {
  int fail = 0;

  printf("Struct sizes:\n");
  printf("  sizeof(nasd_attribute_t) = %d (%d)\n", SizeOf(nasd_attribute_t),
    NASD_ATTRIBUTE_SIZE);
  if (sizeof(nasd_attribute_t) != NASD_ATTRIBUTE_SIZE) {
    fail = 1;
    printf("problem: (size - nasd_attribute_t)\t %3d\n",
      SizeOf(nasd_attribute_t));
  }
  printf("  sizeof(nasd_cookie_t) = %d\n", SizeOf(nasd_cookie_t));
  printf("  sizeof(nasd_capability_args_t) = %d\n", SizeOf(nasd_capability_args_t));
  printf("  sizeof(nasd_security_param_t) = %d\n", SizeOf(nasd_security_param_t));
  printf("  sizeof(nasd_digest_nonce_t) = %d\n", SizeOf(nasd_digest_nonce_t));
  printf("  sizeof(nasd_nfs_identifier_t) = %d\n", SizeOf(nasd_nfs_identifier_t));
  printf("  sizeof(nasd_nfs_lookup_args_t) = %d\n", SizeOf(nasd_nfs_lookup_args_t));
  printf("  sizeof(nasd_nfs_lookup_res_t) = %d\n", SizeOf(nasd_nfs_lookup_res_t));
  printf("  sizeof(nasd_trace_header_t) = %d\n", SizeOf(nasd_trace_header_t));
  printf("\n");

  /* control sizes */
  if (sizeof(nasd_ctrl_drive_info_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_drive_info)\t %3d\n",
	   SizeOf(nasd_ctrl_drive_info_t));
  }

  if (sizeof(nasd_ctrl_part_info_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_part_info)\t %3d\n",
	   SizeOf(nasd_ctrl_part_info_t));
  }

  if (sizeof(nasd_ctrl_drive_opstats_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_drive_opstats)\t %3d\n",
	   SizeOf(nasd_ctrl_drive_opstats_t));
  }

  if (sizeof(nasd_ctrl_cache_stat_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_cache_stat)\t %3d\n",
	   SizeOf(nasd_ctrl_cache_stat_t));
  }

  if (sizeof(nasd_ctrl_layout_stat_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_layout_stat)\t %3d\n",
	   SizeOf(nasd_ctrl_layout_stat_t));
  }

  if (sizeof(nasd_ctrl_ioqueue_stat_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_ioqueue_stat)\t %3d\n",
	   SizeOf(nasd_ctrl_ioqueue_stat_t));
  }

  if (sizeof(nasd_ctrl_io_stat_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_io_stat)\t %3d\n",
	   SizeOf(nasd_ctrl_io_stat_t));
  }

  if (sizeof(nasd_ctrl_io_size_stat_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_io_size_stat)\t %3d\n",
	   SizeOf(nasd_ctrl_io_size_stat_t));
  }

  if (sizeof(nasd_ctrl_trace_basic_t) != NASD_INFO_PAGESIZE) {
    fail = 1;
    printf("problem: (size - ctrl_trace_basic)\t %3d\n",
	   SizeOf(nasd_ctrl_trace_basic_t));
  }

#if __alpha
  /* These checks are only relevant on the Alpha, because that is the
     only architecture where the otw structs will necessarily match
     the in-core structs. */

  /* sizes */
  
  if (sizeof(nasd_p_part_creat_dr_args_t) !=
      sizeof(nasd_p_part_creat_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_part_creat_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_part_creat_dr_args_t),
	   SizeOf(nasd_p_part_creat_dr_args_otw_t),
	   NASD_P_PART_CREAT_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_part_creat_dr_res_t) !=
      sizeof(nasd_p_part_creat_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_part_creat_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_part_creat_dr_res_t),
	   SizeOf(nasd_p_part_creat_dr_res_otw_t),
	   NASD_P_PART_CREAT_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_smpl_op_dr_args_t) !=
      sizeof(nasd_p_smpl_op_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_smpl_op_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_smpl_op_dr_args_t),
	   SizeOf(nasd_p_smpl_op_dr_args_otw_t),
	   NASD_P_SMPL_OP_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_thrtl_op_dr_args_t) !=
      sizeof(nasd_p_thrtl_op_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_thrtl_op_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_thrtl_op_dr_args_t),
	   SizeOf(nasd_p_thrtl_op_dr_args_otw_t),
	   NASD_P_THRTL_OP_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_setattr_dr_args_t) !=
      sizeof(nasd_p_setattr_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_setattr_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_setattr_dr_args_t),
	   SizeOf(nasd_p_setattr_dr_args_otw_t),
	   NASD_P_SETATTR_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_getattr_dr_args_t) !=
      sizeof(nasd_p_getattr_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_getattr_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_getattr_dr_args_t),
	   SizeOf(nasd_p_getattr_dr_args_otw_t),
	   NASD_P_GETATTR_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_getattr_dr_res_t) !=
      sizeof(nasd_p_getattr_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_getattr_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_getattr_dr_res_t),
	   SizeOf(nasd_p_getattr_dr_res_otw_t),
	   NASD_P_GETATTR_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_flush_obj_dr_args_t) !=
      sizeof(nasd_p_flush_obj_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_flush_obj_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_flush_obj_dr_args_t),
	   SizeOf(nasd_p_flush_obj_dr_args_otw_t),
	   NASD_P_FLUSH_OBJ_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_flush_obj_dr_res_t) !=
      sizeof(nasd_p_flush_obj_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_flush_obj_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_flush_obj_dr_res_t),
	   SizeOf(nasd_p_flush_obj_dr_res_otw_t),
	   NASD_P_FLUSH_OBJ_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_eject_obj_dr_args_t) !=
      sizeof(nasd_p_eject_obj_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_eject_obj_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_eject_obj_dr_args_t),
	   SizeOf(nasd_p_eject_obj_dr_args_otw_t),
	   NASD_P_EJECT_OBJ_DR_ARGS_MAX);
  }

  if (sizeof(nasd_p_eject_obj_dr_res_t) !=
      sizeof(nasd_p_eject_obj_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_eject_obj_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_eject_obj_dr_res_t),
	   SizeOf(nasd_p_eject_obj_dr_res_otw_t),
	   NASD_P_EJECT_OBJ_DR_RES_MAX);
  }

  if (sizeof(nasd_p_remove_dr_args_t) !=
      sizeof(nasd_p_remove_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_remove_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_remove_dr_args_t),
	   SizeOf(nasd_p_remove_dr_args_otw_t),
	   NASD_P_REMOVE_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_remove_dr_res_t) !=
      sizeof(nasd_p_remove_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_remove_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_remove_dr_res_t),
	   SizeOf(nasd_p_remove_dr_res_otw_t),
	   NASD_P_REMOVE_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_initialize_dr_args_t) != 
      sizeof(nasd_p_initialize_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_initialize_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_initialize_dr_args_t),
	   SizeOf(nasd_p_initialize_dr_args_otw_t),
	   NASD_P_INITIALIZE_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_initialize_dr_res_t) !=
      sizeof(nasd_p_initialize_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_initialize_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_initialize_dr_res_t),
	   SizeOf(nasd_p_initialize_dr_res_otw_t),
	   NASD_P_INITIALIZE_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_change_key_dr_args_t) !=
      sizeof(nasd_p_change_key_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_change_key_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_change_key_dr_args_t),
	   SizeOf(nasd_p_change_key_dr_args_otw_t),
	   NASD_P_CHANGE_KEY_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_change_key_dr_res_t) !=
      sizeof(nasd_p_change_key_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_change_key_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_change_key_dr_res_t),
	   SizeOf(nasd_p_change_key_dr_res_otw_t),
	   NASD_P_CHANGE_KEY_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_setattr_dr_args_t) !=
      sizeof(nasd_p_setattr_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_setattr_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_setattr_dr_args_t),
	   SizeOf(nasd_p_setattr_dr_args_otw_t),
	   NASD_P_SETATTR_DR_ARGS_MAX);
  }
  
  if (sizeof(nasd_p_setattr_dr_res_t) !=
      sizeof(nasd_p_setattr_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_setattr_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_setattr_dr_res_t),
	   SizeOf(nasd_p_setattr_dr_res_otw_t),
	   NASD_P_SETATTR_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_create_dr_res_t) !=
      sizeof(nasd_p_create_dr_res_otw_t)) {
    fail = 1;
    printf("problem: (size - p_create_res)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_create_dr_res_t),
	   SizeOf(nasd_p_create_dr_res_otw_t),
	   NASD_P_CREATE_DR_RES_MAX);
  }
  
  if (sizeof(nasd_p_create_dr_args_t) !=
      sizeof(nasd_p_create_dr_args_otw_t)) {
    fail = 1;
    printf("problem: (size - p_create_args)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_p_create_dr_args_t),
	   SizeOf(nasd_p_create_dr_args_otw_t),
	   NASD_P_CREATE_DR_ARGS_MAX);
  }

  if (sizeof(nasd_trace_header_t) !=
      sizeof(nasd_trace_header_otw_t)) {
    fail = 1;
    printf("(size - trace_header)\t %3d -> %3d (%3d)\n",
	   SizeOf(nasd_trace_header_t),
	   SizeOf(nasd_trace_header_otw_t),
	   NASD_TRACE_HEADER_SIZE);
  }

#endif /* __alpha */

  if (fail)
    printf("Size check failed.\n\n");

  return fail;
}

int check_drive_info(void) {
  nasd_ctrl_drive_info_t drive_info, drive_info_after;
  nasd_ctrl_drive_info_otw_t drive_info_otw;
  int fail = 0;

  /* ctrl_drive_info */
  drive_info.ctrl_id = 86;
  drive_info.max_parts = 87;
  drive_info.blocksize = 88;
  drive_info.num_parts = 89;
  drive_info.num_blocks = 90;
  drive_info.blocks_allocated = 90;
  
  nasd_drive_info_marshall_ctrl(&drive_info, drive_info_otw);
  nasd_drive_info_unmarshall_ctrl(drive_info_otw, &drive_info_after);

#define CHECKDRIVEINFO(x, y) if (drive_info.##x != drive_info_after.##x) { \
    fail = 1; \
    printf("problem: (%s) %" NASD_64u_FMT " -> %" NASD_64u_FMT "\n", y, \
	   drive_info.##x, drive_info_after.##x);}

  CHECKDRIVEINFO(ctrl_id, "ctrl_id");
  CHECKDRIVEINFO(max_parts, "max_parts");
  CHECKDRIVEINFO(blocksize, "blocksize");
  CHECKDRIVEINFO(num_parts, "num_parts");
  CHECKDRIVEINFO(num_blocks, "num_blocks");
  CHECKDRIVEINFO(blocks_allocated, "blocks_allocated");

#undef CHECKDRIVEINFO

  if (fail)
    printf("Drive info check failed.\n\n");

  return fail;
}

int check_part_info(void)  {
  nasd_ctrl_part_info_t part_info, part_info_after;
  nasd_ctrl_part_info_otw_t part_info_otw;
  int fail = 0;

  part_info.ctrl_id = 2;
  part_info.first_obj = nasd_int64cast(0x12345678abcdef0);
  part_info.num_obj = 3;
  part_info.part_size = 4;
  part_info.blocks_used = 5;
  part_info.blocks_allocated = 6;

  nasd_part_info_marshall_ctrl(&part_info, part_info_otw);
  nasd_part_info_unmarshall_ctrl(part_info_otw, &part_info_after);

#define CHECKPARTINFO(x, y) if (part_info.##x != part_info_after.##x) { \
    fail = 1; \
    printf("problem: (%s) 0x%" NASD_64x_FMT " -> 0x%" NASD_64x_FMT "\n", y, \
	   part_info.##x, part_info_after.##x);}

  CHECKPARTINFO(ctrl_id, "ctrl_id");
  CHECKPARTINFO(first_obj, "first_obj");
  CHECKPARTINFO(num_obj, "num_obj");
  CHECKPARTINFO(part_size, "part_size");
  CHECKPARTINFO(blocks_used, "blocks_used");
  CHECKPARTINFO(blocks_allocated, "blocks_allocated");

#undef CHECKPARTINFO

  if (fail)
    printf("Part info check failed.\n\n");

  return fail;
}


int check_trace_basic(void) {
  nasd_ctrl_trace_basic_t trace_basic, trace_basic_after;
  nasd_ctrl_trace_basic_otw_t trace_basic_otw;
  int fail = 0;

  trace_basic.ctrl_id = 4;
  trace_basic.enabled = 1;
  
  nasd_trace_basic_marshall_ctrl(&trace_basic, trace_basic_otw);
  nasd_trace_basic_unmarshall_ctrl(trace_basic_otw, &trace_basic_after);
  
  if (trace_basic.ctrl_id != trace_basic_after.ctrl_id) {
    fail = 1;
    printf("problem: (ctrl_id) 0x%" NASD_ID_FMT " -> 0x%" NASD_ID_FMT "\n",
	   trace_basic.ctrl_id, trace_basic_after.ctrl_id);
  }

  if (((int) trace_basic.enabled) !=
      ((int) trace_basic_after.enabled)) {
    fail = 1;
    printf("problem: (enabled) %" NASD_64u_FMT " -> %" NASD_64u_FMT "\n",
	   (int)trace_basic.enabled,
	   (int)trace_basic_after.enabled);
  }

  if (fail) {
    printf("Basic trace check failed.\n\n");
  }

  return fail;
}


int check_part_creat(void) {
  nasd_p_part_creat_dr_args_t p_part_creat_args,p_part_creat_args_after;
  nasd_p_part_creat_dr_args_otw_t p_part_creat_args_otw;
  nasd_key_t akey,bkey,ckey;
  int i, fail = 0;
  
  for(i = 0; i < sizeof(nasd_key_t); i++) {
    akey[i] = 2 * i;
    bkey[i] = 2 * i + 1;
    ckey[i] = 2 * i - 1;
  }
        
  p_part_creat_args.in_capability_args.ni = 100;
  p_part_creat_args.in_partnum = 5;
  p_part_creat_args.in_blkcnt = 1234567;
  p_part_creat_args.in_min_protection = 666;
    
  bcopy(akey,p_part_creat_args.in_partition_key,sizeof(nasd_key_t));
  bcopy(bkey,p_part_creat_args.in_red_key,sizeof(nasd_key_t));
  bcopy(ckey,p_part_creat_args.in_black_key,sizeof(nasd_key_t));
  
  nasd_p_part_creat_dr_marshall_args(&p_part_creat_args,
				     p_part_creat_args_otw);
  nasd_p_part_creat_dr_unmarshall_args(p_part_creat_args_otw,
				       &p_part_creat_args_after);
    
#define CHECKPARTCREAT(x,y,z) if (p_part_creat_args.##x != \
				p_part_creat_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   p_part_creat_args.##x, p_part_creat_args_after.##x);}
  
  CHECKPARTCREAT(in_capability_args.ni, "in_capability_args.ni", "0x%" NASD_ID_FMT);
  CHECKPARTCREAT(in_partnum, "in_partnum", "%d");
  CHECKPARTCREAT(in_blkcnt, "in_blkcnt", "%" NASD_64u_FMT);
  CHECKPARTCREAT(in_min_protection, "in_min_protection", "%u");

#undef CHECKPARTCREAT

  if ((((nasd_uint64 *) p_part_creat_args.in_partition_key)[0]) !=
      (((nasd_uint64 *) p_part_creat_args_after.in_partition_key)[0])) {
    fail = 1;
    printf("problem: (in_partition_key) %" NASD_64x_FMT " -> %" NASD_64x_FMT "\n",
	   ((nasd_uint64 *)p_part_creat_args.in_partition_key)[0],
	   ((nasd_uint64 *)p_part_creat_args_after.in_partition_key)[0]);
  }

  if ((((nasd_uint64 *) p_part_creat_args.in_red_key)[0]) !=
      (((nasd_uint64 *) p_part_creat_args_after.in_red_key)[0])) {
    fail = 1;
    printf("problem: (in_red_key) %" NASD_64x_FMT " -> %" NASD_64x_FMT "\n",
	   ((nasd_uint64 *)p_part_creat_args.in_red_key)[0],
	   ((nasd_uint64 *)p_part_creat_args_after.in_red_key)[0]);
  }

  if ((((nasd_uint64 *) p_part_creat_args.in_black_key)[0]) !=
      (((nasd_uint64 *) p_part_creat_args_after.in_black_key)[0])) {
    fail = 1;
    printf("problem: (in_black_key) %" NASD_64x_FMT " -> %" NASD_64x_FMT "\n",
	   ((nasd_uint64 *)p_part_creat_args.in_black_key)[0],
	   ((nasd_uint64 *)p_part_creat_args_after.in_black_key)[0]);
  }

  if (fail)
    printf("Part creat check failed.\n\n");

  return fail;
}

int check_nasd_status(void) {
  int fail = 0;
  nasd_p_part_creat_dr_res_t p_part_creat_res,p_part_creat_res_after;
  nasd_p_part_creat_dr_res_otw_t p_part_creat_res_otw;

  p_part_creat_res.nasd_status = 7;
    
  nasd_p_part_creat_dr_marshall_res(&p_part_creat_res,
				    p_part_creat_res_otw);
  nasd_p_part_creat_dr_unmarshall_res(p_part_creat_res_otw,
				      &p_part_creat_res_after);


  if (p_part_creat_res.nasd_status != p_part_creat_res_after.nasd_status) {
    fail = 1;
    printf("(nasd_status)\t %d -> %d\n",
	   p_part_creat_res.nasd_status,
	   p_part_creat_res_after.nasd_status);
    printf("NASD status check failed.\n\n");
  }

  return fail;
}

int check_smpl_op(void) {
  int fail = 0;
  nasd_p_smpl_op_dr_args_t p_smpl_op_args,p_smpl_op_args_after;
  nasd_p_smpl_op_dr_args_otw_t p_smpl_op_args_otw;

  p_smpl_op_args.in_capability_args.ni = 100;
  p_smpl_op_args.in_partnum = 5;
  p_smpl_op_args.in_identifier = 1234567;
  p_smpl_op_args.in_offset = 37;
  p_smpl_op_args.in_len = 362;
    
  nasd_p_smpl_op_dr_marshall_args(&p_smpl_op_args,p_smpl_op_args_otw);
  nasd_p_smpl_op_dr_unmarshall_args(p_smpl_op_args_otw,&p_smpl_op_args_after);
    
#define CHECKSMPLOPARGS(x,y,z) if (p_smpl_op_args.##x != \
				p_smpl_op_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   p_smpl_op_args.##x, p_smpl_op_args_after.##x);}

  CHECKSMPLOPARGS(in_capability_args.ni, "in_capability_args.ni", "0x%"NASD_ID_FMT);
  CHECKSMPLOPARGS(in_partnum, "in_partnum", "%d");
  CHECKSMPLOPARGS(in_identifier, "in_identifier", "0x%"NASD_ID_FMT);
  CHECKSMPLOPARGS(in_offset, "in_offset", "%"NASD_64u_FMT);
  CHECKSMPLOPARGS(in_len, "in_len", "%d");

#undef CHECKSMPLOPARGS
  
  if (fail)
    printf("smpl_op check failed.\n\n");

  return fail;
}  

int check_thrtl_op(void) {
  int fail = 0;
  nasd_p_thrtl_op_dr_args_t p_thrtl_op_args,p_thrtl_op_args_after;
  nasd_p_thrtl_op_dr_args_otw_t p_thrtl_op_args_otw;

  p_thrtl_op_args.in_capability_args.ni = 100;
  p_thrtl_op_args.in_partnum = 5;
  p_thrtl_op_args.in_identifier = 1234567;
  p_thrtl_op_args.in_offset = 37;
  p_thrtl_op_args.in_bms_targ = nasd_int64cast(991234567890);
  p_thrtl_op_args.in_len = 362;

  nasd_p_thrtl_op_dr_marshall_args(&p_thrtl_op_args,p_thrtl_op_args_otw);
  nasd_p_thrtl_op_dr_unmarshall_args(p_thrtl_op_args_otw,&p_thrtl_op_args_after);

#define CHECKTHRTLOPARGS(x,y,z) if (p_thrtl_op_args.##x != \
				p_thrtl_op_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   p_thrtl_op_args.##x, p_thrtl_op_args_after.##x);}

  CHECKTHRTLOPARGS(in_capability_args.ni, "in_capability_args.ni", "0x%"NASD_ID_FMT);
  CHECKTHRTLOPARGS(in_partnum, "in_partnum", "%d");
  CHECKTHRTLOPARGS(in_identifier, "in_identifier", "0x%"NASD_ID_FMT);
  CHECKTHRTLOPARGS(in_offset, "in_offset", "%"NASD_64u_FMT);
  CHECKTHRTLOPARGS(in_bms_targ, "in_bms_targ", "%"NASD_64u_FMT);
  CHECKTHRTLOPARGS(in_len, "in_len", "%d");

#undef CHECKTHRTLOPARGS

  if (fail)
    printf("thrtl_op check failed.\n\n");

  return fail;
}

int check_getattr_args(void) {
  int fail = 0;
  nasd_p_getattr_dr_args_t getattr_args,getattr_args_after;
  nasd_p_getattr_dr_args_otw_t getattr_args_otw;

  getattr_args.in_capability_args.ni = 100;
  getattr_args.in_identifier = 1001;
  getattr_args.in_partnum = 3;

  nasd_p_getattr_dr_marshall_args(&getattr_args,getattr_args_otw);
  nasd_p_getattr_dr_unmarshall_args(getattr_args_otw,&getattr_args_after);
    
#define CHECKGETATTRARGS(x,y,z) if (getattr_args.##x != \
			      getattr_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   getattr_args.##x, getattr_args_after.##x);}

  CHECKGETATTRARGS(in_capability_args.ni, "in_capability_args.ni", "0x%"NASD_ID_FMT);
  CHECKGETATTRARGS(in_identifier, "in_identifier", "0x%"NASD_ID_FMT);
  CHECKGETATTRARGS(in_partnum, "in_partnum", "%d");

#undef CHECKGETATTRARGS

  if (fail)
    printf("getattr_args check failed.\n\n");

  return fail;
}

int check_getattr_res(void) {
  int fail = 0;
  nasd_p_getattr_dr_res_t p_getattr_res,p_getattr_res_after;
  nasd_p_getattr_dr_res_otw_t p_getattr_res_otw;

  p_getattr_res.out_attribute.object_len = 123;
  p_getattr_res.nasd_status = 5;
    
  nasd_p_getattr_dr_marshall_res(&p_getattr_res,p_getattr_res_otw);
  nasd_p_getattr_dr_unmarshall_res(p_getattr_res_otw,&p_getattr_res_after);

#define CHECKGETATTRRES(x,y,z) if (p_getattr_res.##x != \
				 p_getattr_res_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   p_getattr_res.##x, p_getattr_res_after.##x);}

  CHECKGETATTRRES(out_attribute.object_len, "out_attribute.object_len", "%"NASD_64u_FMT);
  CHECKGETATTRRES(nasd_status, "nasd_status", "0x%x");

#undef CHECKGETATTRRES

  if (fail)
    printf("getattr_res check failed.\n\n");

  return fail;
}

int check_setattr_args(void) {
  int fail = 0;
  nasd_p_setattr_dr_args_t setattr_args,setattr_args_after;
  nasd_p_setattr_dr_args_otw_t setattr_args_otw;

  setattr_args.in_capability_args.ni = 100;
  setattr_args.in_identifier = 1001;
  setattr_args.in_partnum = 3;

  nasd_p_setattr_dr_marshall_args(&setattr_args,setattr_args_otw);
  nasd_p_setattr_dr_unmarshall_args(setattr_args_otw,&setattr_args_after);
    
#define CHECKSETATTRARGS(x,y,z) if (setattr_args.##x != \
				  setattr_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   setattr_args.##x, setattr_args_after.##x);}

  CHECKSETATTRARGS(in_capability_args.ni, "in_capability_args.ni", "0x%"NASD_ID_FMT);
  CHECKSETATTRARGS(in_identifier, "in_identifier", "0x%"NASD_ID_FMT);
  CHECKSETATTRARGS(in_partnum, "in_partnum", "%d");

#undef CHECKSETATTRARGS

  if (fail)
    printf("setattr_args check failed.\n\n");

  return fail;
}
 
int check_setattr_res(void) {
  int fail = 0;
  nasd_p_setattr_dr_res_t p_setattr_res,p_setattr_res_after;
  nasd_p_setattr_dr_res_otw_t p_setattr_res_otw;

  p_setattr_res.out_attribute.object_len = 123;
  p_setattr_res.nasd_status = 5;
    
  nasd_p_setattr_dr_marshall_res(&p_setattr_res,p_setattr_res_otw);
  nasd_p_setattr_dr_unmarshall_res(p_setattr_res_otw,&p_setattr_res_after);

#define CHECKSETATTRRES(x,y,z) if (p_setattr_res.##x != \
				  p_setattr_res_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z " \n", y, \
	   p_setattr_res.##x, p_setattr_res_after.##x);}

  CHECKSETATTRRES(out_attribute.object_len, "out_attribute.object_len", "%"NASD_64u_FMT);
  CHECKSETATTRRES(nasd_status, "nasd_status", "0x%x");

#undef CHECKSETATTRRES
  
  if (fail)
    printf("setattr_res check failed.\n\n");

  return fail;
}
 
int check_create_args(void) {
  int fail = 0;
  nasd_p_create_dr_args_t create_args,create_args_after;
  nasd_p_create_dr_args_otw_t create_args_otw;

  create_args.in_capability_args.ni = 100;
  create_args.in_partnum = 3;

  nasd_p_create_dr_marshall_args(&create_args,create_args_otw);
  nasd_p_create_dr_unmarshall_args(create_args_otw,&create_args_after);
    
#define CHECKCREATEARGS(x,y,z) if (create_args.##x != \
				  create_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   create_args.##x, create_args_after.##x);}

  CHECKCREATEARGS(in_capability_args.ni, "in_capability_args.ni", "0x%"NASD_ID_FMT);
  CHECKCREATEARGS(in_partnum, "in_partnum", "%d");

#undef CHECKCREATEARGS

    if (fail)
      printf("create_args check failed.\n\n");

    return fail;
}

int check_create_res(void) {
  int fail = 0;
  nasd_p_create_dr_res_t p_create_res,p_create_res_after;
  nasd_p_create_dr_res_otw_t p_create_res_otw;

  p_create_res.out_attribute.object_len = 123;
  p_create_res.out_identifier = 1001;
  p_create_res.nasd_status = 5;
    
  nasd_p_create_dr_marshall_res(&p_create_res,p_create_res_otw);
  nasd_p_create_dr_unmarshall_res(p_create_res_otw,&p_create_res_after);

#define CHECKCREATERES(x,y,z) if (p_create_res.##x != \
				p_create_res_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   p_create_res.##x, p_create_res_after.##x);}

  CHECKCREATERES(out_attribute.object_len, "out_attribute.object_len", "%"NASD_64u_FMT);
  CHECKCREATERES(out_identifier, "out_identifier", "0x%"NASD_ID_FMT);
  CHECKCREATERES(nasd_status, "nasd_status", "0x%x");

#undef CHECKCREATERES

  if (fail)
    printf("create_res check failed.\n\n");

  return fail;
}

int check_flush_obj_args(void) {
  int fail = 0;
  nasd_p_flush_obj_dr_args_t flush_obj_args,flush_obj_args_after;
  nasd_p_flush_obj_dr_args_otw_t flush_obj_args_otw;

  flush_obj_args.in_capability_args.ni = 100;
  flush_obj_args.in_identifier = 1001;
  flush_obj_args.in_partnum = 3;
  flush_obj_args.in_digest.nonce.ts_sec=22222;
  flush_obj_args.in_digest.nonce.ts_nsec=1111;
  nasd_p_flush_obj_dr_marshall_args(&flush_obj_args,flush_obj_args_otw);
  nasd_p_flush_obj_dr_unmarshall_args(flush_obj_args_otw,&flush_obj_args_after);
    
#define CHECKFLUSHOBJARGS(x,y,z) if (flush_obj_args.##x != \
				   flush_obj_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   flush_obj_args.##x, flush_obj_args_after.##x);}

  CHECKFLUSHOBJARGS(in_capability_args.ni, "in_capability_args.ni", "0x%"NASD_ID_FMT);
  CHECKFLUSHOBJARGS(in_identifier, "in_identifier", "0x%"NASD_ID_FMT);
  CHECKFLUSHOBJARGS(in_partnum, "in_partnum", "%d");
  CHECKFLUSHOBJARGS(in_digest.nonce.ts_sec, "in_digest.nonce.ts_sec", "%d");
  CHECKFLUSHOBJARGS(in_digest.nonce.ts_nsec, "in_digest.nonce.ts_nsec", "%d");

#undef CHECKFLUSHOBJARGS

  if (fail)
    printf("flush_obj_args check failed.\n\n");

  return fail;
}

int check_flush_obj_res(void) {
  int fail = 0;
  nasd_p_flush_obj_dr_res_t flush_obj_res,flush_obj_res_after;
  nasd_p_flush_obj_dr_res_otw_t flush_obj_res_otw;

  flush_obj_res.nasd_status = NASD_BAD_IDENTIFIER;

  nasd_p_flush_obj_dr_marshall_res(&flush_obj_res,flush_obj_res_otw);
  nasd_p_flush_obj_dr_unmarshall_res(flush_obj_res_otw,&flush_obj_res_after);

  if (flush_obj_res.nasd_status != flush_obj_res_after.nasd_status) {
    fail = 1;
    printf("(nasd_status) 0x%x (%s) -> 0x%x (%s)\n",
	   flush_obj_res.nasd_status,
	   nasd_error_string(flush_obj_res.nasd_status),
	   flush_obj_res_after.nasd_status,
	   nasd_error_string(flush_obj_res_after.nasd_status));

    printf("flush_obj_res check failed.\n\n");
  }

  return fail;
}

int check_eject_obj_args(void) {
  nasd_p_eject_obj_dr_args_t eject_obj_args,eject_obj_args_after;
  nasd_p_eject_obj_dr_args_otw_t eject_obj_args_otw;
  int fail = 0;

  eject_obj_args.in_capability_args.ni = 100;
  eject_obj_args.in_identifier = 1001;
  eject_obj_args.in_partnum = 3;
  eject_obj_args.in_digest.nonce.ts_sec=22222;
  eject_obj_args.in_digest.nonce.ts_nsec=1111;

  nasd_p_eject_obj_dr_marshall_args(&eject_obj_args,eject_obj_args_otw);
  nasd_p_eject_obj_dr_unmarshall_args(eject_obj_args_otw,&eject_obj_args_after);
    
#define CHECKEJECTOBJARGS(x,y,z) if (eject_obj_args.##x != \
				   eject_obj_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) " z " -> " z "\n", y, \
	   eject_obj_args.##x, eject_obj_args_after.##x);}

  CHECKEJECTOBJARGS(in_capability_args.ni, "in_capability_args.ni", "%"NASD_64s_FMT);
  CHECKEJECTOBJARGS(in_identifier, "in_identifier", "0x%"NASD_ID_FMT);
  CHECKEJECTOBJARGS(in_partnum, "in_partnum", "%d");
  CHECKEJECTOBJARGS(in_digest.nonce.ts_sec, "in_digest.nonce.ts_sec", "%d");
  CHECKEJECTOBJARGS(in_digest.nonce.ts_nsec, "in_digest.nonce.ts_nsec", "%d");

#undef CHECKEJECTOBJARGS

  if (fail) {
    printf("eject_obj_args checks failed.\n\n");
  }

  return fail;
}

int check_eject_obj_res(void) {
  nasd_p_eject_obj_dr_res_t eject_obj_res,eject_obj_res_after;
  nasd_p_eject_obj_dr_res_otw_t eject_obj_res_otw;

  eject_obj_res.nasd_status = NASD_BAD_IDENTIFIER;

  nasd_p_eject_obj_dr_marshall_res(&eject_obj_res,eject_obj_res_otw);
  nasd_p_eject_obj_dr_unmarshall_res(eject_obj_res_otw,&eject_obj_res_after);

  if (eject_obj_res.nasd_status != eject_obj_res_after.nasd_status) {
    printf("(nasd_status) 0x%x (%s) -> 0x%x (%s)\n",
	   eject_obj_res.nasd_status,
	   nasd_error_string(eject_obj_res.nasd_status),
	   eject_obj_res_after.nasd_status,
	   nasd_error_string(eject_obj_res_after.nasd_status));
    printf("eject_obj_res checks failed.\n\n");
    return 1;
  } else {
    return 0;
  }
}


int check_initialize_args(void) {
  int fail = 0;
  nasd_p_initialize_dr_args_t initialize_args,initialize_args_after;
  nasd_p_initialize_dr_args_otw_t initialize_args_otw;
  nasd_key_t mkey,dkey;
  int i;

  for(i=0;i<sizeof(nasd_key_t);i++) {
    mkey[i]=2*i;
    dkey[i]=2*i+1;
  }
        
  bcopy(mkey,initialize_args.in_master_key,sizeof(nasd_key_t));
  bcopy(dkey,initialize_args.in_drive_key,sizeof(nasd_key_t));

  nasd_p_initialize_dr_marshall_args(&initialize_args,initialize_args_otw);
  nasd_p_initialize_dr_unmarshall_args(initialize_args_otw,&initialize_args_after);
    
  if ((((nasd_uint64 *) initialize_args.in_master_key)[0]) !=
      (((nasd_uint64 *) initialize_args_after.in_master_key)[0])) {
    fail = 1;
    printf("(in_master_key) %8" NASD_64x_FMT " -> %8" NASD_64x_FMT "\n",
	   ((nasd_uint64 *)initialize_args.in_master_key)[0],
	   ((nasd_uint64 *)initialize_args_after.in_master_key)[0]);
  }

  if ((((nasd_uint64 *) initialize_args.in_drive_key)[0]) !=
      (((nasd_uint64 *) initialize_args_after.in_drive_key)[0])) {
    fail = 1;
    printf("(in_drive_key) %8" NASD_64x_FMT " -> %8" NASD_64x_FMT "\n",
	   ((nasd_uint64 *)initialize_args.in_drive_key)[0],
	   ((nasd_uint64 *)initialize_args_after.in_drive_key)[0]);
  }

  if (fail)
    printf("initialize_args check failed.\n\n");

  return fail;
}

int check_initialize_res(void) {
  int fail = 0;
  nasd_p_initialize_dr_res_t initialize_res,initialize_res_after;
  nasd_p_initialize_dr_res_otw_t initialize_res_otw;

  initialize_res.nasd_status = NASD_BAD_IDENTIFIER;

  nasd_p_initialize_dr_marshall_res(&initialize_res,initialize_res_otw);
  nasd_p_initialize_dr_unmarshall_res(initialize_res_otw,&initialize_res_after);
  

  if (initialize_res.nasd_status != initialize_res_after.nasd_status) {
    fail = 1;
    printf("(nasd_status) 0x%x (%s) -> 0x%x (%s)\n",
	   initialize_res.nasd_status,
	   nasd_error_string(initialize_res.nasd_status),
	   initialize_res_after.nasd_status,
	   nasd_error_string(initialize_res_after.nasd_status));
    printf("initialize_res check failed.\n\n");
  }

  return fail;
}

int check_change_key_args(void) {
  int fail = 0, i = 0;
  nasd_p_change_key_dr_args_t change_key_args,change_key_args_after;
  nasd_p_change_key_dr_args_otw_t change_key_args_otw;
  nasd_key_t tkey;

  change_key_args.in_capability_args.ni = 100;
  change_key_args.in_partnum = 3;

  for(i=0;i<sizeof(nasd_key_t);i++)
    tkey[i]=2*i;
       
  bcopy(tkey,change_key_args.in_new_key,sizeof(nasd_key_t));
    
  nasd_p_change_key_dr_marshall_args(&change_key_args,
				     change_key_args_otw);
  nasd_p_change_key_dr_unmarshall_args(change_key_args_otw,
				       &change_key_args_after);
    
  if (change_key_args.in_capability_args.ni !=
      change_key_args_after.in_capability_args.ni) {
    fail = 1;
    printf("(in_capability_args.ni)\t 0x%" NASD_ID_FMT " -> 0x%" NASD_ID_FMT "\n",
	   change_key_args.in_capability_args.ni,
	   change_key_args_after.in_capability_args.ni);
  }

  if (((nasd_uint64 *)(change_key_args.in_new_key))[0] !=
      ((nasd_uint64 *)(change_key_args_after.in_new_key))[0]) {
    fail = 1;
    printf("(in_new_key)\t %8" NASD_64x_FMT " -> %8" NASD_64x_FMT "\n",
	   ((nasd_uint64 *)(change_key_args.in_new_key))[0],
	   ((nasd_uint64 *)(change_key_args_after.in_new_key))[0]);
  }

  if (change_key_args.in_partnum !=
      change_key_args_after.in_partnum) {
    fail = 1;
    printf("(in_partnum)\t %d -> %d\n",
	   change_key_args.in_partnum,
	   change_key_args_after.in_partnum);
  }

  if (fail)
    printf("change_key_args check failed.\n\n");

  return fail;
}
 
int check_change_key_res(void) {
  int fail = 0;
  nasd_p_change_key_dr_res_t p_change_key_res,p_change_key_res_after;
  nasd_p_change_key_dr_res_otw_t p_change_key_res_otw;

  p_change_key_res.nasd_status = 5;
    
  nasd_p_change_key_dr_marshall_res(&p_change_key_res,
				    p_change_key_res_otw);
  nasd_p_change_key_dr_unmarshall_res(p_change_key_res_otw,
				      &p_change_key_res_after);

  if (p_change_key_res.nasd_status !=
      p_change_key_res_after.nasd_status) {
    fail = 1;
    printf("(nasd_status)\t %d -> %d\n",
	   p_change_key_res.nasd_status,
	   p_change_key_res_after.nasd_status);
    printf("change_key_res check failed.\n\n");
  }

  return fail;
}

int check_trace_header(void) {
  nasd_trace_header_t h1, h2;
  nasd_trace_header_otw_t hotw;
  int fail = 0;

  h1.tr_class = 11;
  h1.tr_type = 3;
  h1.tr_len = 529;
  h1.tr_seq = 0x1234;
  h1.tr_time.ts_sec = 37;
  h1.tr_time.ts_nsec = 3900;

  nasd_trace_header_marshall(&h1, hotw);
  nasd_trace_header_unmarshall(hotw, &h2);

  if (((int) h1.tr_class) != ((int) h2.tr_class)) {
    fail = 1;
    printf("problem: (tr_class)  0x%02x -> 0x%02x\n",
	   h1.tr_class&0xff,
	   h2.tr_class&0xff);
  }

  if (((int) h1.tr_type) != ((int) h2.tr_type)) {
    fail = 1;
    printf("problem: (tr_type)  0x%02x -> 0x%02x\n",
	   h1.tr_type&0xff,
	   h2.tr_type&0xff);
  }

  if (((unsigned short)h1.tr_len) !=
      ((unsigned short)h2.tr_len)) {
    fail = 1;
    printf("(tr_len)   %u -> %u\n",
	   (unsigned short)h1.tr_len,
	   (unsigned short)h2.tr_len);
  }

  if (((unsigned short)h1.tr_seq) !=
      ((unsigned short)h2.tr_seq)) {
    fail = 1;
    printf("(tr_seq)   %u -> %u\n",
	   (unsigned short)h1.tr_seq,
	   (unsigned short)h2.tr_seq);
  }

  if ((h1.tr_time.ts_sec != h2.tr_time.ts_sec) ||
      (h1.tr_time.ts_nsec != h2.tr_time.ts_nsec)) {
    fail = 1;
    printf("(tr_time)  %d:%09d -> %d:%09d\n",
	   h1.tr_time.ts_sec, h1.tr_time.ts_nsec,
	   h2.tr_time.ts_sec, h2.tr_time.ts_nsec);
  }

  if (fail) {
    printf("Trace header checks failed.\n\n");
  }

  return fail;
}

int check_lookup_args(void) {
  int fail = 0;
  nasd_nfs_lookup_args_t lookup_args, lookup_args_after;
  nasd_nfs_lookup_args_otw_t lookup_args_otw;

  lookup_args.in_identifier.nasd_identifier = 1001;
  lookup_args.in_identifier.partnum = 1234;
  lookup_args.in_identifier.disk_identifier = 64;
  lookup_args.in_dirpath[0] = 64;
  lookup_args.in_credential.uid = 12345;
  lookup_args.in_credential.gid = 98765;

  nasd_nfs_lookup_marshall_args(&lookup_args, lookup_args_otw);
  nasd_nfs_lookup_unmarshall_args(lookup_args_otw, &lookup_args_after);

#define CHECKLOOKUPARGS(x,y,z) if (lookup_args.##x != \
				   lookup_args_after.##x) { \
    fail = 1; \
    printf("problem: (%s) %" z " -> %" z "\n", y, \
	   lookup_args.##x, lookup_args_after.##x);}

  CHECKLOOKUPARGS(in_identifier.nasd_identifier,
		  "in_identifier.nasd_identifier", NASD_ID_FMT);
  CHECKLOOKUPARGS(in_identifier.partnum,
		  "in_identifier.partnum", "d");
  CHECKLOOKUPARGS(in_identifier.disk_identifier,
		  "in_identifier.disk_identifier", "d");
  CHECKLOOKUPARGS(in_credential.uid, "in_credential.uid", "d");
  CHECKLOOKUPARGS(in_credential.gid, "in_credential.gid", "d");
  CHECKLOOKUPARGS(in_dirpath[0], "in_dirpath[0]", "d");

#undef CHECKLOOKUPARGS

  if (fail) {
    printf("lookup_args checks failed.\n\n");
  }

  return fail;
}

int check_lookup_res(void) {
  int fail = 0;
  nasd_nfs_lookup_res_t lookup_res, lookup_res_after;
  nasd_nfs_lookup_res_otw_t lookup_res_otw;

  lookup_res.nasd_status = 123456;
  lookup_res.out_identifier.nasd_identifier = 1001;
  lookup_res.out_identifier.partnum = 1234;
  lookup_res.out_identifier.disk_identifier = 64;
  lookup_res.out_attribute.attribute.block_size = 98765;
  lookup_res.out_attribute.attribute.object_len = nasd_int64cast(9876543210);

  nasd_nfs_lookup_marshall_res(&lookup_res, lookup_res_otw);
  nasd_nfs_lookup_unmarshall_res(lookup_res_otw, &lookup_res_after);

#define CHECKLOOKUPRES(x,y,z) if (lookup_res.##x != \
				   lookup_res_after.##x) { \
    fail = 1; \
    printf("problem: (%s) %" z " -> %" z "\n", y, \
	   lookup_res.##x, lookup_res_after.##x);}

  CHECKLOOKUPRES(nasd_status, "nasd_status", "d");
  CHECKLOOKUPRES(out_identifier.nasd_identifier,
		 "out_identifier.nasd_identifier", NASD_ID_FMT);
  CHECKLOOKUPRES(out_identifier.partnum,
		 "out_identifier.partnum", "d");
  CHECKLOOKUPRES(out_identifier.disk_identifier,
		 "out_identifier.disk_identifier", "d");
  CHECKLOOKUPRES(out_attribute.attribute.block_size,
		 "out_attribute.attribute.block_size", "d");
  CHECKLOOKUPRES(out_attribute.attribute.object_len,
		 "out_attribute.attribute.object_len", NASD_64s_FMT);

#undef CHECKLOOKUPRES

  if (fail) {
    printf("lookup_res checks failed.\n\n");
  }

  return fail;
}
