/*
 *  annot.c -- annotations and machine definitions.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

#define MAXSEGS 64	/* maximum segments in an Order struct */
static Order *ostage;	/* construction area for Order struct */

static int omerge(Oseg *p, Oseg g);	/* merge order segments */
static void porder(Order *o);		/* print ordering on stderr */



/*  mkannot(tsize, msize, align, order) -- make annotation structure  */

Annot mkannot(int tsize, int msize, int align, Order *order)
{
    Annot a;
    Oseg g;
    int i, n, end, err;

    err = 0;			/* set nonzero by msize/tsize error */

    if (msize <= 0) {
	fatal("msize must be greater than zero", NULL);
	tsize = msize = 4;	/* just so we can go on... */
	err = 1;
    }
    if (msize < tsize) {
	fatal("tsize exceeds msize", NULL);
	msize = tsize;
	err = 1;
    }
    a.a_tsize = tsize;
    a.a_msize = msize;
    a.a_align = align;
    a.a_order = order;
    a.a_bitorder = NULL;

    if (order == NULL || err != 0)
	return a;

    /* apply various sanity checks to the order specification */
    n = 0;
    for (i = 0; i < order->o_nsegs; i++) {
	g = order->o_oseg[i];
	n += g.o_length;
	end = g.o_start + g.o_dir * g.o_length;
	if ((int) g.o_start >= msize || end > msize) {
	    fatal("order reference is beyond stated msize", NULL);
	    return a;
	}
    }
    if (n != tsize)
	fatal("order spec is inconsistent with stated tsize", NULL);

    return a;
}



/*
 *  setannot(Type *t, annot) -- set the annotation of a type.
 *
 *  Returns zero if an error message should be generated.
 */
int setannot(Type *t, Annot a)
{
    int isstruct;

    if (t->t_tcode == T_FUNC)			/* if function decl */
	t = t->t_utype;				/* use return type */

    isstruct = (t->t_tcode == T_STRUCT) || (t->t_tcode == T_UNION);

    if (a.a_msize != 0) {			/* if explicitly annotated */
	if (a.a_order == NULL) {
	    if (!isstruct) {
		fatal("order specification missing", NULL);
		a.a_order = oclone(ostart(ointerval(0, a.a_tsize)));
	    }
	} else {
	    if (isstruct) {
		fatal("order must be 0 for struct/union", NULL);
		a.a_order = NULL;
	    }
	}
	if ((t->t_tflags & T_INT) && (a.a_msize > MAXINTBYTES)) {
	    fatal("size is too large for an integral type", NULL);
	    a.a_msize = a.a_tsize = MAXINTBYTES;
	}
	t->t_annot = a;
	return 1;
    }

    if (t->t_annot.a_msize != 0)
	return 1;				/* named type was annotated */

    /* pointer/array need no annot; void is illegal and gets a different err */
    if (t->t_tcode == T_PTR || t->t_tcode == T_ARRAY || t->t_tcode == T_VOID)
	return 1;

    /* no annotation available; invent one and return error signal */
    a.a_tsize = a.a_msize = a.a_align = 4;		
    if (isstruct) 
	a.a_order = NULL;
    else
	a.a_order = oclone(ostart(ointerval(0, 3)));
    t->t_annot = a;
    return 0;
}



/*  okalign(a) -- check that the alignment in an annotation is a power of 2.  */

int okalign(Annot a)
{
    if (a.a_msize == 0)
	return 1;			/* ok: unannotated */
    if (a.a_align <= 0)
	return 0;			/* zero or negative alignment */
    if (a.a_align & (a.a_align - 1))
	return 0;			/* not a power of 2 */
    return 1;
}



/*  ointerval(start, end) -- make Oseg structure  */

Oseg ointerval(int start, int end)
{
    Oseg o;

    o.o_start = start;
    if (start < end) {
	o.o_length = end - start + 1;
	o.o_dir = 1;
    } else if (start == end) {
	o.o_length = 1;
	o.o_dir = 0;
    } else /* start > end */ {
	o.o_length = start - end + 1;
	o.o_dir = -1;
    }
    return o;
}



/*
 *  ostart(g) -- start new Order struct beginning with segment g
 *
 *  Order specifications are built in a maximum-sized Order struct,
 *  named "ostage", that is later reused for subsequent specifications.  
 *  When the specification has been fully parsed, it is copied into
 *  a struct no bigger than necessary.  The sequence is
 *	ostart() -- begin new spec
 *	oappend() -- add additional parts (called zero or more times)
 *	oclone() -- copy final structure
 */
Order *ostart(Oseg g)
{
    if (!ostage)
	ostage = zalloc(sizeof(Order) + MAXSEGS * sizeof(Oseg));
    ostage->o_nsegs = 1;
    ostage->o_oseg[0] = g;
    return ostage;
}



/*
 *  oappend(g) -- append segment g to Order segment under construction
 *
 *  If possible, segment g is merged with the previous segment.
 *  If not, it is added at the end.
*/
Order *oappend(Oseg g)
{
    if (ostage->o_nsegs >= MAXSEGS)
	fatal("order specification is too complex to handle", NULL);
    else if (!omerge(&ostage->o_oseg[ostage->o_nsegs - 1], g))
	ostage->o_oseg[ostage->o_nsegs++] = g;	/* could not merge; append */
    return ostage;
}



/*  omerge(p, g) -- merge Oseg g into Oseg *p, returning success or failure  */

static int omerge(Oseg *p, Oseg g)
{
    /* Check for adjacency of two isolated bytes (or bits) */
    if (p->o_dir == 0 && g.o_dir == 0) {
	if (abs(g.o_start - p->o_start) == 1) {
	    p->o_dir = g.o_start - p->o_start;
	    p->o_length = 2;
	    return 1;			/* adjacent */
	} else
	    return 0;			/* not adjacent */
    }

    /* At least one of the segments is a run of 2 or more.  Check directions. */
    if (p->o_dir * g.o_dir == -1)
	return 0;			/* opposite directions */

    /* Handle a run of 2 or more followed by either a run or an isolated byte */
    if (p->o_dir != 0) {
	if (g.o_start == (p->o_start + p->o_dir * p->o_length)) {
	    p->o_length += g.o_length;
	    return 1;			/* adjacent segments combined */
	} else
	    return 0;			/* not adjacent */
    }

    /* *p is an isolated bit or byte; g is a run of 2 or more. */
    if (p->o_start == (g.o_start - g.o_dir)) {
	p->o_dir = g.o_dir;
	p->o_length = 1 + g.o_length;
	return 1;
    } else
	return 0;			/* not adjacent */
}



/*  oclone(o) -- clone an Order struct  */

Order *oclone(Order *o)
{
    int n;

    n = sizeof(Order) + (o->o_nsegs - 1) * sizeof(Oseg);
    /* the casts below satisfy certain misconfigured C compilers */
    return (Order *) memcpy(zalloc(n), (void *) o, n);
}



/*  bitorder(sym, n, o) -- set bitfield ordering of struct member  */

void bitorder(Symptr sym, int n, Order *o)
{
    int i, t;

    sym->s_type->t_annot.a_bitorder = o;
    t = 0;
    for (i = 0; i < o->o_nsegs; i++)
	t += o->o_oseg[i].o_length;
    if (n != t)
	fatal("bit-field order disagrees with stated size", NULL);
}



/*  boff(order, n) -- return offset of nth least significant bit or byte  */

int boff(Order *o, int offset)
{
    Oseg *g;

    g = o->o_oseg;
    while (offset >= (int) g->o_length) {
	offset -= g->o_length;
	g++;
    }
    return g->o_start + g->o_dir * offset;
}



/*  pannot(t) -- print annotation of type t on stderr  */

void pannot(Type *t)
{
    if (t != NULL && t->t_annot.a_tsize != 0) {
	fprintf(stderr, " (%d/%d,%d,", t->t_annot.a_tsize,
	    t->t_annot.a_msize, t->t_annot.a_align);
	porder(t->t_annot.a_order);
	if (t->t_annot.a_bitorder) {
	    fprintf(stderr, ":");
	    porder(t->t_annot.a_bitorder);
	}
	fprintf(stderr, ")");
    }
}



/*  porder(o) -- print ordering fields on stderr  */

static void porder(Order *o)
{
    Oseg g;
    int i;

    if (o == NULL)
	fprintf(stderr, "0");
    else {
	for (i = 0; i < o->o_nsegs; i++) {
	    g = o->o_oseg[i];
	    fprintf(stderr, "%c%d", i == 0 ? '<' : ',', g.o_start);
	    if (g.o_dir != 0)
		fprintf(stderr, "..%d", g.o_start + g.o_dir * (g.o_length - 1));
	}
	fprintf(stderr, ">");
    }
}



/*  mattrib(md, asg) -- update mdef with attribute value  */

Mdef *mattrib(Mdef *old, Symbol *asg)
{
    char *kwd;
    int *p, val;

    if (!old) {				/* if mdef struct not yet created */
	old = zalloc(sizeof(Mdef));	/* make one */
	old->m_read = -1;		/* indicate values uninitialized */
	old->m_write = -1;
    }

    kwd = asg->s_name;
    val = asg->s_val;

    if (strcmp(kwd, "read") == 0)
	p = &old->m_read;
    else if (strcmp(kwd, "write") == 0)
	p = &old->m_write;
    else {
	fatal("unrecognized mdef field", kwd);
	return old;
    }
    if (val < 0 || val > 1) {
	fatal("illegal value for mdef field", asg->s_name);
	return old;
    }
    *p = val;
    return old;
}
