/*
 *  fgrain.c -- fine-grained code generator.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

/*
 *  Chunk: groups bytes for loads and stores; see mkchunk().
 */
typedef struct {
    unsigned int c_size : 15;	/* size of current chunk */
    unsigned int c_done : 1;	/* has chunk already been loaded or stored? */
} Chunk;

/*
 *  Dbits: structure specifying the source of some destination register bits.
 */
typedef struct {
    short r_sxtend;		/* zero if copy; nonzero if sign-extend */
    short r_shift;		/* source shift */
    int r_schunk;		/* source chunk offset */
#if __alpha
    unsigned long r_mask;	/* mask of bits used */
#else /* __alpha */
    unsigned long long r_mask;	/* mask of bits used */
#endif /* __alpha */
} Dbits;

static Chunk *mkchunks(Frame *f, int o, int a, char *s); /* make chunks*/
static Type *regtype(int n);				/* type for reg size */
static int ccount(Chunk *clist, int width);		/* count subchunks */
static void gencalcs(Frame *flist, Chunk *clist);	/* gen calculations */
static void genchunk(Frame *f, int offset, int size);	/* gen one chunk */
static void setbits(Frame *f, int offset, int size);	/* gen set-bit bitop */
static void setbytes(Frame *f, int offset, int size);	/* gen set-byte bitops*/
static int bshift(int w, int n);			/* find byte in chunk */
static void dbop(int s, int sh, int x, int n, int ds);	/* add bitop to list */
static int bopcmp(const void *p1, const void *p2);	/* bitop ordering */
static int bopmerge(Dbits *db1, Dbits *db2);		/* merge bitops */

static Chunk *schunks, *dchunks;	/* source & destination chunk lists */

static Dbits bitops[MAXINTBITS];	/* array of bit operations */
static int nbops;			/* number of operations in array */



/*  gfgrain(flist, lalign, ralign) -- generate fine-grained code  */

void gfgrain(Frame *flist, int lalign, int ralign)
{
    out("    {\n");			/* new C scope for variables */

    /* the framelist is already sorted by source address */
    /* generate list of source chunks, and declare variables as snnn */
    schunks = mkchunks(flist, offsetof(Frame, f_srcoff), ralign, "s");

    /* sort frames in destination order */
    flist = sortframes(flist,
	offsetof(Frame, f_dstoff), offsetof(Frame, f_dstshift));

    /* generate list of destination chunks, and declare variables as dnnn */
    dchunks = mkchunks(flist, offsetof(Frame, f_dstoff), lalign, "d");

    /* generate code to calculate and store destination chunks */
    gencalcs(flist, dchunks);

    free(schunks);
    free(dchunks);
    out("    }\n");
}



/*
 *  mkchunks(flist, offset, align, prefix) -- make chunk list and variables.
 *
 *  flist	the frame list, sorted by source or destination as appropriate
 *  off		offsetof (f_srcoff or f_dstoff) in a Frame struct
 *  align	the alignment guarantee available
 *  prefix	variable prefix ("s" for source or "d" for destination)
 *
 *  mkchunks() constructs a chunk list and generates declarations for the
 *  corresponding variables.
 *
 *  A chunklist is an array of Chunk structs, each corresponding to one byte
 *  of the source or destination.  If that byte of the source/destination
 *  is unused, the chunk size is zero.  If the byte *is* used, the chunk
 *  size contains the size of the corresponding chunk.  (For example, the
 *  size is 2 if a byte is to be accessed as part of a 2-byte short int.)
 *  The address of the appropriate chunk, of course, can be computed from
 *  the offset and the chunk size.
 */
static Chunk *mkchunks(Frame *flist, int off, int align, char *prefix)
{
    Frame *f;
    Type *t;
    Chunk *clist;
    int a, i, n, w, o, maxoff, nchunks, nvar;

    /* find maximum offset used */
    maxoff = 0;
    for (f = flist; f != NULL; f = f->f_next) {
	w = f->f_nbytes;
	if (w == 0)			/* if bit move */
	    w = 1;
	n = IntAt(f, off) + w;
	if (n > maxoff)
	    maxoff = n;
    }

    nchunks = maxoff + MAXINTBYTES - 1;	/* round up to largest full chunk */
    nchunks -= nchunks % MAXINTBYTES;

    /* allocate chunk list */
    clist = zalloc(nchunks * sizeof(Chunk));

    /* start by marking each byte needed as a 1-byte chunk */
    for (f = flist; f != NULL; f = f->f_next) {
	n = IntAt(f, off);
	for (i = 0; i < f->f_nbytes; i++)
	    clist[n + i].c_size = 1;
	clist[n].c_size = 1;		/* in case of bit op (f_nbytes==0) */
	}

    /* coalesce into larger chunks depending on register sizes and alignment */
    for (n = 2; n <= MAXINTBYTES; n *= 2)
	if ((t = regtype(n)) != NULL) {
	    /* there exists a native type of size n */
	    a = t->t_annot.a_align;
	    if (a == 0)
		a = 1;
	    if (align % a != 0)
		continue;			/* alignment not satisfied */	
	    for (o = 0; o < nchunks; o += n) {
		if (o + n > maxoff && *prefix == 'd')
		    break;			/* can't store beyond end */
		if (ccount(clist + o, n) > 1) {
		    /* the chunk encompasses two or more smaller chunks */
		    for (i = 0; i < n; i++)
			clist[o + i].c_size = n;    /* mark chunk members */
		}
	    }
	}

    /* declare variables of each size */
    for (n = 1; n <= MAXINTBYTES; n *= 2)
	if (regtype(n)) {
	    nvar = 0;
	    for (o = 0; o < nchunks; o += n)
		for (i = 0; i < n; i++)
		    if (clist[o + i].c_size == n) {
			if (nvar % 10 == 0)
			    out("\t%s %s%d", native[n]->t_name, prefix, o);
			else
			    out(", %s%d", prefix, o);
			if (++nvar % 10 == 0)
			    out(";\n");
			break;
		    }
	    if (nvar % 10 != 0)
		out(";\n");
	}

    return clist;
}



/*  regtype(n) -- find native type for register size of n bytes  */

static Type *regtype(int n)
{
    if (n == 1)
	return chartype;
    if (inttype->t_annot.a_tsize == n)	/* prefer "int", so try it first */
	return inttype;
    if (shorttype->t_annot.a_tsize == n)
	return shorttype;
    if (longtype->t_annot.a_tsize == n)
	return longtype;
/*
 * Don't return hypertype here on non-alpha because it's
 * not a native type- hurts the compiler instead of helping
 * optimization-wise
 */
    return NULL;
}



/*  ccount(clist, width) -- count smaller chunks used inside chunk  */

static int ccount(Chunk *clist, int width)
{
    unsigned int i, n;

    n = 0;
    i = 0;
    while (i < width) {
	if (clist[i].c_size != 0) {		/* if used by subchunk */
	    n += 1;				/* count subchunk */
	    i += clist[i].c_size - (i % clist[i].c_size);
						/* skip remainder of subchunk */
	} else
	    i++;
    }
    return n;
}



/*
 *  gencalcs(flist, clist) -- generate calculation operations.
 */
static void gencalcs(Frame *flist, Chunk *clist)
{
    Frame *f;
    int i, d, w, nb, o, prevd;

    prevd = -1;
    for (f = flist; f != NULL; f = f->f_next) {
	nb = f->f_nbytes;		/* number of bytes */
	if (nb == 0)			/* if bit move: affects one byte */
	    nb = 1;
	for (i = 0; i < nb; i++) {	/* for each byte of target */
	    o = f->f_dstoff + i;	/* destination offset */
	    w = clist[o].c_size;	/* size of destination chunk */
	    d = o - o % w;		/* chunk offset */
	    if (d != prevd) {		/* if different from previous chunk*/
		genchunk(f, d, w);
		prevd = d;
	    }
	}
    }
}



/*
 *  genchunk(f, offset, size) -- generate code to calculate one output chunk.
 *
 *  f is the first of one or more frames that affect the chunk.
 */
static void genchunk(Frame *f, int offset, int size)
{
    Dbits *db;
    Chunk *c;
    int i, j;
    char *copr, mask[2 * MAXINTBYTES + 5];

    /* generate a list of bit operations */
    if (dbflags[DB_BITOPS])
	fprintf(stderr, "bit operations for d%d:\n", offset);
    nbops = 0;				/* init array to empty */

    for (; f != NULL; f = f->f_next) {

	if (f->f_dstoff >= offset + size)
	    break;			/* if beyond range of this chunk */
	if (f->f_nbytes > 0 && f->f_dstoff + f->f_nbytes <= offset)
	    continue;			/* strange overlap situation */
	if (f->f_flags == MV_ZERO)
	    continue;			/* nothing to do for zero fill */

	if (f->f_nbytes == 0)
	    setbits(f, offset, size);	/* set bits */
	else
	    setbytes(f, offset, size);	/* set full bytes */
    }

    /* sort the list by sxtend flag, source index, and shift */
    qsort(bitops, nbops, sizeof(Dbits), bopcmp);

    /* load any necessary source chunks that haven't been loaded earlier */
    for (i = 0; i < nbops; i++) {
	j = bitops[i].r_schunk;
	c = &schunks[j];
	if (!c->c_done) {
	    out("\ts%d = *(%s*)(src+%d);\n", j, native[c->c_size]->t_name, j);
	    c->c_done = 1;
	}
    }

    /* output the operations */
    out("\td%d = ", offset);
    copr = "";
    for (i = 0; i < nbops; i++) {
	db = &bitops[i];
	if (i < nbops - 1 && bopmerge(db, db + 1))
	    continue;			/* if merge successful */

#if __alpha
	if ((db->r_mask == (unsigned long) -1) ||
	    (size < sizeof(long) && db->r_mask == (1L << BYTE * size) - 1))
	    sprintf(mask, "");		/* using full chunk; no mask needed */
	else {
	    sprintf(mask, "&0x%lXu", db->r_mask);
	}
#else /* __alpha */
	if ((db->r_mask == (unsigned long long) -1) ||
	    (size < sizeof(long long) &&
	     db->r_mask == (1LL << BYTE * size) - 1))
	    sprintf(mask, "");		/* using full chunk; no mask needed */
	else {
	    sprintf(mask, "&0x%LXULL", db->r_mask);
	}
#endif /* __alpha */

	if (db->r_sxtend)
	    out("%s-(s%d>>%d&1)%s", copr, db->r_schunk, db->r_shift, mask);
	else if (db->r_shift == 0)
	    out("%ss%d%s", copr, db->r_schunk, mask);
	else if (db->r_shift > 0)
	    out("%ss%d>>%d%s", copr, db->r_schunk, db->r_shift, mask);
#if __alpha
	else if (schunks[db->r_schunk].c_size < (unsigned) size
		 && db->r_mask > (unsigned) -1) {
	  out("%s(long)s%d<<%d%s", copr, db->r_schunk, -db->r_shift, mask);
	}
#else /* __alpha */
	else if (schunks[db->r_schunk].c_size < (unsigned) size
		 && db->r_mask > (unsigned long long) -1) {
	  out("%s(long long)s%d<<%d%s", copr, db->r_schunk,
	      -db->r_shift, mask);
	}
#endif /* __alpha */
	else
	    out("%ss%d<<%d%s", copr, db->r_schunk, -db->r_shift, mask);
	copr = " | ";
    }
    if (copr[0] == '\0')
	out("0");
    out(";\n");

    /* store the result */
    out("\t*(%s*)(dst+%d) = d%d;\n",
	native[size]->t_name, offset, offset);
}



/*  setbits(frame, offset, size) -- output sub-byte bitstring extraction  */

static void setbits(Frame *f, int offset, int size)
{
    unsigned int soff, doff, sindex, sshift, dshift;

    doff = f->f_dstoff;
    soff = f->f_srcoff;
    sindex = soff - soff % schunks[soff].c_size;
    sshift = f->f_srcshift + bshift(schunks[soff].c_size, soff);
    dshift = f->f_dstshift + bshift(size, doff);

    if (f->f_flags & MV_SIGN)
	dbop(sindex, sshift, 1, f->f_nbits, dshift);
    else
	dbop(sindex, sshift - dshift, 0, f->f_nbits, dshift);
}



/*  setbytes(frame, offset, size) -- output move of full byte(s)  */

static void setbytes(Frame *f, int offset, int size)
{
    unsigned int i, soff, doff, sindex, sshift, dshift;

    for (i = 0; i < size; i++) {		/* one byte at a time */

	doff = offset + i;
	if (f->f_flags & MV_SIGN)
	    soff = f->f_srcoff;
	else
	    soff = f->f_srcoff + (doff - f->f_dstoff);

	if (doff >= f->f_dstoff && doff < f->f_dstoff + f->f_nbytes) {
	    sindex = soff - soff % schunks[soff].c_size;
	    sshift = bshift(schunks[soff].c_size, soff);
	    dshift = bshift(size, doff);
	    if (f->f_flags & MV_SIGN)
		dbop(sindex, sshift + f->f_srcshift, 1, BYTE, dshift);
	    else
		dbop(sindex, sshift - dshift, 0, BYTE, dshift);
	    }
    }
}



/*  bshift(w, n) -- return shift for byte (n % w) out of chunk size w  */

static int bshift(int w, int n)
{
    if (w <= 1)				/* in case native char not annotated */
	return 0;
    n %= w;
    return BYTE * boff(native[w]->t_annot.a_order, n);
}



/*  dbop(schunk, shift, sxtend, nbits, dshift) -- add bit operation to list  */

static void dbop(int schunk, int shift, int sxtend, int nbits, int dshift)
{
    Dbits *db;

    if (nbops >= MAXINTBITS)
	MALFUNC("impossible growth in bitops table");
    if (dbflags[DB_BITOPS])
	fprintf(stderr, "   s%d>>%-3d %c %db <<%d\n",
	    schunk, shift, sxtend ? 'x' : ':', nbits, dshift);

    db = &bitops[nbops++];
    db->r_schunk = schunk;
    db->r_shift = shift;
    db->r_sxtend = sxtend;
#if __alpha
    db->r_mask = ((1L << nbits) - 1) << dshift;
#else /* __alpha */
    db->r_mask = ((1LL << nbits) - 1) << dshift;
#endif /* __alpha */
}



/*  bopcmp(p1, p2) -- compare bit ops for sorting  */

static int bopcmp(const void *p1, const void *p2)
{
    const Dbits *db1, *db2;

    db1 = p1;
    db2 = p2;
    if (db1->r_sxtend != db2->r_sxtend)
	return db1->r_sxtend - db2->r_sxtend;
    if (db1->r_schunk != db2->r_schunk)
	return db1->r_schunk - db2->r_schunk;
    if (db2->r_shift != db1->r_shift)
	return db2->r_shift - db1->r_shift;
    /* mask comparison is slightly subtle because we need an int result */
    if (db1->r_mask >= db2->r_mask)
	return db1->r_mask > db2->r_mask;
    return -1;
}



/*
 *  bopmerge(db1, db2) -- merge bitops.
 *
 *  If the two bit operations can be merged, db2's r_mask field is augmented
 *  with the bits from db1's r_mask field, and 1 is returned.  If the two
 *  bitops cannot be merged, 0 is returned.
 */

static int bopmerge(Dbits *db1, Dbits *db2)
{
    if (db1->r_sxtend != db2->r_sxtend)
	return 0;
    if (db1->r_schunk != db2->r_schunk)
	return 0;
    if (db1->r_shift != db2->r_shift)
	return 0;

    /* they can be merged. */
    db2->r_mask |= db1->r_mask;
    return 1;
}
