/*
 *  gen.c -- code generation.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

/*
 *  Naming conventions for the generated code:
 *
 *  p_xxx	parameter originally named xxx
 *  rv		function return value
 *  src		source for assignment
 *  dst		destination for assignment
 *  i		temporary for holding integer constant
 */

#include "usc.h"

static void macheader(FILE *f, Symbol *func);	/* generate macro header */
static void genproto(FILE *f,Symbol *p,char *s);/* generate function prototype*/
static void gtype(FILE *f, Type *t);		/* generate C type name */
static void genasg(int opr, Type *lt, Type *rt);/* generate assignment */
static void setnative(void);			/* init table of native types */



/*  genheader(func) -- generate header and prototype for function  */

void genheader(Symbol *func)
{
    Type *t;

    t = func->s_type;

    if (dbflags[DB_PARSE] || dbflags[DB_TREE])
	fprintf(stderr, "\n%s():\n", func->s_name);

    if (pfile && !(t->t_tflags & (TF_MACRO | TF_STATIC | TF_INLINE)))
	genproto(pfile, func, ";\n");

    if (t->t_tflags & TF_MACRO) {
	macheader(stdout, func);
	if (pfile)
	    macheader(pfile, func);
    } else {
	printf("\n");
	genproto(stdout, func, "\n{\n");
    }

    if (t->t_utype->t_tcode != T_VOID) {
	out("    ");
	gtype(stdout, t->t_utype);
	if (pfile && (t->t_tflags & TF_MACRO))
	    gtype(pfile, t->t_utype);
	out(" rv;\n");
    }
    out("    char *src, *dst;\n");
    out("    int i;\n");

    setnative();
}



/*  gentail() -- generate end of function  */

void gentail(Symbol *func)
{
    if (func->s_type->t_tflags & TF_MACRO) {
	printf("} while (0);\n");
	if (pfile)
	    fprintf(pfile, "} while (0);\n\n");
    } else
	printf("}\n\n");
}



/*  macheader(f, func) -- generate macro header on file f  */

static void macheader(FILE *f, Symbol *func)
{
    int n;
    Symbol *p;

    fprintf(f, "\n#define %s(", func->s_name);
    for (n = 0, p = func->s_next; p != NULL; p = p->s_next, n++) {
	if (n > 0)
	    fprintf(f, ",");
	fprintf(f, "p_%s", p->s_name);
    }
    fprintf(f, ") do { \\\n");
}



/*  genproto(f, func, s) -- generate function prototype followed by s */

static void genproto(FILE *f, Symbol *func, char *s)
{
    int n;
    Symbol *p;

    gtype(f, func->s_type->t_utype);
    fprintf(f, " %s", func->s_name);
    for (n = 0, p = func->s_next; p != NULL; p = p->s_next, n++) {
	fprintf(f, n == 0 ? "(" : ", ");
	gtype(f, p->s_type);
	fprintf(f, " p_%s", p->s_name);
    }
    if (n == 0)
	fprintf(f, "(void");
    fprintf(f, ")%s", s);
}



/*  gtype(f, t) -- generate C type name, honoring "usc -t" flag  */

static void gtype(FILE *f, Type *t)
{
    if (tflag && (t->t_name != NULL)) {
	fprintf(f, "%s", t->t_name);
	return;
    }
    if (t->t_tflags & TF_SIGNED)
	fprintf(f, "signed ");
    if (t->t_tflags & TF_UNSIGNED)
	fprintf(f, "unsigned ");
    switch (t->t_tcode) {
	case T_VOID:	fprintf(f, "void");	break;
	case T_CHAR:	fprintf(f, "char");	break;
	case T_SHORT:	fprintf(f, "short");	break;
	case T_INT:	fprintf(f, "int");	break;
	case T_LONG:	fprintf(f, "long");	break;
#ifndef __alpha
        case T_HYPER:	fprintf(f, "long long"); break;
#endif /* !__alpha */
	case T_ARRAY:
	case T_PTR:
	    if (tflag) {
		gtype(f, t->t_utype);
		fprintf(f, " *");
	    }
	    else
		fprintf(f, "void *");
	    break;
	default:	fprintf(f, "void");	break;
    }
}


/*
 *  genstmt(tree) -- generate code for one statement in function.
 *
 *  The basic model is to generate the destination and source addresses
 *  followed by a sequence of one or more move operations.
 *
 *  An obvious future optimization would be to remove the generation of
 *  addresses (and resulting extra store operations) when the source or
 *  destination are ultimately registers.
 */

void genstmt(Node *tree)
{
    Node *l, *r;
    Type *rt, *xt;
    int a;

    if (tree->n_type == NULL)
	return;				/* earlier error */

    l = tree->n_l;
    r = newnode(TK_TMPVAL, NULL, tree->n_r);	/* indicate temp available */
    r->n_type = tree->n_r->n_type;

    switch (tree->n_tok) {
	case TK_RETURN:
	    rt = tree->n_type;			/* return type */
	    xt = tree->n_r->n_type;		/* expression type */
	    a = xt->t_annot.a_align;
	    if ((rt->t_tcode == T_PTR
		|| sametype(rt, chartype)
		|| sametype(rt, shorttype)
		|| sametype(rt, inttype)
		|| sametype(rt, longtype)
#ifndef __alpha
		|| sametype(rt, hypertype)
#endif /* !__alpha */
	    ) && (xt->t_tcode == T_PTR
		|| sametype(xt, chartype)  && a >= chartype->t_annot.a_align
		|| sametype(xt, shorttype) && a >= shorttype->t_annot.a_align
		|| sametype(xt, inttype)   && a >= inttype->t_annot.a_align
		|| sametype(xt, longtype)  && a >= longtype->t_annot.a_align
#ifndef __alpha
		|| sametype(xt, hypertype) && a >= hypertype->t_annot.a_align
#endif /* !__alpha */
	    )) {
		/* this can be generated as an expression */
		out("    return %e;\n", tree->n_r);
	    } else {
		/* in general case, generate frame list etc. */
		out("    dst = (char *) &rv;\n");
		out("    src = (char *) %a;\n", r);
		genasg(tree->n_tok, tree->n_type, r->n_type);
		out("    return rv;\n");
	    }
	    break;
	case TK_ASSIGN:
	case TK_ASGPAD:
	    out("    dst = (char *) %a;\n", l);
	    out("    src = (char *) %a;\n", r);
	    genasg(tree->n_tok, l->n_type, r->n_type);
	    break;
	default:
	    MALFUNC("internal confusion in genstmt");
	    break;
    }
}



/*  genasg(opr, lefttype, righttype) -- generate assignment of src to dst  */

static void genasg(int opr, Type *ltype, Type *rtype)
{
    Frame *f, *g;
    int lalign, ralign;			/* alignments */

    lalign = ltype->t_annot.a_align;
    ralign = rtype->t_annot.a_align;

    f = gframes(ltype, rtype);		/* generate frames */
    gcode(f, lalign, ralign);		/* generate code given framelist */

    while (f) {
	g = f->f_next;
	free(f);			 /* free allocated frames */
	f = g;
    }
}



/*  setnative(node) -- initialize table of native types  */

static void setnative(void)
{
    int i;
    Type *t;
    static int cwarn;

    t = NULL;
    native[1] = chartype;		/* size 1 always available as char */

    for (i = 2; i <= MAXINTBYTES; i *= 2)	
	if (inttype->t_annot.a_tsize == i)
	    t = native[i] = inttype;
	else if (shorttype->t_annot.a_tsize == i)
	    t = native[i] = shorttype;
	else if (longtype->t_annot.a_tsize == i)
	    t = native[i] = longtype;
#ifndef __alpha
	else if (hypertype->t_annot.a_tsize == i)
	    t = native[i] = hypertype;
#endif /* !__alpha */

    if (!t & !cwarn++)
	twarning(NULL,
	    "using only char variables; no larger type defined", NULL);
}
