/*
 *  main.c -- overall control and miscellaneous utility functions.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

/*  define globals  */
#define extern
#include "globals.h"
#undef extern

static void helpmsg(void);
static void errmsg(int lno, char *type, char *msg, char *detail);

static int nwarn;			/* warning error count */
static int nfatal;			/* fatal error count */



/*  main program  */

int
main(int argc, char *argv[])
{
    int c;
    char *p, *q;
    extern char *optarg;
    extern int optind;
    extern int yydebug;

    gcode = DEF_GENERATOR;

#define USAGE \
"[-hvt] [-o filename] [-p filename] [-g {nfw}] [-d flags] sourcefile"
    while ((c = getopt(argc, argv, "hvto:p:g:d:")) != EOF)
	switch (c) {
	case 'd':
	    for (p = optarg; c = *p++; )
		if (c == DB_ALL) {
		    memset(dbflags, 1, sizeof(dbflags));
		    dbflags[DB_YACC] = 0;
		}
		else if (c == DB_GENERAL)
		    for (q = DB_G_STR; c = *q++; )
			dbflags[c] ^= 1;
		else if (isascii(c))
		    dbflags[c] ^= 1; 
	    break;
	case 'g':
	    switch (*optarg) {
		case 'n':  gcode = gnaive;   break;
		case 'f':  gcode = gfgrain;  break;
		case 'w':  gcode = gwords;   break;
		default:
		{ fprintf(stderr, "usage: %s %s\n", argv[0], USAGE); exit(1); }
	    }
	    break;
	case 'h':
	    helpmsg();
	    exit(0);
	    /*NOTREACHED*/
	case 'o':
	    if (!freopen(optarg, "w", stdout))
	        { perror(optarg); exit(1); }
	    break;
	case 'p':
	    if (!(pfile = fopen(optarg, "w")))
	        { perror(optarg); exit(1); }
	    break;
	case 't':
	    tflag = 1;
	    break;
	case 'v':
	    fprintf(stderr, "%s, built %s %s\n", VERSION, __TIME__, __DATE__);
	    exit(0);
	    /*NOTREACHED*/
	default:
	    { fprintf(stderr, "usage: %s %s\n", argv[0], USAGE); exit(1); }
	}

    if (optind + 1 < argc)
	{ fprintf(stderr, "usage: %s %s\n", argv[0], USAGE); exit(1); }

    if (optind < argc) {
	iname = argv[optind];
	if (!freopen(iname, "r", stdin))
	    { perror(iname); exit(1); }
    } else
	iname = "stdin";
    iline = 1;

    initkw();
    inittypes();

    yydebug = dbflags[DB_YACC];
    yyparse();

    if (dbflags[DB_SYMTAB])
	dumpsyms();

    if (nfatal > 0)
	return 1;
    else
	return 0;
}



/*  statement(tree) -- process statement given raw parse tree  */

void statement(Node *tree)
{
    if (dbflags[DB_PARSE])
	ptree(tree);			/* print the tree, if debugging */

    typecheck(tree);			/* check types */

    if (dbflags[DB_TREE])
	ptree(tree);			/* print it again */

    /* could fold constants and other such optimizations here */

    genstmt(tree);			/* generate code */

    freenodes();			/* free all parse tree nodes */
}


/*  helpmsg() -- issue help message  */

static void helpmsg(void)
{
    char **p;
    static char *hlines[] = {
	"Command line options:",
	"",
	"-h		-- print this message and exit",
	"-v		-- print version and exit",
	"-t		-- preserve type names in output",
	"-o filename	-- select output file",
	"-p filename	-- select prototype/macro output",
	"-d flags	-- select debugging output",
	0
    };

    for (p = hlines; *p; p++)
	fprintf(stderr, "%s\n", *p);
}



/*
 * setsource(line, echo) -- set filename and line number from cpp information
 *
 * The usual format of a cpp info line is:  # linenum "filename" [flags]
 */
void setsource(char *line, int echo)
{
    char filename[1000];
    int n, lnum;

    n = sscanf(line, "# %d \"%[^\"]", &lnum, filename);

    if (n > 1)				/* if new file name given */
	iname = savename(filename);
		
    if (n > 0) {			/* if valid, w/ or w/o filename */
	iline = lnum - 1;
	if (echo)
	    printf("#line %d \"%s\"", lnum, iname);
    }
}



/*
 *  warning(msg, detail) -- issue warning message
 *  fatal(msg, detail) -- issue fatal error message
 *  tfatal(node, msg, detail) -- issue fatal error using line nbr from node
 *  twarning(node, msg, detail) -- issue warning msg using line nbr from node
 */
void warning(char *msg, char *detail)
{
    errmsg(iline, "warning", msg, detail);
    nwarn++;
}

void fatal(char *msg, char *detail)
{
    errmsg(iline, "fatal", msg, detail);
    nfatal++;
}

void tfatal(Nodeptr n, char *msg, char *detail)
{
    errmsg(n ? n->n_lno : 0, "fatal", msg, detail);
    nfatal++;
}

void twarning(Nodeptr n, char *msg, char *detail)
{
    errmsg(n ? n->n_lno : 0, "warning", msg, detail);
    nwarn++;
}

static void errmsg(int lno, char *type, char *msg, char *detail)
{
    if (lno != 0)
	fprintf(stderr, "\"%s\", line %d: ", iname, lno);
    fprintf(stderr, "%s: %s", type, msg);
    if (detail)
	fprintf(stderr, ": %s\n", detail);
    else
	fprintf(stderr, "\n");
}



/*  malf(file, line, msg) -- abort due to program malfunction  */

void malf(char *file, int line, char *msg)
{
    fprintf(stderr, "program malfunction (%s, line %d): %s\n", file, line, msg);
    exit(1);
}


/*  zalloc(b) -- allocate b bytes of zeroes, with error checking  */

void *zalloc(long b)
{
    void *p;

    if (b == 0)
	b = 1;			/* standard malloc can't handle 0 */
    if ((p = calloc(1, b)) == NULL) {
	fprintf(stderr, "out of memory: can't allocate %ld bytes\n", b);
	exit(1);
    }
    return p;
}
