/*
 *  out.c -- code writing to output files.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

static void oaddr(Node *tree);			/* output address of expressn */
static void oexpr(Node *tree);			/* output value of expression */
static void otype(Node *tree);			/* native type, given USC type*/



/*
 *  out(fmt, ...) -- write code to stdout and/or pfile.
 *
 *  Out() is similar in spirit to printf.  It must be used for ALL generated
 *  stub code to ensure that it is routed to the proper file or files.
 *
 *  The format specifiers are:
 *	spec	arg	meaning
 *	----	----	-------
 *	\n	none	output newline, with preceding backslash if in macro
 *	%a	Node*	generate address of expression given by tree
 *	%b	Frame*	generate zero or sign word specified by frame
 *	%d	int	decimal value
 *	%e	Node*	generate value of expression given by tree
 *	%s	char*	output string
 *	%t	Node*	native type if any corresponding to USC type, else err
 */
void out(char *fmt, ...)
{
    va_list argp;
    FILE *p;			/* file pointer for macro, if any */
    int c, d;
    char *s;
    Frame *f;

    if (pfile != NULL && (curfunc->s_type->t_tflags & TF_MACRO))
	p = pfile;		/* set macro output file */
    else
	p = NULL;		/* no macro output */

    va_start(argp, fmt);
    while (c = *fmt++) {
	if (c == '%') {
	    switch (c = *fmt++) {
		case 'a':		/* %a -- output address */
		    oaddr(va_arg(argp, Node *));
		    break;
		case 'b':		/* %b -- zero or sign bits from frame */
		    f = va_arg(argp, Frame *);
		    if (f->f_flags & MV_ZERO)
			out("0");
		    else if (f->f_flags & MV_SIGN)
			out("-(src[%d]>>%d&1)", f->f_srcoff, f->f_srcshift);
		    else
			MALFUNC("bad out() call");
		    break;
		case 'd':		/* %d -- output decimal value */
		    d = va_arg(argp, int);
		    printf("%d", d);
		    if (p)
			fprintf(p, "%d", d);
		    break;
		case 'e':		/* %e -- output expression value */
		    oexpr(va_arg(argp, Node *));
		    break;
		case 's':		/* %s -- substitute a string */
		    s = va_arg(argp, char *);
		    printf("%s", s);
		    if (p)
			fprintf(p, "%s", s);
		    break;
		case 't':		/* %t -- output native type */
		    otype(va_arg(argp, Node *));
		    break;
		default:		/* anything else: error */
		    fmt--;		/* ignore % and reprocess */
		    break;
	    }
	} else if (c == '\n' && (curfunc->s_type->t_tflags & TF_MACRO)) {
	    printf(" \\\n");		/* add \ to continue macro line */
	    if (p)
		fprintf(p, " \\\n");
	} else {
	    putc(c, stdout);		/* nothing special; just copy */	
	    if (p)
		putc(c, p);
	}
    }
    va_end(argp);
}



/*  oaddr(tree) -- output address of expression  */

static void oaddr(Node *tree)
{
    static int iwarn;
    Node *n;

    n = tree;
    if (n->n_tok == TK_TMPVAL)
	n = tree->n_r;			/* find real node under TMPVAL node */

    switch (n->n_tok) {

	case TK_ID:
	    out("&p_%s", n->n_sym->s_name);
	    return;

	case TK_UASTER:
	    out("%e", n->n_r);
	    return;

	case TK_PERIOD:
	    out("(%d+(char*)%a)", n->n_sym->s_type->t_annot.a_align, n->n_l);
	    return;

	case TK_SUBSCR:
	    out("((char*)%e + %d*%e)",
		n->n_l, tree->n_type->t_annot.a_msize, n->n_r);
	    return;
    }

    /* not legal as tried.  see if we can use a temporary. */
    if (tree->n_tok == TK_TMPVAL) {		/* if originally TMPVAL node */
	if (!inttype->t_annot.a_msize && !iwarn++)
	    tfatal(tree, "illegal constant", "no prior typedef of `int'");
	out("(i = %e, &i)", tree->n_r);		/* gen using temporary */
	return;
    }

    /* catch-all for unimplemented cases */
    if (n->n_tok == TK_ICONST)
	tfatal(n, "int constant in illegal context", NULL);
    else
	tfatal(n, "illegal expression", NULL);
    out("0");
}



/*  oexpr(tree) -- output value of expression  */

static void oexpr(Node *e)
{
    int n;

    switch (e->n_tok) {

	case TK_ID:	out("p_%s", e->n_sym->s_name);	return;
	case TK_UAMPER:	out("%a", e->n_r);		return;

	case TK_ICONST:	out("%d", (int) e->n_long);	return;

	case TK_TMPVAL:	out("%e", e->n_r);		return;

	case TK_UPLUS:	out("%e", e->n_r);		return;
	case TK_UMINUS:	out("(-%e)", e->n_r);		return;

	case TK_PLUS:	out("(%e+%e)", e->n_l, e->n_r);	return;
	case TK_MINUS:	out("(%e-%e)", e->n_l, e->n_r);	return;
	case TK_ASTER:	out("(%e*%e)", e->n_l, e->n_r);	return;
	case TK_SLASH:	out("(%e/%e)", e->n_l, e->n_r);	return;

	case TK_SUBSCR:
	case TK_PERIOD:
	    if (e->n_type->t_tcode == T_ARRAY)
		out("%a", e);		/* result of array expr is its addr */
	    else
		out("*(%t*)%a", e, e);
	    return;

	case TK_SIZEOF:
	    out("%d", n = e->n_r->n_type->t_annot.a_msize);
	    if (!n)
		tfatal(e, "sizeof unknown", NULL);
	    return;

	case TK_UASTER:
	    if (e->n_r->n_tok == TK_UAMPER) {
		out("%e", e->n_r->n_r);
		return;
	    }
	    if (e->n_r->n_tok == TK_ID) {
		out("*p_%s", e->n_r->n_sym->s_name);
		return;
	    }
	    out("*(%t*)%a", e, e->n_r);
	    return;

    }

    /* catch-all for unimplemented cases */
    tfatal(e, "illegal expression", NULL);
    out("0");
}



/*  otype(tree) -- output corresponding native type, issuing error if none  */

static void otype(Node *n)
{
    Type *t;

    t = n->n_type;
    if (t->t_tflags & TF_UNSIGNED)
	out("unsigned ");

    if (sametype(t, inttype))
	out("int");
    else if (sametype(t, longtype))
	out("long");
#ifndef __alpha
    else if (sametype(t, hypertype))
	out("long long");
#endif /* !__alpha */
    else if (sametype(t, shorttype))
	out("short");
    else if (sametype(t, chartype))
	out("char");
    else {
	tfatal(n,"value used in expression does not match a native type",NULL);
	out("int");
    }
}
