/*
 *  wordg.c -- word-oriented code generation.
 *
 *  Copyright 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"

#define VAR_PER_LINE 8

/*
 *  Dbits: structure specifying the source of some destination register bits.
 */
typedef struct {
    short r_sxtend;		/* zero if copy; nonzero if sign-extend */
    short r_shift;		/* source shift */
    int r_sindex;		/* source word offset */
    unsigned long r_mask;	/* mask of bits used */
} Dbits;



static Type *findreg(int n);				/* type for reg size */
static int  ulength(Frame *flist, int offset);		/* find used length */
static void genvars(char *prefix, int length);		/* gen variable decls */
static void gencalcs(Frame *flist);			/* gen calculations */
static void fillgap(int prev, int next);		/* fill output gap */
static void genwd(Frame *f, int offset, int size);	/* gen one output wd */
static void setbits(Frame *f, int offset, int size);	/* gen set-bit bitop */
static void setbytes(Frame *f, int offset, int size);	/* gen set-byte bitops*/
static int  bshift(int w, int n);			/* find byte in word */
static void dbop(int s, int sh, int x, int n, int ds);	/* add bitop to list */
static int  bopcmp(const void *p1, const void *p2);	/* bitop ordering */
static int  bopmerge(Dbits *db1, Dbits *db2);		/* merge bitops */
static void sload(int n);				/* load source word */
static void iload(int n);				/* ld unaligned input */
static void astore(int n);				/* aligned store */
static void ustore(int n);				/* unaligned store */



static int regsize;			/* size of native register */
static Type *regtype;			/* type of native register */
static char *rcast;			/* name of register type, for casting */
static int bigend;			/* is that type big-endian? */

static Dbits bitops[MAXINTBITS];	/* array of bit operations */
static int nbops;			/* number of operations in array */

static int slength, dlength;		/* source & destination length used */
static int salign, dalign;		/* are src/dst register-aligned? */
static char *iloaded;			/* list of input words loaded */
static char *sloaded;			/* list of source words loaded */



/*  gwords(flist, lalign, ralign) -- generate word-oriented moves  */

void gwords(Frame *flist, int lalign, int ralign)
{
    Type *t;
    int n;

    /* figure out register size */
    for (n = 1; n <= MAXINTBYTES; n *= 2)
	if (t = findreg(n)) {
	    regsize = n;
	    regtype = t;
	    rcast = t->t_name;
	}
    if (regsize > 1 && regtype->t_annot.a_order->o_oseg[0].o_start)
	bigend = 1;			/* big-endian */
    else
	bigend = 0;			/* little-endian */

    /* begin declarations */
    out("    {\n");			/* new C scope for variables */

    /* declare input variables*/
    slength = ulength(flist, offsetof(Frame, f_srcoff));
    iloaded = zalloc(slength + regsize);
    sloaded = zalloc(slength);

    if (ralign < regsize) {		/* if need unaligned loads */
	salign = 0;
	out("\tchar *asrc = (char*) (((long)src) & ~%d);\n", regsize - 1);
	out("\tint ishift = %d * (((long)src) & %d);\n", BYTE, regsize - 1);
	genvars("r", slength + regsize);	/* loading registers */
    } else
	salign = 1;

    genvars("s", slength);		/* source words */

    /* declare output variables */
    dlength = ulength(flist, offsetof(Frame, f_dstoff));
    if (lalign < regsize) {		/* if need unaligned stores */
	dalign = 0;
	out("\tchar *adst = (char*) (((long)dst) & ~%d);\n", regsize - 1);
	out("\tint oshift = %d * (((long)dst) & %d);\n", BYTE, regsize - 1);
	genvars("w", dlength);		/* output registers */
    } else
	dalign = 1;

    if (!dalign || dlength % regsize != 0) 
	out("\tunsigned %s m, t;\n", rcast);

    genvars("d", dlength);			/* destination words */


    /* sort frames in destination order */
    flist = sortframes(flist,
	offsetof(Frame, f_dstoff), offsetof(Frame, f_dstshift));

    /* generate code to calculate and store destination words */
    out("\n");
    gencalcs(flist);

    free(iloaded);
    free(sloaded);
    out("    }\n");
}



/*  findreg(n) -- find native type for register size of n bytes  */

static Type *findreg(int n)
{
    if (n == 1)
	return chartype;
    if (inttype->t_annot.a_tsize == n)	/* prefer "int", so try it first */
	return inttype;
    if (shorttype->t_annot.a_tsize == n)
	return shorttype;
    if (longtype->t_annot.a_tsize == n)
	return longtype;
/*
 * Don't return hypertype here on non-alpha because it's
 * not a native type- hurts the compiler instead of helping
 * optimization-wise
 */
    return NULL;
}



/*  ulength(flist, offset) -- return length of structure as actually used  */

static int ulength(Frame *flist, int offset)
{
    Frame *f;
    int n, w, len;

    /* find maximum offset used */
    len = 0;
    for (f = flist; f != NULL; f = f->f_next) {
	w = f->f_nbytes;
	if (w == 0)			/* if bit move */
	    w = 1;
	n = IntAt(f, offset) + w;
	if (n > len)
	    len = n;
    }
    return len;
}



/*  genvars(prefix, length) -- generate variable declarations  */

static void genvars(char *prefix, int length)
{
    int i, n;

    for (i = n = 0; n < length; n += regsize) {
	if (i % VAR_PER_LINE == 0)
	    out("\tunsigned %s %s%d", rcast, prefix, n);
	else
	    out(", %s%d", prefix, n);
	if (++i % VAR_PER_LINE == 0)
	    out(";\n");
    }
    if (i % VAR_PER_LINE != 0)
	out(";\n");
}



/*
 *  gencalcs(flist) -- generate calculation operations.
 */
static void gencalcs(Frame *flist)
{
    Frame *f;
    int i, d, nb, o, prevd;

    prevd = -1;
    for (f = flist; f != NULL; f = f->f_next) {
	nb = f->f_nbytes;		/* number of bytes */
	if (nb == 0)			/* if bit move: affects one byte */
	    nb = 1;
	for (i = 0; i < nb; i++) {	/* for each byte of target */
	    o = f->f_dstoff + i;	/* destination offset */
	    d = o - o % regsize;	/* word offset */
	    if (d != prevd) {		/* if different from previous word */
		fillgap(prevd, d);	/* fill gap, if any */
		genwd(f, d, regsize);	/* generate new word */
		prevd = d;		/* remember previous */
	    }
	}
    }
}



/*
 *  fillgap(prev,next) -- fill gap in output.
 *
 *  If the output is unaligned, and there are gaps between output words,
 *  force out the incomplete fragment of the previous word.  Set the unused
 *  bits to constant zero give determinstic behavior and (it is hoped)
 *  allow better code generation than if left unspecified.
 */
static void fillgap(int prev, int next)
{
    int curr = prev + regsize;

    if (dalign || prev < 0 || curr == next)
	return;
    out("\td%d = 0;\n", curr);
    ustore(curr);
    out("\tw%d = 0;\n", next);
}



/*
 *  genwd(f, offset, size) -- generate code to calculate one output word.
 *
 *  f is the first of one or more frames that affect the word.
 */
static void genwd(Frame *f, int offset, int size)
{
    Dbits *db;
    int i;
    char *copr, mask[2 * MAXINTBYTES + 5];

    /* generate a list of bit operations */
    if (dbflags[DB_BITOPS])
	fprintf(stderr, "bit operations for d%d:\n", offset);
    nbops = 0;				/* init array to empty */

    for (; f != NULL; f = f->f_next) {

	if (f->f_dstoff >= offset + size)
	    break;			/* if beyond range of this word */
	if (f->f_nbytes > 0 && f->f_dstoff + f->f_nbytes <= offset)
	    continue;			/* strange overlap situation */
	if (f->f_flags == MV_ZERO)
	    continue;			/* nothing to do for zero fill */

	if (f->f_nbytes == 0)
	    setbits(f, offset, size);	/* set bits */
	else
	    setbytes(f, offset, size);	/* set full bytes */
    }

    /* sort the list by sxtend flag, source index, and shift */
    qsort(bitops, nbops, sizeof(Dbits), bopcmp);

    /* load any necessary source words */
    for (i = 0; i < nbops; i++)
	sload(bitops[i].r_sindex);

    /* output the operations */
    out("\td%d = ", offset);
    copr = "";
    for (i = 0; i < nbops; i++) {
	db = &bitops[i];
	if (i < nbops - 1 && bopmerge(db, db + 1))
	    continue;			/* if merge successful */

	if ((db->r_mask == (unsigned long) -1)
	|| (size < sizeof(long) && db->r_mask == (1L << BYTE * size) - 1))
	    sprintf(mask, "");		/* using full word; no mask needed */
	else
	    sprintf(mask, "&0x%lXu", db->r_mask);

	if (db->r_sxtend)
	    out("%s-(s%d>>%d&1)%s", copr, db->r_sindex, db->r_shift, mask);
	else if (db->r_shift == 0)
	    out("%ss%d%s", copr, db->r_sindex, mask);
	else if (db->r_shift > 0)
	    out("%ss%d>>%d%s", copr, db->r_sindex, db->r_shift, mask);
	else if (regsize < (unsigned) size && db->r_mask > (unsigned) - 1)
	    out("%s(long)s%d<<%d%s", copr, db->r_sindex, -db->r_shift, mask);
	else
	    out("%ss%d<<%d%s", copr, db->r_sindex, -db->r_shift, mask);
	copr = " | ";
    }
    if (copr[0] == '\0')
	out("0");
    out(";\n");

    /* store the result */
    if (dalign)
	astore(offset);			/* aligned store */
    else
	ustore(offset);			/* unaligned store */
}



/*  setbits(frame, offset, size) -- output sub-byte bitstring extraction  */

static void setbits(Frame *f, int offset, int size)
{
    unsigned int soff, doff, sindex, sshift, dshift;

    doff = f->f_dstoff;
    soff = f->f_srcoff;
    sindex = soff - soff % regsize;
    sshift = f->f_srcshift + bshift(regsize, soff);
    dshift = f->f_dstshift + bshift(size, doff);

    if (f->f_flags & MV_SIGN)
	dbop(sindex, sshift, 1, f->f_nbits, dshift);
    else
	dbop(sindex, sshift - dshift, 0, f->f_nbits, dshift);
}



/*  setbytes(frame, offset, size) -- output move of full byte(s)  */

static void setbytes(Frame *f, int offset, int size)
{
    unsigned int i, soff, doff, sindex, sshift, dshift;

    for (i = 0; i < size; i++) {		/* one byte at a time */

	doff = offset + i;
	if (f->f_flags & MV_SIGN)
	    soff = f->f_srcoff;
	else
	    soff = f->f_srcoff + (doff - f->f_dstoff);

	if (doff >= f->f_dstoff && doff < f->f_dstoff + f->f_nbytes) {
	    sindex = soff - soff % regsize;
	    sshift = bshift(regsize, soff);
	    dshift = bshift(size, doff);
	    if (f->f_flags & MV_SIGN)
		dbop(sindex, sshift + f->f_srcshift, 1, BYTE, dshift);
	    else
		dbop(sindex, sshift - dshift, 0, BYTE, dshift);
	    }
    }
}



/*  bshift(w, n) -- return shift for byte (n % w) out of word size w  */

static int bshift(int w, int n)
{
    if (w <= 1)				/* in case native char not annotated */
	return 0;
    n %= w;
    return BYTE * boff(native[w]->t_annot.a_order, n);
}



/*  dbop(sindex, shift, sxtend, nbits, dshift) -- add bit operation to list  */

static void dbop(int sindex, int shift, int sxtend, int nbits, int dshift)
{
    Dbits *db;

    if (nbops >= MAXINTBITS)
	MALFUNC("impossible growth in bitops table");
    if (dbflags[DB_BITOPS])
	fprintf(stderr, "   s%d>>%-3d %c %db <<%d\n",
	    sindex, shift, sxtend ? 'x' : ':', nbits, dshift);

    db = &bitops[nbops++];
    db->r_sindex = sindex;
    db->r_shift = shift;
    db->r_sxtend = sxtend;
    db->r_mask = ((1L << nbits) - 1) << dshift;
}



/*  bopcmp(p1, p2) -- compare bit ops for sorting  */

static int bopcmp(const void *p1, const void *p2)
{
    const Dbits *db1, *db2;

    db1 = p1;
    db2 = p2;
    if (db1->r_sxtend != db2->r_sxtend)
	return db1->r_sxtend - db2->r_sxtend;
    if (db1->r_sindex != db2->r_sindex)
	return db1->r_sindex - db2->r_sindex;
    if (db2->r_shift != db1->r_shift)
	return db2->r_shift - db1->r_shift;
    /* mask comparison is slightly subtle because we need an int result */
    if (db1->r_mask >= db2->r_mask)
	return db1->r_mask > db2->r_mask;
    return -1;
}



/*
 *  bopmerge(db1, db2) -- merge bitops.
 *
 *  If the two bit operations can be merged, db2's r_mask field is augmented
 *  with the bits from db1's r_mask field, and 1 is returned.  If the two
 *  bitops cannot be merged, 0 is returned.
 */

static int bopmerge(Dbits *db1, Dbits *db2)
{
    if (db1->r_sxtend != db2->r_sxtend)
	return 0;
    if (db1->r_sindex != db2->r_sindex)
	return 0;
    if (db1->r_shift != db2->r_shift)
	return 0;

    /* they can be merged. */
    db2->r_mask |= db1->r_mask;
    return 1;
}



/*  sload(n) -- generate code to load source word n  */

static void sload(int n)
{
    if (sloaded[n])
	return;

    if (salign)
	out("\ts%d = *(%s*)(src+%d);\n", n, rcast, n);
    else {
	iload(n);
	iload(n + regsize);
	if (bigend)
	    out("\ts%d = r%d << ishift | r%d >> %d >> (%d - ishift);\n",
		n, n, n + regsize, BYTE, BYTE * (regsize - 1));
	else
	    out("\ts%d = r%d >> ishift | r%d << %d << (%d - ishift);\n",
		n, n, n + regsize, BYTE, BYTE * (regsize - 1));
	}

    sloaded[n] = 1;
}



/*  iload(n) -- load unaligned input word n  */

static void iload(int n)
{
    if (iloaded[n])
	return;

    if (n >= slength)			/* if final word of unaligned load */
	out("\tr%d = *(unsigned %s*)((long)(src+%d) & ~%d);\n",
	    n, rcast, slength - 1, regsize - 1);
    else
	out("\tr%d = *(unsigned %s*)(asrc+%d);\n", n, rcast, n);
    iloaded[n] = 1;
}



/*  astore(n) -- aligned store of destination word n.  */

static void astore(int n)
{
    Type *t;
    int w;

    w = dlength % regsize;
    if (n + regsize <= dlength) {
	/* easy case: full word store */
	out("\t*(%s*)(dst+%d) = d%d;\n", rcast, n, n);
    } else if (t = findreg(w)) {
	/* can store using a smaller type */
	if (bigend)
	    out("\t*(%s*)(dst+%d) = d%d >> %d;\n",
		t->t_name, n, n, BYTE * (regsize - w));
	else
	    out("\t*(%s*)(dst+%d) = d%d;\n", t->t_name, n, n);
    } else {
	/* general case: store w characters */
	out("\tt = *(%s*)(dst+%d);\n", rcast, n);
	if (bigend) {
	    out("\tm = (unsigned %s)~0L << %d;\n", rcast, BYTE * (regsize - w));
	    out("\td%d = (d%d & m) | (t & ~m);\n", n, n);
	} else {
	    out("\tm = (unsigned %s)~0L << %d;\n", rcast, BYTE * w);
	    out("\td%d = (d%d & ~m) | (t & m);\n", n, n);
	}
	out("\t*(%s*)(dst+%d) = d%d;\n", rcast, n, n);
    }
}



/*  ustore(n) -- unaligned store of destination word n.  */

static void ustore(int n)
{
    int r, k;
    char *lsh = bigend ? "<<" : ">>";	/* shift reg towards low address */
    char *hsh = bigend ? ">>" : "<<";	/* shift reg towards high address */

    if (n + regsize < dlength) {
	/*
	 *  This is not the final output word.
	 *  Start by completing the current output word.
	 */
	if (n == 0)	 /* first word: combine w/ existing word from memory */
	    out(
	"\tw0 = *(%s*)adst & ~((unsigned %s)~0L %s oshift) | d%d %s oshift;\n",
		rcast, rcast, hsh, n, hsh);

	else		 /* interior word: combine with previous fragment */
	    out("\tw%d |= d%d %s oshift;\n", n, n, hsh);

	/* save portion that goes in next word */
	out("\tw%d = d%d %s %d %s (%d-oshift);\n",
	    n + regsize, n, lsh, BYTE, lsh, BYTE * (regsize - 1));

	/* store current word */
	out("\t*(%s*)(adst+%d) = w%d;\n", rcast, n, n);

    } else {
	/*
	 *  This is the final output word.  0 to regsize-1 bytes have already
	 *  been accumulated in w[d], depending on oshift.  1 to regsize bytes
	 *  must be added, depending on dlength % regsize. This may result in
	 *  <1, 1, or >1 words to be stored, and the check must be made at
	 *  run time.
	 */

	if (n == 0)		/* if first AND last word, preload w0 */
	    out("\tw0 = *(%s*)adst & ~((unsigned %s)~0L %s oshift);\n",
		rcast, rcast, hsh);

	k = r = dlength % regsize;
	if (r > 0)
	    k = BYTE * (regsize - r);
	
	if (k > 0)			/* if full-word store is not certain */
	    out("\tif (oshift >= %d) {\n", k);	 /* check for full-word store */

	out("\t    *(%s*)(adst+%d) = w%d | d%d %s oshift;\n", rcast,n,n,n,hsh);

	out("\t    ");
	if (k < BYTE * (regsize - 1))
	    out("if (oshift > %d) ", k);
	out("{\n", k);
	out("\t\tt = *(%s*)(adst+%d) & ((unsigned %s)~0L %s (oshift-%d));\n",
	    rcast, n + regsize, rcast, hsh, k);
	out("\t\t*(%s*)(adst+%d) = t | d%d %s (%d - oshift);\n",
	    rcast, n + regsize, n, lsh, BYTE * regsize);
	out("\t    }\n");

	if (k > 0) {			/* if full-word store is not certain */
	    out("\t} else {\n");
	    out(
	    "\t    t = *(%s*)(adst+%d) & ((unsigned %s)~0L %s (oshift+%d));\n",
		rcast, n, rcast, hsh, BYTE * r);
	    out("\t    *(%s*)(adst+%d) = w%d | t | d%d %s oshift;\n",
		rcast, n, n, n, hsh);
	    out("\t}\n");
	}
    }
}
