/*
 *  imain.c
 *
 *  Copyright 1994, 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"
#include "usit.h"
#include "symbol.h"
#include "buffer.h"
#include "utils.h"
#ifdef LINUX
#include <getopt.h>
#endif /* LINUX */

#ifndef OPTIMIZE
extern void print_type_table (int level);
extern void print_sym_table (int level);
extern void print_struct_table (int level);
extern void print_union_table (int level);
#endif

extern void init_symbols (void);
extern int yydebug;
extern FILE *yyin;

UsitMode usit_output, usit_input;
char *native_compiler;
char *UsitDataFile;
FILE *UsitData;

Buffer *FunctionsBuffer;

static char *HelpMsg[] = {
  "Command line options:\n",
  "-h              -- print this message and exit",
  "-v              -- print version and exit",
  "-n              -- define native types",
  "-C compiler     -- select C compiler",
  "-o filename     -- select output file",
  "-u              -- output the final usc code",
  "-i              -- output inference code (run pass 1 only)",
  "-r filename     -- read inference data (run pass 2 only)",
  "-d level        -- select debugging output",
#ifdef YYDEBUG
  "-t              -- print out yacc trace output",
#endif
  NULL
};

/***************************************************************************/

void usit_output_functions (FILE *out)
{
  if (BUFFER_LEN (FunctionsBuffer) > 0)
    fprintf (out, "%s\n\n", BUFFER_STRING (FunctionsBuffer));
}

char *allocate_new_string (char *string)
{
  char *new;
  int temp;
  if (string != NULL) {
    temp = strmem (string);
    new = (char *)Malloc (temp * sizeof (char));
    dprintf (1000, "Entering allocate_new_string: %s", string);
    strcpy (new, string);
    dprintf (1000, "Exiting allocate_new_string: %s %s", new, string);
  }
  else {
    new = NULL;
  }
  return new;
}

int strmem (char *string)
{
  return strlen (string) + 1;
}

void init (void)
{
  init_mem();
  inityy();
  init_symbols();
  init_gen_utils (OutFile);
  InFileName = NULL;
  OutFile = NULL;
  DefaultCodeType = GEN_PLAIN;
  code_type = DefaultCodeType;
  print_comments = 0;
  speed_test = 0;
  usit_input = USIT_IN_C_CODE;
  usit_output = USIT_OUT_USC_CODE;
  preserve_type_names = 1;            /* talk to dorgival about this */
  FunctionsBuffer = makebuf (128);
}

void finishup(void)
{
  static int done = 0;

  if ( done )
    return;
  else
    done = 1;

  code_type = GEN_PLAIN;
  run_preprocess();

  if (usit_output == USIT_OUT_USC_CODE)
    usit_output_functions (OutFile);
}

void closeup (void)
{
  finishup();
  close_mem();
  if (OutFile != stdout)
    fclose (OutFile);
}

int main(int argc, char **argv)
{
  int  option;
  extern char* optarg;
  extern int   optind;
  char *options;
  char *progname = argv[0];
  char **p;

#ifdef YYDEBUG
  options = "C:d:vhnir:uo:t";
#else
  options = "C:d:vhnir:uo:";
#endif

  init();
  speed_test = 0;
  print_comments = 0;
  native_compiler = NULL;

  InFileName = NULL;

  while ((option = getopt(argc, argv, options)) != EOF)
      
    switch (option) {
      case 'C':
	native_compiler = allocate_new_string (optarg);
	break;
	
      case 'd':
	Debug_Level = atoi (optarg);
	break;
	
      case 'v':
	fprintf(stderr, "%s, built %s %s\n", VERSION, __TIME__, __DATE__);
	exit(0);
	/*NOTREACHED*/
	
      case 'h':
	for (p = HelpMsg; *p; p++) {
	  fprintf (stderr, "%s\n", *p);
	}
	exit (0);
	/*NOTREACHED*/
	
      case 'n':
	define_native_types = 1;
	break;
	
      case 'i':
	usit_output = USIT_OUT_INFER_CODE;
	break;

      case 'u':
	usit_output = USIT_OUT_USC_CODE;
	break;
	
      case 'r':
	UsitDataFile = allocate_new_string (optarg);
	usit_input = USIT_IN_INFER_OUTPUT;
	if ((UsitData = fopen (UsitDataFile, "r")) == NULL) {
	  fprintf (stderr, "%s: Can't open: %s\n", UsitDataFile, progname);
	  exit (1);
	}
	break;
	
#ifdef YYDEBUG
      case 't':
	yydebug = 1;
	break;
#endif
      
      case 'o':
	if (OutFile) {
	  fprintf(stderr, "%s: Two definitions for the output file\n", 
		  progname);
	  exit(1);
	}
	if (!strcmp (optarg, "-"))
	  OutFile = stdout;
	else
	  OutFile = fopen (optarg, "w");
	if (!OutFile) {
	  fprintf (stderr, "%s: Can't open output file: %s\n", 
		   progname, optarg);
	  exit (1);
	}
	break;

      default:
	  fprintf (stderr, "%s: Unknown option\n", progname);
          exit (1);
	  /*NOTREACHED*/

    }

  if (optind + 1 < argc) {
    fprintf (stderr, "%s: Can't have two input files\n", progname);
    exit (1);
  }

  if (optind < argc) {
    InFileName = allocate_new_string (argv[optind]);
    if (strcmp (InFileName, "-")) {
      yyin = fopen (InFileName, "r");
      if (!yyin) {
        fprintf (stderr, "%s: Can't open input file: %s\n",
          progname, argv[optind]);
        exit (1);
      }
    }
  } else
    InFileName = "-";

  if (usit_input == USIT_IN_INFER_OUTPUT &&
      usit_output == USIT_OUT_INFER_CODE) {
    fprintf (stderr, "%s: Can't use -d with -i\n", progname);
    exit (1);
  }

  if (!native_compiler) {
    native_compiler = DEFAULT_NATIVE_COMPILER;
    if (usit_input == USIT_IN_C_CODE && usit_output == USIT_OUT_USC_CODE)
      dprintf(1, "Using default compiler option: %s", native_compiler);
  }

  if (!OutFile)
    OutFile = stdout;

  init_outfile(USIT_OUTPUT);

  yyparse();

  print_sym_table (200);
  print_type_table (200);
  
  print_struct_table (250);
  print_union_table (250);

  closeup();
  return 0;
}
