/*
 *  utils.c -- miscellaneous functions, mostly related to output generation.
 *
 *  Copyright 1994, 1995 by the Arizona Board of Regents.
 *
 *  Permission to use, copy, modify, and distribute this software and its
 *  documentation for any purpose and without fee is hereby granted, provided
 *  that this notice appears in all copies and supporting documentation.
 *  This software is provided "as is", with no warranty.
 */

#include "usc.h"
#include "symbol.h"
#include "utils.h"
#include <ctype.h>

extern char *spec_type_to_string (Type* type);

CodeLevel *code_stack;
NameList *ParamList, *DontDerefList;
FILE *CodeOutFile;

Buffer *func_header = NULL;
Buffer *vars, *body1, *body2, *body3, *body4, *body5;
int code_stack_depth;

void merge_buffers(Buffer *body);


int pos_to_bit (ByteOrder order, int pos)
{
  int bit;
  bit = pos / 8;
  bit = order[bit] - 1;
  bit = bit * 8 + (7 - pos % 8);
  dprintf (20, "pos_to_bit: order: %s pos: %d bit: %d",
		byteorder_to_string(order), pos, bit);
  return bit;
}

void init_gen_utils (FILE *outfile)
{
  code_stack = NULL;
  code_stack_depth = 0;
  ParamList = NULL;
  CodeOutFile = outfile;
  if (func_header == NULL)
    func_header = makebuf (FUNCHEADERBUFFER);
  else
    cleanbuf (func_header);
}

void init_outfile (OutputType output_type)
{  
  char string[64];
  long time1 = time(NULL);
  struct tm *time2 = localtime (&time1);
  int count, n, i;
  char* lfptr;
  FILE* prev_code_out_file = NULL;

  count = strftime (string, 64, "%I:%M:%S %p %b %d, %Y", time2);
  lfptr = strchr(string,'\n');
  if ( lfptr ) {
    *lfptr = '\0';
    count = strlen(string);
  }

  switch( output_type ) {
  case USC_OUTPUT:
  case USIT_OUTPUT:
    init_gen_utils (OutFile);
    break;
  case USC_TYPES:
    prev_code_out_file = CodeOutFile;
    CodeOutFile = HeaderFile;
    break;
  }

  genout ("/**********************************************************/\n");
  genout ("/*                                                        */\n");
  switch( output_type ) {
    case USC_OUTPUT:
      genout ("/*                  USC Output                            */\n");
      break;
    case USIT_OUTPUT:
      genout ("/*                  USIT Output                           */\n");
      break;
    case USC_TYPES:
      genout ("/*                  USC Types                             */\n");
      break;
  }

  genout ("/*");
  n = 56 - count;
  for (i = 0; i < 18; i++)
    genout (" ");
  genout ("%s", string);
  for (; i < n; i++)
    genout (" ");
  genout ("*/\n");
  genout ("/*                  File: ");
  
  if (strcmp (InFileName, "-")) {
    count = strlen (InFileName);
    genout (InFileName);
  }
  else {
    count = strlen ("Standard Input");
    genout ("Standard Input");
  }
  for (i = 0; i < 32 - count; i++)
    genout (" ");
  genout("*/\n");
  genout ("/*                                                        */\n");
  genout ("/**********************************************************/\n\n");
  if ( ( output_type == USC_OUTPUT ) && preserve_type_names) {
    genout ("#define __USC_STUBS__\n");
  }
  if ( output_type == USC_TYPES ) {
    CodeOutFile = prev_code_out_file;
  }
}

void write_eol (Buffer *buf)
{
  if (CODE_IS_MACRO)
    writebuf (buf, "\\\n");
  else
    writebuf (buf, "\n");
}

void gen_eol (void)
{
  if (CODE_IS_MACRO)
    fprintf (CodeOutFile, "\\\n");
  else
    fprintf (CodeOutFile, "\n");
}

void genout (char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  vfprintf (CodeOutFile, fmt, args);
  va_end (args);
}

void push_code_level (void)
{
  CodeLevel *new = (CodeLevel *)TempMalloc (sizeof (CodeLevel));

  /* 'store' the stuff already done - put it BEFORE the push */
  if(code_stack_depth)
    merge_buffers(code_stack->func_body);

  code_stack_depth++;

  /* NOT NEEDED - func_body and func_vars are no longer globals
  func_body = new->func_body = makebuf (FUNCBODYBUFFER);
  func_vars = new->func_vars = makebuf (FUNCVARBUFFER);
  */
  /* However, still allocate the buffers */
  new->func_body = makebuf (FUNCBODYBUFFER);

  /* not used or needed
  new->func_vars = makebuf (FUNCVARBUFFER);
  */

  /* added for new USC phases */
  vars = new->vars = makebuf(FUNCVARBUFFER);
  body1 = new->body1 = makebuf(FUNCBODYBUFFER);
  body2 = new->body2 = makebuf(FUNCBODYBUFFER);
  body3 = new->body3 = makebuf(FUNCBODYBUFFER);
  body4 = new->body4 = makebuf(FUNCBODYBUFFER);
  body5 = new->body5 = makebuf(FUNCBODYBUFFER);
  /* end new for phases */

  new->next = code_stack;
  code_stack = new;
  write_eol (code_stack->vars);

/*
  if(vars){
    catbuf (code_stack->vars, vars);
    cleanbuf(vars);
  }
*/

  if (code_stack_depth == 1 && CODE_IS_MACRO)
    writebuf (code_stack->vars, "do {");
  else
    writebuf (code_stack->vars, "{");
  write_eol (code_stack->vars);
}

void genbuf (Buffer *buffer)
{
  fprintf (CodeOutFile, "%s", BUFFER_STRING (buffer));
}

void merge_buffers(Buffer *body){

  writebuf(body, "\n/* Phase 1 */");
  write_eol(body);
  catbuf(body, body1);
  cleanbuf(body1);

  writebuf(body, "\n/* Phase 2 */");
  write_eol(body);
  catbuf(body, body2);
  cleanbuf(body2);

  writebuf(body, "\n/* Phase 3 */");
  write_eol(body);
  catbuf(body, body3);
  cleanbuf(body3);

  writebuf(body, "\n/* Phase 4 */");
  write_eol(body);
  catbuf(body, body4);
  cleanbuf(body4);

  writebuf(body, "\n/* Phase 5 */");
  write_eol(body);
  catbuf(body, body5);
  cleanbuf(body5);
}

void pop_code_level (void)
{
  code_stack_depth--;
  if (!code_stack)
    crash ("internal error in pop_code_level");

  /* Here's the merge_buffers call */
  merge_buffers(code_stack->func_body);
  if (code_stack->next) {
    catbuf (code_stack->next->func_body, vars);
    catbuf (code_stack->next->func_body, code_stack->func_body);
    writebuf (code_stack->next->func_body, "}");
	 write_eol (code_stack->next->func_body);
  }
  else {
    genbuf (code_stack->vars);
    genbuf (code_stack->func_body);
    genout ("}");
    gen_eol ();
  }

  freebuf (code_stack->func_body);
  freebuf (code_stack->vars);
  freebuf (code_stack->body1);
  freebuf (code_stack->body2);
  freebuf (code_stack->body3);
  freebuf (code_stack->body4);
  freebuf (code_stack->body5);

  code_stack = code_stack->next;
  if (code_stack){
    vars = code_stack->vars;
    body1 = code_stack->body1;
    body2 = code_stack->body2;
    body3 = code_stack->body3;
    body4 = code_stack->body4;
    body5 = code_stack->body5;
  }
}

void gen_func_name_header (char *type, char *ident)
{
  char* static_decl;
  cleanbuf (func_header);

  push_code_level();
  static_decl = (CODE_IS_STATIC) ? "static " : "";
  if (CODE_IS_INLINE)
    writebuf (func_header, "%sinline %s %s ()", static_decl, type, ident);
  else if (CODE_IS_PLAIN)
    writebuf (func_header, "%s%s %s ()", static_decl, type, ident);
  else if (CODE_IS_MACRO)
    writebuf (func_header, "#define %s()", ident);
  else
    crash ("internal error in gen_func_name_header");
}

void gen_func_close (void)
{
  genbuf (func_header);
  pop_code_level();

  if (CODE_IS_MACRO)
    genout (" while (0)\n\n");
  else
    genout ("\n\n");
  CleanGarbage();
}

void dont_deref (char *name)
{
  NameList *new = (NameList *)TempMalloc (sizeof (NameList));

  new->name = name;
  new->next = DontDerefList;
  DontDerefList = new;
}

int do_i_deref (char *name)
{
  NameList *p = DontDerefList;

  int found = 0;
  while (p && !found) {
    if (!strcmp (name, p->name))
      found = 1;
    p = p->next;
  }
  return !found;
}
