/*
 * cp_nasd.c
 *
 * copies all the files in the specified dir to a NASD drive.  outputs
 * to stdout a mapping between filenames and NASD ids.  runs colocated.
 */
/*
 * Copyright (c) 1998,1999 Carnegie-Mellon University.
 * All rights reserved.
 *
 * Author: David Petrou
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting
 * documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie
 * Mellon the rights to redistribute these changes.
 */

#include <nasd/nasd_options.h>
#include <nasd/nasd_threadstuff.h>
#include <nasd/nasd_pdrive.h>
#include <nasd/nasd_pdrive_client.h>
#include <nasd/nasd_timer.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>

typedef int bool;

bool drive_inited;
nasd_drive_handle_t handle;
nasd_key_t op_key;
nasd_cookie_t op_cookie;
nasd_partnum_t partnum;

int main(int argc, char **argv);
void init_drive(char *server_name, nasd_partnum_t partnum,
		char *master_password);
void cp_dir(char *dir_name);
void cp_file(char *filename);
/*void split_filename(char *filename, char **ext);*/
nasd_identifier_t create_nasd_object(void);
void write_nasd(nasd_identifier_t object_id, nasd_byte_t *buf,
		nasd_offset_t off, nasd_len_t in_len);
void usage(char *progname);
void teardown(void);
void exit_bad(const char *fmt, ...);
void exit_good(void);

int
main(int argc, char **argv)
{
    char *progname;
    char *server_name, *dir_name, *master_password;
    nasd_timer_t timer;
    nasd_timespec_t diff;
    double elapsed_time;
    
    drive_inited = 0;

    progname = argv[0];
    if(argc != 5) {
	usage(progname);
    }
    server_name = argv[1];
    partnum = atoi(argv[2]);
    master_password = argv[3];
    dir_name = argv[4];
    
    init_drive(server_name, partnum, master_password);

    NASD_TM_START(&timer);

    cp_dir(dir_name);

    NASD_TM_STOP(&timer);
    NASD_TM_ELAPSED_TS(&timer, &diff);
    elapsed_time = (double)diff.ts_nsec;
    elapsed_time /= (double)NASD_NSEC_PER_SEC;
    elapsed_time += (double)diff.ts_sec;

    fprintf(stderr, "%f seconds\n", elapsed_time);
    
    exit_good();
    /* to shut gcc up */
    exit(0);
}

void
init_drive(char *server_name, nasd_partnum_t partnum, char *master_password)
{
    int retval;
    int i;
    nasd_drive_param_dce_kpdev_t kpdev_args;
    nasd_key_t part_key, master_key, drive_key, red_key;

    assert(server_name);

    nasd_security_password_to_keys(master_password, master_key, drive_key,
				   partnum, part_key, red_key, op_key);

    /* XXX use nasd_security_build_capability() for a non-daily key? */
    bzero((char *)&op_cookie, sizeof(op_cookie));
    op_cookie.capability_args.type = NASD_BLACK_KEY;
    op_cookie.capability_args.min_protection = NASD_NO_PROTECTION;
    bcopy(op_key, op_cookie.key, sizeof(nasd_key_t));

    if((retval = nasd_cl_p_init())) {
	exit_bad("ERROR: cannot init client library, rc = %d, line %d\n",
	       retval, __LINE__);
    }
    drive_inited = 1;
    strcpy(kpdev_args.devname, "/dev/nasdkp0");
    if(nasd_bind_to_drive(server_name, NASD_PDRIVE_PORT, NASD_BIND_COLOCATE,
			  &kpdev_args, sizeof(kpdev_args), &handle)) {
	exit_bad("ERROR: cannot bind to server %s, line %d", server_name,
		 __LINE__);
    }
}

void
cp_dir(char *dir_name)
{
    struct dirent *dirp;
    struct stat statbuf;
    DIR *dp;

    assert(dir_name);

    if(chdir(dir_name) == -1) {
	exit_bad("chdir(), line %d", __LINE__);
    }	

    if((dp = opendir(".")) == 0) {
	exit_bad("opendir(), line %d", __LINE__);
    }

    while((dirp = readdir(dp)) != 0) {

	if(stat(dirp->d_name, &statbuf) == -1) {
	    exit_bad("stat(), line %d", __LINE__);
	}

	if(S_ISDIR(statbuf.st_mode)) {
	    continue;
	}

	cp_file(dirp->d_name);
    }
    
    if(closedir(dp) == -1) {
	exit_bad("closedir(), line %d", __LINE__);
    }
}

void
cp_file(char *filename)
{
    int i;
    char buf[BUFSIZ];
    int fd;
    nasd_status_t nasd_status;
    nasd_rpc_status_t status;
/*    char *ext;*/
    nasd_identifier_t object_id;    
    int cur_pos;

    assert(filename);

/*    split_filename(filename, &ext);*/

    if((fd = open(filename, O_RDONLY, 0)) == -1) {
	exit_bad("open(), line %d", __LINE__);
    }
    object_id = create_nasd_object();

    cur_pos = 0;
    while((i = read(fd, buf, BUFSIZ)) != 0) {
	if(i == -1) {
	    exit_bad("read(), line %d", __LINE__);
	}

	write_nasd(object_id, (nasd_byte_t *)buf, cur_pos, i);
	cur_pos += i;
    }

    if(close(fd) == -1) {
	exit_bad("close(), line %d", __LINE__);
    }

    printf("%s\t0x%" NASD_ID_FMT "\n", filename, object_id);
}

/*

at one point i was considering putting the filename extension into the
uninterpreted field in the nasd object.  this code finds the
extension...

void
split_filename(char *filename, char **ext)
{
    int i;

    assert(filename);
    assert(ext && *ext);

    i = 0;
    while(filename[i] && filename[i] != '.') {
	i++;
    }
    if(filename[i] == '.') {
	*ext = filename + i + 1;
    } else {
	*ext = filename + i;
    }
}*/

nasd_identifier_t
create_nasd_object(void)
{
    nasd_p_create_dr_args_t args;    
    nasd_p_create_dr_res_t res;
    nasd_rpc_status_t status;
    nasd_error_string_t err_str;
    
    bzero((char *)&args, sizeof(args));
    args.in_security_param.type = NASD_BLACK_KEY;
    args.in_security_param.partnum = partnum;
    args.in_security_param.actual_protection = NASD_SECURE_OPS;
    args.in_fieldmask = NASD_ATTR_AV;
    args.in_attribute.av = 7;
    args.in_partnum = partnum;

    nasd_cl_p_create_dr(handle, op_key, &args, &res, &status);
    if(status || res.nasd_status) {
	exit_bad("ERROR: status=0x%x (%s) nasd_status=0x%x (%s)",
		 status, nasd_cl_error_string(handle, status, err_str),
		 res.nasd_status, nasd_error_string(res.nasd_status));
    }

    return res.out_identifier;
}

void
write_nasd(nasd_identifier_t object_id, nasd_byte_t *buf, nasd_offset_t off,
	   nasd_len_t in_len)
{
    nasd_p_smpl_op_dr_args_t wr_args;
    nasd_p_fastwrite_dr_res_t wr_res;
    nasd_rpc_status_t status;
    nasd_len_t out_len;
    nasd_status_t rc;
    nasd_error_string_t err_str;

    assert(buf);

    wr_args.in_partnum = partnum;
    wr_args.in_identifier = object_id;
    wr_args.in_offset = off;
    wr_args.in_len = in_len;
    wr_args.in_capability_args = op_cookie.capability_args;
    wr_args.in_security_param.type = op_cookie.capability_args.type;
    wr_args.in_security_param.partnum = partnum;
    wr_args.in_security_param.actual_protection = op_cookie.capability_args.min_protection;
    nasd_cl_p_write_simple_dr(handle, op_cookie.key, &wr_args, buf,
			   &wr_res, &status);
    out_len = wr_res.out_datalen;
    rc = wr_res.nasd_status;

    if(rc || status) {
	exit_bad("WRITE %d at %" NASD_64u_FMT " from 0x%lx ERROR: "
		 "nasd_status=0x%x (%s) status=0x%x (%s)\n",
		 in_len, off, buf, rc, nasd_error_string(rc),
		 status, nasd_cl_error_string(handle, status, err_str));
    }

    if(out_len != in_len) {
	exit_bad("WRITE %d at %" NASD_64u_FMT " from 0x%lx ERROR: "
		 "in_len=%" NASD_64u_FMT " out_len=%" NASD_64u_FMT "\n",
		 in_len, off, buf, in_len, out_len);
    }
}

void
usage(char *progname)
{
    assert(progname);

    exit_bad("USAGE: %s servername partnum master_password directory",
	     progname);
}

void
teardown(void)
{
    if(drive_inited) {
	nasd_cl_p_shutdown();
    }
}

void
exit_bad(const char *fmt, ...)
{
    int errno_save;
    va_list ap;
    static bool here_already = 0;

    /* eliminate recursive calls.  XXX not signal/thread safe.  */
    if(here_already) {
	return;
    }
    here_already = 1;

    errno_save = errno;

    assert(fmt);

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);

    fprintf(stderr, "\n");

    if(errno_save) {
        fprintf(stderr, "system error: %s\n", strerror(errno_save));
    }

    teardown();
    exit(EXIT_FAILURE);
}

void
exit_good(void)
{
    teardown();
    exit(EXIT_SUCCESS);
}
